/*******************************************************************************

FILE		:   Scan_LM983x.h
Revision	:   1.4
Date		:   10/21/99	

DESCRIPTION	:   The LM983x scanner interface module must conform to the 
                interface presented in this file. The basic functions 
                supported by the scanner module are 
                - Querying Scanner capabilities
                - Setting/Getting current scan parameters
                - Scanning an image using current scan parameters

                The actual interfaces to perform these operations are described 
                in detail below.

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/

#ifndef _Scan_LM983x_H__
#define _Scan_LM983x_H__

/*******************************************************************************
Defines & Macros
*******************************************************************************/
#if _BUILDING_SCAN_MODULE_

#define Scan_LM983x_EXPORT _declspec(dllexport)

#else

#define Scan_LM983x_EXPORT _declspec(dllimport)

#endif

#ifdef __cplusplus
extern "C" {
#endif

#define SCAN_GAMMA_FILENAME     255

/*******************************************************************************
Scanner Module error codes 
*******************************************************************************/
// Methods which return error codes will return the standard scan result 
// error codes - this will allow the error return from the scanner module
// interface functions to be handled uniformly
typedef long SCAN_RESULT;

#define SCAN_OK			    0L
#define SCAN_ERROR		    1L
#define SCAN_DONE           2L
#define SCAN_ABORT          3L
#define SCAN_INVALID_CAPID  4L

/*******************************************************************************
Scanner Module capability types 
*******************************************************************************/
// The scanner module can be interrogated for scanner capabilities for the 
// current scanner. This is done through a single function which allows
/// the type of capability to be specfied. The capabilities are identified 
// using the standard capability ID whose type is defined below
typedef long SCAN_CAPID;

#define SCAN_CAPID_MAXNUM	    0x8000

// Input param will be NULL and output will be a fixed array of ints
#define SCAN_CAP_HRES			0L
// Input param will be NULL and output will be a fixed array of ints
#define SCAN_CAP_VRES			1L
// Input param will be NULL and output will be a fixed array of ints
#define SCAN_CAP_PIXELTYPE		2L
// Input param will be the pixel type (int*) and output will be a fixed 
// array of ints
#define SCAN_CAP_PIXELDEPTH		3L
// Input param will be NULL and the output will be a pointer to an int
#define SCAN_MAX_SCAN_WIDTH     4L
// Input param will be NULL and the output will be a pointer to an int
#define SCAN_MAX_SCAN_HEIGHT    5L

/*******************************************************************************
Scanner Module capability values 
*******************************************************************************/
// These are the supported values for the Pixel Type capability
#define SCAN_PIXELTYPE_BW       0
#define SCAN_PIXELTYPE_GRAY     1
#define SCAN_PIXELTYPE_RGB      2

/*******************************************************************************
Abstract Data Types defined by this module
*******************************************************************************/

typedef struct tagScanParamRec
{
	BOOL	scanToTwain;	// indicates if the scan information has to be passed
							// to the TWAIN application. This determines the 
							// format of the scan data that is returned in response 
							// a read call
	RECT	boundingRect;	// The bounding rect must be the physical size in pixels
							// of the scanner being used
	RECT	scanRect;		// The scan rect is the size in pixels of the area to be 
							// scanned. This must be wholly  contained in the 
							// boundingRect
	long	pixelType;		// The pixel type of the scan data to be acquired
	long	bitsPerPixel;	// The number of bits per pixel of scan data
	long	hRes;			// The horizontal resolution to be used for future scans
	long	vRes;			// The vertical resoltuion to be used for future scans
	BOOL	hInterp;		// This flag indicates if the row data should be 
							// interpolated or discarded to achieve the desired 
							// resolution
	BOOL	vInterp;		// This flag indicates if the column data has to be 
							// interpolated or discarded to achieve the desired 
							// resolution


}SCAN_PARAM_REC;

typedef struct tagScanInfoRec
{
	long totalBytes;		// total bytes that will be returned by a scan request 
							// at this time
	long actBytes;			// The actual bytes that have been returned in this 
							// read
	long bytesPerLine;		// The number of bytes per scan line

	long startTime;			// The time in millisec format when the scan request was 
							// received 
	long scanStartTime;		// The time in millisec format when the scan was started
	long scanEndTime;		// The time in millisec format when the scan was completed
	long endTime;			// The time in millisec format when the last data packet was returned
							// to the scan requester
	char scanStatus[256];	// status/error string

	long rawTotalBytes;		// raw total scan bytes
	long rawActBytes;		// raw scan bytes
}SCAN_INFO_REC;

typedef struct tagScanHardwareRec
{
	long widthOpticalPixels;// Full optical scan width in pixels
	long heightOpticalPixels;//Full optical scan height in pixels
	long opticalRes;		// Optical resolution
	long chip;				// Chip
	char device[256];		// Device string
#define BASE_RESOLUTION_SIZE 16
	long baseResolution[BASE_RESOLUTION_SIZE]; //Base scan resolutions (unused = 0)
}SCAN_HARDWARE_REC;

// Any callback function passed into the scanner module must adhere to the
// signature defined below
typedef SCAN_RESULT (*FN_SCAN_CALLBACK)(BOOL);

/*******************************************************************************
Exported functions 
*******************************************************************************/
// This method asks the scanner to start a scan operation using the current
// scan settings. This call only sets up the scanner for a read. To 
// To actually read in the image data use the ScanRead method.
Scan_LM983x_EXPORT SCAN_RESULT ScanStart();

// This method must be called only after the ScanStart method has been called
// to setup the scanner for a read.
Scan_LM983x_EXPORT SCAN_RESULT ScanRead(void*	scanBuffer,
										long	regCount,
										long*	actCount);

// This method requests the scanner module to abort the current scan
// If no scan is in progress the call will be ignored. This method
// should be invoked only to abnormally terminate the scan in response to 
// a user request or a error condition. This should not be used in normal
// scan operation. Normal scan operation end is signalled by the ScanRead
// method itself.
Scan_LM983x_EXPORT SCAN_RESULT ScanStop();

// This method allows a client to set the current scan paramaters. These
// paramters will be used by the scanner module for all subsequent scan
// until the settings are changed again by a call to this method i.e the
/// scan settings are persistent. 
// The scan settings however are not persistent across module loads i.e
// unloading the scan module and loading it again will reset the scan settings
// to default power on settings defined in the scan module configuration file
Scan_LM983x_EXPORT SCAN_RESULT ScanSetScanParams(SCAN_PARAM_REC* paramRec,
												 FN_SCAN_CALLBACK scanCB);

// This method allows a client to get the current scan paramaters. These
// paramters will be used by the scanner module for any scan that is 
// requested. To change the scan settings use the ScanSetScanParams
// method
Scan_LM983x_EXPORT SCAN_RESULT ScanGetScanParams(SCAN_PARAM_REC* paramRec);

// This method returns the status of the current scan including information
// about the sie of the scan the number of lines scanned etc. If a scan is not
// in progress this method will return the computed values if a scan were
// to be started. In this case some of the field like actual bytes read will
// not be valid and will be set to zero
Scan_LM983x_EXPORT SCAN_RESULT ScanGetScanInfo(SCAN_INFO_REC *infoRec);

// This method returns the hardware settings
Scan_LM983x_EXPORT SCAN_RESULT ScanGetScanHardware(SCAN_HARDWARE_REC *hardwareRec);

// This method allows a client to interrogate the scanner module for the 
// current value of supported capabilities like bit depth, pixel type, 
// resolutions etc
// In order to allow different capabilities to be queried the input and 
// output parameters are defined as void pointers the actual type of the 
// arguments pointed to by these pointers will depend on the capability 
// being queried.
// See the capability ID definitions for the input and output parameters 
// for that capability
Scan_LM983x_EXPORT SCAN_RESULT ScanGetCapability(SCAN_CAPID capId, 
												 void*  input, 
												 void * output,
                                                 long*  numOutputElements);

#ifdef __cplusplus
}
#endif

#endif // _Scan_LM983x_H__
