/*******************************************************************************

FILE		:   Twain_LM983x.h
Revision	:   1.4	
Date		:   10/21/99	

DESCRIPTION	:   The LM983x TWAIN interface module must conform to the 
                interface presented in this file. The basic functions 
                supported by the TWAIN module are 
                - Querying Scanner capabilities
                - Setting/Getting current scan parameters
                - Obtaining image data in requested formats
                The TWAIN module delegates most of the work in these opertions
                to the scanner module but this will be transparent to clients
                of the TWAIN module
                The actual interfaces to perform these operations are described 
                in detail below.

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/

#ifndef _Twain_LM983x_H__
#define _Twain_LM983x_H__

/*******************************************************************************
Defines & Macros
*******************************************************************************/
#if _BUILDING_TWAIN_MODULE_

#include "Scan_LM983x.h"

#define Twain_LM983x_EXPORT _declspec(dllexport)

#else

#define Twain_LM983x_EXPORT _declspec(dllimport)

#endif


#ifdef __cplusplus
extern "C" {
#endif


#define TWAIN_GAMMA_FILENAME     255

/*******************************************************************************
TWAIN Module error codes 
*******************************************************************************/
// Methods which return error codes will return these standard result codes 
// - this will allow the error returns  from this module's interface functions 
// to be handled uniformly
typedef long TWAIN_RESULT;

#define TWAIN_OK				0L
#define TWAIN_ERROR				1L
#define TWAIN_DONE              2L
#define TWAIN_ABORT             3L
#define TWAIN_INVALID_CAPID     4L
#define TWAIN_UNKNOWN_MSG       5L

/*******************************************************************************
TWAIN Module Capability Defines 
*******************************************************************************/
// The TWAIN module can return capabilities for the current scanner. 
/// This is done through a single function which allows
/// the type of capability to be specfied. The capabilities are identified 
// using the standard capability ID whose type is defined below
typedef long TWAIN_CAPID;

// The maximum number of values for any one capabilty 
#define TWAIN_CAPID_MAXNUM		0x8000

// Input param will be NULL and output will be a fixed array of ints
#define TWAIN_CAP_HRES			0L
// Input param will be NULL and output will be a fixed array of ints
#define TWAIN_CAP_VRES			1L
// Input param will be NULL and output will be a fixed array of ints
#define TWAIN_CAP_PIXELTYPE		2L
// Input param will be the pixel type and output will be a fixed array of ints
#define TWAIN_CAP_PIXELDEPTH	3L
// Input param will be NULL and the output will be a pointer to an int
#define TWAIN_MAX_SCAN_WIDTH     4L
// Input param will be NULL and the output will be a pointer to an int
#define TWAIN_MAX_SCAN_HEIGHT    5L


/*******************************************************************************
TWAIN Module capability values 
*******************************************************************************/
// These are the supported values for the Pixel Type capability
#define TWAIN_PIXELTYPE_BW       0
#define TWAIN_PIXELTYPE_GRAY     1
#define TWAIN_PIXELTYPE_RGB      2

/*******************************************************************************
Abstract Data Types defined by this module
*******************************************************************************/
typedef struct tagTwainParamRec
{
	RECT	boundingRect;	// The bounding rect must be the physical size in pixels
							// of the scanner being usef
	RECT	scanRect;		// The scan rect is the size in pixels of the area to be 
							// scanned. This must be wholly  contained in the 
							// boundingRect
	long	pixelType;		// The pixel type of the scan data to be acquired
	long	bitsPerPixel;	// The number of bits per pixel of scan data
	long	hRes;			// The horizontal resolution to be used for future scans
	long	vRes;			// The vertical resoltuion to be used for future scans
	BOOL	hInterp;		// This flag indicates if the row data should be 
							// interpolated or discarded to achieve the desired 
							// resolution
	BOOL	vInterp;		// This flag indicates if the column data has to be 
							// interpolated or discarded to achieve the desired 
							// resolution
    
}TWAIN_PARAM_REC;

typedef struct tagTwainInfoRec
{
	long totalBytes;		// total bytes that will be returned by a scan request 
							// at this time
	long actBytes;			// The actual bytes that have been returned in this 
							// read
	long bytesPerLine;		// The number of bytes per scan line
	long startTime;			// The time in msec format when the scan request was 
							// received 
	long scanStartTime;		// The time in msec format when the scan was started
	long scanEndTime;		// The time in msec format when the scan was completed
	long endTime;			// The time in msec when the last data packet was returned
	char scanStatus[256];	// status/error string
	long rawTotalBytes;		// raw total scan bytes
	long rawActBytes;		// raw scan bytes
}TWAIN_INFO_REC;

typedef struct tagTwainHardwareRec
{
	long widthOpticalPixels;// Full optical scan width in pixels
	long heightOpticalPixels;//Full optical scan height in pixels
	long opticalRes;		// Optical resolution
	long chip;				// Chip
	char device[256];		// Device string
#define BASE_RESOLUTION_SIZE 16
	long baseResolution[BASE_RESOLUTION_SIZE]; //Base scan resolutions (unused = 0)
}TWAIN_HARDWARE_REC;

typedef TWAIN_RESULT (*FN_TWAIN_CALLBACK)(BOOL check_messages);

/*******************************************************************************
Exported Functions 
*******************************************************************************/
// This method should be called to initiate a scan and send the 
// results to the TWAIN application driving this data source
// This method is asynchronous and will return immediately the
// caller must transition to the "Scanning" mode upon successful
// return
// An error return indicates that the scan request could not be 
// completed. The error code specifies the nature of the error
// 
Twain_LM983x_EXPORT TWAIN_RESULT TwainScan();

Twain_LM983x_EXPORT TWAIN_RESULT TwainPreview(LPCSTR pszFileName);

Twain_LM983x_EXPORT TWAIN_RESULT TwainAbort();

Twain_LM983x_EXPORT TWAIN_RESULT TwainSetScanParams(TWAIN_PARAM_REC* paramRec);
Twain_LM983x_EXPORT TWAIN_RESULT TwainGetScanParams(TWAIN_PARAM_REC* paramRec);
Twain_LM983x_EXPORT TWAIN_RESULT TwainGetScanInfo(TWAIN_INFO_REC *infoRec);
Twain_LM983x_EXPORT TWAIN_RESULT TwainGetScanHardware(TWAIN_HARDWARE_REC *hardwareRec);

Twain_LM983x_EXPORT TWAIN_RESULT TwainGetCapability(TWAIN_CAPID capId,
												    void* input,
												    void* output,
                                                    long* numOutputElements);

Twain_LM983x_EXPORT TWAIN_RESULT TwainClose();

#if _BUILDING_TWAIN_MODULE_
Twain_LM983x_EXPORT SCAN_RESULT TwainCallback(BOOL check_messages);
#endif

#ifdef __cplusplus
}
#endif


#endif // _Twain_LM983x_H__
