// FILE		: Scan_Hardware.CPP

// DESCRIPTION	: Implementation of the CHardware class
//			Scan_Hardware interface code

// REVISIONS	: 

#include "StdAfx.h"
#include "Scan_Hardware.h"
#include "Scan_driverusb.h"
#include "math.h"

CDriverUSB DriverUSB; 

#define INIFILE GetFullFilePath("lm983x.def")

////////////////////////////////////////////////////////////////////////////
// constructor
//-------------------------------------------------------
// FUNCTION	:	CHardware class constructor

// DESCRIPTION	:	Create a CHardware object and initialize variables

// RETURNS	:	nothing
//--------------------------------------------------------
CHardware::CHardware()
{
	m_callbackfn = NULL;
	m_home = FALSE;
}
/////////////////////////////////////////////////////////////////////////////
// destructor
//-------------------------------------------------------
// FUNCTION	:	CHardware class destructor

// DESCRIPTION	:	clean up memory allocation and resources

// RETURNS	:	nothing
//--------------------------------------------------------
CHardware::~CHardware()
{
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	delay(count)

// DESCRIPTION	:	process the windows message loop for at least 50ms
//			the clock is in milliseconds, but the resolution
//			is 1/18 of a sec or about 50 ms
//			so delay(1) will delay for 50 ms
//			delay(1000) will dealy 1 sec

// INPUTS	:	count	- number of milliseconds to delay

// RETURNS	:	nothing
//--------------------------------------------------------
void CHardware::delay(long count)
{
	Sleep(count);
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	WriteRegister(unsigned char Address, unsigned char Value)

// DESCRIPTION	:	write a byte to a merlin register

// INPUTS	:	Address	- reg to write
//			Value	- byte to write to merlin

// RETURNS	:	TRUE if success
//--------------------------------------------------------
BOOL CHardware::WriteRegister(unsigned char Address, unsigned char Value)
{
	for (int i = 0; i < 100; i++)
	{
		if (!DriverUSB.BulkOut(Address,&Value,1,FALSE))
		{
			NSCAfxMessageBox("Write Register failed! %02X=%02X",Address,Value);
			return FALSE;
		}
 		if (Address == 0x58)
 		{
 			BYTE val; //flush register 2
 			if (!ReadRegister(2,&val)) return FALSE;
 			if (!ReadRegister(2,&val)) return FALSE;
 		}
		if (Address != 7) 	return TRUE; 
		//verify register 7
		unsigned char val;
		if (!ReadRegister(7, &val)) return FALSE;
		if (val == Value) 
		{
			if (val&7)
			{
				BYTE value;
				if (!ReadRegister(0x45,&value)) return FALSE;
				if (value & 0x10) //motor on
					m_home = (val == 2); //see if homing
			}
			return TRUE;
		}
		NSCStatusOut("Retry 7=%d as 7=%d",Value,val);
	}

	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SendData(int writeaddress,BYTE *bufptr,int len)

// DESCRIPTION	:	write merlin data as a block
//		
// INPUTS	:	bufptr	- data to write
//			len - number of bytes to write

// RETURNS	:	TRUE if success 
//--------------------------------------------------------
BOOL CHardware::SendData(BYTE writeaddress,BYTE *bufptr,int len)
{
	if (!len) return TRUE;
	if (writeaddress == 6)
	{
		if (!WriteRegister(4,0)) return FALSE;
		if (!WriteRegister(5,0)) return FALSE;
	}
	if (!DriverUSB.BulkOut(writeaddress,bufptr,len,FALSE))
	{
		NSCAfxMessageBox("Bulk Out failed! %02X",writeaddress);
		return FALSE;
	}
	if (m_debug && writeaddress > 7)
	{
		unsigned char read_buffer[0x4000*2];
		if (!ReadData(writeaddress,read_buffer,len)) return FALSE;
		for (int i = 0; i < len; i++)
		{
			if (bufptr[i] != read_buffer[i])
			{
				NSCAfxMessageBox("Verify error: addr=%x,write=%x,read=%x",
					writeaddress,bufptr[i],read_buffer[i]);
				return FALSE;
			}
		}
		NSCStatusOut("Verified senddata");
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ReadData(BYTE readaddress,BYTE *bufptr,int maxtoread)

// DESCRIPTION	:	read merlin data as a block

// INPUTS	:	bufptr	- place to store bytes read
//			maxtoread - max number of bytes to read

// RETURNS	:	true if success 
//--------------------------------------------------------
BOOL CHardware::ReadData(BYTE readaddress,BYTE *bufptr,int maxtoread)
{
	if (!maxtoread) return TRUE;
	if (readaddress == 6)
	{
		if (!WriteRegister(4,0)) return FALSE;
		if (!WriteRegister(5,0)) return FALSE;
	}
	if (!DriverUSB.BulkIn(readaddress,bufptr,maxtoread,FALSE))
	{
		NSCAfxMessageBox("Bulk In failed! %02X",readaddress);
		return FALSE;
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ReadRegister(unsigned char Address,BYTE *pValue)

// DESCRIPTION	:	read value from a merlin register

// INPUTS	:	Address	- merlin register to read
//			pValue	- place to store byte read

// RETURNS	:	TRUE if success 
//--------------------------------------------------------
BOOL CHardware::ReadRegister(unsigned char Address,BYTE *pValue)
{
	if (!DriverUSB.BulkIn(Address,pValue,1,FALSE))
	{
		NSCAfxMessageBox("Read Register failed! %02X",Address);
		return FALSE;
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	HomeSensor(BOOL wait)

// DESCRIPTION	:	move sensor to home position

// RETURNS	:	true if ok, false if sensor can't be homed
//--------------------------------------------------------
BOOL CHardware::HomeSensor(BOOL wait /*=TRUE*/)
{
BYTE reg58;		// current value of reg 0x58
BYTE value;		// i/o value for port
BYTE psState;		// bit to clear depending on ini file specification of home_sensor

	if (!ReadRegister(0x7,&value)) return FALSE;
	if (value == 2) return TRUE; //already homing

	// check cmd reg = 0
	if (!CheckScannerReady()) 
		return FALSE;

	// set reg 58 mode to paper sensor level sensitive
	if (!ReadRegister(0x58,&value)) return FALSE;

	reg58 = value;
	if (merlininfo.Home_Sensor == 2) 
	{
		value &= ~0x10; // paper sensor 2 set level sensitive
		psState = 2;
	}
	else 
	{
		value &= ~2; // paper sensor 1 set level sensitive
		psState = 1;
	}
	if (!WriteRegister(0x58, value)) return FALSE;
	if (!ReadRegister(2,&value)) return FALSE;
	if (!WriteRegister(0x58, reg58)) return FALSE;
	if (value & psState)
	{ // says at home
		if (m_home) return TRUE;
		if (!WriteRegister(0x58, 0)) return FALSE; //disable stop at home
		// fast forward for a couple seconds. 
		// if home and at home the sensor bangs
		if (!WriteRegister(7, 1)) return FALSE;
		delay(500); // ff for 1/2 secs
		// stop ff
		if (!WriteRegister(7, 0)) return FALSE;
		delay(2); // brief pause to stop
		if (!WriteRegister(0x58, reg58)) return FALSE;
	}
	// home sensor
	if (!WriteRegister(7, 2)) return FALSE;
	// wait for home complete
	if (wait) if (!CheckScannerReady()) return FALSE;
	return TRUE;
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	CheckScannerReady()

// DESCRIPTION	:	check reg 7 = 0

// RETURNS	:	true if ok, false if error
//--------------------------------------------------------
// return false if cmd reg is not 0
BOOL CHardware::CheckScannerReady()
{//
BYTE value;	// value read/write from/to merlin
DWORD timeout;	// current clock to count time

	// time in ms = 1000*scanner length in inches/max step speed/in
	timeout = GetTickCount() + (int) (1000*(merlininfo.Scanner_Length+5)/merlininfo.Max_Speed);
	while (GetTickCount() < timeout) 
	{ // wait for home
		if (!ReadRegister(7,&value)) return FALSE;
		if (value == 0)
			return TRUE;
		if (value >= 0x20 || value == 3)
			if (!WriteRegister(7,0)) return FALSE;
	}
	NSCAfxMessageBox("Sensor not homed");
	NSCStatusOut("Reg 7 = %d",value);
	m_home = FALSE;
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ComputeMinMCLK(void)

// DESCRIPTION	:	compute the minimum mclk

// INPUTS	:	none

// RETURNS	:	min mclk * MCLKDIV_SCALING
//--------------------------------------------------------
int CHardware::ComputeMinMCLK(void)
{
	merlininfo.minmclk = (int) (1*MCLKDIV_SCALING);
	merlininfo.timeadj = 0;
	return merlininfo.minmclk;
}

//-------------------------------------------------------
// FUNCTION	:	InitUSB(full)

// DESCRIPTION	:	initialize merlin

// RETURNS	:	true if init ok, otherwise false
//-------------------------------------------------------
BOOL CHardware::InitUSB(BOOL full)
{
int regs[0x80];
	//set min clock
	merlininfo.mclk = ComputeMinMCLK();

	FILE *fp = fopen(INIFILE,"r");
	if (fp == NULL)
	{
			NSCAfxMessageBox("%s not found",INIFILE);
			return FALSE;
	}
	if (fclose(fp))
	{
			NSCAfxMessageBox("%s cannot be closed",INIFILE);
			return FALSE;
	}

	//check for debug
	m_debug = GetPrivateProfileInt("SETUP","Debug",0,INIFILE);

	//get scan buff size
	//use temp memory to smooth out scan
	m_membufSize = GetPrivateProfileInt("SETUP","Scan_Buffer_MBytes",0,INIFILE);
	if (!m_membufSize)
	{
		m_membufSize = 8;
		WritePrivateProfileString("SETUP","Scan_Buffer_MBytes","8",INIFILE);
	}
	if ((m_membufSize > 128) || (m_membufSize < 1))
	{
		NSCAfxMessageBox("Scan_Buffer_MBytes must be 1 to 128");
		m_membufSize = 8;
	}
	NSCStatusOut("membufsize=%x",m_membufSize);
	m_membufSize *= 1024*1024; //Mbytes

//strings from .ini file 
	wchar_t Merlin_Manufacturer[MAX_PATH];
	wchar_t Merlin_Device[MAX_PATH];
	if (!(MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,merlininfo.USB_Manufacturer,
			-1,Merlin_Manufacturer,80) && 
		  MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,merlininfo.USB_Device,
			-1,Merlin_Device,80) &&
		DriverUSB.InitDriver(Merlin_Manufacturer,Merlin_Device)))
	{
		NSCAfxMessageBox("Can't connect to %s",merlininfo.USB_Device);
		return FALSE;
	}

	if (!CheckScannerReady()) return FALSE;

	if (!SetupStandardRegisters(regs,FALSE)) return FALSE;

//------------------------------------------
	ComputeStepSize(50,regs); //50dpi for now

//------------------------------------------
	static BOOL lamp_warm = FALSE;
	if (merlininfo.Illumination_Mode == 1) //CCD
	{
		BYTE reg29;
		if (!ReadRegister(0x29,&reg29)) return FALSE;
		if (!reg29)
		{
			lamp_warm = TRUE;	//lamp warm if lamp off
		}
	}

	// home the sensor after wake up
	if (WriteMerlin(regs))
	{
		if (full)
		{
			if (lamp_warm) 	
			{
				NSCAfxMessageBox("Warming lamp");
				delay(10000);		//wait 10 seconds
				lamp_warm = FALSE;
				m_home = FALSE;
				if (!HomeSensor()) return FALSE;
			}
			else if (m_debug) if (!HomeSensor()) return FALSE;
		}
		return TRUE;
	}
//---------------------------------------------------------
	return FALSE; // either write merlin or reset merlin failed
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ComputePauseLimit(int *regs)

// DESCRIPTION	:	compute the pause limit from register settings

// INPUTS	:	regs	- array of merlin registers

// RETURNS	:	none
//--------------------------------------------------------
void CHardware::ComputePauseLimit(int *regs)
{
BYTE value;	// work variable

// bytes/line = valid pixels * (bpp / 8) /DPI
// valid pixels = reg22:23 - reg 24:25
// bpp = reg9 b3:b4 00=1, 01=2, 10=4, 11=8
// dpi = reg9 b0..b2, 0->1, 1->1.5, 2->2, 3->3, 4->4, 5->6, 6->8, 8->12
// reg26 b0..b2 =0 for pixel rate, otherwise line rate
// if pixel rate, then bytes/line = 3 * valid pixels * (bpp/8)/dpi

// high byte of first pixel to scan
	int valpix = 256*regs[0x22];

// low byte of first pixel to scan
	valpix += regs[0x23];
	int startpixel = valpix;

// last pixel to scan
	int endpixel = 256*regs[0x24] + regs[0x25];

// get the bits per pixel
// bpp=bits per pixel, ppb = pixels per bit
	int bpp = 16;
	value = regs[0x9];
	switch (value & 0x38) {
		case 0: bpp = 1; break;
		case 0x8: bpp = 2; break;
		case 0x10: bpp = 4; break;
		case 0x18: bpp = 8; break;
		default: bpp = 16; break;
	}

// compute the dpi (pixels per inch)
	value &= 7;
	double dpi = 1.;
	switch (value) {
		case 0: dpi = 1.; break;
		case 1: dpi = 1.5; break;
		case 2: dpi = 2.; break;
		case 3: dpi = 3.; break;
		case 4: dpi = 4.; break;
		case 5: dpi = 6.; break;
		case 6: dpi = 8.; break;
		case 7: dpi = 12.; break;
	}

// compute the number of bytes per line
	valpix = endpixel - startpixel;
	int x = (int) (valpix/dpi);

// the number of pixels per line = bytes per line * pixels per byte
	int pixelsperline = (x);
	int bytesperline =	(x);

// determine the line rate or pixel rate
	value = regs[0x26];
	value &= 7;
//	int linerate = value; // 0=pixel, 1=line, 4=mode a, 5=mode b
// if line rate, then we get a line of red, a line of green, a line of blue
// if pixel rate, then each line contains rgb rgb rgb, etc.
// if line rate, the buffer must hold 3 lines
	bytesperline *= 3; //handle largest

//ppb divide
	bytesperline *= (bpp/8);

	//bytesperline must be truncated word boundary
	//add two status bytes
	bytesperline &= ~1;
	bytesperline += 2;

//-----------------------------------
//	compute size of coefficient ram
	int coeffsize = 4 + 16 + 16;	//gamma and shading and offset
//  if 16 bit, then not all is used
	if (regs[0x9]&0x20)
	{
		coeffsize = 16 + 16;	//no gamma
	}
	coeffsize *= (2*3); //3 colors and 2 bytes/word

//	compute size of ram
	value = (regs[0x42]/64);
	int ramsize = 512;
	switch (value) {
		case 0: ramsize = 512; break;
		case 1: ramsize = 2048;break;
	}

//	pause limit = size of sram - size of coefficient ram - memory for 1 scan line
	int pauselimit = ramsize - coeffsize - (bytesperline/1024 + 1);
	if (regs[0x50] == 0) // if not reversing, take into account the steps to reverse
		pauselimit -= ((regs[0x54] & 7)*bytesperline+1023)/1024;
//	for some values of data pixel end, like 5399, we are getting
//	over runs, this is an attempt to fix that. i dont think it works
	pauselimit *= 1024;
	int kcount = KCountFromSize(pauselimit);
	if (kcount>2) kcount-=2;//--kcount;

	regs[0x4E] = kcount;
	regs[0x4F] = regs[0x4e] - 1;

    NSCStatusOut("5. pause limit=%d",kcount);
	return;
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ComputeStepSize(int vertres,int tr, int *regs)

// DESCRIPTION	:	compute the step size reg 46:47
//			tr * vertical resolution
//			------------------------
//			4 * motor steps/inch

// INPUTS	:	vertres	- vertical resolution
//			tr	- tr computed from dpd equation

// RETURNS	:	none
//--------------------------------------------------------
void CHardware::ComputeStepSize(int vertres,int *regs)
{
	int tr = ComputeTR(regs); // compute tr w cm
	int stepsize;
	if (merlininfo.timeadj) 
		stepsize = (int)ceil((double)tr * vertres * (merlininfo.timeadj+1)) / 
					(4 * merlininfo.Motor_Steps_Per_Inch * merlininfo.timeadj);
	else 
		stepsize = (int)ceil((double)tr * vertres) / 
					(4 * merlininfo.Motor_Steps_Per_Inch);
	LimitStepSize(stepsize,regs);
	NSCStatusOut("1. step size for mclk calc=%d",stepsize);
	return;
}

//align stepsize with tr
void CHardware::LimitStepSize(int& stepsize,int *regs)
{
	stepsize = max(stepsize,2); 
	stepsize = min(stepsize,65535);
	regs[0x47] = stepsize & 0xFF;
	regs[0x46] = stepsize/256;
//compute new DPD
    ComputeDPD(regs); 
	NSCStatusOut("dpd=%d",(regs[0x51]&0x3)*256*256+regs[0x52]*256+regs[0x53]);
//compute pause limit
	ComputePauseLimit(regs); 
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ComputeDPD(int *regs,
//					BOOL bWantCM /*=FALSE*/)

// DESCRIPTION	:	compute dpd reg 52:53

// INPUTS	:	regs	- merlin registers
//			bWantCm	- TRUE if calculated tr contains cm component
//				- FALSE if calculated tr does not include cm component

// RETURNS	:	tr	- calculated tr
//--------------------------------------------------------
// bWantCM = TRUE by default, if bWantCM=0, then tr is returned with cm removed
int CHardware::ComputeDPD(int *regs,BOOL bWantCM /*=TRUE*/)
{
int linend;	// line end value reg 20:21
int actpst;	// active pixel start reg 1e:1f
int tpspd;	// turbo/preview mode speed reg 0a b2..3
int tpsel;	// turbo/preview mode select reg 0a b0..1
int gbnd;	// guardband duration reg 0e b4..7
int dur;	// pulse duration reg 0e b0..3
int ntr;	// number of tr pulses reg 0d b7
int afeop;	// scan mode, 0=pixel rate, 1=line rate, 
		// 4=1 channel mode a, 5=1 channel mode b, reg 26 b0..2
int ctmode;	// CIS tr timing mode reg 19 b0..1
int qtcnt;	// quarter speed count count reg 51 b2..3
int hfcnt;	// half speed count reg 51 b0..1
int strev;	// steps to reverse reg 50
int st;		// step size reg 46:47
int dpd;	// calculated dpd reg 52:53
int tp;		// tpspd or 1 if tpsel=0
int cm;	// 3 if 3 channel line rate or 1 channel mode b
		// 1 if pixel rate or 1 channel mode a
int afelat;	// 4 if afeop=0, 7 otherwise
int b;	// if ctmode=0, (ntr+1)*((2*gbnd)+dur+1), otherwise 1
int a;	// if vps != 0 tcdp*(linend - actpst - afelat)/vps, otherwise 0
int tr;
//new following

//want tr/(4*stepsize) == integer if possible
//force TR to be multiple of 48 so aligned with
//standard 4*step sizes for /1,/2,/3,/4,/6,/12
int multiple = 48;
while (1)
{
	// compute linend
	int value = regs[0x21]; // line end lsb
	linend = value;
	value = regs[0x20]; // line end msb
	linend += 256 * value;

	// compute active pixel start
	value = regs[0x1F]; // active pixel start lsb
	actpst = value;
	value = regs[0x1E]; // active pixel start msb
	actpst += 256 * value;

	value = regs[0x0A]; // turbo/preview mode speed
	tpspd = (value & 0xC)/4;

	tpsel = value & 3; // turbo/preview mode select

	value = regs[0x0E]; // guardband/duration
	gbnd = (value & 0xF0)/16;

	dur = (value & 0xF);

	value = regs[0x0D]; // number of tr pulses
	ntr = value/128;

	// afeop = 0 if pixel rate, 1 if line rate, 4 if 1 channel mode a,
	//		5 if 1 channel mode b
	value = regs[0x26]; // afe op
	afeop = value & 7;

	value = regs[0x19]; // cis tr timing mode
	int tradj;
	tradj = value & 0x7F;
	value = regs[0xb]/8;

	ctmode = value & 3;

	value = regs[0x51]; // quarter speed count
	value /=4;
	qtcnt = (value & 0xC)/4;
	hfcnt = (value & 0x30)/16; // half speed count
	if (LM9832() || LM9833())
	{
		if (qtcnt == 3) qtcnt = 8;
		if (hfcnt == 3) hfcnt = 8;
	}

	value = regs[0x50]; // steps to reverse
	strev = value & 0x3F;

	value = regs[0x47]; // scan step size lsb
	st = value;
	value = regs[0x46]; // scan step size msb
	st += 256 * (int) value;

	//new
	BOOL ch3_pix = (afeop == 0);
	int en_tradj = 0;
	if (tradj) en_tradj = 1;

	// calculate dpd
	cm = 1;
	if (afeop == 1 || afeop == 5) cm = 3; // if 3 channel line or 1 channel mode b

	afelat = 7;
	if (afeop == 0) afelat = 4;

	if (tpsel == 0) tp = 1;
	else
	{
		tp = tpspd +2;
		if (tp==5) tp++;
	}

	b = 1;
	if (ctmode == 0)
	{
		b = (ntr+1)*((2*gbnd)+dur+1);
		b+= (1-ntr)*en_tradj;
	}
	if (ctmode == 2) b = 3;

	a = linend;

	tr = cm * (a + tp*(b + 3 - ntr));

	{
		//new
		int TRadj = tradj;
		if (tradj == 0)
		{
			if (ctmode == 0) tr += cm;
		}
		else
		{
			int le_phi,num_byteclk,num_mclkf,tr_fast_pix,extra_pix;
			if (!ch3_pix)
			{
				le_phi = (TRadj+1)/2 + 1 + 6;
				num_byteclk = ((le_phi + 8*a + 8*b + 4)/(8*TRadj)) + 1;
				num_mclkf = 8 * TRadj * num_byteclk;
				tr_fast_pix = num_byteclk;
				extra_pix = (num_mclkf - le_phi) % 8;
			}
			else
			{
				le_phi = (TRadj+1)/2 + 1 + 10 + 12;
				num_byteclk = ((le_phi + 3*8*a + 3*8*b + 3*4)/(3*8*TRadj)) + 1;
				num_mclkf = 3*8 * TRadj * num_byteclk;
				tr_fast_pix = num_byteclk;
				extra_pix = (num_mclkf - le_phi) % (3*8);
			}
			tr = b + a + 4 + tr_fast_pix;
			if (extra_pix == 0) tr+=1;
			tr *= cm;
		}
	}
	if ((tr%multiple)==0) 
		break; 
	linend++;
	regs[0x20] = linend/256;
	regs[0x21] = linend&0xff;
}

//dpd
	if (tr == 0) dpd = 0;
	else {
		dpd = (((qtcnt*4) + (hfcnt*2) + strev)*4*st) % tr;
		dpd = tr - dpd;
	}

	if (!bWantCM && cm != 0) tr /= cm; // remove cm component if requested

	regs[0x53] = dpd & 0xFF;
	regs[0x52] = (dpd>>8) & 0xFF;
	regs[0x51] |= ((dpd>>16) & 3);

	return tr;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ComputeTR(int *regs,
//					BOOL bWantCM)

// DESCRIPTION	:	compute tr

// INPUTS	:	regs	- merlin registers
//			bWantCm	- TRUE if calculated tr contains cm component
//				- FALSE if calculated tr does not include cm component

// RETURNS	:	tr
//--------------------------------------------------------
// bWantCM = TRUE by default, if bWantCM=0, then tr is returned with cm removed
int CHardware::ComputeTR(int *regs,BOOL bWantCM)
{
	int tr = ComputeDPD(regs,bWantCM);
	NSCStatusOut ("tr=%d,withcm=%d",tr,bWantCM);
	return tr;
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	WriteMerlin(int *regs)

// DESCRIPTION	:	write merlin registers
//			if a register value > 0xFF, dont write to merlin

//			regs		- array of reg values to write to merlin

// RETURNS	:	true if regs written, false if error
//--------------------------------------------------------
BOOL CHardware::WriteMerlin(int *regs)
{

	{
		//for debug,verify that regs are valid mclk*hdpi*timeadj >=6
		//mclk scaled by 2
		//hdpi by 2
		int hdpi = 2;
		if (regs[0x9]&1) hdpi = 3;
		if (regs[0x9]&2) hdpi*=2;
		if (regs[0x9]&4) hdpi*=4;

		int mclk = regs[0x8] + 2;

		int timeadj = regs[0x19];
		if (!timeadj) timeadj = 1;

		if (hdpi*mclk*timeadj < 2*2*6)
		{
			NSCAfxMessageBox("Invalid clock/hpdi combination");
			return FALSE;
		}

		//set test registers = 0
		regs[0x28] = 0;
		for (int i=0x5c; i < 0x80; i++)
		{
			regs[i] = 0;
		}

		const size=0x80;
		BYTE breg[size+1];
		//need to mod reg 0x18 before reset
 		if (!WriteRegister(0x18,24)) return FALSE;
		delay(2);
 		if (!WriteRegister(0x7,0x20)) return FALSE;
 		breg[0x7] = 0x20;

		int start=3,end=size;
		for (i=3; i<size+1; i++)
		{
			//skip write 0x100 entries
			if ((i < size) && (regs[i] != 0x100)) breg[i] = regs[i];
			else
			{
				end = i;
				if (start < end)
					if (!DriverUSB.BulkOut(start,breg+start,end-start,TRUE))
					{
						NSCAfxMessageBox("BulkOut failed! %02X",start);
						return FALSE;
					}
				start=i+1;
			}
		}
		delay(10);
 		if (!WriteRegister(0x7,0x0)) return FALSE;
	}
	return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ReadMerlin(int *regs)

// DESCRIPTION	:	read merlin registers

//			regs		- array to fill with register values

// RETURNS	:	true if regs read, false if error
//--------------------------------------------------------
BOOL CHardware::ReadMerlin(int *regs)
{
	// init regs to 0
	memset(regs,0,0x80*sizeof(int));

	{
		const size=0x80-0x7;
		BYTE breg[size];
		if (!DriverUSB.BulkIn(7,breg,size,TRUE))
		{
			NSCAfxMessageBox("BulkIn failed! %02X",7);
			return FALSE;
		}
		for (int i=0; i<size; i++) regs[i+7] = breg[i];
	}
	return TRUE;
}

//-------------------------------------------------------
// FUNCTION	:	ReadMerlinINI(filename,scansettings)

// DESCRIPTION	:	read ini and fill in merlininfo structure

// INPUTS	:	filename	- name of ini file

// RETURNS	:	TRUE if ini file found, FALSE otherwise
//--------------------------------------------------------
BOOL CHardware::ReadMerlinINI(char *filename,SCAN_PARAM_REC *pCurrentScanSettings)
{
//new table interface for .ini
//chip supported
typedef enum
{
	CHIPALL=-1,	//ALL
	CHIP31 = 1,
	CHIP32 = 2+4, //Both 32 and 33
} CHIP;

//types of variables
typedef enum
{
	STRING,			//string values (any value if no arg)
	INDEX,			//integer index based on string argument
	INTEGER,		//integer values (any value if no arg)
	INTEGER_RANGE,	//integer range based on two arguments
	FLOATING,		//floating values (any value if no arg)
	FLOATING_RANGE	//floating range based on two arguments
} TYPE;

//one table entry
typedef struct
{	
	CHIP	chip;		//chips
	char*	section;	//ini section
	char*	name;		//ini string
	void*	var;		//pter to variable
	TYPE	type;	//type of variable
#define MAX_ARGS 8
	char*	arg[MAX_ARGS];	//max of 8 separate values
} INIENTRY;

//ini table
INIENTRY initable[]=
{
//File Settings
	//Chip name
	CHIPALL,"File Settings","Chip",
			&merlininfo.chip,INDEX,
			"LM9831","LM9832","LM9833","","","","","",
	//File version
	CHIPALL,"File Settings","File_Version",
			&merlininfo.file_version,INTEGER_RANGE,
			INIFILE_VERSION,"100","","","","","","",
	//USB device name
	CHIPALL,"File Settings","USB_Device",
			&merlininfo.USB_Device,STRING,
			"","","","","","","","",
	//USB manufacturer name
	CHIPALL,"File Settings","USB_Manufacturer",
			&merlininfo.USB_Manufacturer,STRING,
			"","","","","","","","",
//Scanner Software Settings
	//Crystal frequency
	CHIPALL,"Scanner Software Settings","Crystal_Frequency",
			&merlininfo.Crystal_Frequency,INTEGER,
			"48000000","","","","","","","",
	//Red target code for calibrated white pixels
	CHIPALL,"Scanner Software Settings","Red_Target",
			&merlininfo.red_target,INTEGER_RANGE,
			"0x8000","0xffff","","","","","","",
	//Green target code for calibrated white pixels
	CHIPALL,"Scanner Software Settings","Green_Target",
			&merlininfo.green_target,INTEGER_RANGE,
			"0x8000","0xffff","","","","","","",
	//Blue target code for calibrated white pixels
	CHIPALL,"Scanner Software Settings","Blue_Target",
			&merlininfo.blue_target,INTEGER_RANGE,
			"0x8000","0xffff","","","","","","",
	//Red target code for calibrated black pixels
	CHIPALL,"Scanner Software Settings","Min_Red_Target",
			&merlininfo.min_red_target,INTEGER_RANGE,
			"0","0x8000","","","","","","",
	//Green target code for calibrated black pixels
	CHIPALL,"Scanner Software Settings","Min_Green_Target",
			&merlininfo.min_green_target,INTEGER_RANGE,
			"0","0x8000","","","","","","",
	//Blue target code for calibrated black pixels
	CHIPALL,"Scanner Software Settings","Min_Blue_Target",
			&merlininfo.min_blue_target,INTEGER_RANGE,
			"0","0x8000","","","","","","",
	//Black fine calibration enable
	CHIPALL,"Scanner Software Settings","Black_Fine_Calibration",
			&merlininfo.Black_Fine_Calibration,INDEX,
			"No","Yes","","","","","","",
//[Scanner Hardware Settings]
	//Full motor steps per inch
	CHIPALL,"Scanner Hardware Settings","Motor_Steps_Per_Inch",
			&merlininfo.Motor_Steps_Per_Inch,INTEGER_RANGE,
			"10","10000","","","","","","",
	//Max motor speed in inches per second
	CHIPALL,"Scanner Hardware Settings","Max_Speed",
			&merlininfo.Max_Speed,FLOATING_RANGE,
			".001","1000","","","","","","",
	//Full steps to skip at start of calibration
	CHIPALL,"Scanner Hardware Settings","Start_of_Calibration",
			&merlininfo.Start_of_Calibration,INTEGER_RANGE,
			"0","10000","","","","","","",
	//Height of calibration strip in full steps, after skip above
	CHIPALL,"Scanner Hardware Settings","Calibration_Strip_Height",
			&merlininfo.Calibration_Strip_Height,INTEGER_RANGE,
			"1","10000","","","","","","",
	//Full steps to skip at start of scan, from home position (reg 4a:4b, bit 14:0)
	CHIPALL,"Scanner Hardware Settings","Start_Of_Scan",
			&merlininfo.Start_of_Scan,INTEGER_RANGE,
			"0","10000","","","","","","",
	//Display scan as a mirror image
	CHIPALL,"Scanner Hardware Settings","Mirror_Image",
			&merlininfo.Mirror_Image,INDEX,
			"No","Yes","","","","","","",
	//Image width in pixels
	CHIPALL,"Scanner Hardware Settings","Image_Pixel_Width",
			&merlininfo.Image_Pixel_Width,INTEGER_RANGE,
			"0","16383","","","","","","",
	//Scanner Length in Inches
	CHIPALL,"Scanner Hardware Settings","Scanner_Length",
			&merlininfo.Scanner_Length,FLOATING_RANGE,
			".001","1000","","","","","","",
	//Minimum static gain so saturation causes max pixel value
	CHIPALL,"Scanner Hardware Settings","Minimum_Static_Gain",
			&merlininfo.Minimum_Static_Gain,FLOATING_RANGE,
			"1.","10.","","","","","","",
	//Bias current setting
	CHIP32,"Scanner Hardware Settings","Bias_Current",
			&merlininfo.Bias_Current,INDEX,
			"100","80","70","50","","","","",
//Sensor Settings
	//Input polarity (reg 0b, bit 0)
	CHIPALL,"Sensor Settings","Input_Polarity",
			&merlininfo.Input_Polarity,INDEX,
			"Negative","Positive","","","","","","",
	//CDS (reg 0b, bit 1)
	CHIPALL,"Sensor Settings","CDS",
			&merlininfo.CDS,INDEX,
			"Off","On","","","","","","",
	//Sensor type (reg 0b, bit 2)
	CHIPALL,"Sensor Settings","Sensor_Type",
			&merlininfo.Sensor_Type,INDEX,
			"Standard","EvenOdd","","","","","","",
	//Optical resolution in dpi
	CHIPALL,"Sensor Settings","Optical_Resolution",
			&merlininfo.Optical_Resolution,INTEGER_RANGE,
			"10","10000","","","","","","",
	//Max base resolution in dpi
	CHIPALL,"Sensor Settings","Max_Sensor_Resolution",
			&merlininfo.Max_Sensor_Resolution,INTEGER_RANGE,
			"10","10000","","","","","","",
	//Line separation between red and green
	CHIPALL,"Sensor Settings","Line_Separation",
			&merlininfo.Line_Separation,FLOATING_RANGE,
			"-1000","1000","","","","","","",
	//Line separation between even and odd
	CHIPALL,"Sensor Settings","EvenOdd_Line_Separation",
			&merlininfo.EvenOdd_Line_Separation,FLOATING_RANGE,
			"-1000","1000","","","","","","",
	//Total pixels in line (reg 20:21, bit 13:0)
	CHIPALL,"Sensor Settings","Total_Pixels",
			&merlininfo.Total_Pixels,INTEGER_RANGE,
			"10","16383","","","","","","",
	//Max integration time in milliseconds (lamp on full)
	CHIPALL,"Sensor Settings","Maximum_Integration_Time_Highlamp",
			&merlininfo.Maximum_Integration_Time_Highlamp,FLOATING_RANGE,
			".001","1000","","","","","","",
	//Max integration time in milliseconds (lamp on lowest setting)
	CHIPALL,"Sensor Settings","Maximum_Integration_Time_Lowlamp",
			&merlininfo.Maximum_Integration_Time_Lowlamp,FLOATING_RANGE,
			".001","1000","","","","","","",
	//Min integration time in milliseconds, horizontal div <=2
	CHIPALL,"Sensor Settings","Minimum_Integration_Time_HighRes",
			&merlininfo.Minimum_Integration_Time_HighRes,FLOATING_RANGE,
			".001","1000","","","","","","",
	//Min integration time in milliseconds, horizontal div > 2
	CHIPALL,"Sensor Settings","Minimum_Integration_Time_LowRes",
			&merlininfo.Minimum_Integration_Time_LowRes,FLOATING_RANGE,
			".001","1000","","","","","","",
	//Max integration time adjust before switch to line rate
	CHIPALL,"Sensor Settings","Maximum_IntTimeAdj",
			&merlininfo.Maximum_IntTimeAdj,INTEGER_RANGE,
			"0","127","","","","","","",
	//Phi1 Polarity (reg 0c, bit 0)
	CHIPALL,"Sensor Settings","Phi1_Polarity",
			&merlininfo.Phi1_Polarity,INDEX,
			"Positive","Negative","","","","","","",
	//Phi2 Polarity (reg 0c, bit 1)
	CHIPALL,"Sensor Settings","Phi2_Polarity",
			&merlininfo.Phi2_Polarity,INDEX,
			"Positive","Negative","","","","","","",
	//RS Polarity (reg 0c, bit 2)
	CHIPALL,"Sensor Settings","RS_Polarity",
			&merlininfo.RS_Polarity,INDEX,
			"Positive","Negative","","","","","","",
	//CP1 Polarity (reg 0c, bit 3)
	CHIPALL,"Sensor Settings","CP1_Polarity",
			&merlininfo.CP1_Polarity,INDEX,
			"Positive","Negative","","","","","","",
	//CP2 Polarity (reg 0c, bit 4)
	CHIPALL,"Sensor Settings","CP2_Polarity",
			&merlininfo.CP2_Polarity,INDEX,
			"Positive","Negative","","","","","","",
	//TR1 Polarity (reg 0c, bit 5)
	CHIPALL,"Sensor Settings","TR1_Polarity",
			&merlininfo.TR1_Polarity,INDEX,
			"Positive","Negative","","","","","","",
	//TR2 Polarity (reg 0c, bit 6)
	CHIPALL,"Sensor Settings","TR2_Polarity",
			&merlininfo.TR2_Polarity,INDEX,
			"Positive","Negative","","","","","","",
	//Phi1 status (reg 0d, bit 0)
	CHIPALL,"Sensor Settings","Phi1_Status",
			&merlininfo.Phi1_Status,INDEX,
			"Disabled","Active","","","","","","",
	//Phi2 status (reg 0d, bit 1)
	CHIPALL,"Sensor Settings","Phi2_Status",
			&merlininfo.Phi2_Status,INDEX,
			"Disabled","Active","","","","","","",
	//RS status (reg 0d, bit 2)
	CHIPALL,"Sensor Settings","RS_Status",
			&merlininfo.RS_Status,INDEX,
			"Disabled","Active","","","","","","",
	//CP1 status (reg 0d, bit 3)
	CHIPALL,"Sensor Settings","CP1_Status",
			&merlininfo.CP1_Status,INDEX,
			"Disabled","Active","","","","","","",
	//CP2 status (reg 0d, bit 4)
	CHIPALL,"Sensor Settings","CP2_Status",
			&merlininfo.CP2_Status,INDEX,
			"Disabled","Active","","","","","","",
	//TR1 status (reg 0d, bit 5)
	CHIPALL,"Sensor Settings","TR1_Status",
			&merlininfo.TR1_Status,INDEX,
			"Disabled","Active","","","","","","",
	//TR2 status (reg 0d, bit 6)
	CHIPALL,"Sensor Settings","TR2_Status",
			&merlininfo.TR2_Status,INDEX,
			"Disabled","Active","","","","","","",
	//TR_Pulse (reg 0d, bit 7)
	CHIPALL,"Sensor Settings","TR_Pulse",
			&merlininfo.TR_Pulse,INDEX,
			"1","2","","","","","","",
	//TR_Pulse_Duration (reg 0e, bit 3:0)
	CHIPALL,"Sensor Settings","TR_Pulse_Duration",
			&merlininfo.TR_Pulse_Duration,INTEGER_RANGE,
			"0","15","","","","","","",
	//TR_PHI_Guardband_Duration (reg 0e, bit 7:4)
	CHIPALL,"Sensor Settings","TR_PHI_Guardband_Duration",
			&merlininfo.TR_PHI_Guardband_Duration,INTEGER_RANGE,
			"0","15","","","","","","",
//Pixel Rate Timing
	//Pixl_Optical_Black_Clamp_Start (reg 0f, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_Optical_Black_Clamp_Start",
			&merlininfo.Pixl_Optical_Black_Clamp_Start,INTEGER_RANGE,
			"0","23","","","","","","",
	//Pixl_Optical_Black_Clamp_End (reg 10, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_Optical_Black_Clamp_End",
			&merlininfo.Pixl_Optical_Black_Clamp_End,INTEGER_RANGE,
			"0","23","","","","","","",
	//Pixl_Reset_Pulse_Start (reg 11, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_Reset_Pulse_Start",
			&merlininfo.Pixl_Reset_Pulse_Start,INTEGER_RANGE,
			"0","23","","","","","","",
	//Pixl_Reset_Pulse_End (reg 12, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_Reset_Pulse_End",
			&merlininfo.Pixl_Reset_Pulse_End,INTEGER_RANGE,
			"0","23","","","","","","",
	//Pixl_CP1_Pulse_Start (reg 13, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_CP1_Pulse_Start",
			&merlininfo.Pixl_CP1_Pulse_Start,INTEGER_RANGE,
			"0","23","","","","","","",
	//Pixl_CP1_Pulse_End (reg 14, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_CP1_Pulse_End",
			&merlininfo.Pixl_CP1_Pulse_End,INTEGER_RANGE,
			"0","23","","","","","","",
	//Pixl_CP2_Pulse_Start (reg 15, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_CP2_Pulse_Start",
			&merlininfo.Pixl_CP2_Pulse_Start,INTEGER_RANGE,
			"0","23","","","","","","",
	//Pixl_CP2_Pulse_End (reg 16, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_CP2_Pulse_End",
			&merlininfo.Pixl_CP2_Pulse_End,INTEGER_RANGE,
			"0","23","","","","","","",
	//Pixl_Reference_Sample (reg 17, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_Reference_Sample",
			&merlininfo.Pixl_Reference_Sample,INTEGER_RANGE,
			"0","23","","","","","","",
	//Pixl_Signal_Sample (reg 18, bit 4:0)
	CHIPALL,"Pixel Rate Timing","Pixl_Signal_Sample",
			&merlininfo.Pixl_Signal_Sample,INTEGER_RANGE,
			"0","23","","","","","","",
//Line Rate Timing
	//Line_Optical_Black_Clamp_Start (reg 0f, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_Optical_Black_Clamp_Start",
			&merlininfo.Line_Optical_Black_Clamp_Start,INTEGER_RANGE,
			"0","7","","","","","","",
	//Line_Optical_Black_Clamp_End (reg 10, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_Optical_Black_Clamp_End",
			&merlininfo.Line_Optical_Black_Clamp_End,INTEGER_RANGE,
			"0","7","","","","","","",
	//Line_Reset_Pulse_Start (reg 11, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_Reset_Pulse_Start",
			&merlininfo.Line_Reset_Pulse_Start,INTEGER_RANGE,
			"0","7","","","","","","",
	//Line_Reset_Pulse_End (reg 12, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_Reset_Pulse_End",
			&merlininfo.Line_Reset_Pulse_End,INTEGER_RANGE,
			"0","7","","","","","","",
	//Line_CP1_Pulse_Start (reg 13, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_CP1_Pulse_Start",
			&merlininfo.Line_CP1_Pulse_Start,INTEGER_RANGE,
			"0","7","","","","","","",
	//Line_CP1_Pulse_End (reg 14, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_CP1_Pulse_End",
			&merlininfo.Line_CP1_Pulse_End,INTEGER_RANGE,
			"0","7","","","","","","",
	//Line_CP2_Pulse_Start (reg 15, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_CP2_Pulse_Start",
			&merlininfo.Line_CP2_Pulse_Start,INTEGER_RANGE,
			"0","7","","","","","","",
	//Line_CP2_Pulse_End (reg 16, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_CP2_Pulse_End",
			&merlininfo.Line_CP2_Pulse_End,INTEGER_RANGE,
			"0","7","","","","","","",
	//Line_Reference_Sample (reg 17, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_Reference_Sample",
			&merlininfo.Line_Reference_Sample,INTEGER_RANGE,
			"0","7","","","","","","",
	//Line_Signal_Sample (reg 18, bit 4:0)
	CHIPALL,"Line Rate Timing","Line_Signal_Sample",
			&merlininfo.Line_Signal_Sample,INTEGER_RANGE,
			"0","7","","","","","","",
//Misc Timing
	//CIS_TR_Timing (reg 0b, bit 4:3)
	CHIPALL,"Misc Timing","CIS_TR_Timing",
			&merlininfo.CIS_TR_Timing,INDEX,
			"Off","Mode1","Mode2","","","","","",
	//Fake_Optical_Black_Pixels (reg 0b, bit 5)
	CHIPALL,"Misc Timing","Fake_Optical_Black_Pixels",
			&merlininfo.Fake_Optical_Black_Pixels,INDEX,
			"Off","On","","","","","","",
	//TR_Stepper_Correction (reg 1a:1b, bit 13:0) 
	CHIPALL,"Misc Timing","TR_Stepper_Correction",
			&merlininfo.TR_Stepper_Correction,INTEGER_RANGE,
			"1","16383","","","","","","",
	//Optical_Black_Pixels_Start (reg 1c, bit 7:0)
	CHIPALL,"Misc Timing","Optical_Black_Pixels_Start",
			&merlininfo.Optical_Black_Pixels_Start,INTEGER_RANGE,
			"0","255","","","","","","",
	//Optical_Black_Pixels_End (reg 1d, bit 7:0)
	CHIPALL,"Misc Timing","Optical_Black_Pixels_End",
			&merlininfo.Optical_Black_Pixels_End,INTEGER_RANGE,
			"0","255","","","","","","",
	//Color_Mode (reg 26, bit 2:0) map 0-3 to 0,1,4,5
	CHIPALL,"Misc Timing","Color_Mode",
			&merlininfo.Color_Mode,INTEGER_RANGE,
			"0","3","","","","","","",
	//Gray_Channel (reg 26, bit 4:3)
	CHIPALL,"Misc Timing","Gray_Channel",
			&merlininfo.Gray_Channel,INDEX,
			"Red","Green","Blue","","","","","",
//3 Channel Line Rate Settings
	//TR_Red (reg 26, bit 5)
	CHIPALL,"3 Channel Line Rate Settings","TR_Red",
			&merlininfo.TR_Red,INTEGER_RANGE,
			"0","1","","","","","","",
	//TR_Green (reg 26, bit 6)
	CHIPALL,"3 Channel Line Rate Settings","TR_Green",
			&merlininfo.TR_Green,INTEGER_RANGE,
			"0","1","","","","","","",
	//TR_Blue (reg 26, bit 7)
	CHIPALL,"3 Channel Line Rate Settings","TR_Blue",
			&merlininfo.TR_Blue,INTEGER_RANGE,
			"0","1","","","","","","",
	//TR_Red_Drop (reg 27, bit 1:0)
	CHIPALL,"3 Channel Line Rate Settings","TR_Red_Drop",
			&merlininfo.TR_Red_Drop,INTEGER_RANGE,
			"0","2","","","","","","",
	//TR_Green_Drop (reg 27, bit 3:2)
	CHIPALL,"3 Channel Line Rate Settings","TR_Green_Drop",
			&merlininfo.TR_Green_Drop,INTEGER_RANGE,
			"0","2","","","","","","",
	//TR_Blue_Drop (reg 27, bit 5:4)
	CHIPALL,"3 Channel Line Rate Settings","TR_Blue_Drop",
			&merlininfo.TR_Blue_Drop,INTEGER_RANGE,
			"0","2","","","","","","",
	//Triple_TR (reg 27, bit 6)
	CHIPALL,"3 Channel Line Rate Settings","Triple_TR",
			&merlininfo.Triple_TR,INDEX,
			"Off","On","","","","","","",
//Illumination Settings
	//Illumination_Mode (reg 29, bit 1:0)
	CHIPALL,"Illumination Settings","Illumination_Mode",
			&merlininfo.Illumination_Mode,INDEX,
			"Off","Mode1","Mode2","Mode3","","","","",
	//LampB_Output_ITA (reg 29, bit 2)
	CHIP32,"Illumination Settings","LampB_Output_ITA",
			&merlininfo.LampB_Output_ITA,INDEX,
			"Off","On","","","","","","",
	//Green_PWM_Duty_Cycle_High (reg 2a:2b, bit 11:0)
	CHIPALL,"Illumination Settings","Green_PWM_Duty_Cycle_High",
			&merlininfo.Green_PWM_Duty_Cycle_High,INTEGER_RANGE,
			"0","4095","","","","","","",
	//Green_PWM_Duty_Cycle_Low (reg 2a:2b, bit 11:0)
	CHIPALL,"Illumination Settings","Green_PWM_Duty_Cycle_Low",
			&merlininfo.Green_PWM_Duty_Cycle_Low,INTEGER_RANGE,
			"0","4095","","","","","","",
	//Red_Lamp_On (reg 2c:2d, bit 13:0)
	CHIPALL,"Illumination Settings","Red_Lamp_On",
			&merlininfo.Red_Lamp_On,INTEGER_RANGE,
			"0","16383","","","","","","",
	//Red_Lamp_Off (reg 2e:2f, bit 13:0)
	CHIPALL,"Illumination Settings","Red_Lamp_Off",
			&merlininfo.Red_Lamp_Off,INTEGER_RANGE,
			"0","16383","","","","","","",
	//Green_Lamp_On (reg 30:31, bit 13:0)
	CHIPALL,"Illumination Settings","Green_Lamp_On",
			&merlininfo.Green_Lamp_On,INTEGER_RANGE,
			"0","16383","","","","","","",
	//Green_Lamp_Off (reg 32:33, bit 13:0)
	CHIPALL,"Illumination Settings","Green_Lamp_Off",
			&merlininfo.Green_Lamp_Off,INTEGER_RANGE,
			"0","16383","","","","","","",
	//Blue_Lamp_On (reg 34:35, bit 13:0)
	CHIPALL,"Illumination Settings","Blue_Lamp_On",
			&merlininfo.Blue_Lamp_On,INTEGER_RANGE,
			"0","16383","","","","","","",
	//Blue_Lamp_Off (reg 36:37, bit 13:0)
	CHIPALL,"Illumination Settings","Blue_Lamp_Off",
			&merlininfo.Blue_Lamp_Off,INTEGER_RANGE,
			"0","16383","","","","","","",
//I/O Settings
	//RAM_WordSize (reg 42, bit 6)
	CHIPALL,"I/O Settings","RAM_WordSize",
			&merlininfo.RAM_WordSize,INTEGER,
			"256","1024","","","","","","",
//[Stepper Motor Settings]	
	//Stepping (reg 45, bit 1:0)
	CHIPALL,"Stepper Motor Settings","Stepping",
			&merlininfo.Stepping,INDEX,
			"Full","Micro","","","","","","",
	//Phase_A_Polarity (reg 45, bit 2)
	CHIPALL,"Stepper Motor Settings","Phase_A_Polarity",
			&merlininfo.Phase_A_Polarity,INDEX,
			"Positive","Negative","","","","","","",
	//Phase_B_Polarity (reg 45, bit 3)
	CHIPALL,"Stepper Motor Settings","Phase_B_Polarity",
			&merlininfo.Phase_B_Polarity,INDEX,
			"Positive","Negative","","","","","","",
	//Reverse_Motor_Direction (reg 45, bit 5)
	CHIP32,"Stepper Motor Settings","Reverse_Motor_Direction",
			&merlininfo.Reverse_Motor_Direction,INDEX,
			"Off","On","","","","","","",
	//Fullstep_During_Fastfeed (reg 45, bit 6)
	CHIP32,"Stepper Motor Settings","Fullstep_During_Fastfeed",
			&merlininfo.Fullstep_During_Fastfeed,INDEX,
			"Off","On","","","","","","",
	//Steps_to_Scan_After_PS2 (reg 4c:4d, bit 13:0)
	CHIPALL,"Stepper Motor Settings","Steps_to_Scan_After_PS2",
			&merlininfo.Steps_to_Scan_After_PS2,INTEGER_RANGE,
			"0","16383","","","","","","",
	//Steps_To_Reverse (reg 50, bit 5:0)
	CHIP31,"Stepper Motor Settings","Steps_To_Reverse",
			&merlininfo.Steps_To_Reverse,INTEGER_RANGE,
			"0","63","","","","","","",
	//Steps_To_Reverse (reg 50, bit 7:0)
	CHIP32,"Stepper Motor Settings","Steps_To_Reverse",
			&merlininfo.Steps_To_Reverse,INTEGER_RANGE,
			"0","255","","","","","","",
	//Steps_Stopped (reg 51, bit 3:2)
	CHIP31,"Stepper Motor Settings","Steps_Stopped",
			&merlininfo.Steps_Stopped,INDEX,
			"0","1","2","3","","","","",
	//Steps_25 (reg 51, bit 5:4)
	CHIP31,"Stepper Motor Settings","Steps_25",
			&merlininfo.Steps_25,INDEX,
			"0","1","2","3","","","","",
	//Steps_50 (reg 51, bit 7:6)
	CHIP31,"Stepper Motor Settings","Steps_50",
			&merlininfo.Steps_50,INDEX,
			"0","1","2","3","","","","",
	//Steps_Stopped (reg 51, bit 3:2)
	CHIP32,"Stepper Motor Settings","Steps_Stopped",
			&merlininfo.Steps_Stopped,INDEX,
			"0","1","2","8","","","","",
	//Steps_25 (reg 51, bit 5:4)
	CHIP32,"Stepper Motor Settings","Steps_25",
			&merlininfo.Steps_25,INDEX,
			"0","1","2","8","","","","",
	//Steps_50 (reg 51, bit 7:6)
	CHIP32,"Stepper Motor Settings","Steps_50",
			&merlininfo.Steps_50,INDEX,
			"0","1","2","8","","","","",
	//Non_Reversing_Extra_Lines (reg 54, bit 2:0)
	CHIPALL,"Stepper Motor Settings","Non_Reversing_Extra_Lines",
			&merlininfo.Non_Reversing_Extra_Lines,INTEGER_RANGE,
			"0","7","","","","","","",
	//KickStart_Steps (reg 55, bit 2:0)
	CHIPALL,"Stepper Motor Settings","KickStart_Steps",
			&merlininfo.KickStart_Steps,INTEGER_RANGE,
			"0","7","","","","","","",
	//Hold_Current_Timeout (reg 55, bit 7:3)
	CHIPALL,"Stepper Motor Settings","Hold_Current_Timeout",
			&merlininfo.Hold_Current_Timeout,INTEGER_RANGE,
			"0","31","","","","","","",
	//Stepper_PWM_Frequency (reg 56, bit 7:0)
	CHIPALL,"Stepper Motor Settings","Stepper_PWM_Frequency",
			&merlininfo.Stepper_PWM_Frequency,INTEGER_RANGE,
			"0","255","","","","","","",
	//Stepper_PWM_Duty_Cycle (reg 57, bit 5:0)
	CHIPALL,"Stepper Motor Settings","Stepper_PWM_Duty_Cycle",
			&merlininfo.Stepper_PWM_Duty_Cycle,INTEGER_RANGE,
			"0","63","","","","","","",
//Paper Sensor
	//Home_Sensor
	CHIPALL,"Paper Sensor","Home_Sensor",
			&merlininfo.Home_Sensor,INTEGER_RANGE,
			"1","2","","","","","","",
	//Paper_Sensor_1_Polarity (reg 58, bit 0)
	CHIPALL,"Paper Sensor","Paper_Sensor_1_Polarity",
			&merlininfo.Paper_Sensor_1_Polarity,INDEX,
			"Low","High","","","","","","",
	//Paper_Sensor_1_Trigger (reg 58, bit 1)
	CHIPALL,"Paper Sensor","Paper_Sensor_1_Trigger",
			&merlininfo.Paper_Sensor_1_Trigger,INDEX,
			"Level","Edge","","","","","","",
	//Paper_Sensor_1_Stop_Scan (reg 58, bit 2)
	CHIPALL,"Paper Sensor","Paper_Sensor_1_Stop_Scan",
			&merlininfo.Paper_Sensor_1_Stop_Scan,INDEX,
			"No","Yes","","","","","","",
	//Paper_Sensor_2_Polarity (reg 58, bit 3)
	CHIPALL,"Paper Sensor","Paper_Sensor_2_Polarity",
			&merlininfo.Paper_Sensor_2_Polarity,INDEX,
			"Low","High","","","","","","",
	//Paper_Sensor_2_Trigger (reg 58, bit 4)
	CHIPALL,"Paper Sensor","Paper_Sensor_2_Trigger",
			&merlininfo.Paper_Sensor_2_Trigger,INDEX,
			"Level","Edge","","","","","","",
	//Paper_Sensor_2_Stop_Scan (reg 58, bit 5)
	CHIPALL,"Paper Sensor","Paper_Sensor_2_Stop_Scan",
			&merlininfo.Paper_Sensor_2_Stop_Scan,INDEX,
			"No","Yes","","","","","","",
//Misc I/O	
	//MiscIO_1_Type (reg 59, bit 0)
	CHIPALL,"Misc I/O","MiscIO_1_Type",
			&merlininfo.MiscIO_1_Type,INDEX,
			"Input","Output","","","","","","",
	//MiscIO_1_Polarity (reg 59, bit 1)
	CHIPALL,"Misc I/O","MiscIO_1_Polarity",
			&merlininfo.MiscIO_1_Polarity,INDEX,
			"Low","High","","","","","","",
	//MiscIO_1_Trigger (reg 59, bit 2)
	CHIPALL,"Misc I/O","MiscIO_1_Trigger",
			&merlininfo.MiscIO_1_Trigger,INDEX,
			"Level","Edge","","","","","","",
	//MiscIO_1_Output_State (reg 59, bit 3)
	CHIPALL,"Misc I/O","MiscIO_1_Output_State",
			&merlininfo.MiscIO_1_Output_State,INDEX,
			"Low","High","","","","","","",
	//MiscIO_2_Type (reg 59, bit 4)
	CHIPALL,"Misc I/O","MiscIO_2_Type",
			&merlininfo.MiscIO_2_Type,INDEX,
			"Input","Output","","","","","","",
	//MiscIO_2_Polarity (reg 59, bit 5)
	CHIPALL,"Misc I/O","MiscIO_2_Polarity",
			&merlininfo.MiscIO_2_Polarity,INDEX,
			"Low","High","","","","","","",
	//MiscIO_2_Trigger (reg 59, bit 6)
	CHIPALL,"Misc I/O","MiscIO_2_Trigger",
			&merlininfo.MiscIO_2_Trigger,INDEX,
			"Level","Edge","","","","","","",
	//MiscIO_2_Output_State (reg 59, bit 7)
	CHIPALL,"Misc I/O","MiscIO_2_Output_State",
			&merlininfo.MiscIO_2_Output_State,INDEX,
			"Low","High","","","","","","",
	//MiscIO_3_Type (reg 5a, bit 0)
	CHIPALL,"Misc I/O","MiscIO_3_Type",
			&merlininfo.MiscIO_3_Type,INDEX,
			"Input","Output","","","","","","",
	//MiscIO_3_Polarity (reg 5a, bit 1)
	CHIPALL,"Misc I/O","MiscIO_3_Polarity",
			&merlininfo.MiscIO_3_Polarity,INDEX,
			"Low","High","","","","","","",
	//MiscIO_3_Trigger (reg 5a, bit 2)
	CHIPALL,"Misc I/O","MiscIO_3_Trigger",
			&merlininfo.MiscIO_3_Trigger,INDEX,
			"Level","Edge","","","","","","",
	//MiscIO_3_Output_State (reg 5a, bit 3)
	CHIPALL,"Misc I/O","MiscIO_3_Output_State",
			&merlininfo.MiscIO_3_Output_State,INDEX,
			"Low","High","","","","","","",
	//MiscIO_4_Type (reg 5a, bit 4)
	CHIPALL,"Misc I/O","MiscIO_4_Type",
			&merlininfo.MiscIO_4_Type,INDEX,
			"Input","Output","","","","","","",
	//MiscIO_4_Polarity (reg 5a, bit 5)
	CHIPALL,"Misc I/O","MiscIO_4_Polarity",
			&merlininfo.MiscIO_4_Polarity,INDEX,
			"Low","High","","","","","","",
	//MiscIO_4_Trigger (reg 5a, bit 6)
	CHIPALL,"Misc I/O","MiscIO_4_Trigger",
			&merlininfo.MiscIO_4_Trigger,INDEX,
			"Level","Edge","","","","","","",
	//MiscIO_4_Output_State (reg 5a, bit 7)
	CHIPALL,"Misc I/O","MiscIO_4_Output_State",
			&merlininfo.MiscIO_4_Output_State,INDEX,
			"Low","High","","","","","","",
	//MiscIO_5_Type (reg 5b, bit 0)
	CHIPALL,"Misc I/O","MiscIO_5_Type",
			&merlininfo.MiscIO_5_Type,INDEX,
			"Input","Output","","","","","","",
	//MiscIO_5_Polarity (reg 5b, bit 1)
	CHIPALL,"Misc I/O","MiscIO_5_Polarity",
			&merlininfo.MiscIO_5_Polarity,INDEX,
			"Low","High","","","","","","",
	//MiscIO_5_Trigger (reg 5b, bit 2)
	CHIPALL,"Misc I/O","MiscIO_5_Trigger",
			&merlininfo.MiscIO_5_Trigger,INDEX,
			"Level","Edge","","","","","","",
	//MiscIO_5_Output_State (reg 5b, bit 3)
	CHIPALL,"Misc I/O","MiscIO_5_Output_State",
			&merlininfo.MiscIO_5_Output_State,INDEX,
			"Low","High","","","","","","",
	//MiscIO_6_Type (reg 5b, bit 4)
	CHIPALL,"Misc I/O","MiscIO_6_Type",
			&merlininfo.MiscIO_6_Type,INDEX,
			"Input","Output","","","","","","",
	//MiscIO_6_Polarity (reg 5b, bit 5)
	CHIPALL,"Misc I/O","MiscIO_6_Polarity",
			&merlininfo.MiscIO_6_Polarity,INDEX,
			"Low","High","","","","","","",
	//MiscIO_6_Trigger (reg 5b, bit 6)
	CHIPALL,"Misc I/O","MiscIO_6_Trigger",
			&merlininfo.MiscIO_6_Trigger,INDEX,
			"Level","Edge","","","","","","",
	//MiscIO_6_Output_State (reg 5b, bit 7)
	CHIPALL,"Misc I/O","MiscIO_6_Output_State",
			&merlininfo.MiscIO_6_Output_State,INDEX,
			"Low","High","","","","","","",

/* add register addr/bit field later*/
};

	//search for any valid additional sections of the form [hdpi x:y] 
	//where horizontal resolution is in the range x:y
	FILE *fp = fopen(filename,"r");
	if (fp == NULL)
	{
			NSCAfxMessageBox("%s not found",filename);
			return FALSE;
	}
	char line[LINESIZ] = "";
	char specialsection[LINESIZ] = "";
	while (fgets(line,sizeof(line),fp) != NULL)
	{
		char keyword[LINESIZ];
		int low,high;
		int ct = sscanf(line," [ %s %d:%d ]",&keyword,&low,&high);
		if (ct <= 0) continue;
		if (stricmp(keyword,"hdpi")) continue;
		if ((ct < 3) || (low > high))
		{
			NSCAfxMessageBox("section syntax must be [hdpi low:high]");
			return FALSE;
		}
		if (pCurrentScanSettings->hRes < low) continue;
		if (pCurrentScanSettings->hRes > high) continue;
		//save line without []
		int start = 0,end;
		while (line[start++] != '[');
		end = start;
		while (line[++end] != ']');
		line[end] = '\0';
		strcpy(specialsection,line+start);
		break;
	}
	if (fclose(fp))
	{
			NSCAfxMessageBox("%s cannot be closed",filename);
			return FALSE;
	}

	//loop for all table
	memset(&merlininfo,0,sizeof(merlininfo));
//	merlininfo.chip = 0;
	for (int index = 0; index < sizeof(initable)/sizeof(INIENTRY); index++)
	{
		INIENTRY *tptr = &initable[index];
		//see if this chip
		if (!(tptr->chip & (1<<merlininfo.chip))) continue;
		//get argument
		char LineBuf[LINESIZ];	// file input buffer

		//check special section first
		if (!GetPrivateProfileString(specialsection,tptr->name,"",
			LineBuf,sizeof(LineBuf),filename))
		{
			if (!GetPrivateProfileString(tptr->section,tptr->name,"",
				LineBuf,sizeof(LineBuf),filename))
			{
				NSCAfxMessageBox("%s= missing",tptr->name);
				return FALSE;
			}
		}
		//remove comments starting with //
		char *ptr = LineBuf;
		const char ch = '/';
		while (ptr = strchr(ptr,ch))
		{
			ptr++;
			if (*ptr == ch)
			{
				ptr--;
				*ptr = '\0'; //end of string
				break;
			}
		}
		//remove trailing white space
		for (ptr = LineBuf+strlen(LineBuf)-1; ptr >= LineBuf; ptr--)
		{
			if (*ptr != ' ') 
				if (*ptr != '\t') break;
		}
		ptr++;
		*ptr = '\0';
		//get first argument
		char argument[LINESIZ];
		if (1 != sscanf(LineBuf,"%s",argument))
		{
			NSCAfxMessageBox("%s= missing value",tptr->name);
			return FALSE;
		}
		int integer;
		double floating;
		switch(tptr->type)
		{
		case INTEGER:
		case INTEGER_RANGE:
			if (!GetIntegerArg(argument,integer))
			{
				NSCAfxMessageBox("%s= invalid decimal value",tptr->name);
				return FALSE;
			}
			break;
		case FLOATING:
		case FLOATING_RANGE:
			if (!GetFloatingArg(argument,floating))
			{
				NSCAfxMessageBox("%s= invalid float value",tptr->name);
				return FALSE;
			}
			break;
		}
		//get value and compare to argument
		//if no arguments, assume any value ok (except for INDEX)
		int argct;
		switch (tptr->type)
		{
		char* string;
		case STRING:
		case INDEX:
		case INTEGER:
		case FLOATING:
			for (argct = 0; argct < MAX_ARGS; argct++)
			{
				string = tptr->arg[argct];
				if (!strcmp(string,"")) break; //done
				if (tptr->type == STRING)
				{
					if (!stricmp(string,LineBuf))
						break; //match whole line
				}
				else if (tptr->type == INDEX)
				{
					if (!stricmp(string,argument))
						break; //match first string only
				}
				else if (tptr->type == INTEGER)
				{
					int i;
					if (GetIntegerArg(string,i)) 
						if (i == integer)
							break; //match integer value
				}
				else if (tptr->type == FLOATING)
				{
					double f;
					if (GetFloatingArg(string,f))
						if (f == floating)
							break; //match float value
				}
			}
			//error if no match (except if no argument at all
			//except for INDEX
			if ((argct == MAX_ARGS) || (!strcmp(string,"")))
			{
				if ((tptr->type == INDEX) || argct)
				{
					char validargs[LINESIZ] = "";
					for (argct = 0; argct < MAX_ARGS; argct++)
					{
						if (!strcmp(tptr->arg[argct],"")) break;
						strcat(validargs,tptr->arg[argct]);
						strcat(validargs," ");
					}
					NSCAfxMessageBox("%s must be: %s",
						tptr->name,validargs);
					return FALSE;
				}
			}
			break;
		case INTEGER_RANGE:
			{
				int low,high;
				if (!GetIntegerArg(tptr->arg[0],low) ||
					!GetIntegerArg(tptr->arg[1],high) ||
					(integer < low) || (integer > high)) 
				{
					NSCAfxMessageBox("%s must be >= %s AND <= %s",
						tptr->name,tptr->arg[0],tptr->arg[1]);
					return FALSE;
				}
			}
			break;
		case FLOATING_RANGE:
			{
				double low,high;
				if (!GetFloatingArg(tptr->arg[0],low) ||
					!GetFloatingArg(tptr->arg[1],high) ||
					(floating < low) || (floating > high)) 
				{
					NSCAfxMessageBox("%s must be >= %s AND <= %s",
						tptr->name,tptr->arg[0],tptr->arg[1]);
					return FALSE;
				}
			}
			break;
		}
		switch (tptr->type)
		{
		case STRING:
			strcpy((char*)(tptr->var),LineBuf); //string
			break;
		case INDEX:
			*((int*)tptr->var) = argct; //index
			break;
		case INTEGER:
		case INTEGER_RANGE:
			*((int*)tptr->var) = integer; //index
			break;
		case FLOATING:
		case FLOATING_RANGE:
			*((double*)tptr->var) = floating; //index
			break;
		}
	}
	merlininfo.Green_PWM_Duty_Cycle = merlininfo.Green_PWM_Duty_Cycle_High;
	merlininfo.first_white_pixel = merlininfo.Optical_Black_Pixels_Start;
	return TRUE;
//end new code
}

//get hex or decimal integer value
//return false if missing or extra characters in arg
BOOL CHardware::GetIntegerArg(char* arg,int& integer)
{
	char excess[LINESIZ];
	if (!strnicmp("0x",arg,2)) //hex
		return (1 == sscanf(arg,"%x%s",&integer,excess)); 
	else //decimal
		return (1 == sscanf(arg,"%d%s",&integer,excess)); 
}

//get floating value
//return false if missing or extra characters in arg
BOOL CHardware::GetFloatingArg(char* arg,double& floating)
{
	char excess[LINESIZ];
	return (1 == sscanf(arg,"%lf%s",&floating,excess));
}

//-------------------------------------------------------
// FUNCTION	:	BOOL OnStartupSendGamma( )

// DESCRIPTION	:	send gamma on ui startup
//			use gamma file specified in .ini
//			if not specified or not found, try default.gam
//			if not found, create default.gam and send

// INPUTS	:	none

// RETURNS	:	TRUE  if send gamma ok,
//			FALSE if not
//--------------------------------------------------------
BOOL CHardware::OnStartupSendGamma(int bpp)
{
char filename[MAX_PATH];	// name of gamma file to send
FILE *fptr;		// gamma file handle
int iRetVal;		// return value

	if (bpp > 8) return TRUE; //not for 16 bit

//override gamma file name here based on pixel bit count
	char string[MAX_PATH];
	sprintf(string,"%i",bpp);
	strcpy(m_gammafilename,"lm983x_");
	strcat(m_gammafilename,string);
	strcat(m_gammafilename,".gam");
	strcpy(m_gammafilename,GetFullFilePath(m_gammafilename));
	strcpy(filename,m_gammafilename);
	if (strlen(filename) == 0)
		strcpy(filename,"default.gam");
	NSCStatusOut(filename);

	// open or create gamma table
	if(!(fptr = fopen((LPCSTR) filename,"r")))
	{
		// if INI gamma file not available, try default.gam
		NSCStatusOut("gamma file not found, using default.");
		strcpy(filename,"default.gam");
		if(!(fptr = fopen((LPCSTR) filename,"r")))
		{
			// if default.gam not available, create it
			if(!(fptr = fopen((LPCSTR) filename,"w"))) 
			{
diskerr:
				if (fptr) fclose(fptr);
				NSCAfxMessageBox("Can't create default Gamma table!");
				return FALSE;//GAMMA_DISK_ERROR;
			}
			if (!SaveDefaultGamma(fptr)) goto diskerr;
			fclose(fptr);
			if(!(fptr = fopen((LPCSTR) filename,"r"))) goto diskerr;
		}
	}

	iRetVal = TRUE; //GAMMA_SUCCESS;
	// write gamma w/o addresses
	if (!SetGammaCurves(fptr)) { // == GAMMA_VERIFY_ERROR) {
		fclose(fptr);
		if(!(fptr = fopen((LPCSTR) filename,"r"))) {
			iRetVal = FALSE; //GAMMA_DISK_ERROR;
			NSCAfxMessageBox("Can't read Gamma table file!");
		}
		else {
			// next try write gamma with addresses
			if (!SetGammaCurves(fptr)) { // == GAMMA_VERIFY_ERROR) {
				NSCAfxMessageBox("Error verifying gamma table!");
				iRetVal = FALSE; //GAMMA_VERIFY_ERROR;
			}
		}
	}

	if (fptr) fclose(fptr);

	return iRetVal;
}

//-------------------------------------------------------
// FUNCTION	:	SaveDefaultGamma(FILE *fptr)

// DESCRIPTION	:	save a default gamma file

// INPUTS	:	fptr	- file handle of file to write

// RETURNS	:	TRUE if file stored ok
//			FALSE otherwise
//--------------------------------------------------------
BOOL CHardware::SaveDefaultGamma(FILE* fptr)
{
int address;	// loop counter
int iValue;	// gamma value to write
int color;	// color to write
int BUFFER_SIZE = 4096;

	for (color = 0; color < 3; color++) {

		switch(color) {
			case 0: fprintf(fptr, ";Gamma Data (red)\n");
				break;
			case 1: fprintf(fptr, ";Gamma Data (green)\n");
				break;
			case 2: fprintf(fptr, ";Gamma Data (blue)\n");
				break;
		}

		for (address = 0; address < BUFFER_SIZE; address++)
		{
			iValue = address*256/BUFFER_SIZE;
			if (fprintf(fptr, "%X\t%X\n", address, iValue) < 0)
				return FALSE;
		} // for addr=0..1023

	} // for color=0..2

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SetGammaCurves(FILE* fptr)

// DESCRIPTION	:	read a file of gamma values and send to merlin

// INPUTS	:	fptr	- file handle for file containing gamma values


// RETURNS	: TRUE = ok
//			else FALSE
//--------------------------------------------------------
BOOL CHardware::SetGammaCurves(FILE* fptr)
{
int address;		// gamma table address read from file/sent to merlin
int iValue;		// gamma value read from file/sent to merlin
int i,color;		// loop counters
char LineBuf[MAX_PATH];	// file input buff
BYTE value;		// value transmitted to port
BOOL iError = TRUE; // return value
int numErrors = 0;	// number of verify errors
BYTE reg8 = 0;		// value on reg 8 at start

	int BUFFER_SIZE = 4096; //for now read the same
	// allocate a buffer for the gamma table, 1024 entries for 3 colors
	BYTE pGamma[3*4096];

	// init gamma values to 0
	memset(pGamma,0,3*BUFFER_SIZE*sizeof(BYTE));

	// start by writing a 0 to reg 7
	i = !WriteRegister(7,0);
	if (i != SUCCESS) {
ioerr:
		iError = FALSE; //GAMMA_IO_ERROR;
		goto cleanup;
	}

	// read 3 gamma color tables from file
	for (color = 0; color < 3; color++) {

	// read a line of address value
    int max_address = 0;	//save max address read
	while(fgets(LineBuf, MAX_PATH, fptr))
	{
		LPSTR ptr = LineBuf; // skip comments and blank lines
		while (isspace(*ptr) && *ptr != '\n') ++ptr;
		if (*ptr == '\n') continue;
		if (*ptr == ';')
		{
			if (max_address == 0) continue;
			if (max_address < 1024)
			{
				//old 10 bit format
				for (int i = max_address; i >= 0; i--)
					for (int j=0; j<4; j++)
						pGamma[color*BUFFER_SIZE+i*4+j] = pGamma[color*BUFFER_SIZE+i];
				break; 
			}
		}
		// each line must contain 2 values
		if(sscanf(ptr, "%X %X", &address, &iValue) < 2)
		{
			if ((ptr = strchr(LineBuf,'\n')) != NULL) *ptr = '\0';
			NSCAfxMessageBox("Invalid entry: %s",LineBuf);
			continue;
		}
		// make sure the address is in range
		if (address < 0 || address > BUFFER_SIZE-1) {
			NSCAfxMessageBox("Invalid offset %x",address);
		}
		else 
		{ // good value read from file, so save in gamma table
			{
				pGamma[color*BUFFER_SIZE+address] = iValue;
			}
		}
		// done with this color when max addr is read
		if (address > max_address) max_address = address;
		if (address == BUFFER_SIZE-1) break;
	}
	// select color
	value = (color << 2)+2;
	i = !WriteRegister(3,value); // set gamma color selector
	if (i != SUCCESS) goto ioerr;

		// write the gamma table entry to merlin
	if (!SendData(6,pGamma+color*BUFFER_SIZE,BUFFER_SIZE))
		goto ioerr;

	} // for each color

cleanup:
	// clean up and display status if no error
	if (iError) { // == GAMMA_SUCCESS) {
		i = !WriteRegister(7,0);
		if (i != SUCCESS) {
			iError = FALSE; //GAMMA_IO_ERROR;
		}
	}
	return iError;
}


//for debug
//-------------------------------------------------------
// FUNCTION	:	NSCStatusOut(char *format, ...)

// DESCRIPTION	:	output a message to the status window and file
//			works like printf

// INPUTS	:	format	- message to output
//		:	...	- parameters for format string

// RETURNS	:	nothing
//--------------------------------------------------------
void CHardware::NSCStatusOut(const char *format, ...)
{
va_list ap;		// variable parameter list
char sBuf[200];		// buffer to create string to output
static FILE *pInfoFile = NULL;

    va_start(ap, format);
    vsprintf(sBuf,format,ap);
    va_end(ap);

	//for file
	char filename[MAX_PATH];
	strcpy(filename,GetFullFilePath("lm983x_status.txt"));

	char mode[MAX_PATH];
	if (!pInfoFile) strcpy(mode,"wt");
	else strcpy(mode,"at");
	pInfoFile = fopen(filename,mode); //open text file 
	if (pInfoFile)
	{
		fprintf(pInfoFile,"%s\n",sBuf);	//write to file
		fclose(pInfoFile); //close so keep if crash
	}
}

//pass error string to exe
void CHardware::NSCAfxMessageBox(const char *format, ...)
{
va_list ap;		// variable parameter list
char sBuf[LINESIZ];		// buffer to create string to output
	va_start(ap, format);
    vsprintf(sBuf,format,ap);
    va_end(ap);
	//error string
	for (int i = 0; i < LINESIZ; i++) m_scanStatus[i] = 0;
	strcpy(m_scanStatus,(LPCTSTR)sBuf);
	NSCStatusOut((LPCSTR)sBuf);
	Callback(FALSE);
}

//setup all standard registers
BOOL CHardware::SetupStandardRegisters(int *regs,BOOL calibration)
{
//	note that some values below are overwritten later
	regs[0]=regs[1]=regs[2]=0x100; //skip
	regs[3]=regs[4]=regs[5]=0;
	regs[6]=0x100;					//skip
	regs[7]=0;

	regs[0x8] = ComputeReg8Value(merlininfo.mclk);
	regs[0x9] = 0x3f;
	regs[0x9] |= 64*merlininfo.Bias_Current;
	regs[0xA] = 0;
	regs[0xB] = 0;
	if (merlininfo.Input_Polarity)
		regs[0xB] |= 1;
	if (merlininfo.CDS)
		regs[0xB] |= 2;
	if (merlininfo.Sensor_Type)
		regs[0xB] |= 4;
	{
		if (merlininfo.CIS_TR_Timing == 1)
			regs[0xb] |= 8;
		if (merlininfo.CIS_TR_Timing == 2)
			regs[0xb] |= 16;
		if (merlininfo.Fake_Optical_Black_Pixels)
			regs[0xb] |= 32;
	}

	regs[0xC] = 0;
	if (merlininfo.Phi1_Polarity)
		regs[0xC] |= 1;
	if (merlininfo.Phi2_Polarity)
		regs[0xC] |= 2;
	if (merlininfo.RS_Polarity)
		regs[0xC] |= 4;
	if (merlininfo.CP1_Polarity)
		regs[0xC] |= 8;
	if (merlininfo.CP2_Polarity)
		regs[0xC] |= 0x10;
	if (merlininfo.TR1_Polarity)
		regs[0xC] |= 0x20;
	if (merlininfo.TR2_Polarity)
		regs[0xC] |= 0x40;

	regs[0xD] = 0;
	if (merlininfo.Phi1_Status)
		regs[0xD] |= 1;
	if (merlininfo.Phi2_Status)
		regs[0xD] |= 2;
	if (merlininfo.RS_Status)
		regs[0xD] |= 4;
	if (merlininfo.CP1_Status)
		regs[0xD] |= 8;
	if (merlininfo.CP2_Status)
		regs[0xD] |= 0x10;
	if (merlininfo.TR1_Status)
		regs[0xD] |= 0x20;
	if (merlininfo.TR2_Status)
		regs[0xD] |= 0x40;
	if (merlininfo.TR_Pulse)
		regs[0xD] |= 0x80;

	regs[0xE] = merlininfo.TR_Pulse_Duration;
	regs[0xE] |= 16*merlininfo.TR_PHI_Guardband_Duration;

	// set pixel rate/line rate
	int color_mode;
	switch (merlininfo.Color_Mode) {
		default:
		case 0: color_mode = 0; break;	// pixel rate
		case 1: color_mode = 1; break;	// 3 channel line rate
		case 2: color_mode = 4; break;	// gray scale (1 channel mode A)
		case 3: color_mode = 5; break;	// 1 channel line rate 
	}

	if (!m_colorMode)
	{
		color_mode = 4; //set gray linerate
		if (merlininfo.Illumination_Mode == 2)
			merlininfo.Illumination_Mode = 3; //set gray for CIS
	}

	if (color_mode == 0) { // 0 for pixel rate, else line rate
	//Use these parameters for all pixel rate scanning
		regs[0xF] = merlininfo.Pixl_Optical_Black_Clamp_Start;
		regs[0x10] = merlininfo.Pixl_Optical_Black_Clamp_End;
		regs[0x11] = merlininfo.Pixl_Reset_Pulse_Start;
		regs[0x12] = merlininfo.Pixl_Reset_Pulse_End;
		regs[0x13] = merlininfo.Pixl_CP1_Pulse_Start;
		regs[0x14] = merlininfo.Pixl_CP1_Pulse_End;
		regs[0x15] = merlininfo.Pixl_CP2_Pulse_Start;
		regs[0x16] = merlininfo.Pixl_CP2_Pulse_End;
		regs[0x17] = merlininfo.Pixl_Reference_Sample;
		regs[0x18] = merlininfo.Pixl_Signal_Sample;
	}
	else {
	//Use these parameters for all line rate scanning
		regs[0xF] = merlininfo.Line_Optical_Black_Clamp_Start;
		regs[0x10] = merlininfo.Line_Optical_Black_Clamp_End;
		regs[0x11] = merlininfo.Line_Reset_Pulse_Start;
		regs[0x12] = merlininfo.Line_Reset_Pulse_End;
		regs[0x13] = merlininfo.Line_CP1_Pulse_Start;
		regs[0x14] = merlininfo.Line_CP1_Pulse_End;
		regs[0x15] = merlininfo.Line_CP2_Pulse_Start;
		regs[0x16] = merlininfo.Line_CP2_Pulse_End;
		regs[0x17] = merlininfo.Line_Reference_Sample;
		regs[0x18] = merlininfo.Line_Signal_Sample;
	}

	regs[0x19] = merlininfo.timeadj;	//LM9831 int time adj

	regs[0x1A] = 0;	
	regs[0x1B] = 0;	
	{
		regs[0x1A] = merlininfo.TR_Stepper_Correction/256;	//LM9831 tr to stepper
		regs[0x1B] = merlininfo.TR_Stepper_Correction & 0xff;	//ditto
	}

	regs[0x1C] = merlininfo.Optical_Black_Pixels_Start;
	regs[0x1D] = merlininfo.Optical_Black_Pixels_End; //diff for cal?

	regs[0x1E] = 0;//merlininfo.first_white_pixel/256;
	regs[0x1F] = 0;//merlininfo.first_white_pixel & 0xFF;
	//allow for adjust later
	merlininfo.Total_Pixels = max(merlininfo.Total_Pixels,
									merlininfo.Image_Pixel_Width+212);
	regs[0x20] = merlininfo.Total_Pixels/256;
	regs[0x21] = merlininfo.Total_Pixels & 0xFF;
	regs[0x22] = merlininfo.first_white_pixel/256; // data pixel start = active pixel start
	regs[0x23] = merlininfo.first_white_pixel & 0xFF;
	int linend;
	linend = merlininfo.first_white_pixel + merlininfo.Image_Pixel_Width;
	linend = min(linend,merlininfo.Total_Pixels-20);
	regs[0x24] = linend/256;
	regs[0x25] = linend & 0xFF;

	regs[0x26] = color_mode;
	regs[0x26] |= 8*merlininfo.Gray_Channel;
	if (merlininfo.TR_Red)	//are these swapped with calibration?
		regs[0x26] |= 0x20;
	if (merlininfo.TR_Green)
		regs[0x26] |= 0x40;
	if (merlininfo.TR_Blue)
		regs[0x26] |= 0x80;
	regs[0x27] = merlininfo.TR_Red_Drop;
	regs[0x27] |= 4*merlininfo.TR_Green_Drop;
	regs[0x27] |= 16*merlininfo.TR_Blue_Drop;
	regs[0x27] |= 64*merlininfo.Triple_TR;
	regs[0x28] = 0;

	regs[0x29] = merlininfo.Illumination_Mode;
	regs[0x29] |= 4*merlininfo.LampB_Output_ITA; //9832 only
	regs[0x2A] = merlininfo.Green_PWM_Duty_Cycle/256;
	regs[0x2B] = merlininfo.Green_PWM_Duty_Cycle & 0xFF;
	if ((color_mode == 4) && (merlininfo.Illumination_Mode != 1))
	{
		//gray CIS has only green channel
		merlininfo.Red_Lamp_Off = merlininfo.Red_Lamp_On;
		merlininfo.Blue_Lamp_Off = merlininfo.Blue_Lamp_On;
	}
	regs[0x2C] = merlininfo.Red_Lamp_On/256;
	regs[0x2D] = merlininfo.Red_Lamp_On & 0xFF;
	regs[0x2E] = merlininfo.Red_Lamp_Off/256;
	regs[0x2F] = merlininfo.Red_Lamp_Off & 0xFF;
	regs[0x30] = merlininfo.Green_Lamp_On/256;
	regs[0x31] = merlininfo.Green_Lamp_On & 0xFF;
	regs[0x32] = merlininfo.Green_Lamp_Off/256;
	regs[0x33] = merlininfo.Green_Lamp_Off & 0xFF;
	regs[0x34] = merlininfo.Blue_Lamp_On/256;
	regs[0x35] = merlininfo.Blue_Lamp_On & 0xFF;
	regs[0x36] = merlininfo.Blue_Lamp_Off/256;
	regs[0x37] = merlininfo.Blue_Lamp_Off & 0xFF;

	//leave 38 to 42 alone if not calibration
	if (!calibration) for (int i = 0x38; i < 0x42; i++)
		regs[i] = 0x100;

	{
		BYTE value;
		if (!ReadRegister(0x42,&value)) return FALSE;
		regs[0x42] = value;
		// init reg 42 to make sure we can r/w merlin
		regs[0x42] |= 0x20;/*(merlininfo.Parallel_Port_Current?1:0);*/
		regs[0x42] |= 0x40*(merlininfo.RAM_WordSize/1024);
	}

	regs[0x43] = 0;  

	regs[0x44] = 0;

	regs[0x45] = merlininfo.Stepping;
	// If it's Micro, current sensing = 2 phases,		reg 45 b1
	// If it's Full, current sensing = 1 phase
	if (merlininfo.Stepping)
		regs[0x45] |= 2;
	regs[0x45] |= 4*merlininfo.Phase_A_Polarity;
	regs[0x45] |= 8*merlininfo.Phase_B_Polarity;
	regs[0x45] |= 0x10; // set motor on
	regs[0x45] |= 32*merlininfo.Reverse_Motor_Direction;  //9832 only
	regs[0x45] |= 64*merlininfo.Fullstep_During_Fastfeed; //9832 only

	//FastFeed Stepsize
	int r = 8; // line rate
	if (color_mode == 0) r = 24; // pixel rate
	//--- calculate the fast feed step size
	double x = merlininfo.Crystal_Frequency/((double)merlininfo.mclk/MCLKDIV_SCALING);
	x /= merlininfo.Max_Speed * 4 * merlininfo.Motor_Steps_Per_Inch * r;
	if (merlininfo.timeadj) x /= (merlininfo.timeadj);
	//Scan Stepsize
	regs[0x46] = (BYTE) (x/256);
	regs[0x47] = (BYTE) x;
	regs[0x48] = (BYTE) (x/256);
	regs[0x49] = (BYTE) x;

	//------- lines to skip before scanning


	if (calibration) 
	{
	regs[0x4a] = merlininfo.Start_of_Calibration/256;
	regs[0x4b] = merlininfo.Start_of_Calibration&0xff;
	}
	else
	{
	regs[0x4a] = merlininfo.Start_of_Scan/256;
	regs[0x4b] = merlininfo.Start_of_Scan&0xff;
	}

	regs[0x4C] = merlininfo.Steps_to_Scan_After_PS2/256;
	regs[0x4D] = merlininfo.Steps_to_Scan_After_PS2 & 0xFF;
	regs[0x4E] = 0;//compute later 
	regs[0x4F] = 0;
	regs[0x50] = merlininfo.Steps_To_Reverse; 

	regs[0x51]  = 4*merlininfo.Steps_Stopped;
	regs[0x51] |= 16*merlininfo.Steps_25;
	regs[0x51] |= 64*merlininfo.Steps_50;
	regs[0x52] = 0; 
	regs[0x53] = 1; 
	regs[0x54] = merlininfo.Non_Reversing_Extra_Lines;


	regs[0x55] = merlininfo.KickStart_Steps;
	regs[0x55] |= 8*merlininfo.Hold_Current_Timeout;
	regs[0x56] = merlininfo.Stepper_PWM_Frequency;
	regs[0x57] = merlininfo.Stepper_PWM_Duty_Cycle;

	regs[0x58] = merlininfo.Paper_Sensor_1_Polarity;
	regs[0x58] |= 2*merlininfo.Paper_Sensor_1_Trigger;
	regs[0x58] |= 4*merlininfo.Paper_Sensor_1_Stop_Scan;
	regs[0x58] |= 8*merlininfo.Paper_Sensor_2_Polarity;
	regs[0x58] |= 16*merlininfo.Paper_Sensor_2_Trigger;
	regs[0x58] |= 32*merlininfo.Paper_Sensor_2_Stop_Scan;

	regs[0x59] = merlininfo.MiscIO_1_Type;
	regs[0x59] |= 2*merlininfo.MiscIO_1_Polarity;
	regs[0x59] |= 4*merlininfo.MiscIO_1_Trigger;
	regs[0x59] |= 8*merlininfo.MiscIO_1_Output_State;
	regs[0x59] |= 16*merlininfo.MiscIO_2_Type;
	regs[0x59] |= 32*merlininfo.MiscIO_2_Polarity;
	regs[0x59] |= 64*merlininfo.MiscIO_2_Trigger;
	regs[0x59] |= 128*merlininfo.MiscIO_2_Output_State;

	regs[0x5a] = 0;
	regs[0x5b] = 0;

	regs[0x5a] = merlininfo.MiscIO_3_Type;
	regs[0x5a] |= 2*merlininfo.MiscIO_3_Polarity;
	regs[0x5a] |= 4*merlininfo.MiscIO_3_Trigger;
	regs[0x5a] |= 8*merlininfo.MiscIO_3_Output_State;
	regs[0x5a] |= 16*merlininfo.MiscIO_4_Type;
	regs[0x5a] |= 32*merlininfo.MiscIO_4_Polarity;
	regs[0x5a] |= 64*merlininfo.MiscIO_4_Trigger;
	regs[0x5a] |= 128*merlininfo.MiscIO_4_Output_State;

	regs[0x5b] = merlininfo.MiscIO_5_Type;
	regs[0x5b] |= 2*merlininfo.MiscIO_5_Polarity;
	regs[0x5b] |= 4*merlininfo.MiscIO_5_Trigger;
	regs[0x5b] |= 8*merlininfo.MiscIO_5_Output_State;
	regs[0x5b] |= 16*merlininfo.MiscIO_6_Type;
	regs[0x5b] |= 32*merlininfo.MiscIO_6_Polarity;
	regs[0x5b] |= 64*merlininfo.MiscIO_6_Trigger;
	regs[0x5b] |= 128*merlininfo.MiscIO_6_Output_State;

	for (int i = 0x5c; i <= 0x7F; i++)
		regs[i] = 0;

//TEST!!!!!!!!!!
//	regs[0x5e] = 0x80;	//CDS Signal

	return TRUE;
}


BOOL CHardware::LM9831()
{
	return merlininfo.chip == 0;
}

BOOL CHardware::LM9832()
{
	return merlininfo.chip == 1;
}

BOOL CHardware::LM9833()
{
	return merlininfo.chip == 2;
}

//mclk value to Reg8 value
int CHardware::ComputeReg8Value(int mclk)
{
	return (int) (((double)mclk/MCLKDIV_SCALING - 1.) * 2.);
}

//value to size
int CHardware::SizeOfKCount(int count)
{
	int size = count * 1024;
	size *= merlininfo.RAM_WordSize/128;
	return size;
}

//size to kcount
int CHardware::KCountFromSize(int size)
{
	int KCount = size/1024;
	KCount = KCount*128/merlininfo.RAM_WordSize;
	return KCount;
}

//enable paper sensor
//disable accel
BOOL CHardware::EnableStopScan()
{
	BYTE reg58;
	reg58 = merlininfo.Paper_Sensor_1_Polarity;
	reg58 |= 2*merlininfo.Paper_Sensor_1_Trigger;
	reg58 |= 4*merlininfo.Paper_Sensor_1_Stop_Scan;
	reg58 |= 8*merlininfo.Paper_Sensor_2_Polarity;
	reg58 |= 16*merlininfo.Paper_Sensor_2_Trigger;
	reg58 |= 32*merlininfo.Paper_Sensor_2_Stop_Scan;
	if (!WriteRegister(0x58,reg58)) return FALSE; //do stop
	return TRUE;
}

//disable paper sensor
//enable accel
BOOL CHardware::DisableStopScan()
{
	if (!WriteRegister(0x58,0)) return FALSE; //no stop
	return TRUE;
}

BOOL CHardware::CloseDriver()
{
	return DriverUSB.CloseDriver();
}

//Do callback
SCAN_RESULT CHardware::Callback(BOOL check_messages)
{
	if (m_callbackfn)
	{
		if (m_callbackfn(check_messages) != SCAN_OK) return SCAN_ERROR;
	}
	return SCAN_OK;
}

/*******************************************************************************
DESCRIPTION:	Returns full path name
PARAMETERS:		none

RETURN VALUE:	char* containing the path.  If error, then a null
				string is returned.
*******************************************************************************/
char* CHardware::GetFullFilePath(char* file)
{
	static char	path[MAX_PATH];

	if (!GetWindowsDirectory(path,MAX_PATH))
	{
		return "";
	}
	if (path[strlen(path)-1] != '\\') strcat(path,"\\");
	strcat(path,"twain_32\\lm983x\\");
	strcat(path,file);
	return( path );
}
