// FILE		: scan_scan.CPP

// DESCRIPTION	: Implementation of the CScan class
//			scan interface code

// REVISIONS	: 

#include "stdafx.h"
#include "Scan_scan.h"
#include "Scan_hardware.h"
#include "Scan_scan2bmp.h"
#include "math.h"
#include "io.h"
#include "fcntl.h"
#include "sys\stat.h"

CHardware Hardware;
CScan2Bmp Scan2Bmp;

#define INIFILE Hardware.GetFullFilePath("lm983x.def")

//#define TESTMODE


/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	CScan class constructor

// DESCRIPTION	:	Create a CHardware object and initialize variables

// RETURNS	:	nothing
//--------------------------------------------------------
CScan::CScan()
{
	npts = 0;		// number of points in the pRed, pGreen, etc arrays
	m_reentry = FALSE;
	m_firstscan = TRUE;
	m_callback = FALSE;
	PCTransferRate = 1000000; //typical rate
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	CScan class destructor

// DESCRIPTION	:	clean up memory allocation and resources

// RETURNS	:	nothing
//--------------------------------------------------------
CScan::~CScan()
{
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	Startup()

// DESCRIPTION	:	Do startup - setup port

// RETURNS	:	FALSE if fail
//--------------------------------------------------------
BOOL CScan::Startup()
{
	m_currentScanInfo.startTime = m_currentScanInfo.endTime = 
		m_currentScanInfo.scanStartTime = m_currentScanInfo.scanEndTime = 
		0;
	Hardware.m_colorMode = 
		(m_currentScanSettings.pixelType == SCAN_PIXELTYPE_RGB);
	char fname[MAX_PATH];		// path to windows directory
	char sBuf[MAX_PATH];
	GetPrivateProfileString("SETUP","ScannerFile","",sBuf,sizeof(sBuf),INIFILE);
	if (!strlen(sBuf))
	{
		Hardware.NSCAfxMessageBox("ScannerFile= not found in LM983x.def");
		return FALSE;
	}
	strcpy(fname,Hardware.GetFullFilePath(sBuf));

	memset(&m_currentScanHardware,0,sizeof(m_currentScanHardware));
	
	if (Hardware.ReadMerlinINI(fname,&m_currentScanSettings))
	{
		m_currentScanHardware.opticalRes = 
			Hardware.merlininfo.Optical_Resolution;
		m_currentScanHardware.widthOpticalPixels = 
			Hardware.merlininfo.Image_Pixel_Width;
		m_currentScanHardware.heightOpticalPixels = 
			Hardware.merlininfo.Scanner_Length
			*Hardware.merlininfo.Optical_Resolution;
		m_currentScanHardware.chip = 
			Hardware.merlininfo.chip;
		strcpy(m_currentScanHardware.device,
			Hardware.merlininfo.USB_Device);
		//determine base resolutions
		int base_optical_resolution = Hardware.merlininfo.Optical_Resolution;
		if (Hardware.merlininfo.EvenOdd_Line_Separation)
			base_optical_resolution /= 2;
		for (int i = 0; i < BASE_RESOLUTION_SIZE; i++)
		{
			int res;
			int j = 7 - i;
			if (j >= 0)
			{
				//calculate divider scaled by 2
				int hdiv = ((j&1)+2)*(j&2?2:1)*(j&4?4:1);	
				res = base_optical_resolution*2/hdiv;
			}
			else
			{
				//do 2x until max
				res = 2*m_currentScanHardware.baseResolution[i-1];
				if (res > Hardware.merlininfo.Max_Sensor_Resolution)
				{
					res = 0;
				}
			}
			m_currentScanHardware.baseResolution[i] = res;
		}
		return TRUE;
	}
	return FALSE;
}

////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SetupScan(full)

// DESCRIPTION: setup scan
// full is TRUE if full setup, else just calculate

// RETURNS	:	FALSE if fail
//--------------------------------------------------------
BOOL CScan::SetupScan(BOOL full)
{
char filename[MAX_PATH];		// file name for pxl file
	m_currentScanInfo.rawActBytes = 0;
	if (!Hardware.InitUSB(full)) return FALSE;

	if (full)
	{
		//get transferrate and file name
		strcpy(filename,ComputePCTransferRate("finalscn"));
		if (!strcmp(filename,"")) return FALSE;
	}
	else
	{
		strcpy(filename,"compute.pxl");
//		PCTransferRate = 1000*1000; //typical rate
//		Hardware.NSCAfxMessageBox("Idle");
	}

	int regs[0x80];

	if (!SetupScanRes(regs,filename)) return FALSE;

	// setup regs for final scan

	Hardware.NSCStatusOut("actual yres=%d",scaninfo.actualyres);
	Hardware.NSCStatusOut("desired yres=%d",scaninfo.desiredyres);
	Hardware.NSCStatusOut("scanparms yres=%d",ScanParms.yres);
	strcpy(m_scanfilename,filename);

	//setup for first scan
	BOOL done;
	if (!ScanSome(&done,TRUE)) return FALSE; //cancel last 
	if (!ScanSome(&done,FALSE)) return FALSE;//init	
//update scan size
	if (SubString(filename,"compute.") || SubString(filename,"finalscn."))
	{
		m_currentScanInfo.rawTotalBytes = 
			ScanParms.lines*ScanParms.bytesperline;
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ComputePCTransferRate()

// DESCRIPTION	:	processor for PCTransferRate
//					perform PCtransferrate

// RETURNS	:	filename
//--------------------------------------------------------
char *CScan::ComputePCTransferRate(char *basename)
{
	static BOOL PCTransfer = FALSE;
	static char filename[MAX_PATH];
	PCTransfer = !PCTransfer;
	if (PCTransfer)
	{
		m_currentScanInfo.startTime = m_currentScanInfo.endTime = GetTickCount();
		Hardware.NSCAfxMessageBox("PC Transfer Rate"); //status message
		PCTransferRate = 2000*1000; //force high clock rate
		if (!SetupScan()) return "";

		if (!DoScan(m_scanfilename)) return "";
		//compute PC rate
		Hardware.NSCStatusOut("Transfer bytes = %d",ScanParms.bytesread);
		Hardware.NSCStatusOut("Millisecs = %d",ScanParms.time);
		PCTransferRate = (ScanParms.bytesread/ScanParms.time)*1000; // time in bytes/sec
		Hardware.NSCStatusOut("PCTransferRate=%d",(int)PCTransferRate);
	}
	else
	{
		basename =  "pcrate";
	}
	// create pxl filename in windows directory
	strcpy(filename,Hardware.GetFullFilePath(basename));
	strcat(filename,".pxl");
	return filename;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SetupScanRes()

// DESCRIPTION	:	set merlin registers to do a final scan

// INPUTS	:	regs	- array of merlin regs to compute

// RETURNS	:	1 if ok, 0 otherwise
//--------------------------------------------------------
BOOL CScan::SetupScanRes(int *regs,char *filename)
{
int pauselimit;			// calculated pause limit

//--- these register settings are set in set merlin registers.
//------ set merlin regs portion
restart:
	if (!Hardware.SetupStandardRegisters(regs,FALSE)) return FALSE;

//---------------------------------------------------------
	if (!SubString(filename,"pcrate.")) 
	{
		SetScanRes(regs,m_currentScanSettings.hRes,m_currentScanSettings.vRes,
			m_currentScanSettings.bitsPerPixel);	// calculate resolution (use preview for now)!!!!
		if (!Hardware.m_debug) ScanParms.autohome = 1;		// home when done
		else 
		{
			ScanParms.autohome = 0;
			regs[0x45] &= ~0x10; // set motor off
		}
	}
	else 
	{
		regs[0x45] &= ~0x10; // set motor off
		SetScanRes(regs,m_currentScanSettings.hRes,m_currentScanSettings.vRes,16);
		ScanParms.autohome = 0;		// don't home when done
	}
	CalculateScanRect(regs,filename); // calculate rectangle to scan

	// indicate the actual resolution used for scanning
	Hardware.merlininfo.mclk = ComputeCalibrationMClk(regs);
	if (Hardware.merlininfo.mclk < 0) goto restart; //need new calculation

// convert mclk to reg 8 value
	regs[0x8] = Hardware.ComputeReg8Value(Hardware.merlininfo.mclk);
//
	// setup for removing line separation
	scaninfo.roffset = 0;
	scaninfo.goffset = 0;
	scaninfo.boffset = 0;
	// have even/odd offset for pixel and line rate
	scaninfo.eoffset = Hardware.merlininfo.EvenOdd_Line_Separation*
							(scaninfo.actualyres
							/(double)Hardware.merlininfo.Optical_Resolution);
	int color_mode = regs[0x26]&7;
	if (color_mode == 5) color_mode = 1;
	if (color_mode != 4)
	{
		if (Hardware.merlininfo.Illumination_Mode == 1) //CCD
		{
			scaninfo.roffset = -Hardware.merlininfo.Line_Separation*
							(scaninfo.actualyres
							/(double)Hardware.merlininfo.Optical_Resolution);
			scaninfo.boffset = -scaninfo.roffset;
		}
		if (Hardware.merlininfo.Color_Mode & 1) //color line rate
		{
			scaninfo.roffset += (1./3.);
			scaninfo.boffset -= (1./3.);
		}
	}

	// number of lines to scan = image ht in inches * vert res in lines/inch
	ScanParms.lines = (int) (ScanParms.htinches * ScanParms.yres);
	if (ScanParms.lines <= 0) ScanParms.lines = 1;
//	if (color_mode == 1) ScanParms.lines *= 3;	// line rate
	Hardware.NSCStatusOut("lines to be returned=%d",ScanParms.lines);

//	Hardware.merlininfo.mclk = ScanParms.clkdivider; // computed above in call to ComputeCalibrationMClk
	// the clock divider and ffss are calculated here
	// there's no need to do it in set scan parms

		int r = 8; // line rate
		if (color_mode == 0) r = 24; // pixel rate
	//--- calculate the fast feed step size
		double x = Hardware.merlininfo.Crystal_Frequency/((double)Hardware.merlininfo.mclk/MCLKDIV_SCALING);
		x /= Hardware.merlininfo.Max_Speed * 4 * Hardware.merlininfo.Motor_Steps_Per_Inch * r;
		if (Hardware.merlininfo.timeadj)
			x /= (Hardware.merlininfo.timeadj);
		regs[0x48] = (BYTE) (x/256);
		regs[0x49] = (BYTE) x;

	if (!SubString(filename,"pcrate.")) 
		pauselimit = regs[0x4E];
	else
	{
		 pauselimit = 255;
	}

	//if calculate pcrate, don't forward to region
	if (SubString(filename,"pcrate."))
	{
		regs[0x4a] = 0;
		regs[0x4b] = 0; //skip acceleration
		regs[0x51] = 0;
	}

//--- calc pause limit
	regs[0x4E] = pauselimit;
	regs[0x4f] = regs[0x4e] - 1;
//----------------------------------------------------

	// save scan info during calibration
	SCANINFO savescaninfo = scaninfo;
	SCANPARMS saveparms = ScanParms;
	if (!SubString(filename,"pcrate.") && !SubString(filename,"compute."))
	{
		if (!DoCalibration(regs[0x9])) 
		{
			return FALSE;
		}
	}
	// save scan info during calibration
	scaninfo = savescaninfo;
	ScanParms = saveparms;

	BYTE value;
	if (!Hardware.ReadRegister(0x42,&value)) return FALSE;
	regs[0x42] = value;
	
	CalculateScanRect(regs,filename); // calculate rectangle to scan
//dont reset lamp settings from calibration
	for (int i = 0x2a; i < 0x38; i++) regs[i] = 0x100;

	if (!Hardware.WriteMerlin(regs))return FALSE;

	// calc bytes/line
	if (!CalcBytesPerLine()) return FALSE;

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	CalcBytesPerLine()

// DESCRIPTION	:	read merlin regs and calculate scanning parameters

// INPUTS	:	none

// RETURNS	:	TRUE if ok, otherwise error cause
//--------------------------------------------------------
// determine the bytesperline
// ret 0 if ok, otherwise error cause
BOOL CScan::CalcBytesPerLine()
{
int i;		// work variable
BYTE value;	// byte read from merlin port
int regs[0x80];	// merlin regs

	// read merlin regs
	if (!Hardware.ReadMerlin(regs)) return FALSE;

// bytes/line = valid pixels * (bpp / 8) /DPI
// valid pixels = reg22:23 - reg 24:25
// bpp = reg9 b3:b4 00=1, 01=2, 10=4, 11=8
// dpi = reg9 b0..b2, 0->1, 1->1.5, 2->2, 3->3, 4->4, 5->6, 6->8, 8->12
// reg26 b0..b2 =0 for pixel rate, otherwise line rate
// if pixel rate, then bytes/line = 3 * valid pixels * (bpp/8)/dpi

// read high byte of first pixel to scan
	int valpix = 256*regs[0x22];

// read low byte of first pixel to scan
	valpix += regs[0x23];
	ScanParms.startpixel = valpix;

// read high byte of last pixel to scan
	i = 256*regs[0x24];

// read low byte of last pixel to scan
	i += regs[0x25];
	ScanParms.endpixel = i;

// get the bits per pixel
// bpp=bits per pixel, ppb = pixels per byte
	value = regs[0x9];
	ScanParms.bpp = 16;
	switch (value & 0x38) {
		case 0: ScanParms.bpp = 1; break;
		case 0x8: ScanParms.bpp = 2; break;
		case 0x10: ScanParms.bpp = 4; break;
		case 0x18: ScanParms.bpp = 8; break;
		default: ScanParms.bpp = 16;break;
	}

// compute the horizontal dpi (pixels per inch)
	value &= 7;
	ScanParms.dpi = ((value&1)*.5+1)*(value&2?2:1)*(value&4?4:1);	
	ScanParms.xres = 
		(int) (Hardware.merlininfo.Optical_Resolution/ScanParms.dpi);

// calculate vertical resolution
	int stepsize = 256 * regs[0x46] + regs[0x47];

// determine the line rate or pixel rate
	value = regs[0x26];
	value &= 7;
	if (value == 5) value = 1; 
	ScanParms.linerate = value; // 0=pixel, 1=line, 4=mode a, 5=mode b
	int pm = 1;
// if line rate, then we get a line of red, a line of green, a line of blue
// if pixel rate, then each line contains rgb rgb rgb, etc.
	if (value == 0) { // pixel rate
		pm = 3;
		ScanParms.linerate = 0;
	}

// compute the number of bytes per line
	valpix = ScanParms.endpixel - ScanParms.startpixel;
	int x = (int) (valpix/ScanParms.dpi);

// the number of pixels per line = bytes per line * pixels per byte
	ScanParms.pixelsperline = x;
	ScanParms.bytesperline = x*pm*ScanParms.bpp/8;

	{
		//bytesperline must be truncated word boundary'
		//add two status bytes
		ScanParms.bytesperline &= ~1;
		if (pm == 1) //adjust pixelsperline only for line rate for now
		{
			ScanParms.pixelsperline = (ScanParms.bytesperline * 8)/(ScanParms.bpp*pm);
		}
		ScanParms.bytesperline += 2;
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SetScanRes(int *regs,hres,vres,bpp)

// DESCRIPTION	:	set resolution to scan

// INPUTS	:	regs	- merlin registers

// RETURNS	:	void
//--------------------------------------------------------
void CScan::SetScanRes(int *regs,int horresval,int verresval,int bpp)
{
int j;		// work variables

	// calculate dpi divider for specified horizontal resolution
	// set dpi divider to be next step greater than requested
	double dpi;
	dpi = (double)Hardware.merlininfo.Optical_Resolution/(double) horresval;
	if (Hardware.merlininfo.EvenOdd_Line_Separation) dpi = 1.; //no divider 
	j = 1;
	if (dpi < 1.5) j = 0;
	if (dpi >= 1.5) j = 1;
	if (dpi >= 2) j = 2;
	if (dpi >= 3) j = 3;
	if (dpi >= 4) j = 4;
	if (dpi >= 6) j = 5;
	if (dpi >= 8) j = 6;
	if (dpi >= 12) j = 7;
	dpi = ((j&1)*.5+1)*(j&2?2:1)*(j&4?4:1);	

	// scan at the resolution that is just greater than desired
	scaninfo.actualxres = (int) (Hardware.merlininfo.Optical_Resolution/dpi);
	// but save in specified resolution
	scaninfo.desiredxres = horresval;

	if (!(bpp%3)) bpp/=3; //in case passed bits*3
	if (bpp > 8) bpp = 16;
	ScanParms.bpp = bpp; // 1,2,4,8
	int i;
	switch (bpp) {
		case 1: i = 0; break;
		case 2: i = 8; break;
		case 4: i = 0x10; break;
		case 8: i = 0x18; break;
		case 16: i = 0x20; break;
	}
	regs[0x9] = i + j; // 8 bit image data, bpp, dpi
	regs[0x9] |= 64*Hardware.merlininfo.Bias_Current;

	ScanParms.xres = 
		(int) (Hardware.merlininfo.Optical_Resolution/ScanParms.dpi);

	scaninfo.desiredyres = verresval;
	if (Hardware.merlininfo.Color_Mode & 1)
	{
		//min vertical scan res for color line rate
// 		while (verresval < 75) verresval *= 2;
	}
	ScanParms.yres = verresval;
	scaninfo.actualyres = verresval;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	CalculateScanRect(regs)

// DESCRIPTION	:	compute regs for zoom rect
//			set startpixel, endpixel to zoom rect or entire scan area

// INPUTS	:	regs	- array of merlin regs

// RETURNS	:	regs set
//--------------------------------------------------------
void CScan::CalculateScanRect(int *regs,char *filename)
{
//RECT irect;	// zoom rectangle mapped to scanner
//RECT rect;	// image rectangle

	if (!regs) return;

	ScanParms.xres = 
		(int) (Hardware.merlininfo.Optical_Resolution/ScanParms.dpi);

	int linend;
	linend = Hardware.merlininfo.first_white_pixel + Hardware.merlininfo.Image_Pixel_Width;
	linend = min(linend,Hardware.merlininfo.Total_Pixels-20);

	// setup full screen
	ScanParms.startpixel = Hardware.merlininfo.first_white_pixel;
	ScanParms.endpixel = linend;

	// calculate the number of lines in the full scan area
	ScanParms.lines = (int) (Hardware.merlininfo.Scanner_Length * scaninfo.actualyres);
	if (ScanParms.lines <= 0) ScanParms.lines = 1;

	//mod for selected rect
	long rectwidth = m_currentScanSettings.boundingRect.right - m_currentScanSettings.boundingRect.left;
	long rectheight = m_currentScanSettings.boundingRect.bottom - m_currentScanSettings.boundingRect.top;

	int newstartpixel = (int)(((double)(ScanParms.endpixel - ScanParms.startpixel)*
						(m_currentScanSettings.scanRect.left - m_currentScanSettings.boundingRect.left) 
						/rectwidth) + ScanParms.startpixel);

	int newendpixel   = (int)(((double)(ScanParms.endpixel - ScanParms.startpixel)*
						(m_currentScanSettings.scanRect.right - m_currentScanSettings.boundingRect.left) 
						/rectwidth) + ScanParms.startpixel);

	int newlinestart  = (int)(((double)ScanParms.lines) * 
						(m_currentScanSettings.scanRect.top - m_currentScanSettings.boundingRect.top)
						/rectheight);

	int newlineend  =  (int)(((double)ScanParms.lines) * 
						(m_currentScanSettings.scanRect.bottom - m_currentScanSettings.boundingRect.top)
						/rectheight);

	//be sure that there is at least 8k for LM9831 data to scan!!!!
	if (Hardware.LM9831())
	if ((newendpixel-newstartpixel)*(newlineend - newlinestart) < 
			Hardware.SizeOfKCount(1))
	{
		//force bigger
		newstartpixel = max(ScanParms.startpixel,newstartpixel-500);
		newendpixel   = min(ScanParms.endpixel,newendpixel+500);
		newlinestart  = max(0,newlinestart-20);
		newlineend	  = min(ScanParms.lines,newlineend+20);
	}

	if (SubString(filename,"pcrate."))		//keep full if pcrate calculation
	{
		newstartpixel = ScanParms.startpixel;
		newendpixel = ScanParms.endpixel;
		newlinestart = 0;
		newlineend = ScanParms.lines;
	}

//check bytesperline, and adjust pixel length if needed so that
//we are guaranteed to get all pixels 
//bytes per line must be even also
	int bpp = m_currentScanSettings.bitsPerPixel;
	if (m_currentScanSettings.pixelType == SCAN_PIXELTYPE_RGB)
	{
		if (bpp > 24) bpp = 48;
	}
	else if (bpp > 8) bpp = 16;
	int value = regs[0x9] & 7;
	double dpi = ((value&1)*.5+1)*(value&2?2:1)*(value&4?4:1);	

	int pixelswanted = (int) ceil(((double)m_currentScanSettings.scanRect.right - 
		m_currentScanSettings.scanRect.left)*scaninfo.actualxres/
		scaninfo.desiredxres);
	value = pixelswanted*bpp;
	int byteswanted = value/8;
	if (value%8) byteswanted++;	//round up
	if (!byteswanted) byteswanted++; //must be non zero
	if (byteswanted&1) byteswanted++; //must be even

	int opticalpixelsperline = (int) (byteswanted*8*dpi/bpp);
	newendpixel = newstartpixel + opticalpixelsperline;

	if (!Hardware.merlininfo.Mirror_Image) 
	{
		ScanParms.startpixel = newstartpixel;
		ScanParms.endpixel = newendpixel;
	}
	else
	{
		//mirror image 
		int start = ScanParms.startpixel,end = ScanParms.endpixel;
		ScanParms.startpixel = start + (end - newendpixel);
		ScanParms.endpixel =   end - (newstartpixel - start);
	}

	//force startpixel to be even, so know where even start
	if (ScanParms.startpixel&1)
	{
		ScanParms.startpixel++;
		ScanParms.endpixel++;
	}

	regs[0x25] = ScanParms.endpixel & 0xFF;
	regs[0x24] = ScanParms.endpixel / 256;
	regs[0x23] = ScanParms.startpixel & 0xFF;
	regs[0x22] = ScanParms.startpixel / 256;

	int stepstoskip = Hardware.merlininfo.Start_of_Scan; 
	stepstoskip += (newlinestart*Hardware.merlininfo.Motor_Steps_Per_Inch
					/scaninfo.actualyres);

	regs[0x4A] = stepstoskip/256;
	regs[0x4B] = stepstoskip & 0xFF;

	// calculate the number of lines in the scan area
	ScanParms.htinches = Hardware.merlininfo.Scanner_Length*
						 (newlineend-newlinestart)/ScanParms.lines;
	ScanParms.lines = (int) (newlineend - newlinestart);
	if (ScanParms.lines <= 0) ScanParms.lines = 1;

}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ComputeCalibrationMClk(int *regs)

// DESCRIPTION	:	calculate the mclk for calibration and scanning
// this function was removed from the CCalib class so that the scanning
// routines could use it to calculate the clk divider before scans.

//if the pctransfer rate has not been computed, compute it
//compute ideal_step_size
//calculate ideal mclk divider from average pc transfer rate
//calculate mclk divider for max integration time
//calculate mclk divider for max motor speed
//check if stepsize must be adjusted because max integration time
//   gives too fast motor speed
//   else limit mclk divider between max integration time and max motor speed

// INPUTS	:	regs[0x80] - merlin configuration registers
// scaninfo.yres is the vertical res that may be changed

// RETURNS	:	calculated mclk
// registers set: 19
//			46:47 step size
//			4e pause limit
//			52:53 dpd
//--------------------------------------------------------
int CScan::ComputeCalibrationMClk(int *regs)
{

	Hardware.merlininfo.timeadj = regs[0x19] = 0; //assume off now until needed

// the first calibration needs to have the pc transfer rate
//0. if the pctransfer rate has not been computed

	Hardware.NSCStatusOut("0. vres=%d",scaninfo.actualyres);

//1. compute ideal_step_size
// step size = tr*vres/(4*motor steps/in)
// tr is from the dpd calculation but does include the cm component
// vres is the vertical resolution and is stored in scaninfo.yres
// motor steps/in is from ini file
	if (scaninfo.actualyres == 0 || scaninfo.actualyres > Hardware.merlininfo.Optical_Resolution) 
		scaninfo.actualyres = Hardware.merlininfo.Optical_Resolution;
	ScanParms.yres = scaninfo.actualyres;
	Hardware.ComputeStepSize(scaninfo.actualyres,regs);
	int stepsize = 256 * regs[0x46] + regs[0x47];

// 1.5, calculate the min mclk divider based on min integration time
//	save mclk divider as int*2 value
//	calculate the max mclk divider and save as *2
// r=24 for pixel rate, 8 for line rate
	int r;
	r = 8; // line rate
	if ((regs[0x26] & 7) == 0) r = 24; // pixel rate

	//use high or low res min integration time
	double min_integration_time = ((regs[0x9]&7) > 2 ?
		Hardware.merlininfo.Minimum_Integration_Time_LowRes:
		Hardware.merlininfo.Minimum_Integration_Time_HighRes);
	int linend = Hardware.ComputeTR(regs,FALSE); //linend wo cm
	int minmclk = (int) ceil((double) MCLKDIV_SCALING * 
					Hardware.merlininfo.Crystal_Frequency * 
					min_integration_time
					/((double) 1000. * r * linend)); 

	minmclk = max(minmclk,Hardware.merlininfo.minmclk);

	minmclk = max(minmclk,1*MCLKDIV_SCALING); 

	int maxmclk = (int) (32.5*MCLKDIV_SCALING + .5); //round to be sure proper int 

	Hardware.NSCStatusOut("2.5 lower mclkdiv limit=%f",(double)minmclk/MCLKDIV_SCALING);
	Hardware.NSCStatusOut("2.5 upper mclkdiv limit=%f",(double)maxmclk/MCLKDIV_SCALING);

// 2. calculate ideal mclk divider from average pc transfer rate
// ideal mclkdiv = pixelsinline*clock in/(8*line width*data rate)
// *2 and round up
// byte in line from calculation
// line width = tr from dpd calculation

// get the bits per pixel
	int pixelbits;
	switch (regs[0x9] & 0x38) {
		case 0:	    pixelbits=1; break;
		case 0x8:   pixelbits=2; break;
		case 0x10:  pixelbits=4; break;
		case 0x18:  pixelbits=8; break;
		default:	pixelbits=16;break;
	}
// compute the horizontal dpi (pixels per inch)
	double dpi;
	int j = regs[0x9] & 0x7;
	dpi = ((j&1)*.5+1)*(j&2?2:1)*(j&4?4:1);	

	int pixelsperline;
	pixelsperline = (int)((256*regs[0x24] + regs[0x25] - 256*regs[0x22] - regs[0x23])	    
			     *pixelbits/(dpi * 8));

	int mclkdiv = (int) ceil ((double) MCLKDIV_SCALING * 
					pixelsperline * 
					Hardware.merlininfo.Crystal_Frequency
					/((double) 8. * linend * PCTransferRate));

	Hardware.NSCStatusOut("2. mclkdiv before limit=%f",(double)mclkdiv/MCLKDIV_SCALING);

	mclkdiv = max(mclkdiv,minmclk);
	mclkdiv = min(mclkdiv,maxmclk);

	if (PCTransferRate == 2000000)
	{
		{
			int mult,timeadj;
			mult = timeadj =Hardware.merlininfo.timeadj;
			if (!mult) mult++;
			while (mclkdiv*dpi*mult < 6.*MCLKDIV_SCALING) 
			{
				mclkdiv++; //for now
			}
			Hardware.merlininfo.timeadj = regs[0x19] = timeadj;		
		}
		Hardware.NSCStatusOut("PCrate mclkdiv=%f",(double)mclkdiv/MCLKDIV_SCALING);
		return mclkdiv; //computing rate
	}

	// write scan info to status window
	Hardware.NSCStatusOut("2. mclkdiv=%f",(double)mclkdiv/MCLKDIV_SCALING);
	Hardware.NSCStatusOut("2. pixel bytes per line=%d",pixelsperline*r/8);
	Hardware.NSCStatusOut("2. linewidth=%d",linend);


// 3. calculate max mclk divider based on max integration time
// max mclkdiv = (clock in * max integration time)/(r * line width)
// *2 and round down
// limit to min,max

	int max_mclkdiv = (int) floor((double) MCLKDIV_SCALING * 
					Hardware.merlininfo.Crystal_Frequency * 
					Hardware.merlininfo.Maximum_Integration_Time_Lowlamp
					/((double) 1000. * r * linend));

	Hardware.NSCStatusOut("3. max mclkdiv before limit=%f",(double)max_mclkdiv/MCLKDIV_SCALING);

	max_mclkdiv = max(max_mclkdiv,minmclk);
	max_mclkdiv = min(max_mclkdiv,maxmclk);

	Hardware.NSCStatusOut("3. max mclkdiv=%f",(double)max_mclkdiv/MCLKDIV_SCALING);

//4. calculate min mclk divider based on max motor speed
//   min mclkdiv = clock in/(stepsize*max motor speed*r*step_per_inch*4)
//   *2 and round up
//   limit to min,max

	int min_mclkdiv = (int) ceil((double) MCLKDIV_SCALING * 
				    Hardware.merlininfo.Crystal_Frequency/
				    ((double) stepsize * 
					Hardware.merlininfo.Max_Speed *
					r * Hardware.merlininfo.Motor_Steps_Per_Inch * 4.));

	Hardware.NSCStatusOut("4. min mclkdiv before limit=%f",(double)min_mclkdiv/MCLKDIV_SCALING);

	min_mclkdiv = max(min_mclkdiv,minmclk);
	min_mclkdiv = min(min_mclkdiv,maxmclk);

	Hardware.NSCStatusOut("4. min mclkdiv=%f",(double)min_mclkdiv/MCLKDIV_SCALING);

//5.Check if stepsize must be adjusted because max integration time gives too
//fast motor speed

	if (min_mclkdiv > max_mclkdiv)
	{
//	    mclkdiv = min(mclkdiv,max_mclkdiv); //limit by max integration time
		mclkdiv = max_mclkdiv;
		min_mclkdiv = mclkdiv;				//use below
	    double x = ((double)Hardware.merlininfo.Crystal_Frequency/
			((double)Hardware.merlininfo.Motor_Steps_Per_Inch * 4. *
			 Hardware.merlininfo.Max_Speed * r * mclkdiv/MCLKDIV_SCALING));
		stepsize = (int) ceil(x);
		Hardware.LimitStepSize(stepsize,regs);
	    Hardware.NSCStatusOut("5. step size for mclk calc=%d",stepsize);
		int tr = Hardware.ComputeTR(regs); // recompute for this step size
	    scaninfo.actualyres = stepsize * 4 * Hardware.merlininfo.Motor_Steps_Per_Inch/tr;
	    ScanParms.yres = scaninfo.actualyres;

	    Hardware.NSCStatusOut("5. vres=%d",scaninfo.actualyres);
	}
	//else limited by motor speed and integration time
	{
		int org_mclkdiv = mclkdiv;
		int timeadj = Hardware.merlininfo.timeadj;
		int org_timeadj = timeadj;
		int mult = timeadj;
		if (!mult) mult++;
		while ((mclkdiv*dpi*mult < 6.*MCLKDIV_SCALING) ||
			   (mclkdiv > max_mclkdiv) ||
			   (mclkdiv*(timeadj+1) < org_mclkdiv*(org_timeadj+1)))
		{
			if (!timeadj && (mclkdiv < max_mclkdiv))
				mclkdiv++;
			else
			{
				mclkdiv = min_mclkdiv;
				timeadj = ++mult;
			}
		}
		Hardware.merlininfo.timeadj = regs[0x19] = timeadj;		
		Hardware.NSCStatusOut("5. time adj=%d",timeadj);
		double x = stepsize;
		if (timeadj)
			x = x * (timeadj+1)/(double)timeadj;
		stepsize = (int) ceil(x);
	}
	mclkdiv = max(mclkdiv,min_mclkdiv);
	Hardware.NSCStatusOut("5. final mclkdiv=%f",(double)mclkdiv/MCLKDIV_SCALING);

	Hardware.LimitStepSize(stepsize,regs);
	Hardware.NSCStatusOut("5. step size for mclk calc=%d",stepsize);

	if (Hardware.merlininfo.timeadj > Hardware.merlininfo.Maximum_IntTimeAdj)  
	{
		if (Hardware.merlininfo.Color_Mode == 0)
		{
			Hardware.merlininfo.Color_Mode = 1;
			ScanParms.linerate = 1;
			regs[0x26] |= 1;
			Hardware.merlininfo.timeadj = regs[0x19] = 0;		
			return -1; //force calculation
		}
	}

	return (mclkdiv);
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	DoScan(char *filename)

// DESCRIPTION	:	perform a scan 
//			all merlin registers except clock divider and fast feed
//			step size must be set before calling doscan.

// INPUTS	:
//			filename	- pxl file name
//					  false to not write

// RETURNS	:	true if ok, false if error
//--------------------------------------------------------
BOOL CScan::DoScan(char *filename)
{
	BOOL done = FALSE;
	strcpy(m_scanfilename,filename);

	while (!done)
	{
		if (!ScanSome(&done,FALSE/*,a*/)) return FALSE;
	}

	return TRUE;
}

static BYTE *pBMPBuf = NULL;
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ScanSome(BOOL *done)

// DESCRIPTION	:	perform a scan of at least 20 lines
//			all merlin registers except clock divider and fast feed
//			step size must be set before calling doscan.

// INPUTS	:
// OUTPUTS	: done flag

// RETURNS	:	true if ok, false if error
//--------------------------------------------------------
BOOL CScan::ScanSome(BOOL *done,BOOL gCancel/*,int a*/)
{
static int i,j;		// misc ints
static int multiccd;		// dummy var needed for historical purposes
static BYTE value;		// value read/write from/to merlin
static char sBuf[MAX_PATH];	// temp pxl file if no file specified
static int fh;
static int pauselimit; //pauselimit
static int length;
static BYTE *pBuf;
static BYTE *bufptr;
static int numInBuf;
static int bytesmoved;
static int read_bufsum;
static int linestoscan;
static int bytestoskip;   
static int linesread;
static int bytesread;
static BOOL bError;
static int counter;
static int no_read_ct;
static BOOL scanisdone;
static int initiallinesread;
static BOOL skipmem;
static BOOL pause;
static int maxtoread;
static int bytesavail;
static int wStepCounter;

#define READ_BUF_SIZE (256*1024)
static BYTE read_buf[READ_BUF_SIZE],*read_bufstart,*read_bufend;

	char filename[MAX_PATH];
	strcpy(filename,m_scanfilename);  
	
	m_callback = FALSE;

	if (m_firstscan) //first time
	{
	if (gCancel)
	{
		*done = TRUE; //already done
		return TRUE;
	}
	m_firstscan = FALSE;
	fh = NULL;		// file handle for pxl file
	pBuf = NULL;
	pBMPBuf = NULL;
	read_bufstart=read_buf;
	read_bufend=read_buf;

	// init line ct for BMP
	scaninfo.initialized = FALSE;
	scaninfo.xInterpolate = m_currentScanSettings.hInterp; //no y interpolate??
	scaninfo.yInterpolate = m_currentScanSettings.vInterp;

	// init scanning time
	ScanParms.time = 0;

	// calculate bytes/line
	if (!CalcBytesPerLine()) return FALSE;
	// set clock divider/fast feed step size
	if (!SetScanParms()) return FALSE;

	// the scanned data will be saved in a file, 
	// get the name of the file to save
	if (strlen(filename) == 0) {
		strcpy(sBuf,"");
		GetTempFileName(".","",0,sBuf);	// no file specified, so make one up
		strcpy(filename,sBuf);
		if (strlen(filename) == 0)
			strcpy(filename,"nsscan.tmp");
	}

	if (!SubString(filename,"compute."))
	{
		// get a buffer to store the scanned data
		// if scanning to memory, allocate and integer number of lines
#define MIN_MEMBUFSIZE (3*1024*1024)
		if (SubString(filename,"finalscn."))
			length = max(MIN_MEMBUFSIZE,Hardware.m_membufSize);
		else
			length = MIN_MEMBUFSIZE;
		pBuf = new BYTE[length];
		if (!pBuf) 
		{
			Hardware.NSCAfxMessageBox("Insufficient memory!");
			return FALSE;
		}
	}
	bufptr = pBuf;
	numInBuf = 0;	// number of bytes currently in scan buffer
	bytesmoved = 0;
	read_bufsum = 0;
	linestoscan = ScanParms.lines;
	if (ScanParms.linerate&1)
		linestoscan *= 3;

//	m_currentScanInfo.totalLines = linestoscan;

//------------------------------------
	bytestoskip = 0;
	//get info for determining transfer buffer size
	//get pauselimit
	if (!Hardware.ReadRegister(0x4e,&value)) goto iofailmsg;
	pauselimit = Hardware.SizeOfKCount(value);

	if (!Hardware.WriteRegister(0x4f,value - 1))
		goto iofailmsg; 

	linesread = 0;	// number of lines read
	bytesread = 0;	// number of bytes read

//	m_currentScanInfo.actBytes = 0;

	bError = FALSE;	// error indicator
	counter = 0;	// value indicating we need to check the msg loop

	// ask caller if cancel scan has been pressed
	//setup for fast read (restore at end)

	no_read_ct = 0;
	scanisdone = FALSE;
	m_reentry = FALSE;
	*done = FALSE;

	if (SubString(filename,"finalscn.") || SubString(filename,"compute."))
	{
		//init for transfer, and do sure all the lines are there
		scaninfo.initialized = FALSE;
		if (!ScantoBMPfile(filename))
			return FALSE;
		if (ScanParms.linerate&1)
			scaninfo.scanlines *= 3;
		linestoscan = ScanParms.lines = scaninfo.scanlines; //get lines to scan
		Hardware.NSCStatusOut("actual lines to scan = %d", linestoscan);

		if (SubString(filename,"compute.")) return TRUE;

		pBMPBuf = new BYTE[scaninfo.maxBMPLinesPerCall*scaninfo.bmpbytesperline];
		if (pBMPBuf == NULL)
		{
			Hardware.NSCAfxMessageBox("Insufficient memory!");
			return FALSE;
		}
	}

	wStepCounter = 0;
	if (!SubString(filename,"pcrate.") &&
		!Hardware.LM9831())
	{
		int regs[0x80];
		if (!Hardware.ReadMerlin(regs)) goto iofailmsg;
		int tr = Hardware.ComputeTR(regs,FALSE);
		int stepsize = 256 * regs[0x46] + regs[0x47];
		//write line count to 4c,4d for autostop
		//must have at least 16 extra bytes left in scan buffer
		//and at least one extra line
		int extralines = (int)ceil(16.0/ScanParms.bytesperline)+1;
		int totallines = linestoscan+extralines;
		double x = stepsize;
//		if (Hardware.merlininfo.timeadj)
//			x = (floor) (x * Hardware.merlininfo.timeadj / (Hardware.merlininfo.timeadj+1));
		wStepCounter = (int)ceil((double)totallines*tr/(4*x)); 

		if (wStepCounter && (wStepCounter < 0x4000)) 
		{
			regs[0x4c] = wStepCounter/256;
			regs[0x4d] = wStepCounter&0xff; 

			// Compute stepper phase correction
			ComputeStepperPhaseCorrection(regs,tr);
			if (!Hardware.WriteMerlin(regs)) goto iofailmsg;
		}
		else wStepCounter = 0;
	}

		// open the file to store the scan lines
		DeleteFile((LPCSTR)filename);
		fh = _open((LPCSTR)filename,_O_RDWR|_O_CREAT|_O_BINARY,_S_IREAD|_S_IWRITE);
		if (fh == -1)
		{
			wsprintf(sBuf,"Can't open file: %s",(LPCSTR)filename);
			Hardware.NSCAfxMessageBox(sBuf);
			fh = NULL;
			goto failmsg;
		}

		// the PXL file contains a few parameters so the scan line to 
		// dib converter knows how to convert the scan lines to a dib
		strcpy(sBuf,"v2\n");
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) {
writeerr:
				wsprintf(sBuf,"Can't write file: %s",(LPCSTR)filename);
				Hardware.NSCAfxMessageBox(sBuf);
				goto failmsg;
		}
		wsprintf(sBuf, "%d\n", ScanParms.pixelsperline);
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) goto writeerr;
		sprintf(sBuf, "%.2f\n", 0.0);
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) goto writeerr;
/////////////////////

		// this parameter is not used in the conversion and is stored for
		// historical purposes
		multiccd = 0;
		wsprintf(sBuf, "%d\n", multiccd);
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) goto writeerr;
	
		wsprintf(sBuf, "%d\n", ScanParms.linerate);
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) goto writeerr;
	
		int pixelpack = 0;
		switch (ScanParms.bpp)
		{
		case 16: pixelpack = 4; break;
		case 8: pixelpack = 3; break;;
		case 4: pixelpack = 2; break;
		case 2: pixelpack = 1; break;
		default: pixelpack = 0; break;
		}
		wsprintf(sBuf, "%d\n", pixelpack);
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) goto writeerr;
	
		wsprintf(sBuf, "%d\n", linestoscan);
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) goto writeerr;
	
		wsprintf(sBuf, "%d\n", ScanParms.xres); // dpi in x
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) goto writeerr;
	
		wsprintf(sBuf, "%d\n", ScanParms.yres); // dpi in y
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) goto writeerr;
	
		wsprintf(sBuf, "%d\n", 2); // status bytes
		if (_write(fh, sBuf, strlen(sBuf)) != (int)strlen(sBuf)) goto writeerr;

	//disable home during scan
	if (!Hardware.DisableStopScan()) goto iofailmsg;
	return TRUE;
	}
	else if (gCancel) 
	{
		goto exit;
	}
	else if (m_reentry) 
	{
		goto reentry;
	}
/////

	if (!bytesread)
	{
		//test for request size valid
		if (SubString(filename,"finalscn."))
			if (ScanReadParms.scanReadReqCount < 
				scaninfo.bmpbytesperline)
		{
			Hardware.NSCAfxMessageBox("Too many pixels in image line");
			goto failmsg;
		}		//reset
		BYTE oldkbytes;
		oldkbytes = 0;
		if (Hardware.LM9831())
		{

			// flush cache
			int toread;
			if (!Hardware.ReadRegister(1,&value)) goto iofailmsg;
			toread = Hardware.SizeOfKCount(value);
			if (!Hardware.ReadData(0,bufptr,toread)) goto iofailmsg; //flush all
			if (!Hardware.ReadRegister(1,&oldkbytes)) goto iofailmsg;
		}
		{
			if (Hardware.m_debug) Hardware.NSCStatusOut("filename=%s",filename);
			if (SubString(filename,"finalscn.")) //for calib verify
			{
				//load offset/gain again
 				OffsetGainToMerlin("WhiteFine");
				//load gamma
				int bpp = m_currentScanSettings.bitsPerPixel;
				if (!(bpp%3)) bpp/=3;
				if (!Hardware.OnStartupSendGamma(bpp)) goto failmsg;
		        Hardware.NSCAfxMessageBox("Scan"); //status message
				int regs[0x80];	
				if (!Hardware.ReadMerlin(regs)) goto iofailmsg;
				for (int i = 7; i < 0x60; i++) 
					Hardware.NSCStatusOut("%x\t%x",i,regs[i]);
			}
//display registers for debug
			else if (Hardware.m_debug) 
			{
				int regs[0x80];
				if (!Hardware.ReadMerlin(regs)) goto iofailmsg;
				for (int i = 7; i < 0x60; i++) 
					Hardware.NSCStatusOut("%x\t%x",i,regs[i]);
			}
		}
		//start scan
		if (!Hardware.WriteRegister(7, 0)) goto iofailmsg;

		if (!Hardware.WriteRegister(7, 3))
		{
iofailmsg:
failmsg:
			bError = TRUE;
			goto exit;
		}
		if (SubString(filename,"finalscn."))
		{
			m_currentScanInfo.scanStartTime = 
				m_currentScanInfo.scanEndTime = GetTickCount();
		}
		BYTE data[3];
		if (!wStepCounter) 
			while (1)
		{
			//wait for new data
			if (!Hardware.ReadData(1,data,3)) goto iofailmsg;
			value = data[1];
			if ((data[0] != value) && (value != data[2])) continue;
			if (value > oldkbytes) break;
 			if (!Hardware.ReadRegister(3,&value)) goto iofailmsg;
			if (value&0x10) break; //pause
			if (!Hardware.ReadRegister(7,&value)) goto iofailmsg;
			if (!value) break; //stopped
		}
	}

	initiallinesread = linesread;
	while (!scanisdone)
	{
		//update display value and do callback
		long time;
		time = GetTickCount();
		if (m_currentScanInfo.endTime)
			m_currentScanInfo.endTime = time;
		if (m_currentScanInfo.scanEndTime) 
			m_currentScanInfo.scanEndTime = time;
		if (SubString(filename,"finalscn."))
			m_currentScanInfo.rawActBytes = bytesread;

		m_callback = TRUE;
		int res;
		res = Hardware.Callback(TRUE);
		if (!m_callback) return FALSE; //done
		m_callback = FALSE;
		if (res) goto failmsg; //new error

		//check k bytes if 9831 and 16 bit scan
		if (Hardware.LM9831() && (ScanParms.bpp == 16))
		{
		// check number of k bytes available in buffer
		while (1) //wait for good data which is > 1
		{
			{
				// read 3 values to get 2 same
				BYTE data[3];
				if (!Hardware.ReadData(1,data,3)) goto iofailmsg;
				value = data[1];
				if ((data[0] == value) || (value == data[2]))
				{
					if (value > 0)
					{
						//check if cal, then stop scan
						if (SubString(filename,"finalscn.") || 
							SubString(filename,"pcrate."))
							break;
						if (!Hardware.WriteRegister(0x4f,0)) //dont start again
							bError = TRUE;
						break;
					}
					else
					{
						//be sure resume
						if (!Hardware.WriteRegister(0x4f,1)) //continue
						{					
							bError = TRUE;
							break;
						} 
					}
				}
			}
		}
		}
		else
		{
			value = Hardware.KCountFromSize(4*64*1024); //always read 4*64K
		}
		if (bError) goto iofailmsg;
		if (!ScanParms.time)
		{
			ScanParms.time = GetTickCount(); //start scanning
		}
		bytesavail = Hardware.SizeOfKCount(value - 1); 
		bytesavail = min(bytesavail,READ_BUF_SIZE);
		int bytesleft;
		bytesleft = linestoscan*ScanParms.bytesperline - bytesread;
		if (!bytesleft) 
		{
			if (!SubString(filename,"finalscn.") ||
				m_currentScanInfo.actBytes >= m_currentScanInfo.totalBytes)
					scanisdone = TRUE;
			while(wStepCounter)
			{
				BYTE value;
				if (!Hardware.ReadRegister(0x7,&value)) 
				{
					bError = TRUE;
					break;
				}
				if (!value)
				{
					break;	//done
				}
				if (!Hardware.ReadRegister(0x3,&value)) 
				{
					bError = TRUE;
					break;
				}
				if (value&0x10)
				{
					//this should not happen, but just in case
					Hardware.NSCStatusOut("Scan Pause!! for %s",filename);
					break;  //pause
				}
			}

			if (!Hardware.WriteRegister(7,0))
			{
				bError = TRUE;
				break;				
			} 

			if (wStepCounter) 
			{
				Hardware.NSCStatusOut("end scan for %s",filename);
				Hardware.ReadRegister(1,&value);//test!!
				Hardware.NSCStatusOut("reg 1=%d",value);
				Hardware.NSCStatusOut("lines read=%d,lines left=%d",
					bytesread/ScanParms.bytesperline,
					Hardware.SizeOfKCount(value)/ScanParms.bytesperline);
			}

			// restore paper sensor stop scan
			if (!Hardware.EnableStopScan())
			{
				bError = TRUE;
				break;				
			}
			if (SubString(filename,"finalscn."))
			{
				if (ScanParms.autohome && !Hardware.HomeSensor(FALSE))
				{
					bError = TRUE;
					break;				
				}
				PositionTest();
			}
		}
		bytesavail = min(bytesavail, bytesleft);
		// when the buffer contains some data, read it
		while (bytesavail > bytestoskip) //was 1024*hwinfo.minbufferlimit) 
		{   // have some data to read
			// after writing address 0 the first time, all following data reads
			// will be read from address 0
			maxtoread = min(length - numInBuf,
									bytesavail); // was -1024*hwinfo.minbufferlimit);
			maxtoread = min(maxtoread,pBuf+length - bufptr);
			if (maxtoread == 0) break; 

//			// read a buffer full of data
			int read_bufsize;
			pause = FALSE;
			int totalct;
			totalct = 0;
			while (maxtoread > (read_bufsize = (read_bufend - read_bufstart))) 
			{
				//move rest of read buffer down before read new
				memmove(read_buf,read_bufstart,read_bufsize);
				read_bufstart = read_buf;
				read_bufend = read_buf + read_bufsize;
			
				// read a buffer full of data
				int bytestoread = min(bytesavail,
									READ_BUF_SIZE-read_bufsize-256); //256 status bytes
				bytestoread = min(bytestoread,maxtoread);
				if (bytestoread == 0) break;
				
				bytestoread += bytestoskip; //read skip bytes, skip later

				if (!Hardware.ReadData(0,read_bufend,bytestoread))
				{
					bError = TRUE;
					break;
				}

				if (Hardware.m_debug) if (!bytesread) 
					if (SubString(filename,"finalscn."))
				{
					UINT average[3],max[3],min[3],number=0;
					for (int j=0;j<3;j++)
					{
						average[j]=max[j]=0;
						min[j] = 0xffff;
					}
					for (int i=0;i < ScanParms.bytesperline;)
					{
//						if (!i) Hardware.NSCStatusOut("8 bit scan=ct,color,value");
						for (int j=0;j<3;j++)
						{
							UINT value=0;
							if (ScanParms.bpp == 16)
								value = read_bufend[i++]*256;
							value += read_bufend[i++];
							average[j] += value;
							if (value > max[j]) max[j] = value;
							if (value > 0x80) if (value < min[j]) min[j] = value;
						}
						number++;
					}
					for (j=0;j<3;j++) average[j]/=number;
					Hardware.NSCStatusOut("average rgb first line=%x,%x,%x",
						average[0],average[1],average[2]);
					Hardware.NSCStatusOut("max rgb first line=%x,%x,%x",
						max[0],max[1],max[2]);
					Hardware.NSCStatusOut("min rgb first line=%x,%x,%x",
						min[0],min[1],min[2]);
				}

				//remove dummy bytes
				if (bytestoskip) 
				{
					bytestoread -= bytestoskip;
					memmove(read_bufend,read_bufend+bytestoskip,bytestoread);
					bytestoskip = 0;
				}

				read_bufend +=bytestoread;
				read_bufsum +=bytestoread;
				bytesread   +=bytestoread;
				totalct +=bytestoread;
				if (pause) break;
			}
	
			//return data from buffer
			memcpy(bufptr,read_bufstart,totalct);
			bufptr += totalct;
			read_bufstart +=totalct;
		
			numInBuf += totalct;//
			bytesavail -=totalct;//
			bytesmoved += totalct; //for debug

			if (maxtoread > (read_bufend - (read_bufstart-totalct))) break; //must have paused


			// when the buffer is full, write the buffer to the disk file
			//number of lines
			i = bytesread/ScanParms.bytesperline;

			if (i != linesread) 
			{ // got a change in the number of lines read
				linesread = i;
		
				if (
					((SubString(filename,"pcrate.")) && ((GetTickCount() - ScanParms.time)	> 1000)))
				{
					scanisdone = TRUE;
					{
						ScanParms.time = GetTickCount() - ScanParms.time;
						if (SubString(filename,"finalscn."))
						{
								m_currentScanInfo.scanEndTime = 
									m_currentScanInfo.endTime = GetTickCount();
						}
					}
				}
			}
		}
		{
			skipmem = (!SubString(filename,"finalscn.") && 
							!SubString(filename,"pcrate."));
			if ((numInBuf >= length-Hardware.SizeOfKCount(255)) || scanisdone || skipmem || pause || !bytesleft)
			{
				if (numInBuf > 0) 
				{ //new
					if (SubString(filename,"finalscn."))
					{
						scaninfo.pScanlinebuffer = pBuf;
						m_reentry = FALSE;//TRUE; //exit first time for updates
						while (numInBuf >= ScanParms.bytesperline)
						{
							if (!m_reentry) 
							{
								if (!ScantoBMPfile(filename))
								{
									if (m_currentScanInfo.actBytes < m_currentScanInfo.totalBytes)
										m_reentry = TRUE;
								}
								if (!scaninfo.returnbytecount)	//done with scan line
								{
									if (Hardware.m_debug)
										_write(fh, scaninfo.pScanlinebuffer, ScanParms.bytesperline); 
									numInBuf -= ScanParms.bytesperline;
									scaninfo.pScanlinebuffer += ScanParms.bytesperline;
								}
							}
							if (m_reentry)
							{
								*done = FALSE;
								return TRUE;
							}
reentry:
							m_reentry = FALSE;
							if (scanisdone) break;
						}
						//move any reminder down
						if (numInBuf)
							memmove(pBuf,scaninfo.pScanlinebuffer,numInBuf);
						bufptr = pBuf+numInBuf;
					}
					else
					{
						if (!SubString(filename,"pcrate."))
							if (_write(fh, (LPSTR)pBuf,numInBuf) != numInBuf)
						{
							Hardware.NSCAfxMessageBox("Error writing file!");
							bError = TRUE;
							break;
						}
						numInBuf = 0;
						bufptr = pBuf;
					}
				}
				if (skipmem) break;
			}
			if (bError) break;
			if (scanisdone) break;

		} // while there are bytes to read
		// check msg loop
	} // while scanning
	if (!scanisdone && !bError) 
	{
		*done = FALSE;
		return TRUE;
	}
exit:
	Scan2Bmp.generate_bmp_line(&scaninfo,TRUE); //close bmp memory
	if (fh) _close(fh);	// close pxl file
	fh = NULL;
	if (pBuf) delete pBuf;
	pBuf = NULL;
	if (pBMPBuf) delete pBMPBuf;
	pBMPBuf = NULL;

	ScanParms.bytesread = bytesread;
	
	if (bError || gCancel)
	{
		Hardware.WriteRegister(7,0);// goto iofailmsg;

		// restore paper sensor stop scan
		Hardware.EnableStopScan();// goto iofailmsg;

		if (ScanParms.autohome) {
			Hardware.HomeSensor();
		} // if auto home checked
	}

	if (SubString(filename,"finalscn."))
	{
		if (!bError) Hardware.NSCAfxMessageBox("Idle"); //status message
	}
	m_firstscan = TRUE;	
	*done = TRUE;
	return !bError;	
}

////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SetScanParms()

// DESCRIPTION	:	compute the clock divider and fast feed step size

// INPUTS	:
// RETURNS	:	TRUE if ok, FALSE if error
//--------------------------------------------------------
BOOL CScan::SetScanParms()
{
BYTE value;	// value written to merlin register
int iValue;	// clock divider value
int j;		// read register return value

//-- clear bit so sensor is skipped so scan will not stop
// need to use merlininfo.Home_Sensor to determine which bit to use

	if (!Hardware.DisableStopScan())
	{
ioerr:
		return FALSE;
	}

	//-- set clock divider
	int mclk = Hardware.merlininfo.mclk;
	iValue = Hardware.ComputeReg8Value(mclk);
	if (iValue >= 0) {
		j = !Hardware.ReadRegister(0x26,&value);
		if (j != 0) goto ioerr;
		int r = 8; // line rate
		if ((value & 7) == 0) r = 24; // pixel rate

		//--- calculate the fast feed step size
		double x = Hardware.merlininfo.Crystal_Frequency
					/((double)Hardware.merlininfo.mclk/MCLKDIV_SCALING);
		x /= Hardware.merlininfo.Max_Speed * 4
				* Hardware.merlininfo.Motor_Steps_Per_Inch * r;
		if (Hardware.merlininfo.timeadj)
			x /= (Hardware.merlininfo.timeadj);
		if (Hardware.m_debug) 
			Hardware.NSCStatusOut("setting fast feed step size=%.0f",x);
		value = (BYTE) (x/256);
		j = !Hardware.WriteRegister(0x48,value); 
		if (j != 0) goto ioerr;
		value = (BYTE) x;
		j = !Hardware.WriteRegister(0x49,value); 
		if (j != 0) goto ioerr;
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	DoCalibration()

// DESCRIPTION	:	do calibration

// RETURNS	:	TRUE if success, FALSE if fails
//--------------------------------------------------------
BOOL CScan::DoCalibration(int reg9)
{

    char type[30];
	Hardware.NSCAfxMessageBox("Calibration"); //status message
	if (Hardware.m_debug) Hardware.NSCStatusOut("Start Docalibration");

	// allocate mem for register settings
	int regs[0x80]; 

	if (!DoCoarseCalibration(reg9,regs)) 
		return FALSE;

	strcpy(type,"BlackFine");
	if (Hardware.merlininfo.Black_Fine_Calibration)
	{
		if (!DoFineCalibration(reg9,regs,type))
			return FALSE;
		blackFine = TRUE;
	}
	else blackFine = FALSE;
    OffsetGainToMerlin(type);

	strcpy(type,"WhiteFine");
	if (!DoFineCalibration(reg9,regs,type))
		return FALSE;
    OffsetGainToMerlin(type);

	if (!Hardware.m_debug) if (!Hardware.HomeSensor()) return FALSE;

	if (Hardware.m_debug) Hardware.NSCStatusOut("End DoCalibration");
	return TRUE;
}


/////////////////////////////////////////////////////////////////////////////
//-----------------------------------------------------------------
// FUNCTION  :  DoCoarseCalibration(int reg9,int regs[0x80])

// DESCRIPTION: Setup paramenter for doing coarse calibration scan
//              Scan one line.  Determine the  static gain/offset 
//              value for each color so that the max pixel value near 
//              the target code and the min pixel value close to 
//              min target code

// RETURNS	:	TRUE if success
//------------------------------------------------------------------

BOOL CScan::DoCoarseCalibration(int reg9,int regs[0x80])
{
int i,j,k;		  // misc ints
BYTE value;		  // byte to write to parallel port
char sBuf[MAX_PATH];		// msg buffer
BOOL bHi[6];		// TRUE if no change to gain/offset coeff, FALSE otherwise
int GainHi[3];   // max static gain values reg 3B,3C,3D
int OffLow[3];   // min static offset values reg 38,39,3A
BOOL bReturn=FALSE;
BOOL bFirstTime;	// flag indicating whether this is the first coarse calibration scan
int maxval[3],minval[3]; // max/min pixel values for r,g,b
char type[30];     // type of the calibration

//	Hardware.NSCAfxMessageBox("Calibration"); //status message
	if (Hardware.m_debug) Hardware.NSCStatusOut("Start Docalibration");

	if (!Hardware.SetupStandardRegisters(regs,TRUE)) return FALSE;

	//------ set merlin regs portion
	regs[0x9] = reg9; //same divider and bias
	regs[0x9] |= 0x20; 	// not process data

	// optical black pixel start/end
	regs[0x1C] = regs[0x1d] = 0;
	if (Hardware.merlininfo.Optical_Black_Pixels_Start > 0)
		regs[0x1D] = Hardware.merlininfo.Optical_Black_Pixels_Start - 1;

	//do up to 1/2" pass end of image
	int linend = Hardware.merlininfo.first_white_pixel + 
				Hardware.merlininfo.Image_Pixel_Width +
				Hardware.merlininfo.Optical_Resolution/2;
	linend = min(linend,Hardware.merlininfo.Total_Pixels-20);
	regs[0x24] = linend/256;
	regs[0x25] = linend & 0xFF;

	//------- init static gain/offset
	i = 0;
	regs[0x38] = i; // init static offset
	regs[0x39] = i;
	regs[0x3A] = i;
	int min_gain = (int) ((Hardware.merlininfo.Minimum_Static_Gain - .933)/.067);
	regs[0x3B] = min_gain; // init static gain
	regs[0x3C] = min_gain;
	regs[0x3D] = min_gain;

	//------- set gain/offset mode
	{
		regs[0x3e] = 0;
		regs[0x3f] = 0;		// digital offset = 0
		regs[0x40] = 0x40;
		regs[0x41] = 0x00;	// digital gain = x1
		regs[0x42] &= ~7;  //use fixed offset and gain coeff
	}

	// compute tr for step size calculation
	ScanParms.yres = scaninfo.actualyres = 300;

	// calc the step size based on the specified resolution
	Hardware.ComputeStepSize(scaninfo.actualyres,regs);
	
//---------------------------------------------------------
	// write the computed reg values to merlin
	if (!Hardware.WriteMerlin(regs)) goto cleanup;

	if (!Hardware.HomeSensor()) goto cleanup;

	if (!Hardware.CheckScannerReady()) goto cleanup;

	// init the min/max values for the gain/offset for setting reg 38..3D
	for (i = 0; i < 3; i++) {
		GainHi[i] = 63;
		OffLow[i] = -31;
	}

	// setup the parameters for scanning
	// scan 1 line
	ScanParms.lines = 1;
	// if 3 channel line rate, read a line of each color
//	if ((regs[0x26]&3) == 1) ScanParms.lines *= 3;
	// don't scan any additional lines for line separation
//	ScanParms.lines = scaninfo.lines;
	// tell everybody what the scanning parameters are

	// init the scan parameters
	bFirstTime = TRUE;

//	ScanParms.datamode = 1;	// scan 10 bit data
	ScanParms.autohome = 0;	// don't home after scan

//this loop does the coarse calibration
	BOOL lampchange;
	lampchange = TRUE;
rescan:
	if (lampchange)
 		Hardware.delay(100); //stablize any lamp changes
	lampchange = FALSE;
	// make sure scanner is done with home and is ready to scan
	if (!Hardware.CheckScannerReady()) goto cleanup;

	wsprintf(sBuf,"COARSE.PXL");
	strcpy(m_scanfilename,sBuf);

	if (!CalcBytesPerLine()) goto cleanup;

	if (!DoScan(m_scanfilename)) goto cleanup;

	if (!Hardware.WriteRegister(0x45, regs[0x45] & ~0x10)) goto cleanup; //motor off
//	if (!Hardware.WriteRegister(0x4a,0)) goto cleanup; //no step forward
//	if (!Hardware.WriteRegister(0x4b,1)) goto cleanup; //no step forward
//	if (!Hardware.WriteRegister(0x51,0)) goto cleanup; //acceleration off

	// compute the line for this coarse calibration scan
   strcpy(type,"Coarse");
	i = 0;
	if (!ProcessCalibrationLines(m_scanfilename,type,&i,0)) goto cleanup;

    // read the user specified minimum target code
	int minTargetCode[3];
	minTargetCode[0] = Hardware.merlininfo.min_red_target;	
	minTargetCode[1] = Hardware.merlininfo.min_green_target;	
	minTargetCode[2] = Hardware.merlininfo.min_blue_target;


	// read the user specified target codes
	int targetcode[3];
	targetcode[0] = Hardware.merlininfo.red_target;	
	targetcode[1] = Hardware.merlininfo.green_target;	
	targetcode[2] = Hardware.merlininfo.blue_target;	

	// set min values read during this scan into an array for loop processing
	minval[0] = minBlackred;
	minval[1] = minBlackgreen;
	minval[2] = minBlackblue;
	maxval[0] = maxWhitered;
	maxval[1] = maxWhitegreen;
	maxval[2] = maxWhiteblue;


	// if this is the first coarse scan, compute the next gain/offset guess
	// for reg 38..3D
    // - static gain calculation (regs 3B-3D):
	//   Gain = 0.93 + 0.067*j         for b5=0 & j=0to31
	//     j  = (Gain - 0.93)/0.067 
	//        = 15 * (targetcode/(max pixel - min pixel) - .933)
    // - static offset calculation (regs 38-3A) 
    //   ADC:  Dout   = (Vin/Vref)*2^k
	//        -minval = (Vin/2)*2^16
	//   Offset = n * 0.0093 * gain = Vin     for b5=0 & n=0to31 
	//        n = (-minval/2^15) / (0.0093*gain)
	if (bFirstTime) {
		for (i = 0; i < 3; i++) { // compute gain/offset for r,g,b
			double gain = (double) targetcode[i]/(maxval[i]-minval[i]);
			gain = min(gain,9.); //max gain
			j = (int) (15.*(gain - 0.933));
            k = (int) (-1.*(double)minval[i]/(gain*32768.*0.0093));
			if (j > 31) {
				gain /= 3.;
				j = (int) (15.*(gain - 0.933));
				if (j > 31) j = 31;
				j += 0x20;
                k = (int) (-1.*(double)minval[i]/(gain*32768.*0.0093));
			}
			if (j < min_gain) j = min_gain;
			if (j > 63) j = 63;
			if (k < -31) k = -31;
			if (k > 31) k = 31;
			regs[0x3B+i] = j;
			if (!Hardware.WriteRegister(0x3B+i, j)) goto writeerr;
			if (k < 0) k = (-k & 0x1F) + 0x20;
			regs[0x38+i] = k;
			if (!Hardware.WriteRegister(0x38+i, k)) goto writeerr;								
		}
		bFirstTime = FALSE;
		goto rescan;
	}

//	bFirstTime = FALSE;


	// adjust gain based on max pixel read during coarse scan
    // so that the pixel values is close to the target codes
	for (i = 3; i < 6; i++) 
	{ // check max white for r,g,b
		bool FirstLamp;
		if (i == 3)	FirstLamp = true;
		bHi[i] = FALSE;
		// if the max pixel value > target code, decrement the gain setting
		if (maxval[i-3] > targetcode[i-3]) { // check if biggest color > target code
			value = regs[0x38+i];
			if (value > min_gain) {
				--value;
				GainHi[i-3] = value;
				regs[0x38+i] = value;
				if (!Hardware.WriteRegister(0x38+i, value)) goto writeerr;
			}
			else 
			{
				lampchange = TRUE;
				//for CIS scanner: if pixel value>targetcode & the adjusted gain<min_gain,
                //the period of the lamp on is decreased by 10%
				//for CCD scanner: the lamp PWM is reduced by 3%
				//subject to lower limit
				if (Hardware.merlininfo.Illumination_Mode & 2)
				{
					if (Hardware.m_debug)
						Hardware.NSCStatusOut("Exceed integration time, color=%x",i-3);
					int lampon;
					switch (i-3)
					{
					case 0: 
						lampon = (Hardware.merlininfo.Red_Lamp_Off - 
							Hardware.merlininfo.Red_Lamp_On);
						lampon = (lampon * 9)/10;
						Hardware.merlininfo.Red_Lamp_Off = 
							lampon + Hardware.merlininfo.Red_Lamp_On;
						if (!Hardware.WriteRegister(0x2e,
							Hardware.merlininfo.Red_Lamp_Off/256)) return FALSE;
						if (!Hardware.WriteRegister(0x2f,
							Hardware.merlininfo.Red_Lamp_Off&0xff)) return FALSE;
						break;
					case 1:
						lampon = (Hardware.merlininfo.Green_Lamp_Off - 
							Hardware.merlininfo.Green_Lamp_On);
						lampon = (lampon * 9)/10;
						Hardware.merlininfo.Green_Lamp_Off = 
							lampon + Hardware.merlininfo.Green_Lamp_On;
						if (!Hardware.WriteRegister(0x32,
							Hardware.merlininfo.Green_Lamp_Off/256)) return FALSE;
						if (!Hardware.WriteRegister(0x33,
							Hardware.merlininfo.Green_Lamp_Off&0xff)) return FALSE;
						break;
					case 2:
						lampon = (Hardware.merlininfo.Blue_Lamp_Off - 
							Hardware.merlininfo.Blue_Lamp_On);
						lampon = (lampon * 9)/10;
						Hardware.merlininfo.Blue_Lamp_Off = 
							lampon + Hardware.merlininfo.Blue_Lamp_On;
						if (!Hardware.WriteRegister(0x36,
							Hardware.merlininfo.Blue_Lamp_Off/256)) return FALSE;
						if (!Hardware.WriteRegister(0x37,
							Hardware.merlininfo.Blue_Lamp_Off&0xff)) return FALSE;
						break;
					}
				}
				else
				{
					//CCD
					if (!FirstLamp) continue;
					FirstLamp = false;
					Hardware.merlininfo.Green_PWM_Duty_Cycle = 
						(Hardware.merlininfo.Green_PWM_Duty_Cycle*90)/100;
					if (Hardware.merlininfo.Green_PWM_Duty_Cycle <
						Hardware.merlininfo.Green_PWM_Duty_Cycle_Low)
					{
						Hardware.NSCAfxMessageBox("Image too bright!");
						return FALSE;
					}
					if (!Hardware.WriteRegister(0x2a,
						Hardware.merlininfo.Green_PWM_Duty_Cycle/256)) return FALSE;
					if (!Hardware.WriteRegister(0x2b,
						Hardware.merlininfo.Green_PWM_Duty_Cycle&0xff)) return FALSE;
				}
			}
			continue; // to next color
		}
		else {
			// otherwise, the max pixel value < target code, so increment gain
			value = regs[0x38+i];			
			if (value < GainHi[i-3]) {
				++value;
				regs[0x38+i] = value;
				if (!Hardware.WriteRegister(0x38+i, value)) goto writeerr;
			}
			else if (value == 63)
			{
					Hardware.NSCAfxMessageBox("Image too dim!");
					return FALSE;			
			}
			else bHi[i] = TRUE;

		}
	} // for i = 3..5


	// after the gain is set, increment/decrement the offset based on the
	// min pixel value read during the previous scan so that
	// the max pixel value is close to the target code & the min
	// pixel value is close to the min target code
	for (i = 0; i < 3; i++) { // check min black for r,g,b
		bHi[i] = FALSE;
	
	if (minval[i] < minTargetCode[i] /*== 0*/) { // check if min black is 0
		// if the min pixel value = 0, increment the offset
			j = regs[0x38+i];
			if (j & 0x20) j = -(j & 0x1F);
			if (j < 31) {
				++j;
				if (bHi[i+3]) OffLow[i] = j; //if gain stable
				value = abs(j) & 0x1F;
				if (j < 0) value |= 0x20;
				regs[0x38+i] = value;

				if (!Hardware.WriteRegister(0x38+i, value)) {
writeerr:
					goto cleanup;
				}
			}
			else if (bHi[i+3]) {
				Hardware.NSCAfxMessageBox("Cannot adjust black level above zero!");
				return FALSE;
			}
			continue;
		}
		// if the min pixel value > 0, decrement offset
		j = regs[0x38+i];
		if (j & 0x20) j = -(j & 0x1F);
		if (j > OffLow[i]) {
			--j;
			value = abs(j) & 0x1F;
			if (j < 0) value |= 0x20;
			regs[0x38+i] = value;
			if (!Hardware.WriteRegister(0x38+i, value)) goto writeerr;
		}
		else if ((j == -31) && (bHi[i+3]))  {
			Hardware.NSCAfxMessageBox("Cannot adjust black level to zero!");
			return FALSE;
		}
		else bHi[i] = TRUE;
	} // for i = 0..2

	// if no error and gain/offset still changing for some color, rescan
		for (i = 0; i < 6; i++)
			if (!bHi[i]) goto rescan;

		bReturn = TRUE;

cleanup:
	// delete the pxl files if not saving reg files
	DeleteFile((LPCSTR)m_scanfilename);

	return bReturn;

}


/////////////////////////////////////////////////////////////////////////////
//-----------------------------------------------------------------
// FUNCTION  :  DoFineCalibration(int reg9,int regs[0x80], char type[])

// DESCRIPTION: Setup paramenter for doing fine calibration scan
//              Determine the gain/offset coefficient for each pixel
//              so that the pixel values near the target codes 
//              For type = blackFineCalibration, the offset coefficients
//              are determined when the LED is off and the gain coefficients
//              calculation is ignored

// RETURNS	:	TRUE if success
//------------------------------------------------------------------
BOOL CScan::DoFineCalibration(int reg9, int regs[0x80], char type[])
{

//int stepsize;		// calculated step size
int looplines;    // total lines to read 
int i;
BOOL bReturn=FALSE;


	if (!Hardware.SetupStandardRegisters(regs,TRUE)) return FALSE;	


    // set reg values for fine calibration

	{
		regs[0x9] = reg9; //same HDPI divider and bias
		regs[0x9] |= 0x20; 	// not process data 

        if (strncmp(type,"BlackFine",9)==0)
		   regs[0x42] &= ~7;  //use fixed coeff
		else
           regs[0x42] |= 4;   //use external DRAM offset

		CalculateScanRect(regs,"fine"); // calculate rectangle to scan
		//restore start of cal
		regs[0x4a] = Hardware.merlininfo.Start_of_Calibration/256;
		regs[0x4b] = Hardware.merlininfo.Start_of_Calibration&0xff;
	}

	//------- set step size
	// compute tr for step size calculation
	ScanParms.yres = scaninfo.actualyres = 300;

	// calc the step size based on the specified resolution
	Hardware.ComputeStepSize(scaninfo.actualyres,regs);
	int stepsize = 256 * regs[0x46] + regs[0x47];

	// write these regs to merlin
	if (!Hardware.WriteMerlin(regs)) goto cleanup;

	if (!Hardware.HomeSensor()) goto cleanup;

	// calculate the number of lines to scan
	int tr;
	tr = Hardware.ComputeTR(regs);
	int cm;
	cm = 1;
	if ((regs[0x26]&3) == 1) cm = 3;
	i = Hardware.merlininfo.Calibration_Strip_Height * stepsize * 4/tr;
	if (Hardware.merlininfo.timeadj)
		i = i * Hardware.merlininfo.timeadj/(Hardware.merlininfo.timeadj+1);
	if (i < 4) i = 4;
	looplines = i;	//how many total lines to read
	//compute lines to read w/o wrapping DRAM
	if (!CalcBytesPerLine()) goto cleanup;
 	if (Hardware.LM9831())
	{
		ScanParms.lines = 
			(Hardware.SizeOfKCount(regs[0x4e])+ScanParms.bytesperline-1)
			/(ScanParms.bytesperline*cm); //get how many lines before pause
	}
	else
	{
		ScanParms.lines = looplines;

	}

	// don't scan any extra lines for line separation
//	ScanParms.lines = scaninfo.lines;
	if (Hardware.m_debug) 
		Hardware.NSCStatusOut("fine lines/scan = %x",ScanParms.lines);

	// setup scan parms
	ScanParms.autohome = 0;	// dont home after scan
	DeleteFile((LPCSTR)m_scanfilename);

	if (strncmp(type,"WhiteFine",9)==0)
 	   strcpy(m_scanfilename,"FINE.PXL");
	else
 	   strcpy(m_scanfilename,"BlackFINE.PXL");

	static BYTE reg29;
	if (regs[0x29]) reg29 = regs[0x29]; //save on value
    if(strncmp(type,"BlackFine",9)==0)
	{
//       regs[0x45] &= ~0x10;   //motor off
	   regs[0x29] &= ~0x3;			//lamp off
	}
	else regs[0x29] = reg29;
	if (Hardware.m_debug) regs[0x45] &= ~0x10;   //motor off

	// write these regs to merlin
	if (!Hardware.WriteMerlin(regs)) goto cleanup;
	Hardware.delay(100); //allow for lamp change

    //The following tasks are repeated untill all the lines are scanned
    //Scan some lines, stop before pause, do fine calibration of those lines
	//(throw away the first 3 lines since the pixel values are not good)
	for (i = -3; i < looplines-3; )
	{
		if (!Hardware.CheckScannerReady()) goto cleanup;
		if (!CalcBytesPerLine()) goto cleanup;
		ScanParms.lines = min(ScanParms.lines,looplines -3 - i);

		// do a fine calibration scan
		if (!DoScan(m_scanfilename)) goto cleanup;
        if (!ProcessCalibrationLines(m_scanfilename,type,&i,looplines-3)) goto cleanup;
	}
 
	bReturn = TRUE;
cleanup:
	// delete the pxl files if not saving reg files
	DeleteFile((LPCSTR)m_scanfilename);
		return bReturn;
}

/////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------
// FUNCTION	:	ProcessCalibrationLines(char *filename, char type[],int *actuallyavg,int totalct
// DESCRIPTION:	 This routine reads the pixel values from the file
//               Averaging the pixel value is more than one lines are read 
//               type = Coarse  -> find the max & min pixel value
//                      and find the first white pixel 
//                    = WhiteFine -> call function to compute gain coefficient
//                    = ComputeNormal -> compute the max, min, and the total 
//                      number of pixel for pixel values < .5target code and
//                      pixel value > .5target code 

// RETURNS	:	nothing
//--------------------------------------------------------
BOOL CScan::ProcessCalibrationLines(char *filename, char type[],int *actuallyavg,int totalct) 
{
HFILE hFile = NULL;	// file handle for pxl file
int width;		// pixels per line
int j,k;		// misc ints
SCANINFO localscaninfo;	// local structure for reading pxl file info
int offset;		// length of pxl file header
double redtargetcode;	// target codes 
double greentargetcode;
double bluetargetcode;
int address;		// index   
BYTE *scanbuf = NULL;	// pxl file input buffer

	// open the file containing the scan line
	if ((hFile = open((LPCSTR)filename, _O_BINARY)) == -1) // open input file
	{
		Hardware.NSCAfxMessageBox("Error opening scan line file!");
		return FALSE;
	}

	// read the header info from the file
	if (!ReadScanParms(hFile,&localscaninfo,&offset)) {
		_close(hFile);
		return FALSE;
	}

	width = localscaninfo.width; // pixels per line

	if (!width) goto fileerror;

	// width is the number of pixels per line.
	// the assumption is that each line contains 2 bytes/color for each pixel
	// there are 2 status bytes at the end of each line
	int bytesperline;
	bytesperline = width;
	if (ScanParms.linerate == 0) bytesperline *= 3; // pixel rate
	if (ScanParms.bpp == 16) bytesperline *= 2;
	{
		//bytesperline must be truncated word boundary
		//add two status bytes
		bytesperline &= ~1;
		bytesperline += 2;
	}

			// allocate a buffer for the scan line
	if ((strncmp(type,"WhiteFine",9)) != 0)
	   scanbuf = new BYTE[bytesperline*3];
	else
       scanbuf = new BYTE[0x30000];

	if (!scanbuf) {
nomem:
		_close(hFile);
		Hardware.NSCAfxMessageBox("Insufficient memory");
		return FALSE;
	}

	// if the width of the current line > than lines previously graphed,
	// the graph routine will run out of data, so only graph the data where
	// values exist for all graphs
	//if (width > npts && npts > 0) width = npts;

	// make sure memory was really allocated
	if (!pRed || !pGreen || !pBlue) goto nomem;

	if (*actuallyavg <= 0)
	{ 
	// initialize each buffer to 0
	memset(pRed,0,width*sizeof(int));
	memset(pGreen,0,width*sizeof(int));
	memset(pBlue,0,width*sizeof(int));
	}

	// move the file pointer to the start of the scan line data
	if (_lseek(hFile,offset,SEEK_SET) < 0) {
fileerror:
		_close(hFile);
		if (scanbuf) delete scanbuf;
		Hardware.NSCAfxMessageBox("Error reading scan line file");
		return FALSE;
	}
	int maxlines = filelength(hFile) - offset;
	maxlines /= bytesperline;
	if (ScanParms.linerate == 1) {
		maxlines /= 3;
		maxlines *= 3;
	}

	// add up the pixels for each color in each scan line
	int color = 0;
	int multiplier = 1;
 
	redtargetcode   = Hardware.merlininfo.red_target;	
	greentargetcode = Hardware.merlininfo.green_target;	
	bluetargetcode  = Hardware.merlininfo.blue_target;

	if (Hardware.m_debug) Hardware.NSCStatusOut("linerate = %d",ScanParms.linerate);
	while (maxlines && _read(hFile, scanbuf, bytesperline) == bytesperline)
	{
		int l,m,n;
		int ct = *actuallyavg + 1;
		for (j = 0; j < width && ct>0; j++) {
			if (ScanParms.linerate == 0) { // pixel rate
				// there are 2 bytes for each color of each pixel
				l = (256*scanbuf[6*j] + scanbuf[6*j+1])*multiplier;
		        SumFineCalValue(&pRed[j],ct,l & 0xFFFF,type);
				m = (256*scanbuf[6*j+2] + scanbuf[6*j+3])*multiplier;
				SumFineCalValue(&pGreen[j],ct,m & 0xFFFF,type);
				n = (256*scanbuf[6*j+4] + scanbuf[6*j+5])*multiplier;
				SumFineCalValue(&pBlue[j],ct,n & 0xFFFF,type);		
if (Hardware.m_debug) if (j<10)
   Hardware.NSCStatusOut("ct,red,green,blue = %x, %x, %x, %x",j*2,l,m,n);
	
			}			
			  
			if (ScanParms.linerate == 1) { // line rate
				switch (color) {
					case 0:
						l = (256*scanbuf[2*j] + scanbuf[2*j+1])*multiplier;
						SumFineCalValue(&pRed[j],ct,l & 0xFFFF,type);
			            break;
					case 1:
						m = (256*scanbuf[2*j] + scanbuf[2*j+1])*multiplier;
						SumFineCalValue(&pGreen[j],ct,m & 0xFFFF,type);
						break;
					case 2:
						n = (256*scanbuf[2*j] + scanbuf[2*j+1])*multiplier;
						SumFineCalValue(&pBlue[j],ct,n & 0xFFFF,type);
if (Hardware.m_debug) if (j<10)
   Hardware.NSCStatusOut("ct,red,green,blue = %x, %x, %x, %x",j*2,l,m,n);
						break;

				} // switch color

			} // if line rate
			if (ScanParms.linerate == 4) { // 1 channel mode a
				// there are 2 bytes for each color of each pixel
				k = (256*scanbuf[2*j] + scanbuf[2*j+1])*multiplier;
				SumFineCalValue(&pRed[j],ct,k & 0xFFFF,type);
				SumFineCalValue(&pGreen[j],ct,k & 0xFFFF,type);
				SumFineCalValue(&pBlue[j],ct,k & 0xFFFF,type);
if (Hardware.m_debug) if (j<10)
   Hardware.NSCStatusOut("ct,red,green,blue = %x, %x, %x, %x",j*2,k,k,k);
			} // if 1 channel mode a

		    if ((strncmp(type,"BlackFine",9)==0) && (*actuallyavg == 0))
			   if ((pRed[j] > redtargetcode/4) && (pGreen[j] > greentargetcode/4) && (pBlue[j] > bluetargetcode/4)) 
			   {
				   *actuallyavg = totalct;
                   blackFine = FALSE; 
	               // close the file and remove temporary memory
	               _close (hFile);
	               hFile = NULL;
	               delete scanbuf;
	               scanbuf = NULL;
				   return TRUE;
				}

		
		} // for each pixel
		--maxlines;

		if (ScanParms.linerate == 0 || ScanParms.linerate == 4) {
			++(*actuallyavg);
		}
		else {
			++color;
			if (color > 2) {
				color = 0;
				++(*actuallyavg);
			}

		}
		if (Hardware.m_debug) Hardware.NSCStatusOut("line = %d",*actuallyavg);
    } // for each row

	_close (hFile);

	if ((strncmp(type,"WhiteFine",9) == 0) || (strncmp(type,"BlackFine",9) == 0))
	{ 
		//if there are still lines to be read, continue reading the them
		if(*actuallyavg != totalct)  
	   {
		   return TRUE; //use outer loop
	   }
	}
	// check if the lines need averaging
	if (*actuallyavg > 1)
		for (j = 0; j < width; j++) {
			pRed[j] /= *actuallyavg;
			pGreen[j] /= *actuallyavg;
			pBlue[j] /= *actuallyavg;   
		}
  
     if ((strncmp(type,"Coarse",6)) == 0)
	 {
     // init the min/max pixel values
	     maxWhitered = 0;
	     maxWhitegreen = 0;
	     maxWhiteblue = 0;
	     minBlackred = 0xFFFF;
	     minBlackgreen = 0xFFFF;
	     minBlackblue = 0xFFFF;


	// find the first white pixel. first find the first pixel
	// greater than .2*targetcode. then find the first pixel where the
	// slope is less than .05*targetcode.

	     int minval = 0;
         double targetCode[3];
		 for (int i=0; i<width; i++)
		 {
             pColor[0][i] = pRed[i];
             pColor[1][i] = pGreen[i];
			 pColor[2][i] = pBlue[i];
		 }


		 // get the target codes specified on the calibration window
         targetCode[0] = Hardware.merlininfo.red_target;
         targetCode[1] = Hardware.merlininfo.green_target;
         targetCode[2] = Hardware.merlininfo.blue_target;

	     //use slope offset
	     int slope_offset = (int) ((12/ScanParms.dpi)+.1);
	     // init min value of each color
	     int minwhite[3];
	     for (i=0; i < 3; i++) minwhite[i] = slope_offset;


	     // find first pixel for each color > .2*targetcode & the slope<.05
         for(i = 0; i<3; i++)
		 {
            for (address = slope_offset; address < width; address++) 
			{
		      j = pColor[i][address] - pColor[i][address-slope_offset];
			  if ( (j > 0) && (j < (int) (targetCode[i]/20)) && 
					(pColor[i][address] > (int) (targetCode[i]/5)) ) 
			  {
			     minwhite[i] = address;
			     break;
			  }
			}
		 }


	  // the first pixel is the max which meets the above conditions
	     minval = max(minwhite[0],minwhite[1]);
	     minval = max(minval,minwhite[2]);

		 Hardware.merlininfo.first_white_pixel = 
		    minval*Hardware.merlininfo.Optical_Resolution/localscaninfo.actualxres +
		    Hardware.merlininfo.Optical_Black_Pixels_Start;

		 // compute the min/max pixel for each color
	     for (j = 0; j < width; j++) {
		   if (pRed[j] < minBlackred) minBlackred = pRed[j];
		   if (pGreen[j] < minBlackgreen) minBlackgreen = pGreen[j];
		   if (pBlue[j] < minBlackblue) minBlackblue = pBlue[j];
		   if (pRed[j] > maxWhitered) maxWhitered = pRed[j];
		   if (pGreen[j] > maxWhitegreen) maxWhitegreen = pGreen[j];
		   if (pBlue[j] > maxWhiteblue) maxWhiteblue = pBlue[j];
		 }

	    // display the min/max 
	     if (Hardware.m_debug) Hardware.NSCStatusOut("maxWhitered=%X",maxWhitered);
 	     if (Hardware.m_debug) Hardware.NSCStatusOut("maxWhitegreen=%X",maxWhitegreen);
	     if (Hardware.m_debug) Hardware.NSCStatusOut("maxWhiteblue=%X",maxWhiteblue);
	     if (Hardware.m_debug) Hardware.NSCStatusOut("minBlackred=%X",minBlackred);
	     if (Hardware.m_debug) Hardware.NSCStatusOut("minBlackgreen=%X",minBlackgreen);
         if (Hardware.m_debug) Hardware.NSCStatusOut("minBlackblue=%X",minBlackblue);
     }
	 else if ((strncmp(type,"WhiteFine",9)) == 0)
        ComputeGainCoefficents();
     else if((strncmp(type,"ComputeNormalCoefficient",24)) == 0)
	 {
	    long avwhite[3],avblack[3];
	    int maxwhite[3],minwhite[3],maxblack[3],minblack[3];
	    int numwhite[3],numblack[3];
	    for (k = 0; k < 3; k++)
		{
		   avwhite[k]=avblack[k]=maxwhite[k]=maxblack[k]=0;
		   minwhite[k]=minblack[k]=0xffff;
		   numwhite[k]=numblack[k]=0;
		}
	    int targetcode[3] = 
		{
		   Hardware.merlininfo.red_target,
		   Hardware.merlininfo.green_target,
		   Hardware.merlininfo.blue_target
		};
	    for (j = 0; j < width; j++) 
		{
		   int data[3];
		   data[0] = pRed[j];
		   data[1] = pGreen[j];
		   data[2] = pBlue[j];

		   for (k = 0; k < 3; k++)
		   {
			  if (data[k] >= targetcode[k]/2) //1/2 range
			  {
				numwhite[k]++;
				avwhite[k] += data[k];
				maxwhite[k] = max(maxwhite[k],data[k]);
				minwhite[k] = min(minwhite[k],data[k]);	
			  }
			  else
			  {
				numblack[k]++;
				avblack[k] += data[k];
				maxblack[k] = max(maxblack[k],data[k]);
				minblack[k] = min(minblack[k],data[k]);
			  }
		   }
		}

	    for (k = 0; k < 3; k++)
		{
		   if (numwhite[k]) avwhite[k] /= numwhite[k];
		   if (numblack[k]) avblack[k] /= numblack[k];
		}
    
		Hardware.NSCStatusOut("ComputeNormalCoeff");
	    Hardware.NSCStatusOut("target code rgb=%x,%x,%x",
		  targetcode[0],targetcode[1],targetcode[2]);
	    Hardware.NSCStatusOut("num white rgb=%x,%x,%x",
		  numwhite[0],numwhite[1],numwhite[2]);
	    Hardware.NSCStatusOut("fine cal average white rgb=%x,%x,%x",
		  avwhite[0],avwhite[1],avwhite[2]);
	    Hardware.NSCStatusOut("fine cal max white rgb=%x,%x,%x",
		  maxwhite[0],maxwhite[1],maxwhite[2]);
	    Hardware.NSCStatusOut("fine cal min white rgb=%x,%x,%x",
		  minwhite[0],minwhite[1],minwhite[2]);
	    Hardware.NSCStatusOut("num black rgb=%x,%x,%x",
		  numblack[0],numblack[1],numblack[2]);
	    Hardware.NSCStatusOut("fine cal average black rgb=%x,%x,%x",
		  avblack[0],avblack[1],avblack[2]);
	    Hardware.NSCStatusOut("fine cal max black rgb=%x,%x,%x",
		  maxblack[0],maxblack[1],maxblack[2]);
	    Hardware.NSCStatusOut("fine cal min black rgb=%x,%x,%x",
		  minblack[0],minblack[1],minblack[2]);
	
	}
  
	// close the file and remove temporary mem
	_close (hFile);
	hFile = NULL;
	delete scanbuf;
	scanbuf = NULL;


	// send the arrays to the graph control for display
	npts = width;	
	return TRUE;
 
}

/////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------
// FUNCTION	:	SumFineCalValue(int *sum,int ct,int value,char type[])
// DESCRIPTION:	 This function adds the pixel values if more than one 
//               line scanned.  For type = WhiteFine, this routine filters
//               out the non-white pixel

//--------------------------------------------------------
void CScan::SumFineCalValue(int *sum,int ct,int value,char type[])
{
	
	if (strncmp(type,"WhiteFine",9) == 0)
	{
	   int average;
	   if (ct <= 0) return; //ignore leading values
	   if (ct > 1)
	   {
			average = *sum/(ct-1);
			int limit = average/min(100,ct); //min 1% limit
			if (value < (average - limit))  //filter out non-white
			{
				value = average; //ignore this dark value
			}
			else if (value > (average + limit*10)) //filter out too high
			{
				value = average; //ignore this light value
			}
	   }
	}
	*sum += value; //average
}


/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OffsetGainToMerlin() 

// DESCRIPTION	:	message processor for the To Merlin button
//			send the computed gain/offset coefficients to merlin

// RETURNS	:	nothing
//--------------------------------------------------------
void CScan::OffsetGainToMerlin(const char type[]) 
{
{ 
int address;		// index for gain/offset, 0..5460
int iGain,iOffset; // gain/offset coefficients to send to merlin
int iMin;		// min black pixel value
int color;		// index
BYTE value;		// value to write to merlin
int numToWrite;		// number of gain/offset coefficients, 2730/5460
int *ptr;		// pointer into gain arrays (pRedGain, pGreenGain, pBlueGain)
BYTE reg8 = 0;		// value on reg 8 at start
unsigned char write_buffer[0x4000*2]; //hold all offset,gain data for write
unsigned char write_buffer2[0x4000*2];	
int buffer_len;		//buffer length
int *blackPtr;

	//---------------------------------------------------

	{
		numToWrite = (int) (0x4000);// /ScanParms.dpi); //use divider
		//always 16 bit, select sram
		if (!Hardware.ReadRegister(0x42,&value)) 
		{
	writeerr:
			goto cleanup;
		}


        value &= ~7;        //set fixed offset and gain
	if (Hardware.m_debug < 2)
	{
        if (strncmp(type,"BlackFine",9)==0)
			value |=4;      //set external offset  
		else if (strncmp(type,"WhiteFine",9)==0)
   			value |=6;     //set external gain & offset
	}
	else if (Hardware.m_debug == 2) value |= 4; //black offset only
	else value = value; //none 

		
		if (!Hardware.WriteRegister(0x42,value)) goto writeerr;
	}

	//store the offset coeffient value
	if (strncmp(type,"BlackFine",9)==0)
		for (int i=0; i<MAX_PIXELS; i++)
	   {
          pColor[0][i] = pRed[i];
          pColor[1][i] = pGreen[i];
	      pColor[2][i] = pBlue[i];
	   }

	for (color = 0; color < 3; color++) 
	{
	// write offset/gain values for current color
		switch (color) {
			case 0: 
				ptr = pRedGain;
				iMin = minBlackred;
				blackPtr = pColor[0];
				break; // red
			case 1: 
				ptr = pGreenGain;   
				iMin = minBlackgreen; 
				blackPtr = pColor[1];
				break; // green
			case 2: 
				ptr = pBlueGain;
				iMin = minBlackblue; 
				blackPtr = pColor[2];
				break; // blue
		}

	// for each coefficient address, send the gain/offset
		for (address = 0,buffer_len=0; address < numToWrite; address++)
		{
			//use the iMin from course calibration routine as the offset
		    //coeff if the black fine calibration can't be performed
			iOffset = iMin;

			if (address >= npts) break;
			if (blackFine)
	       			iOffset = blackPtr[address];
			iGain = ptr[address];

			{
				iOffset = min(iOffset,0xffff);
				iGain = min(iGain,0xffff);
				if (Hardware.m_debug)
				{
					long avOffset,avGain;
					int maxGain,minGain;
					if (address == 0)
					{
						maxGain = avOffset = avGain = 0;
						minGain = 0xffff;
					}
					maxGain = max(maxGain,iGain);
					minGain = min(minGain,iGain);
					avOffset += iOffset;
					avGain += iGain;
				}
				write_buffer[buffer_len++] = iOffset/256;
				write_buffer[buffer_len--] = iOffset&0xff;
				write_buffer2[buffer_len++] = iGain/256;
				write_buffer2[buffer_len++] = iGain&0xff;
			}
           
		} // for each addr

	// write the address then the offset coefficient
		{   
			int startpixel = 0;
			int len = buffer_len;
			if (Hardware.m_debug)
				Hardware.NSCStatusOut("offset,gain color=%x,startpixel=%x,len=%x",color,startpixel,len);

			if ((strncmp(type,"BlackFine",9)==0) || (strncmp(type,"WhiteFine",9)==0))
			{
				if (!Hardware.WriteRegister(3,color*4)) goto writeerr;
				if (!Hardware.SendData(6,write_buffer+startpixel,len)) goto writeerr;
				if (!Hardware.WriteRegister(3,color*4+1)) goto writeerr;
				if (!Hardware.SendData(6,write_buffer2+startpixel,len)) goto writeerr;
			}
 
			if (Hardware.m_debug)
			{
				for (int i = startpixel; i < 10; i+=2)
				Hardware.NSCStatusOut("ct,offset,gain=%x,%x,%x",i,
						write_buffer[i]*256+write_buffer[i+1],
						write_buffer2[i]*256+write_buffer2[i+1]);
			}
		}
	} // for color=0..2

cleanup:;
}
}


/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ReadScanParms(HFILE hFile,SCANINFO *pScanInfo,int *len)

// DESCRIPTION	:	read the header info from a pxl file

// INPUTS	:	hFile	- handle to pxl file
//			pScanInfo - structure to fill with pxl file parameters
//			*len	- lenth of the header info
//		

// RETURNS	:	true if ok, false if file error
//--------------------------------------------------------
// read the header info from a pxl file,
// set len=offset to start of image data
// if bFillSI = TRUE, the scaninfo structure is filled with the pxl file parms
// otherwise, the pxl file parms are not set
BOOL CScan::ReadScanParms(HFILE hFile,SCANINFO *pScanInfo,int *len)
{
char LineBuf[MAX_PATH];	// file input buffer
int width;		// number of pixels per line
double linespacing;	// line separation (no longer used)
int tripleccd;		// triple ccd indicator (no longer used)
int linerate;		// 0=pixel rate, 1=line rate, 4=1 channel mode a, 5=1 channel mode b
int pixelpack;		// indicates number of pixels per byte
int lines;		// number of lines scanned
LPSTR sptr,eptr;	// pointers into file input buff
int actualxres;		// actual x res in pxl file
int actualyres;		// actual y res in pxl file
int statusbytes;	// number of status bytes per line
int version = 0;	// version number in pxl file

	// read the pxl file header
	if (_read(hFile, (LPSTR) LineBuf, sizeof (LineBuf)-1) != sizeof (LineBuf)-1)
	{
fileerror:
		Hardware.NSCAfxMessageBox("Error reading pixel file!");
		return FALSE;
	}
	LineBuf[sizeof(LineBuf)-1] = 0;
	// the first entry in the pxl file is the version number
	if (LineBuf[0] != 'v') goto fileerror;

	// get version number
	sptr = LineBuf+1;
	eptr = strchr(sptr,'\n');
	if (!eptr) goto fileerror;
	*eptr = 0;
	sscanf(sptr, "%d", &version);
	sptr = eptr+1;

	// get pixels per line
	eptr = strchr(sptr,'\n');
	if (!eptr) goto fileerror;
	*eptr = 0;
	sscanf(sptr, "%d", &width);
	sptr = eptr+1;
	pScanInfo->width = width;

	// get line separation
	eptr = strchr(sptr,'\n');
	if (!eptr) goto fileerror;
	*eptr = 0;
	sscanf(sptr, "%lf", &linespacing);
	linespacing = 0;
	sptr = eptr+1;

	// get triple ccd
	eptr = strchr(sptr,'\n');
	if (!eptr) goto fileerror;
	*eptr = 0;
	sscanf(sptr, "%d", &tripleccd);
	sptr = eptr+1;

	// get line rate
	eptr = strchr(sptr,'\n');
	if (!eptr) goto fileerror;
	*eptr = 0;
	sscanf(sptr, "%d", &linerate);
	sptr = eptr+1;
	pScanInfo->linerate = linerate;

	// get pixel pack
	eptr = strchr(sptr,'\n');
	if (!eptr) goto fileerror;
	*eptr = 0;
	sscanf(sptr, "%d", &pixelpack);
	sptr = eptr+1;
	int bpp = 1 << pixelpack;
	pScanInfo->bpp = bpp;

	// get lines scanned
	eptr = strchr(sptr,'\n');
	if (!eptr) goto fileerror;
	*eptr = 0;
	sscanf(sptr, "%d", &lines);
	sptr = eptr+1;

	// get horizontal res
	eptr = strchr(sptr,'\n');
	if (!eptr) goto fileerror;
	*eptr = 0;
	sscanf(sptr, "%d", &actualxres);
	sptr = eptr+1;
	pScanInfo->actualxres = actualxres;

	// get vertical res
	eptr = strchr(sptr,'\n');
	if (!eptr) goto fileerror;
	*eptr = 0;
	sscanf(sptr, "%d", &actualyres);
	sptr = eptr+1;
	{
		pScanInfo->actualyres = actualyres;
		// set line separation info
		pScanInfo->roffset = 0;
		pScanInfo->goffset = 0; 
		pScanInfo->boffset = 0;
		// have even/odd offset for pixel and line rate
		scaninfo.eoffset = Hardware.merlininfo.EvenOdd_Line_Separation*
							(scaninfo.actualyres
							/(double)Hardware.merlininfo.Optical_Resolution);
		if (pScanInfo->linerate != 4)  // mode a
		{
			if (Hardware.merlininfo.Illumination_Mode == 1) //CCD
			{
				scaninfo.roffset = -Hardware.merlininfo.Line_Separation*
							(scaninfo.actualyres
							/(double)Hardware.merlininfo.Optical_Resolution);
				scaninfo.boffset = -scaninfo.roffset;
			}
			if (Hardware.merlininfo.Color_Mode & 1) //color line rate
			{
				scaninfo.roffset -= (1./3.);
				scaninfo.boffset += (1./3.);
			}
		}
	}

	// get status bytes
	if (version > 1) {
		eptr = strchr(sptr,'\n');
		if (!eptr) goto fileerror;
		*eptr = 0;
		sscanf(sptr, "%d", &statusbytes);
		sptr = eptr+1;
	}
	// return the length of the pxl file header info
	if (len) *len = sptr - LineBuf;
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ComputeGainCoefficents()

// DESCRIPTION	:	fine calibration scan has been done, use the fine 
//			calibration scan data to compute the gain coefficients
//          for each pixel 
// 		 	For 9831: Gain = 0x4000*targetcode/(pixel value-offset coeff);
//          Pixel values are avg of lines scanned during final calibration
//          Offset is the min black value if the black fine calibration can't 
//          be performed, otherwise it is the black fine calibration pixel value          

// INPUTS	:	nothing

// RETURNS	:	nothing
//--------------------------------------------------------
void CScan::ComputeGainCoefficents()
{
int address;		// index for gain coefficients 0..5460
int iGain;			// computed gain and offset for a pixel
int iMin;			// the minimum black pixel value for a color
int i,color;		// loop counters
BYTE value;			// register 3 value of color to write (not used)
BOOL bError = FALSE;	// i/o error indicator
int *ptr;			// pointer into pixel values (pRed, pGreen, pBlue)
int *sptr;			// pointer into gain coeff (pRedGain, pGreenGain, pBlueGain)
double redtargetcode;	// target codes as specified in window
double greentargetcode;
double bluetargetcode;
double targetcode;		// target code for the current color

	// make sure there is fine calibration data to work with
	if (!pRed || !pGreen || !pBlue) return;

	// make sure the arrays actually got allocated
	if (!pRedGain || !pGreenGain || !pBlueGain) return;

	// init each array to 0
	memset(pRedGain,0,npts*sizeof(int));
	memset(pGreenGain,0,npts*sizeof(int));
	memset(pBlueGain,0,npts*sizeof(int));

	// get the target codes specified on the calibration window
	redtargetcode = Hardware.merlininfo.red_target;	
	greentargetcode = Hardware.merlininfo.green_target;	
	bluetargetcode = Hardware.merlininfo.blue_target;

	// compute the min white value and the max targetcode/min white error
//	minwhite = CalcWhiteError(&ratio);

	// for each color compute the gain coefficients (correction codes)
	for (color = 0; color < 3; color++) {

	//---------------------------------------------------
	// initialize pointers/values for this color
		switch (color) {
			case 0: value = 1; 
				iMin = minBlackred;
				ptr = pRed;
				sptr = pRedGain;
				targetcode = redtargetcode;
				break; // red
			case 1: value = 3; 
				iMin = minBlackgreen; 
				ptr = pGreen;
				sptr = pGreenGain;
				targetcode = greentargetcode;
				break; // green
			case 2: value = 5; 
				iMin = minBlackblue; 
				ptr = pBlue;
				sptr = pBlueGain;
				targetcode = bluetargetcode;
				break; // blue
		}
//---------------------------------------------------
// pixel values are avg of lines scanned during final calibration

		for (address = 0; address < npts; address++)
		{
            i = ptr[address];
			{
				if (i > 0) 
				   iGain = 	(int) (0x4000*targetcode/i);
		    	else 
					iGain = 0;
			}
			sptr[address] = iGain;
		} // for each addr
	} // for each color
}

//scan line to BMP
BOOL CScan::ScantoBMPfile(char *fname)
{
	BOOL		retcode = FALSE;
	BYTE		tempBuffer[sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD)];
	static int fhBMP;
	static char filename[MAX_PATH];
	static int color_ct; 

	strcpy(filename,fname);
	strcpy(strrchr(filename,'.'),".bmp");

	if( !scaninfo.initialized || scaninfo.scanlinecount == 0 )
	{
		scaninfo.bmpbytesperline =0 ; // Nymber of BYTEs in a returned BMP line, determined by me
		scaninfo.bmpreturnlinecount = 0 ;	// The BMP line number being returned.
								// Set to zero until the first BMP line is returned
		scaninfo.bytesperscanlinepixel = 1 ; // one byte per pixel or two bytes per pixel
		if (ScanParms.bpp == 16)
			scaninfo.bytesperscanlinepixel = 2;
		scaninfo.n = 0 ; // not used
		scaninfo.m = 0 ; // not used
		scaninfo.bIsTwain = !m_currentScanSettings.scanToTwain; //RGB if true
		scaninfo.returnbytecount = 0 ; // Returns the byte count 

		//Initialize return buffer for the BMP file headers
		scaninfo.dibwidth = m_currentScanSettings.scanRect.right - 
			m_currentScanSettings.scanRect.left ;	// number of pixels in each bmp line
		scaninfo.width = (int) ceil((double)scaninfo.dibwidth*
			scaninfo.actualxres/scaninfo.desiredxres);
		scaninfo.bmpoutlines = m_currentScanSettings.scanRect.bottom - 
			m_currentScanSettings.scanRect.top;	// number of lines in bmp file
		scaninfo.bpp = ScanParms.bpp;
		scaninfo.bytesperscanline = ScanParms.bytesperline-2;
		if ((ScanParms.bpp == 16) && !scaninfo.bIsTwain) //bmp not support 16
		{
			scaninfo.bpp /= 2;
			scaninfo.bytesperscanline /= 2;
			scaninfo.bytesperscanlinepixel = 1 ; 
		}
		scaninfo.linerate = ScanParms.linerate;
		if (ScanParms.linerate&1)
		{
			//force color line rate modes to look like RBG pixel rate
			scaninfo.bytesperscanline *= 3;
			scaninfo.linerate = 0;
		}

		if( !scaninfo.initialized )
		{
			scaninfo.initialized = TRUE;
			scaninfo.scanlinecount = 0;
			scaninfo.pBmplinebuffer = tempBuffer; //throw away header
			Scan2Bmp.generate_bmp_line(&scaninfo,FALSE);
			scaninfo.scanlinecount = 0;
			color_ct = 0; //ct 3 color lines
			m_currentScanInfo.bytesPerLine = scaninfo.bmpbytesperline;
			m_currentScanInfo.actBytes = 0;
			m_currentScanInfo.totalBytes = scaninfo.bmpoutlines*scaninfo.bmpbytesperline;
			if (m_currentScanInfo.totalBytes/scaninfo.bmpoutlines
				!= scaninfo.bmpbytesperline)
			{
				//too big for long so encode as -Kbytes
				m_currentScanInfo.totalBytes = 
					(long)((double)scaninfo.bmpoutlines*
								scaninfo.bmpbytesperline/-1024.);
				Hardware.NSCAfxMessageBox("Image size is too large");
				return FALSE;
			}
			return TRUE;
		}
		else
		{
			// Call generate_bmp_line the first time with scaninfo.scanlinecount = 0 
			// to force it to initialize its internal data
			scaninfo.pBmplinebuffer = tempBuffer;
			Scan2Bmp.generate_bmp_line(&scaninfo,FALSE);
			ScanReadParms.scanReadActCount = 0;
			if (Hardware.m_debug)
			{
				DeleteFile((LPCSTR) filename);
				fhBMP = open(filename,_O_RDWR|_O_CREAT|_O_BINARY,_S_IREAD|_S_IWRITE);
				_write (fhBMP, scaninfo.pBmplinebuffer,
							scaninfo.returnbytecount);
			}
			scaninfo.returnbytecount = 0;
		}
	}
	
	if( scaninfo.bmpreturnlinecount < scaninfo.bmpoutlines )
	{
		if (scaninfo.returnbytecount == 0)
		{
		scaninfo.pBmplinebuffer = pBMPBuf;
		BYTE *tmp_pScanlinebuffer = scaninfo.pScanlinebuffer; //save
		if ((ScanParms.bpp == 16) && !scaninfo.bIsTwain) //bmp not support 16
		{
			static BYTE data[0x30000]; //max size (3*16K)
			//remove low byte
			for (int i = 0; i < scaninfo.bytesperscanline; i++)
			{
				data[i] = scaninfo.pScanlinebuffer[2*i];
			}
			scaninfo.pScanlinebuffer = data;
		}
		if ((Hardware.merlininfo.Color_Mode & 1) //color line rate to RGB
			&& (scaninfo.linerate != 4))
		{
			static BYTE data[0x30000]; //max size (3*16K)
			if (!color_ct)
				for (int i = 0; i < scaninfo.bytesperscanline; i++) data[i] = 0;
			int num_colors = 3;
			//bits per scan line ignore partial at end of the line
			int bitsperpixel = scaninfo.bpp*num_colors;
			int bitsperscanline = 
				((scaninfo.bytesperscanline*8)/bitsperpixel)*bitsperpixel;
			int bitin,bitout;
			BYTE bitmask = (1 << scaninfo.bpp) - 1;
			//convert to pixel rate data 
			for (bitin = 0,bitout = color_ct*scaninfo.bpp;
					bitout < bitsperscanline; bitin+=scaninfo.bpp,bitout+=bitsperpixel)
			{
				BYTE value = scaninfo.pScanlinebuffer[bitin/8];
				if (scaninfo.bpp < 8)
				{
					value >>= (8 - scaninfo.bpp - (bitin%8));
					value &= bitmask;
					value <<= (8 - scaninfo.bpp - (bitout%8));
				}
				data[bitout/8] += value;
				if (scaninfo.bpp == 16)
				{
					BYTE value = scaninfo.pScanlinebuffer[(bitin/8)+1];
					data[(bitout/8)+1] = value;
				}
			}
			if (++color_ct != 3) return TRUE; //get next line
			color_ct = 0;
			scaninfo.pScanlinebuffer = data;
		}
		if (Hardware.merlininfo.Mirror_Image) //Mirror
		{
			static BYTE data[0x30000]; //max size (3*2^16)
			for (int i = 0; i < scaninfo.bytesperscanline; i++) data[i] = 0;
			int num_colors = 3;
			if (scaninfo.linerate == 4)	//color to gray
				num_colors = 1;
			//bits per scan line ignore partial at end of the line
			int bitsperpixel = scaninfo.bpp*num_colors;
			int bitsperscanline = 
				((scaninfo.bytesperscanline*8)/bitsperpixel)*bitsperpixel;
			int bitin,bitout;
			BYTE bitmask = (1 << scaninfo.bpp) - 1;
			for (int color_ct = 0; color_ct < num_colors; color_ct++)
				for (bitin = color_ct*scaninfo.bpp,bitout = bitsperscanline - bitsperpixel
					+ color_ct*scaninfo.bpp; 
						bitout >= 0; bitin+=bitsperpixel,bitout-=bitsperpixel)
			{
				BYTE value = scaninfo.pScanlinebuffer[bitin/8];
				if (scaninfo.bpp < 8)
				{
					value >>= (8 - scaninfo.bpp - (bitin%8));
					value &= bitmask;
					value <<= (8 - scaninfo.bpp - (bitout%8));
				}
				data[bitout/8] += value;
				if (scaninfo.bpp == 16)
				{
					BYTE value = scaninfo.pScanlinebuffer[(bitin/8)+1];
					data[(bitout/8)+1] = value;
				}
			}
			scaninfo.pScanlinebuffer = data;
		}
		++scaninfo.scanlinecount;	// line number of scan line in buffer
		Scan2Bmp.generate_bmp_line(&scaninfo,FALSE) ;
		scaninfo.pScanlinebuffer = tmp_pScanlinebuffer; //restore
		}
	}
	if( scaninfo.returnbytecount > 0 )
	{
        if (!m_currentScanInfo.actBytes)
		{
			Hardware.NSCStatusOut("raw scanReq=%d,bmpbytesperline=%d,lines/req=%d",
				ScanReadParms.scanReadReqCount,
				scaninfo.bmpbytesperline,
				ScanReadParms.scanReadReqCount/scaninfo.bmpbytesperline);
		}


		//be sure line aligned
		ScanReadParms.scanReadReqCount /= scaninfo.bmpbytesperline;
		ScanReadParms.scanReadReqCount *= scaninfo.bmpbytesperline;

		//move to buffer line size available
		int nowmove = min(scaninfo.returnbytecount,
							ScanReadParms.scanReadReqCount-
							  ScanReadParms.scanReadActCount);
		scaninfo.returnbytecount -= nowmove;

		if (Hardware.m_debug)
		{
//			_lseek(fhBMP,scaninfo.fileOffset,SEEK_SET);
			_write (fhBMP, pBMPBuf,	nowmove);
		}

		memcpy(ScanReadParms.scanReadBuffer+ScanReadParms.scanReadActCount,
				pBMPBuf,nowmove);
		memmove(pBMPBuf,pBMPBuf+nowmove,scaninfo.returnbytecount);
		m_currentScanInfo.actBytes += nowmove;
		ScanReadParms.scanReadActCount +=  nowmove;

		if (m_currentScanInfo.actBytes >= m_currentScanInfo.totalBytes)
		{
			if (Hardware.m_debug) _close(fhBMP);
			return FALSE; //ScanRead done
		}
		if (ScanReadParms.scanReadActCount >= 
			ScanReadParms.scanReadReqCount)
		{
			return FALSE; //ScanRead done
		}
	}

	return TRUE;
}


//find substring
BOOL CScan::SubString(char *string,char *substring)
{
	int string_len = strlen(string);
	int substring_len = strlen(substring);
	for (int i = 0; i < string_len; i++)
		if (!strnicmp(string+i,substring,substring_len)) return TRUE;
	return FALSE;
}

//Clean up
void CScan::Cleanup()
{
	Hardware.CloseDriver(); //close USB
}

//Update status message
void CScan::UpdateScanStatus()
{
	strcpy(m_currentScanInfo.scanStatus,Hardware.m_scanStatus);	
}

//Save callback
void CScan::SaveCallback(FN_SCAN_CALLBACK scanCB)
{
	Hardware.m_callbackfn = scanCB;
}

//1a,1b phase correction
int CScan::ComputeStepperPhaseCorrection(int *m_byRegister,int m_wLineLength)
{
	WORD wScanStepSize = m_byRegister[0x46]*256+m_byRegister[0x47];
	WORD wFastFeedStepSize = m_byRegister[0x48]*256+m_byRegister[0x49];
	WORD wFullsteps2Skip = m_byRegister[0x4A]*256+m_byRegister[0x4B];
	WORD wStepCounter = m_byRegister[0x4C]*256+m_byRegister[0x4D];

	WORD wFullstepsAtQuarterSpeed = (m_byRegister[0x51] & 0x30) >> 4;
	if ((wFullstepsAtQuarterSpeed == 3))
	{
		wFullstepsAtQuarterSpeed = 8;
	}
	WORD wFullstepsAtHalfSpeed = (m_byRegister[0x51] & 0xC0) >> 6;
	if ((wFullstepsAtHalfSpeed == 3))
	{
		wFullstepsAtHalfSpeed = 8;
	}
	WORD wAccelSteps = wFullstepsAtQuarterSpeed+wFullstepsAtHalfSpeed;
	DWORD dwFPCNT;
	if ((wFullsteps2Skip-1 >= wAccelSteps) && (wStepCounter >= 1))
	{
		dwFPCNT = wFullstepsAtQuarterSpeed*wFastFeedStepSize*16
				 +wFullstepsAtHalfSpeed*wFastFeedStepSize*8
				 +(wFullsteps2Skip-wFullstepsAtQuarterSpeed-wFullstepsAtHalfSpeed-1)*wFastFeedStepSize*4
				 +wFastFeedStepSize+wScanStepSize*3+wStepCounter*wScanStepSize*4;
	}
	else if ((wFullsteps2Skip-1 < wAccelSteps) && (wFullsteps2Skip-1 >= wFullstepsAtQuarterSpeed))
	{
		WORD wFFRCNT = (wFullstepsAtQuarterSpeed+wFullstepsAtHalfSpeed-(wFullsteps2Skip-1));
		dwFPCNT = wFullstepsAtQuarterSpeed*wFastFeedStepSize*16
				 +(wFullstepsAtHalfSpeed-wFFRCNT)*wFastFeedStepSize*8
				 +(wFastFeedStepSize+3*wScanStepSize)*2+(wFFRCNT-1)*wScanStepSize*8
				 +(wStepCounter-wFFRCNT+1)*wScanStepSize*4;
	}
	else if ((wFullsteps2Skip-1 < wAccelSteps) && (wFullsteps2Skip-1 < wFullstepsAtQuarterSpeed))
	{
		dwFPCNT = (wFullsteps2Skip-1)*wFastFeedStepSize*16
				 +(wFastFeedStepSize+3*wScanStepSize)*4
				 +(wFullstepsAtQuarterSpeed-wFullsteps2Skip)*wScanStepSize*16
				 +wFullstepsAtHalfSpeed*wScanStepSize*8
				 +(wStepCounter-wFullstepsAtHalfSpeed-wFullstepsAtQuarterSpeed+wFullsteps2Skip)*wScanStepSize*4;
	}

	int nStepperPhaseCorrection = m_wLineLength-(dwFPCNT%m_wLineLength)-
									Hardware.merlininfo.TR_Stepper_Correction;//2*wScanStepSize;//-5
	int n = 0;
	while (n*m_wLineLength+nStepperPhaseCorrection < 0) 
	{
		n++;
	}
	WORD wStepperPhaseCorrection = n*m_wLineLength+nStepperPhaseCorrection;
	m_byRegister[0x1A] = (BYTE)(wStepperPhaseCorrection >> 8);
	m_byRegister[0x1B] = (BYTE)(wStepperPhaseCorrection & 0xFF);

	// Output stepper phase correction to log file 
	//lines transferred
	int linestrans = (wStepCounter*wScanStepSize*4 - 
						Hardware.merlininfo.TR_Stepper_Correction)/
						m_wLineLength;
	linestrans++;
	Hardware.NSCStatusOut ("lines trans=%d",linestrans);

	return wStepperPhaseCorrection;
}

//test motor position
void CScan::PositionTest()
{
#if 0
			//sensor test!!!
			//forward half page
			if (!Hardware.CheckScannerReady()) return;
			if (!Hardware.WriteRegister(0x4a,10)) return;
			if (!Hardware.WriteRegister(0x4b,0)) return;
			if (!Hardware.WriteRegister(7,5)) return;
			if (!Hardware.CheckScannerReady()) return;
			//read ff and step
			BYTE val;
			if (!Hardware.ReadRegister(0x46,&val)) return;
			int scanstep = val*256;
			if (!Hardware.ReadRegister(0x47,&val)) return;
			scanstep += val;
			if (!Hardware.ReadRegister(0x48,&val)) return;
			int ffstep = val*256;
			if (!Hardware.ReadRegister(0x49,&val)) return;
			ffstep += val;
			//loop
			for (int i=0; i < 1; i++)
			{
				Hardware.delay(5000);
#define UP 200
				int up;
				up=UP;
#define DOWN (UP-100)
				int down;
				down=DOWN;
				int total;
				total = 0;
				while (up >= DOWN)
				{
				//forward 
//					if (!Hardware.WriteRegister(0x46,ffstep/256)) break;
//					if (!Hardware.WriteRegister(0x47,ffstep&0xff)) break;
#define SCAN_STEPS 20
#define DECEL 16
#define ACCEL DECEL
//					if (!Hardware.WriteRegister(0x4a,(up-SCAN_STEPS-DECEL-ACCEL)/256)) break;
//					if (!Hardware.WriteRegister(0x4b,(up-SCAN_STEPS-DECEL-ACCEL)&0xff)) break;
//
//					if (!Hardware.WriteRegister(7,5)) break;
//					if (!Hardware.CheckScannerReady()) break;  
//					Hardware.delay(500);	
				//scan
//					if (!Hardware.WriteRegister(0x46,scanstep/256)) break;
//					if (!Hardware.WriteRegister(0x47,scanstep&0xff)) break;

					if (!Hardware.WriteRegister(0x4a,(up-SCAN_STEPS)/256)) break;
					if (!Hardware.WriteRegister(0x4b,(up-SCAN_STEPS)&0xff)) break;

					if (!Hardware.WriteRegister(0x4c,SCAN_STEPS/256)) break;
					if (!Hardware.WriteRegister(0x4d,SCAN_STEPS&0xff)) break;

					if (!Hardware.WriteRegister(7,3)) break;
					if (!Hardware.CheckScannerReady()) break; 
					Hardware.delay(1500);
				//reverse
					if (!Hardware.WriteRegister(0x4a,down/256)) break;
					if (!Hardware.WriteRegister(0x4b,down&0xff)) break;

					if (!Hardware.WriteRegister(7,6)) break;
					if (!Hardware.CheckScannerReady()) break;
					Hardware.delay(500);
				//calculate
					total += up - down;
					Hardware.NSCStatusOut("up=%d,down=%d,total=%d",
						up,down,total);
					up--;
					down++;
				}
			}
#endif
}
