/*******************************************************************************
FILE        : Twain_CapHelpers.cpp
REVISION    : 1.0
DATE        : 11/1/99

DESCRIPTION :

This module contains the implementtation of the capability helper mthods
which are simply methods that encapsulate frequently used tasks involving
capabilities. 

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/
#include "stdafx.h"
#include "Twain_CapHelpers.h"

#include "Twain_Utility.h"
#include "Twain_ScanInterface.h"
#include "Twain_CapHolder.h"
#include "Twain_GenCaps.h"
#include "Twain_ImageCaps.h"



/*******************************************************************************
DESCRIPTION:

GetCurrentXResolution _ this methods returns the current horizontal res. as
a float instead od the TWAIN FIX32 format. The method provides a wrapper over 
directly using the CICapXresolution  capability  object

PARAMETERS: 

float& xResolution

RETURN VALUE:
TWAIN_RESULT  - status return

NOTES:

*******************************************************************************/
TWAIN_RESULT  GetCurrentXResolution(float& xResolution )
{
    CCapBase *          capObject;
    CCapHolder::GetInstance()->GetCapObject(ICAP_XRESOLUTION, capObject);

    CICapXResolution*   pCapXRes = dynamic_cast<CICapXResolution*>(capObject);

    if (!pCapXRes) 
        return TWAIN_ERROR;

    TW_FIX32        xRes;
    TWAIN_RESULT    result = pCapXRes->GetCurrentValue(xRes);

    if (result == TWAIN_OK)
    {
        xResolution = FIX32ToFloat(xRes);
    }

    return result;
}


/*******************************************************************************
DESCRIPTION:

GetCurrentYResolution - this method returns the current vertical res. as
a float instead of the TWAIN FIX32 format. The method provides a wrapper over 
directly using the CICapYresolution  capability  object

PARAMETERS: 

float& yResolution

RETURN VALUE:
TWAIN_RESULT  - status code that can be TWAIN_OK or TWAIN_ERROR

NOTES:

*******************************************************************************/
TWAIN_RESULT  GetCurrentYResolution(float& yResolution )
{
    CCapBase *          capObject;
    CCapHolder::GetInstance()->GetCapObject(ICAP_YRESOLUTION, capObject);

    CICapYResolution*   pCapYRes = dynamic_cast<CICapYResolution*>(capObject);

    if (!pCapYRes) 
        return TWAIN_ERROR;

    TW_FIX32        yRes;
    TWAIN_RESULT    result = pCapYRes->GetCurrentValue(yRes);

    if (result == TWAIN_OK)
    {
        yResolution = FIX32ToFloat(yRes);
    }

    return result;
}


/*******************************************************************************
DESCRIPTION:

GetCurrentPixelType - returns the current pixel type as one of the enumerations
defined by the TWAIN spec i.e. TWPT_BW, TWPT_GRAY or TWPT_RGB

PARAMETERS: 

TW_UINT16& pixelType

RETURN VALUE:
TWAIN_RESULT - status return TWAIN_OK or TWAIN_ERROR

NOTES:

*******************************************************************************/
TWAIN_RESULT GetCurrentPixelType(TW_UINT16& pixelType)
{
    CCapBase *          capObject;
    CCapHolder::GetInstance()->GetCapObject(ICAP_PIXELTYPE, capObject);

    CICapPixelType* pCapPixelType = dynamic_cast<CICapPixelType*>(capObject);

    if (!pCapPixelType) 
        return TWAIN_ERROR;

    TWAIN_RESULT    result = pCapPixelType->GetCurrentValue(pixelType);

    return result;
}


/*******************************************************************************
DESCRIPTION:

GetCurrentBitDepth - returns the current setting for the bitsPerPixel parameter
queried from the correspnding capability  object

PARAMETERS: 

TW_UINT16& bitDepth

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT GetCurrentBitDepth(TW_UINT16& bitDepth )
{
    CCapBase *          capObject;
    CCapHolder::GetInstance()->GetCapObject(ICAP_BITDEPTH, capObject);

    CICapBitDepth* pBitDepth = dynamic_cast<CICapBitDepth*>(capObject);

    if (!pBitDepth) 
        return TWAIN_ERROR;

    TWAIN_RESULT  result = pBitDepth->GetCurrentValue(bitDepth);

    return result;
}


/*******************************************************************************
DESCRIPTION:

GetCurrentUnits - returns the currently used UNITS which is one of the 
TWAIN enumerations TWUN_PIXELS or TWUN_INCHES since those that are the 
only ones this module supports

PARAMETERS: 

TW_UINT16& currentUnits

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT GetCurrentUnits(TW_UINT16& currentUnits)
{
    CCapBase *          capObject;
    CCapHolder::GetInstance()->GetCapObject(ICAP_UNITS, capObject);

    CICapUnits* pUnits = dynamic_cast<CICapUnits*>(capObject);

    if (!pUnits) 
        return TWAIN_ERROR;

    TWAIN_RESULT  result = pUnits->GetCurrentValue(currentUnits);

    return result;
}


/*******************************************************************************
DESCRIPTION:

GetCurrentXferMech - returns the currently negotiated Xfer mech for image data 
transfer. This can be the TWAIN enumerations for native or memory only since we
dont support file transfers. 

PARAMETERS: 

TW_UINT16& xferMech

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT GetCurrentXferMech(TW_UINT16& xferMech)
{
    CCapBase *          capObject;
    CCapHolder::GetInstance()->GetCapObject(ICAP_UNITS, capObject);

    CICapXferMech * pXferMech = dynamic_cast<CICapXferMech*>(capObject);

    if (!pXferMech) 
        return TWAIN_ERROR;

    TWAIN_RESULT  result = pXferMech->GetCurrentValue(xferMech);

    return result;
}

bool operator ==(TW_FIX32 first, TW_FIX32 second)
{
    return ((first.Frac == second.Frac) && (first.Whole == second.Whole));
}


/*******************************************************************************
DESCRIPTION:

GetScanWidthInPixels - returns the current scan rectangle width in pixels 
irrespective of the units currently being used.

PARAMETERS: 

int& scanWidth

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT GetScanWidthInPixels(int& scanWidth)
{
    TWAIN_PARAM_REC scanRec;
    
    TWAIN_RESULT result = 
        CScanInterface::GetInstance()->ScanGetScanParams(&scanRec);
    
    if (result == TWAIN_OK)
    {
        scanWidth = scanRec.scanRect.right - scanRec.scanRect.left;
    }

    return result;
}


/*******************************************************************************
DESCRIPTION:

GetScanHeightInPixels - returns the current scan rectangle height in pixels 
irrespective of the units currently being used.

PARAMETERS: 

int& scanHeight

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT GetScanHeightInPixels(int& scanHeight)
{
    TWAIN_PARAM_REC scanRec;
    
    TWAIN_RESULT result = 
        CScanInterface::GetInstance()->ScanGetScanParams(&scanRec);
    
    if (result == TWAIN_OK)
    {
        scanHeight = scanRec.scanRect.bottom - scanRec.scanRect.top;
    }

    return result;
}



/*******************************************************************************
DESCRIPTION:

GetCurrentScanWidth - This method will return the scan width in the current 
units which may be pixels or inches. The scan module always works in pixels
do this routine will convert the values appropriately using the scan resolution

PARAMETERS: 

float& scanWidth

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT GetCurrentScanWidth(float& scanWidth)
{
    int widthPixels;
    TWAIN_RESULT twainResult = GetScanWidthInPixels(widthPixels);
    if (twainResult != TWAIN_OK) 
        return twainResult;
    
    twainResult = CalculateWidthInCurrentUnits(widthPixels, scanWidth);

    return twainResult;
}



/*******************************************************************************
DESCRIPTION:

GetCurrentScanHeight - This method will return the scan height in the current 
units which may be pixels or inches. Since the scan module always works in 
pixels, this routine may do some conbersion based on the current 
scan resolution

PARAMETERS: 

float& scanHeight

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT GetCurrentScanHeight(float& scanHeight)
{
    int heightPixels;
    TWAIN_RESULT twainResult = GetScanHeightInPixels(heightPixels);
    if (twainResult != TWAIN_OK) 
        return twainResult;

    twainResult = CalculateHeightInCurrentUnits(heightPixels, scanHeight);

    return twainResult;
}


/*******************************************************************************
DESCRIPTION:

GetMaxScanWidth - This routine returns the maximum scan width in the current
units i.e. pixels or inches

PARAMETERS: 

float& scanWidth

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT GetMaxScanWidth(float& scanWidth)
{
    int     maxScanWidth;
    long    itemCount;

    TWAIN_RESULT twainResult = CScanInterface::GetInstance()->
            ScanGetCapability(  TWAIN_MAX_SCAN_WIDTH,
                                NULL,
                                &maxScanWidth,
                                &itemCount);

    if (twainResult == TWAIN_OK)
    {
        twainResult = CalculateWidthInCurrentUnits(maxScanWidth, scanWidth);
    }

    return twainResult;
}


/*******************************************************************************
DESCRIPTION:

GetMaxScanHeight - This routine returns the maximum image height supported by 
the scanner in the current units i.e. pixels or inches

PARAMETERS: 

float& scanHeight

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT GetMaxScanHeight(float& scanHeight)
{
    int     maxScanHeight;
    long    itemCount;

    TWAIN_RESULT twainResult = CScanInterface::GetInstance()->
            ScanGetCapability(  TWAIN_MAX_SCAN_HEIGHT,
                                NULL,
                                &maxScanHeight,
                                &itemCount);

    if (twainResult == TWAIN_OK)
    {
        twainResult = CalculateHeightInCurrentUnits(maxScanHeight, scanHeight);
    }

    return twainResult;
}


/*******************************************************************************
DESCRIPTION:

CalculateWidthInCurrentUnits - This is a helper method that converts the
passed in width in pixels to the width in the current units (whatever that 
may be ) using the horizontal resolution.

PARAMETERS: 

int pixelWidth
float& scanWidth

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT CalculateWidthInCurrentUnits(int pixelWidth, float& scanWidth)
{
    // If the current units is not pixels then we need to 
    // use the resolution to figure out the width in the current
    // units
    TW_UINT16 currentUnits;
    TWAIN_RESULT  twainResult = GetCurrentUnits(currentUnits);
    if (twainResult != TWAIN_OK)
        return twainResult;

    if (currentUnits != TWUN_PIXELS)
    {
        // Another assumption that the only other units we support are 
        /// inches so if we are not doing pixels we are doing inches
        // so we get the current YRes and figure out the height in inches
        float xRes;
        twainResult = GetCurrentXResolution(xRes);
        if ((twainResult != TWAIN_OK) || (xRes <= 0))
            return TWAIN_ERROR;

        scanWidth = pixelWidth / xRes;
    }
    else
    {
        scanWidth = pixelWidth ;
    }

    return twainResult;
}


/*******************************************************************************
DESCRIPTION:

CalculateHeightInCurrentUnits- This is a helper method that converts the
passed in height in pixels to the height in the current units (whatever that 
may be ) using the vertical resolution.

PARAMETERS: 

int pixelHeight
float& scanHeight

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT CalculateHeightInCurrentUnits(int pixelHeight, float& scanHeight)
{
    // If the current units is not pixels then we need to 
    // use the resolution to figure out the height in the current
    // units
    TW_UINT16 currentUnits;
    TWAIN_RESULT twainResult = GetCurrentUnits(currentUnits);
    if (twainResult != TWAIN_OK)
        return twainResult;

    if (currentUnits != TWUN_PIXELS)
    {
        // Another assumption that the only other units we support are 
        // inches so if we are not doing pixels we are doing inches
        // so we get the current YRes and figure out the height in inches
        float yRes;
        twainResult = GetCurrentYResolution(yRes);
        if ((twainResult != TWAIN_OK) || (yRes <= 0))
            return TWAIN_ERROR;

        scanHeight = pixelHeight / yRes;
    }
    else
    {
        scanHeight = pixelHeight ;
    }

    return twainResult;
}


/*******************************************************************************
DESCRIPTION:

CalculateWidthInPixels - This methods takes the width specified in the current
units (inches or pixels) and calculates the width in pixels based on the 
current scan resolution.

PARAMETERS: 

float scanWidth
int& pixelWidth

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT CalculateWidthInPixels(float scanWidth, int& pixelWidth)
{
    // If the current units is not pixels then we need to 
    // use the resolution to figure out the width in pixels
    TW_UINT16 currentUnits;
    TWAIN_RESULT  twainResult = GetCurrentUnits(currentUnits);
    if (twainResult != TWAIN_OK)
        return twainResult;

    if (currentUnits != TWUN_PIXELS)
    {
        // Another assumption that the only other units we support are 
        /// inches so if we are not doing pixels we are doing inches
        // so we get the current YRes and figure out the height in inches
        float xRes;
        twainResult = GetCurrentXResolution(xRes);
        if ((twainResult != TWAIN_OK) || (xRes <= 0))
            return TWAIN_ERROR;

        pixelWidth = int(scanWidth *  xRes);
    }
    else
    {
        pixelWidth = int(scanWidth);
    }

    return twainResult;
}


/*******************************************************************************
DESCRIPTION:

CalculateHeightInPixels - This methods takes the height specified in the current
units (inches or pixels) and calculates the height in pixels based on the 
current scan resolution.

PARAMETERS: 

float scanHeight
int& pixelHeight

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT CalculateHeightInPixels(float scanHeight,  int& pixelHeight)
{
    // If the current units is not pixels then we need to 
    // use the resolution to figure out the height in pixels
    TW_UINT16 currentUnits;
    TWAIN_RESULT twainResult = GetCurrentUnits(currentUnits);
    if (twainResult != TWAIN_OK)
        return twainResult;

    if (currentUnits != TWUN_PIXELS)
    {
        // Another assumption that the only other units we support are 
        // inches so if we are not doing pixels we are doing inches
        // so we get the current YRes and figure out the height in inches
        float yRes;
        twainResult = GetCurrentYResolution(yRes);
        if ((twainResult != TWAIN_OK) || (yRes <= 0))
            return TWAIN_ERROR;

        pixelHeight = int(pixelHeight * yRes);
    }
    else
    {
        pixelHeight = int(scanHeight);
    }

    return twainResult;
}
