/*******************************************************************************
FILE        : Twain_CapHolder.cpp
REVISION    : 1.0
DATE        : 9/28/99

DESCRIPTION :

This module contains the implementation of the CCapHolder class which is the 
singleton container class which holds references to all the capability objects
in the system.

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/

#include "stdafx.h"
#include "Twain_CapHolder.h"
#include "Twain_session.h"
#include "Twain_GenCaps.h"
#include "Twain_ImageCaps.h"

//////////////////////////////////////////////////////////////////////
// The single static instance of the class 
//////////////////////////////////////////////////////////////////////
CCapHolder* CSingleton<CCapHolder>::m_theInstance=0;

/*******************************************************************************
DESCRIPTION:

CCapHolder::CCapHolder is the class constructor

PARAMETERS: 
none

RETURN VALUE:
none

NOTES:
Loading the capability objects is not done in the constrcutor since we 
do not control the exact point where this object is instantiated. this 
object will be created automatically when a client calls the GetInstance
static access function
*******************************************************************************/
CCapHolder::CCapHolder()
{

}


/*******************************************************************************
DESCRIPTION:

CCapHolder::~CCapHolder is the class destructor

PARAMETERS: 
none

RETURN VALUE:
none

NOTES:
Just like construction the destruction of the object takes place sometime
during the module unloading process so we dont do any cleanup here. Instead
a seperate method is provided which will be called by the TWAIN module
during the processing of the CloseDS request from the TWAIN source manager
*******************************************************************************/
CCapHolder::~CCapHolder()
{

}


/*******************************************************************************
DESCRIPTION:

CCapHolder::LoadCapabilities is the function that actually creates all the 
capability objects and thus kickstarts the capability engine.

PARAMETERS: 
none

RETURN VALUE:
void 

NOTES:
This method has to be called before the module enters TWAIN state 4 since at 
that time capability negotiaion can occur at any point. So the appropriate
place to call this is as soon as the OpenDS call is made by the TWAIN 
data source manager by sending a DG_CONTROL / DAT_IDENTITY / MSG_OPENDS 
operation triplet and the TWAIN module moves into state 4
*******************************************************************************/
void CCapHolder::LoadCapabilities()
{
    // Create the capability objects and add them to our list
    CCapBase* pCapability;

    // First the general capability objects    
    pCapability = new CCapDeviceOnline();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CCapSupportedCaps();
    m_capabilities[pCapability->GetCapID()] = pCapability;
    
    pCapability = new CCapUIControllable();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CCapXferCount();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    // then the Image specfific capabilities
    pCapability = new CICapBitDepth();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CICapBitOrder();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CICapCompression();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CICapPhysicalHeight();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CICapPhysicalWidth();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CICapPixelFlavor();
    m_capabilities[pCapability->GetCapID()] = pCapability;


    pCapability = new CICapPixelType();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CICapPlanarChunky();
    m_capabilities[pCapability->GetCapID()] = pCapability;
    
    pCapability = new CICapUnits();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CICapXferMech();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CICapXResolution();
    m_capabilities[pCapability->GetCapID()] = pCapability;

    pCapability = new CICapYResolution();
    m_capabilities[pCapability->GetCapID()] = pCapability;
}


/*******************************************************************************
DESCRIPTION:

CCapHolder::UnloadCapabilities is the method that "cleans up" by deleting all
capability objects thus shutting down the capability engine.

PARAMETERS: 


RETURN VALUE:
void 

NOTES:
This method has to be called before the object gets destroyed but after all 
possbile capability queries and negotiations are over. The ideal place for this
is in response to the TWAIN source manager's request to close the data source
i.e. in response to the DG_CONTROL / DAT_IDENTITY / MSG_CLOSEDS operation
triplet
*******************************************************************************/
void CCapHolder::UnloadCapabilities()
{
    // Delete all the capability objects in the map and then clear the map
    CAP_ITERATOR it;
    for(it = m_capabilities.begin(); it != m_capabilities.end(); it++)
    {
        delete (*it).second;
    }

    m_capabilities.clear();
}


/*******************************************************************************
DESCRIPTION:

CCapHolder::HandleCapabilityMsg handles all capability messages i.e operation
triplets of the form DG_CONTROL / DAT_CAPABILITY / MSG_XXXXX. When the main
dispatch code in DS_Entry receives these messages it obtains the pointer to 
the singleton instance of this class using the GetInstance method and then 
invokes this function to handle the capability message

PARAMETERS: 

PTWMSG pTWMSG - The pointer to the message that was passed into DS_Entry

RETURN VALUE:
TW_UINT16  - The TWAIN status response code

NOTES:
This method receives all capability messages and then searches for a capability
object that can handle the message based on the CapID . If it finds one it call
the handleer function of that object. If it does not then it returns the 
cap unsupported error.
*******************************************************************************/
TW_UINT16 CCapHolder::HandleCapabilityMsg(PTWMSG pTWMSG)
{
    // We assume the worst and set the return code to failure and the 
    // condition code (maintained in the session object to capunsupported
    // if an object if found it will change the condition code in its
    // handler and will return the appropriate reponse which will 
    // override our assumptions
    TW_UINT16 CapID = ((pTW_CAPABILITY)pTWMSG->pData)->Cap;
    TW_UINT16 twRc  = TWRC_FAILURE;
    CSession::GetInstance()->SetStatusCode(TWCC_CAPUNSUPPORTED);

    CAP_ITERATOR it = m_capabilities.find(CapID);

    if (it != m_capabilities.end())
    {
        TRACE("Twain_LM983x : Handling Capability Msg - %s \n", 
                (*it).second->GetCapName().c_str());

        twRc = (*it).second->HandleCapMessage(pTWMSG);
    }
    
    return twRc;
}


/*******************************************************************************
DESCRIPTION:

CCapHolder::GetSupportedCapList returns a vector of capIDs which represents
all the capability objects currently held in this container. 

PARAMETERS: 

vector<TW_UINT16>& capIDs - The vector which has to be filled in with the 
capIDs of the currently created capability objects. This will be first 
cleared before being returned with the current list.

RETURN VALUE:
void 

NOTES:
The vector contains CapsIDS as defined by the TWAIN spec i.e. UINTS which 
represent the capability objects not pointers to the objects themselves!
*******************************************************************************/
void CCapHolder::GetSupportedCapList(vector<TW_UINT16>& capIDs)
{
    capIDs.clear();

    CAP_ITERATOR it;
    for(it = m_capabilities.begin(); it != m_capabilities.end(); it++)
    {
        capIDs.push_back((*it).first);
    }
}


/*******************************************************************************
DESCRIPTION:

CCapHolder::GetCapObject method allows the called to retrieve a pointer to 
a capability object based on the CAPID of the object

PARAMETERS: 

TW_UINT16 capID - The capid for which the capability object  has to be returned

CCapBase*& capObject - The object pointer will be returned in this paramter if 
an object is found correspoding to the cap Id

RETURN VALUE:
bool - false indicates that no such capid is supported, true indicates that 
can object was found

NOTES:
*** Warning ** the pointer to the object has to be used within the scope of 
the calling function and not retained across function invocations. In the 
current scheme the capability objects are created and destroyed only during 
module initialization and deinitialization but this could change so object
pointers queried using this method should not be stored across function 
invocations.
*******************************************************************************/
bool CCapHolder::GetCapObject(TW_UINT16 capID, CCapBase*& capObject)
{
    CAP_ITERATOR it = m_capabilities.find(capID);

    if (it != m_capabilities.end())
    {
        capObject = (*it).second;
        return true;
    }
    
    return false;
}
