/*******************************************************************************
FILE        : Twain_CapImplementation.h
REVISION    : 1.0
DATE        : 9/28/99

DESCRIPTION :

This file contains the defintion of the CCapImplementation class which 
is the concrete base for all capability classes. In this data source the
capabilities are represented by objects which can handle messages to 
get and set capabilties and also act as the store for the supported
cap values and current and default settings. 
This class is parametrized on the data type of the capability value.

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/
#ifndef _TWAIN_CAP_IMPLEMENTATION_H__
#define _TWAIN_CAP_IMPLEMENTATION_H__

#include "Twain_CapBase.h"
#include "Twain_types.h"
#include "Twain_Session.h"
#include "Twain_Containers.h"

template <class T> 
class  CCapImplementation : public CCapBase
{
public:
    CCapImplementation(){};

    // These functions will be used for querying the capability attributes
    // for use within the TWAIN module itself
    TWAIN_RESULT GetCapValues(vector<T>&   capValues)
    { 
        if (UpdateCapList() != TWAIN_OK)
            return TWAIN_ERROR;
        capValues = m_capValues;
        return TWAIN_OK; 
    }

    // When using these functions with derived classes it is important to 
    // remember for some capabilities the current and default values might
    // not make sense (like CAP_SUPPORTEDCAPS). So the presence of these 
    // functions here does not mean that they will return correct results
    // for all derived classes. They are meant to provide a standard way
    // of querying for these values ** where they are applicable ***
    TWAIN_RESULT GetCurrentValue(T& value)   
    { 
        if (UpdateIndices() != TWAIN_OK)
            return TWAIN_ERROR;
        value =  m_capValues[m_currentValueIndex];
        return TWAIN_OK;
    }
    
    TWAIN_RESULT GetDefaultValue(T& value)   
    { 
        if (UpdateIndices() != TWAIN_OK)
            return TWAIN_ERROR;
        value = m_capValues[m_defaultValueIndex];
        return TWAIN_OK;
    }

protected:
    // This function is used by accessor function implementations
    // above to make sure that the data members declared in this 
    // class namely the list of cap values and the indexes for the 
    // current and default values are valid.
    // Derived classes will implement this function and either return 
    // an error if the data values are not valid or ensure that when 
    // the function returns successfully the data valus are valid
    // The default implementation assumes that all the data is valid
    // right after construction.
    virtual TWAIN_RESULT UpdateCapList(){return TWAIN_OK;}
    virtual TWAIN_RESULT UpdateIndices(){return TWAIN_OK;}

    // The implementation for this message does not vary for the 
    // capability classes so it is implemented in this base class
    TW_UINT16 HandleQuerySupportedCaps(pTW_CAPABILITY pCapability)
    {
        TW_UINT16 twRc = TWRC_SUCCESS;

        pCapability->ConType = TWON_ONEVALUE;
        
        TW_UINT16 conditionCode = BuildUpOneValue(  pCapability, 
                                                    TWTY_INT32, 
                                                    GetSupportedCaps(),
                                                    true);

        if (conditionCode != TWCC_SUCCESS)
        {
            CSession *pSession = CSession::GetInstance();
		    twRc = pSession->HandleError(E_CAPABILITYMSG, conditionCode);
        }

        return twRc;
    }

    // The default implementation of these functions simply 
    // returns FAILURE with a condition code indicating that this
    // operation is not supported on this  capability. Therefore
    // If an operation is supported the corresponding function 
    // must be overridden by the derived class
    TW_UINT16 HandleGetCapValues(pTW_CAPABILITY)
    {
	    return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                TWCC_CAPBADOPERATION);
    }

    TW_UINT16 HandleGetCurrentCap(pTW_CAPABILITY)
    {
	    return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_CAPBADOPERATION);
    }

    TW_UINT16 HandleGetDefaultCap(pTW_CAPABILITY)
    {
	    return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_CAPBADOPERATION);
    }

    TW_UINT16 HandleResetCap(pTW_CAPABILITY)
    {
	    return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_CAPBADOPERATION);
    }

    TW_UINT16 HandleSetCap (pTW_CAPABILITY)
    {
	    return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_CAPBADOPERATION);
    }


    // Implementation support helpers which factor a lot of common
    // code into this base class 
    TW_UINT16 ReturnOneValue(pTW_CAPABILITY pCapability, const T& value)
    {
        TW_UINT16 conditionCode = BuildUpOneValue(  pCapability, 
                                                    GetItemType(), 
                                                    value,
                                                    true );

        if (conditionCode != TWCC_SUCCESS)
        {
            CSession *pSession = CSession::GetInstance();
		    return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
        }

        return TWRC_SUCCESS;
    }

    
    TW_UINT16 ReturnCapValuesAsEnumeration(pTW_CAPABILITY pCapability)
    {
        TW_UINT16 conditionCode = BuildUpEnumeration(   pCapability, 
                                                        GetItemType(), 
                                                        m_currentValueIndex,
                                                        m_defaultValueIndex,
                                                        m_capValues,
                                                        true );

        if (conditionCode != TWCC_SUCCESS)
        {
            CSession *pSession = CSession::GetInstance();
		    return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
        }

        return TWRC_SUCCESS;
    }

    TW_UINT16 ReturnCapValuesAsArray(pTW_CAPABILITY pCapability)
    {
        TW_UINT16 conditionCode = BuildUpArray( pCapability, 
                                                GetItemType(), 
                                                m_capValues,
                                                true );

        if (conditionCode != TWCC_SUCCESS)
        {
            CSession *pSession = CSession::GetInstance();
		    return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
        }

        return TWRC_SUCCESS;
    }

    int GetIndexOfValue(T& value)
    {
        for (int i=0; i < m_capValues.size(); i++)
        {
            if (m_capValues[i] == value)
                return i;
        }
        
        return -1;
    }

    TW_UINT16 HandleSetWithEnumeration(pTW_CAPABILITY pCapability)
    {
        vector<TW_UINT16> enums;
        TW_UINT16 conditionCode = ExtractEnumerationToVector(pCapability,
                                                                enums);
        if (conditionCode == TWCC_SUCCESS)
            conditionCode = RestrictCapValues(enums, m_capValues);

        if (conditionCode != TWCC_SUCCESS)
        {
            CSession *pSession = CSession::GetInstance();
		    return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
        }

        return TWRC_SUCCESS;
    }

    TW_UINT16 RestrictCapValues(vector<T>& source, vector<T>& destination)
    {
        // Every element in the source has to be in the destination 
        // otherwise we will flag a bad value
        for (int i = 0; i < source.size(); i++)
        {
            if (find(destination.begin(), destination.end(), source[i])
                == destination.end())
            {
                return TWCC_BADVALUE;
            }
        }

        destination.clear();
        destination = source;
        return TWCC_SUCCESS;
    }
    
    // Data common to all capability objects
    vector<T>   m_capValues;              // The array containing the values 
                                        // for this capability
    int         m_defaultValueIndex;    // The index of the default value for this cap
    int         m_currentValueIndex;    // The index of the current value for this cap


};

#endif // _TWAIN_CAP_IMPLEMENTATION_H__