/*******************************************************************************
FILE        : Twain_Containers.h
REVISION    : 1.0
DATE        : 9/22/99

DESCRIPTION :

This file contains the function prototype for the TWAIN container handling 
functions. These functions provide services for packing TWAIN containers 
using data from STL vectors and the reverse function of unpacking 
data into STL vectors or simple data types 

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/
#ifndef TWAIN_CONTAINERS_H__
#define TWAIN_CONTAINERS_H__

#include "Twain_Utility.h"

/*******************************************************************************
DESCRIPTION:

The ExtractEnumerationToVector method essentially copies the given enumeration
into a STL vector . This is a replacement for the C type void pointer based
container handling found in the TWAIN toolkit data source 

PARAMETERS: 

pTW_CAPABILITY pCapability
vector<T>&     capValues

RETURN VALUE:
TW_UINT16  - TWAIN errror code  - returns TRC_FAILURE if locking the container
memory fails

NOTES:

*******************************************************************************/
template <class T>  
TW_UINT16 ExtractEnumerationToVector(pTW_CAPABILITY pCapability, 
                                     vector<T>&     capValues)
{
	ASSERT(pCapability);
	ASSERT(pCapability->ConType == TWON_ENUMERATION);

    TW_UINT16 conditionCode = TWCC_SUCCESS;

    pTW_ENUMERATION pEnumeration    = NULL;
	LPVOID          pItemList       = NULL;

    pEnumeration = (pTW_ENUMERATION)GlobalLock(pCapability->hContainer);

	if (pEnumeration != NULL)
	{
		// assign base address of ItemList array to 'generic' pointer
		pItemList = (LPVOID)pEnumeration->ItemList;
        // Clear any existing data in the vector
        capValues.clear();
        // Copy the data from the enumeration list into the vector
        for (int i = 0; i < pEnumeration->NumItems; i++)
        {
            capValues.push_back(*((T*)pItemList+i));
        }

        // If we successfully locked the memory then we have to unlock it 
        // too, however we dont have to free the memory, the TWAIN application 
        // will do that for us 
        GlobalUnlock(pCapability->hContainer);
	} 
    else
    {
        conditionCode = TWCC_BADVALUE;
    }
    
    return conditionCode;
}

template <class T>
TW_UINT16 ExtractOneValue(pTW_CAPABILITY pCapability,
                          T&             capValue)
{
	ASSERT(pCapability);
	ASSERT(pCapability->ConType == TWON_ONEVALUE);

    TW_UINT16 conditionCode = TWCC_SUCCESS;

    pTW_ONEVALUE pOneValue = NULL;
    
    pOneValue = (pTW_ONEVALUE)GlobalLock(pCapability->hContainer);

	if (pOneValue != NULL)
	{
		capValue = (T)pOneValue->Item;

        GlobalUnlock(pCapability->hContainer);
    }
    else
    {
        conditionCode = TWCC_BADVALUE;
    }

    return conditionCode;
}


template <class T> 
TW_UINT16 BuildUpOneValue(pTW_CAPABILITY pCapability,
                          TW_UINT16      itemType,
                          const T&       capValue,
                          bool           allocateMemory)
{
	ASSERT(pCapability);

	if (allocateMemory)	
	{
		pCapability->hContainer = (TW_HANDLE)GlobalAlloc(GMEM_MOVEABLE, 
                                        sizeof(TW_ONEVALUE)+sizeof(T) );

            if (pCapability->hContainer == NULL)
		{																				
			return(TWCC_LOWMEMORY);
		}																				
	}

	// tell APP the ConType returning
	pCapability->ConType = TWON_ONEVALUE;

    pTW_ONEVALUE pOneValue = (pTW_ONEVALUE)GlobalLock(pCapability->hContainer);

	if (pOneValue != NULL)
	{
		pOneValue->ItemType         = itemType; 
		*(T*)&pOneValue->Item       = capValue;	 

        GlobalUnlock(pCapability->hContainer);
	}
    else
    {
        // We could not lock the memory (this should never happen in Win32)
        // so free the memory that was allocated and return an error
        GlobalFree(pCapability->hContainer);

        return TWCC_OPERATIONERROR;
    
    }
	
    return  TWCC_SUCCESS;
}

template <class T> 
TW_UINT16 BuildUpArray  (pTW_CAPABILITY   pCapability, 
                           TW_UINT16        itemType,
                           const vector<T>& capValues,
                           bool             allocateMemory)
{
	ASSERT(pCapability);

	if (allocateMemory)	
	{
		pCapability->hContainer = (TW_HANDLE)GlobalAlloc(GMEM_MOVEABLE,
		                            (sizeof(TW_ARRAY) - sizeof(TW_UINT8))+
		                               capValues.size()*sizeof(T) );


        if (pCapability->hContainer == NULL)
		{																				
			return(TWCC_LOWMEMORY);
		}																				
	}

	pCapability->ConType = TWON_ARRAY;  	
    pTW_ARRAY pArray = (pTW_ARRAY)GlobalLock(pCapability->hContainer);

    if (pArray != NULL)
	{
		pArray->ItemType = itemType;    // TWTY_XXXX
		pArray->NumItems = capValues.size();

		// assign base address of ItemList array to 'generic' pointer
		// i.e. reposition the struct pointer to overlay the allocated block
		LPVOID pVoid = (LPVOID)pArray->ItemList;

		for (int j=0; j < capValues.size(); j++)
		{
			
            *((T*)pVoid+j) = capValues[j];

		} // end of for loop

		GlobalUnlock(pCapability->hContainer);
	}
    else
    {
        // We could not lock the memory (this should never happen in Win32)
        // so free the memory that was allocated and return an error
        GlobalFree(pCapability->hContainer);

        return TWCC_OPERATIONERROR;
    
    
    }

    return TWCC_SUCCESS;
}


template <class T> 
TW_UINT16 BuildUpEnumeration(pTW_CAPABILITY     pCapability, 
                            TW_UINT16           itemType,
                            TW_UINT32           currentValueIndex, 
                            TW_UINT32           defaultValueIndex,
                            const vector<T>&    capValues,
                            bool                allocateMemory)
{
	ASSERT(pCapability);

    if (allocateMemory)
	{
		// allocate a block large enough for struct and complete enumeration array
        pCapability->hContainer = (TW_HANDLE)GlobalAlloc(GMEM_MOVEABLE,
    		                        (sizeof(TW_ENUMERATION) - sizeof(TW_UINT8)) +
                                        capValues.size()*sizeof(T) );

		if (pCapability->hContainer == NULL)
		{
			return TWCC_LOWMEMORY;
		}
	}

	// fill in contype
	pCapability->ConType = TWON_ENUMERATION;
    
    pTW_ENUMERATION pEnumeration = (pTW_ENUMERATION)GlobalLock(pCapability->hContainer);

    if (pEnumeration != NULL)
	{
		pEnumeration->ItemType     = itemType;        	    // TWTY_XXXX
		pEnumeration->NumItems     = capValues.size();      // TWPT_XXXX...
		pEnumeration->CurrentIndex = currentValueIndex;	    // current index setting
		pEnumeration->DefaultIndex = defaultValueIndex;     // default index setting

		// assign base address of ItemList array to 'generic' pointer
		// i.e. reposition the struct pointer to overlay the allocated block
		LPVOID pVoid = (LPVOID)pEnumeration->ItemList;
        
        // The size of the vector and the NumItems in the passed in 
        // enumeration must be the same
		for (int j=0; j < capValues.size() ; j++)
		{
			*((T*)pVoid+j) = capValues[j];
		} // end of for loop

        GlobalUnlock(pCapability->hContainer);

    }	 //if pEnumeration != NULL
    else
    {
        // We could not lock the memory (this should never happen in Win32)
        // so free the memory that was allocated and return an error
        GlobalFree(pCapability->hContainer);

        return TWCC_OPERATIONERROR;
    }


  // now you have a dynamically sized enumeration array WITHIN a struct
  return TWCC_SUCCESS;
}

#endif // TWAIN_CONTAINERS_H__