/*******************************************************************************
FILE        : Twain_DSEntry.cpp
REVISION    : 1.0
DATE        : 9/12/99

DESCRIPTION :

This module contains the implementaion of the main entry point function 
for  the TWAIN data source as required by the TWAIN spec - DS_Entry, and
the main dispatch function for operations triplet handling - Dispatch

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/

/*******************************************************************************
// Required Includes
*******************************************************************************/
#include "stdafx.h"
#include "Twain_OperationTriplets.h"
#include "Twain_UIInterface.h"
#include "Twain_Session.h"

/*******************************************************************************
//	The operation triplet dispatch tables 
*******************************************************************************/
MAPPER ControlMap[] =
{
	{	DAT_IDENTITY,		IdentityMsg},
	{	DAT_USERINTERFACE,	InterfaceMsg},
	{	DAT_CAPABILITY, 	CapabilityMsg},
	{	DAT_STATUS, 		StatusMsg},
	{	DAT_PENDINGXFERS,	PendingXferMsg},
	{	DAT_SETUPMEMXFER,	SetupMemXferMsg},
	{	DAT_XFERGROUP,		XferGroupMsg},
	// THIS MUST BE LAST
	{	0,					NULL }
};

MAPPER ImageMap[] =
{
	{	DAT_IMAGEINFO,		ImageInfoMsg},
	{	DAT_IMAGELAYOUT,	ImageLayoutMsg},
	{	DAT_IMAGEMEMXFER,	ImageMemXferMsg},
	{	DAT_IMAGENATIVEXFER,ImageNativeXferMsg},
	// THIS MUST BE LAST
	{	0,					NULL}
};

/*******************************************************************************
//	Implementation
*******************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
DESCRIPTION:

DS_Entry - The main entry point into the TWAIN data source as required by the
TWAIN spec

PARAMETERS: 

pTW_IDENTITY pSrc   - ptr to caller, source of message
TW_UINT32    DG     - data group ID, DG_IMAGE, DG_xxxx
TW_UINT16    DAT    - data argument type, DAT_xxxx
TW_UINT16    MSG    - message ID, MSG_xxxx
TW_MEMREF    pData  - pointer to the data

RETURN VALUE:
TW_UINT16 - The Twain status code

NOTES:
This function is normally declared as the single entry point to the Source. In 
our case the TWAIN module does export other functions the TWain_xxx functions 
used by the UI module. 
However this function is exported as Ordinal function 1 (@1), 
(see Twain_LM983x.def) as required by the TWAIN data source manager

*******************************************************************************/
TW_UINT16 FAR PASCAL DS_Entry(pTW_IDENTITY pSrc, TW_UINT32	  DG, 
							  TW_UINT16    DAT,  TW_UINT16	  MSG, 
							  TW_MEMREF    pData)
{
	TWMSG	  dcMsg;
	TW_UINT16 dcRc;
	
	// TW apps are sending down all their messages prior to process them
	// to give us a chance to extract messages such as dialog keyboard
	// handling, accelerator translation or MDI processing in a Source
	// while the Source is enabled. We are doing this filtering prior to process
	// real Source messages to minimize overhead.
	// Relay windows messages down to the Source via the SM.  The SM will send
	// a MSG_PROCESSEVENT to the Source to flag the arrival of a message relayed
	// from the APP that was originally a windows message.
	//
	// Valid states: 5 -- 7.  As soon as the application has enabled the Source it
	// must being sending the Source events.  This allows the Source to recieve events
	// to update it's user interface and to return messages to the application.
	// The app sends down ALL message, the Source decides which ones apply to it.
	
	// added to keep stop Apps from talking to us until properly connected
	// NOTE: Since SM peeks into Sources pSrc may be invalid (NULL) on some calls
	// so the order of evaluation is key. 

    TW_IDENTITY appIdentity = CSession::GetInstance()->GetAppIdentity();    
	if ((!appIdentity.Id) || (appIdentity.Id == pSrc->Id))
	{
		if (MSG == MSG_PROCESSEVENT) 
		{
            // Send event messages directly to the UI interface module
            // and if it is not handled then ask the app to handle it
            // This will work correctly even when there is no UI module
            if (CUIInterface::GetInstance()->
                    HandleMessage((LPMSG)(((pTW_EVENT)pData)->pEvent)))
			{
				dcRc = TWRC_DSEVENT;
				((pTW_EVENT)pData)->TWMessage = MSG_NULL;
			}		
			else 
			{
				dcRc = TWRC_NOTDSEVENT;
			}
		}
		else 
		{
			// Reset the condition code
			if (DG != DG_CONTROL && DAT != DAT_STATUS)
			{
                CSession::GetInstance()->SetStatusCode(TWCC_SUCCESS);
			}
			
			// Build the message for the dispatch routine
			dcMsg.pSrc	= pSrc;
			dcMsg.DG	= DG;
			dcMsg.DAT	= DAT;
			dcMsg.MSG	= MSG;
			dcMsg.pData = pData;
			
			// Route the Source message.
			if (DG == DG_CONTROL)
            {
				dcRc = Dispatch(&dcMsg, ControlMap, DAT);
			}
			else if (DG ==	DG_IMAGE)
            {
				dcRc = Dispatch(&dcMsg, ImageMap, DAT);
			}
			else 
            {
                CSession::GetInstance()->SetStatusCode(TWCC_BADPROTOCOL);
                dcRc = TWRC_FAILURE;
			}

		}
	}	
	
	return dcRc;
} // end DS_Entry()


#ifdef __cplusplus
}
#endif

/*******************************************************************************
DESCRIPTION:

Dispatch

PARAMETERS: 

PTWMSG pTWMsg - ptr to the message struct passed to DS_ENTRY
PMAPPER pMapper - ptr to a struct containing ptr's to functions
TW_UINT16 DAT  - DAT_xxxx that you are 'switching' on 

RETURN VALUE:
TW_UINT16  - condition code

NOTES:
This function will dispatch the correct function based on the
struct identifier.  Searches down through the given Map for the
matching DAT_xxxx type.

*******************************************************************************/
TW_UINT16 Dispatch(PTWMSG pTWMsg, PMAPPER pMapper, TW_UINT16 DAT)
{
	while (pMapper->pFunc != NULL)
	{
		if (pMapper->ID == DAT)
		{
			return(*pMapper->pFunc)(pTWMsg);
		}
		++pMapper;
	}
	
    CSession::GetInstance()->SetStatusCode(TWCC_BADPROTOCOL);
	
    return (TWRC_FAILURE);

} // Dispatch


