/*******************************************************************************
FILE        : Twain_GenCaps.cpp
REVISION    : 1.0
DATE        : 9/12/99

DESCRIPTION :

Class or Function set description - Implementation View

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/
#include "stdafx.h"
#include "Twain_GenCaps.h"
#include "Twain_Session.h"
#include "Twain_Containers.h"
#include "Twain_CapHolder.h"

CCapXferCount::CCapXferCount() 
{
    // We can transfer as many images as the application 
    // wants however when the user is controlling the UI he may 
    // decide to transfer a different number of images 
    m_capValues.push_back(-1);

    // since we support one value the current and default values
    // are the same 
    m_defaultValueIndex = 0;
    m_currentValueIndex = 0;
}

// Overridden handlers for the supported Cap operations 
TW_UINT16 CCapXferCount::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CCapXferCount::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CCapXferCount::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_defaultValueIndex]);
}

TW_UINT16 CCapXferCount::HandleResetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    // Change it our default values and then return the current 
    // value in a one value container
    m_capValues[0]        = -1;
    /// This is redundant since they should be 0 but we are just 
    // doing this to make it clear what the current settings are
    m_currentValueIndex = 0;
    m_defaultValueIndex = 0;

    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CCapXferCount::HandleSetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    // We only accept one value container for this particular 
    // capability 
    if (pCapability->ConType != TWON_ONEVALUE)
    {
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_CAPBADOPERATION);
    }

    // Extract the value 
    TW_UINT16 newValue;
    TW_UINT16 conditionCode =  ExtractOneValue(pCapability, newValue);

    if (conditionCode == TWCC_SUCCESS)
    {
        // Change our single value to that one
        m_capValues[0] = newValue;
        m_currentValueIndex = 0;
        m_defaultValueIndex = 0;
    }
    else
    {
		return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
    }

    return TWRC_SUCCESS;
}


CCapDeviceOnline::CCapDeviceOnline()
{
    // We assume that the device is online . This has to be changed 
    // to a call to an actual function to determine if the scanner 
    // is online
    m_capValues.push_back(TRUE);

    // since we support one value the current and default values
    // are the same 
    m_defaultValueIndex = 0;
    m_currentValueIndex = 0;
    
    return;
}

// Overridden handlers for the supported Cap operations 
TW_UINT16 CCapDeviceOnline::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

CCapUIControllable::CCapUIControllable()
{
    // We can start up with our UI disabled so we return 
    // TRUE for UI controllable
    m_capValues.push_back(TRUE);

    // since we support one value the current and default values
    // are the same 
    m_defaultValueIndex = 0;
    m_currentValueIndex = 0;
    
    return;
}

// Overridden handlers for the supported Cap operations 
TW_UINT16 CCapUIControllable::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

CCapSupportedCaps::CCapSupportedCaps()
{
    // *** WARNING *** - We do  not do anything here - everytime we
    // receive a request we will get the current capability
    // list from the cap holder and copy it into our array
    // Attempting to do it now will cause all sorts of havoc
    // with the integrity of the data since other capability
    // objects which are not yet instantaied will be left out of
    // the reckoning

    // The default and current values don't make sense for this capability
    // so we just initialize to zeros.
    m_defaultValueIndex = 0;
    m_currentValueIndex = 0;
}

// Overridden handlers for the supported Cap operations 
TW_UINT16 CCapSupportedCaps::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    // get the data from the Cap holder and fill up our array
    CCapHolder::GetInstance()->GetSupportedCapList(m_capValues);

    return ReturnCapValuesAsArray(pCapability);
}

TW_UINT16 CCapSupportedCaps::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    // get the data from the Cap holder and fill up our array
    CCapHolder::GetInstance()->GetSupportedCapList(m_capValues);

    return ReturnCapValuesAsArray(pCapability);
}

TW_UINT16 CCapSupportedCaps::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    // get the data from the Cap holder and fill up our array
    CCapHolder::GetInstance()->GetSupportedCapList(m_capValues);

    return ReturnCapValuesAsArray(pCapability);
}
