/*******************************************************************************
FILE        : Twain_ImageCaps.cpp
REVISION    : 1.0
DATE        : 9/29/99

DESCRIPTION :

Class or Function set description - Implementation View

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/
#include "stdafx.h"
#include "Twain_ImageCaps.h"
#include "Twain_Session.h"
#include "Twain_Containers.h"
#include "Twain_CapHelpers.h"
#include "Twain_ScanInterface.h"
#include "Twain_utility.h"

CICapCompression::CICapCompression()
{
    m_capValues.push_back(TWCP_NONE);
    m_defaultValueIndex = 0;
    m_currentValueIndex = 0;
}

// Overridden handlers for the supported Cap operations 
TW_UINT16 CICapCompression::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    return ReturnCapValuesAsEnumeration(pCapability);
}

TW_UINT16 CICapCompression::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CICapCompression::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_defaultValueIndex]);
}

CICapPlanarChunky::CICapPlanarChunky()
{
    m_capValues.push_back(TWPC_CHUNKY);

    m_defaultValueIndex = 0;
    m_currentValueIndex = 0;
}

TW_UINT16 CICapPlanarChunky::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    return ReturnCapValuesAsEnumeration(pCapability);
}

TW_UINT16 CICapPlanarChunky::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CICapPlanarChunky::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_defaultValueIndex]);
}

CICapPhysicalWidth::CICapPhysicalWidth()
{
    // We will determine the physical width when somebody actually 
    // inquires about it
    m_defaultValueIndex = -1;
    m_currentValueIndex = -1;
}

TWAIN_RESULT CICapPhysicalWidth::UpdateCapList()
{
    if (m_currentValueIndex < 0)
    {
        TW_FIX32 physicalWidth;

        // get the physical width from the scanner module
        // which will always be in pixels 
        // Then figure out what unit we will need to report
        // it in and then convert it based on the current units
        int     maxScanWidth;
        long    itemCount;

        TWAIN_RESULT result = CScanInterface::GetInstance()->
                ScanGetCapability(  TWAIN_MAX_SCAN_WIDTH,
                                    NULL,
                                    &maxScanWidth,
                                    &itemCount);

        if (result != TWAIN_OK)
        {
            return TWAIN_ERROR;
        }

        physicalWidth = FloatToFIX32(float(maxScanWidth));

        // If there is no error then return the values in a 
        // one value container if not then return an error code
        // that is approprriate 
        m_capValues.clear();
        m_capValues.push_back(physicalWidth);
        m_currentValueIndex = 0;
        m_defaultValueIndex = 0;
    }

    return TWAIN_OK;
}

TWAIN_RESULT CICapPhysicalWidth::UpdateIndices()
{
    // Same logic as UpdateCapList - both getting the list and updating
    // the indices are done in the same method
    return UpdateCapList();
}

TW_UINT16 CICapPhysicalWidth::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    TW_FIX32        physicalWidth;
    TWAIN_RESULT    twainResult = GetCurrentValue(physicalWidth);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    // The current value is always stored in pixels so if we 
    // are currently using some other units we need to do a
    // conversion here to that units
    TW_UINT16 currentUnits;
    twainResult = GetCurrentUnits(currentUnits);
    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    if (currentUnits != TWUN_PIXELS)
    {
        // Another assumption that the only other units we support are 
        /// inches so if we are not doing pixels we are doing inches
        // so we get the current YRes and figure out the height in inches
        float xRes;
        twainResult = GetCurrentYResolution(xRes);
        if ((twainResult != TWAIN_OK) || (xRes <= 0))
        {
            return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                        TWCC_OPERATIONERROR);
        }

        physicalWidth = ::FloatToFIX32(FIX32ToFloat(physicalWidth) / xRes);
         
    }

    return ReturnOneValue(pCapability, physicalWidth);
}

TW_UINT16 CICapPhysicalWidth::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    // Since we only have one value - the same routine is used for 
    // get, getcurrent and getdefault
    return HandleGetCapValues(pCapability);
}

TW_UINT16 CICapPhysicalWidth::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    // Since we only have one value - the same routine is used for 
    // get, getcurrent and getdefault
    return HandleGetCapValues(pCapability);
}

CICapPhysicalHeight::CICapPhysicalHeight()
{
    // We will get this on demand so for now we simply mark as values 
    // as unknown
    m_currentValueIndex = -1;
    m_defaultValueIndex = -1;
}

TWAIN_RESULT CICapPhysicalHeight::UpdateCapList()
{
    if (m_currentValueIndex < 0)
    {
        TW_FIX32 physicalHeight;

        // get the physical width from the scanner module
        // which will always be in pixels - [VASSUME]
        int     maxScanHeight;
        long    itemCount;

        TWAIN_RESULT result = CScanInterface::GetInstance()->
                ScanGetCapability(  TWAIN_MAX_SCAN_HEIGHT,
                                    NULL,
                                    &maxScanHeight,
                                    &itemCount);

        if (result != TWAIN_OK)
        {
            return TWAIN_ERROR;
        }

        physicalHeight = FloatToFIX32(float(maxScanHeight));

        // If there is no error then return the values in a 
        // one value container if not then return an error code
        // that is approprriate 
        m_capValues.clear();
        m_capValues.push_back(physicalHeight);
        m_currentValueIndex = 0;
        m_defaultValueIndex = 0;
    }
    
    return TWAIN_OK;
}

TWAIN_RESULT CICapPhysicalHeight::UpdateIndices()
{
    // In this case the UpdateCapList and UpdateIndices basically do the 
    // same thing
    return UpdateCapList();
}

TW_UINT16 CICapPhysicalHeight::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    TW_FIX32        physicalHeight;
    TWAIN_RESULT    twainResult = GetCurrentValue(physicalHeight);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    // The current value is always stored in pixels so if we 
    // are currently using some other units we need to do a
    // conversion here to that units
    TW_UINT16 currentUnits;
    twainResult = GetCurrentUnits(currentUnits);
    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    if (currentUnits != TWUN_PIXELS)
    {
        // Another assumption that the only other units we support are 
        /// inches so if we are not doing pixels we are doing inches
        // so we get the current YRes and figure out the height in inches
        float yRes;
        twainResult = GetCurrentYResolution(yRes);
        if ((twainResult != TWAIN_OK) || (yRes <= 0))
        {
            return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                        TWCC_OPERATIONERROR);
        }

        physicalHeight = FloatToFIX32(FIX32ToFloat(physicalHeight) / yRes);
                
    }

    return ReturnOneValue(pCapability, physicalHeight);
}

TW_UINT16 CICapPhysicalHeight::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    // Since we only have one value - the same routine is used for 
    // get, getcurrent and getdefault
    return HandleGetCapValues(pCapability);
}

TW_UINT16 CICapPhysicalHeight::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    // Since we only have one value - the same routine is used for 
    // get, getcurrent and getdefault
    return HandleGetCapValues(pCapability);
}

CICapPixelFlavor::CICapPixelFlavor()
{
    // We currently only support pixels with one sense that is 
    // 0 means black and the maximum value (like 255) means white
    m_capValues.push_back(TWPF_CHOCOLATE);
    m_currentValueIndex = 0;
    m_defaultValueIndex = 0;
}

TW_UINT16 CICapPixelFlavor::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CICapPixelFlavor::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CICapPixelFlavor::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

CICapBitOrder::CICapBitOrder() 
{
    m_capValues.push_back(TWBO_MSBFIRST);
    m_capValues.push_back(TWBO_LSBFIRST);

    // The first value is the default as specfied in the TWAIN spec
    m_defaultValueIndex = 0;
    m_currentValueIndex = 0;
}

// Overridden handlers for the supported Cap operations 
TW_UINT16 CICapBitOrder::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    return ReturnCapValuesAsEnumeration(pCapability);
}

TW_UINT16 CICapBitOrder::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CICapBitOrder::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_defaultValueIndex]);
}

TW_UINT16 CICapBitOrder::HandleResetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    // Change our current value back to our default value gs are
    m_currentValueIndex = m_defaultValueIndex ;

    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CICapBitOrder::HandleSetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    // We only accept one value container for this particular 
    // capability 
    if (pCapability->ConType != TWON_ONEVALUE)
    {
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_CAPBADOPERATION);
    }

    // Extract the value 
    TW_UINT16 newValue;
    TW_UINT16 conditionCode =  ExtractOneValue(pCapability, newValue);

    if (conditionCode == TWCC_SUCCESS)
    {
        // Check the value to see if it something we recognize - we store 
        // MSBFirst as our first value and LSBFirst as our second
        if (newValue == TWBO_MSBFIRST)
        {
            m_currentValueIndex = 0;
        }
        else if (newValue == TWBO_LSBFIRST)
        {
            m_currentValueIndex = 1;
        }
        else // we have an unrecognized value
        {
            return pSession->HandleError(E_CAPABILITYMSG, TWCC_BADVALUE);
        }
    }
    else
    {
		return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
    }

    return TWRC_SUCCESS;
}

/*******************************************************************************
    Implementation of the  CICapBitDepth capability class 
*******************************************************************************/

CICapBitDepth::CICapBitDepth() 
{
    // The bit depths supported at aby time are limited by the pixel type
    // that is currently in use therfore we first need to determine 
    // the pixel type and then form a list of bit depths. To make sure
    // that we are always returning the correct value for the current 
    // settings we always query the scanner module for the currently supported
    // bit depths whenever somebody asks us for the bit depths so we dont do 
    // anything here
    // If we can successfully get a list then the default value is simply the 
    // zeroth one but till then the default value index is invalid
    m_currentValueIndex = -1;
    m_defaultValueIndex = -1;
}

TWAIN_RESULT CICapBitDepth::UpdateCapList()
{
    // First get the current pixel type
    TW_UINT16       currentPixelType;
    TWAIN_RESULT    twainResult =  GetCurrentPixelType(currentPixelType);

    if (twainResult != TWAIN_OK)
    {
        return TWAIN_ERROR;
    }

    // Get the current list from the scanner for this pixel type
    int     bitDepths[SCAN_CAPID_MAXNUM];
    long    numOutputElements;
    int     pixelType = currentPixelType;

    TWAIN_RESULT  resultCode =     
        CScanInterface::GetInstance()->ScanGetCapability(SCAN_CAP_PIXELDEPTH, 
                                                        &pixelType, 
                                                        bitDepths, 
                                                        &numOutputElements);    

    if ((resultCode != TWAIN_OK) || (numOutputElements <= 0))
    {
        return TWAIN_ERROR;
    }

    // now we need to store this list in our list of cap values  
    m_capValues.clear();
    for (int i=0; i < numOutputElements; i++)
        m_capValues.push_back(bitDepths[i]);

    return TWAIN_OK;
}

TWAIN_RESULT CICapBitDepth::UpdateIndices()
{
    if (UpdateCapList() != TWAIN_OK)
        return TWAIN_ERROR;

    TWAIN_PARAM_REC scanRec;
    
    TWAIN_RESULT result = 
        CScanInterface::GetInstance()->ScanGetScanParams(&scanRec);
    
    // We are only interested in figuring out which value from our
    // list is the current value used by the scanner. We then store
    // the index of this value in our member variables. This is 
    // necessary so that functions which use this index can be 
    // assured of correct results
    if (result == TWAIN_OK)
    {
        TW_UINT16 bitDepth  = scanRec.bitsPerPixel;
        m_defaultValueIndex = 0;
        m_currentValueIndex = GetIndexOfValue(bitDepth);
        if (m_currentValueIndex < 0)
            result = TWAIN_ERROR;
    }

    return result;
}

// Overridden handlers for the supported Cap operations 
TW_UINT16 CICapBitDepth::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    TWAIN_RESULT    twainResult =  UpdateIndices();

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnCapValuesAsEnumeration(pCapability);
}

TW_UINT16 CICapBitDepth::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    TW_UINT16       currentBitDepth;
    TWAIN_RESULT    twainResult =  GetCurrentValue(currentBitDepth);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnOneValue(pCapability, currentBitDepth);
}

TW_UINT16 CICapBitDepth::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    TW_UINT16       defaultBitDepth;
    TWAIN_RESULT    twainResult =  GetDefaultValue(defaultBitDepth);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnOneValue(pCapability, defaultBitDepth);
}

TW_UINT16 CICapBitDepth::HandleResetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow resetting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    TW_UINT16       defaultBitDepth;
    TWAIN_RESULT    twainResult =  GetDefaultValue(defaultBitDepth);
    
    if (twainResult == TWAIN_OK)
    {
        twainResult = SetCurrentBitDepth(defaultBitDepth);
    }

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnOneValue(pCapability, defaultBitDepth);
}

TW_UINT16 CICapBitDepth::HandleSetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    // We will accept one value and enumeration containers however we will
    // only change the value for one value containers. For enumeration 
    // containers we will restrict the list of available choice if the 
    // enumeration is a subset of out supported cap values - This is the 
    // behavior required by the TWAIN spec
    switch (pCapability->ConType)
    {
        case TWON_ONEVALUE:
            {
                // Extract the value 
                TW_UINT16 newValue;
                TW_UINT16 conditionCode =  ExtractOneValue(pCapability, newValue);
                if (conditionCode == TWCC_SUCCESS)
                {
                    TWAIN_RESULT  twainResult = SetCurrentBitDepth(newValue);

                    if (twainResult != TWAIN_OK)
                    {
                        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                                    TWCC_OPERATIONERROR);
                    }
                }
                else
                {
		            return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
                }
            }
            break;
        case TWON_ENUMERATION:
            return HandleSetWithEnumeration(pCapability);
        default:
            return pSession->HandleError(E_CAPABILITYMSG, TWCC_CAPBADOPERATION);
    }

    return TWRC_SUCCESS;
}


/*******************************************************************************
DESCRIPTION:

CICapBitDepth::SetCurrentBitDepth

PARAMETERS: 

TW_UINT16  bitDepth

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT CICapBitDepth::SetCurrentBitDepth(TW_UINT16  bitDepth)
{
    TWAIN_PARAM_REC scanRec;
    
    // Get the current settings and change just the scan bit depth
    TWAIN_RESULT result = 
        CScanInterface::GetInstance()->ScanGetScanParams(&scanRec);

    if (result == TWAIN_OK)
    {
        scanRec.bitsPerPixel = bitDepth;
        
        result = CScanInterface::GetInstance()->ScanSetScanParams(&scanRec);
    }

    return result;
}


/*******************************************************************************
    Implementation of the  CICapPixelType capability class 
*******************************************************************************/


/*******************************************************************************
DESCRIPTION:

CICapPixelType::CICapPixelType

PARAMETERS: 


RETURN VALUE:


NOTES:

*******************************************************************************/
CICapPixelType::CICapPixelType()
{
    // The pixel types supported and current and default values will be 
    // queried on demand from the scan module
    m_currentValueIndex = -1;
    m_defaultValueIndex = -1;
}


TWAIN_RESULT CICapPixelType::UpdateCapList()
{
    // The pixel types supported cannot suddenly change during a TWAIN
    // session so we will query the supported types from the scanner module 
    // only once when we are  first asked for them and the first one in the list 
    // returned from the scanner module will be the default
    if (m_currentValueIndex < 0)
    {
        int     pixelTypes[SCAN_CAPID_MAXNUM];
        long    numOutputElements;

        TWAIN_RESULT  resultCode =     
            CScanInterface::GetInstance()->ScanGetCapability(SCAN_CAP_PIXELTYPE, 
                                                            NULL, 
                                                            pixelTypes, 
                                                            &numOutputElements);    

        if ((resultCode != TWAIN_OK) || (numOutputElements <= 0))
        {
            return TWAIN_ERROR;
        }

        // now we need to store this list in our list of cap values  
        m_capValues.clear();
        for (int i=0; i < numOutputElements; i++)
            m_capValues.push_back(pixelTypes[i]);

    }

    return TWAIN_OK;
}


TWAIN_RESULT CICapPixelType::UpdateIndices()
{
    if (UpdateCapList() != TWAIN_OK)
        return TWAIN_ERROR;

    // In order to make our base class code work correctlywith the
    // cap base functions we will have to make sure that the 
    // indices are correct 
    TWAIN_PARAM_REC scanRec;
    
    TWAIN_RESULT result = 
        CScanInterface::GetInstance()->ScanGetScanParams(&scanRec);

    if (result == TWAIN_OK)
    {
        TW_UINT16 pixelType = scanRec.pixelType;
        m_defaultValueIndex = 0;
        m_currentValueIndex = GetIndexOfValue(pixelType);
        if (m_currentValueIndex < 0)
            result = TWAIN_ERROR;
    }

    return result;
}

/*******************************************************************************
DESCRIPTION:

CICapPixelType::HandleGetCapValues

PARAMETERS: 

pTW_CAPABILITY pCapability

RETURN VALUE:
TW_UINT16 

NOTES:

*******************************************************************************/
TW_UINT16 CICapPixelType::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    if (m_currentValueIndex < 0)
    {
        TWAIN_RESULT twainResult = UpdateIndices();

        if (twainResult != TWAIN_OK)
        {
            return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                        TWCC_OPERATIONERROR);
        }
    }
    
    return ReturnCapValuesAsEnumeration(pCapability);
}


/*******************************************************************************
DESCRIPTION:

CICapPixelType::HandleGetCurrentCap

PARAMETERS: 

pTW_CAPABILITY pCapability

RETURN VALUE:
TW_UINT16 

NOTES:

*******************************************************************************/
TW_UINT16 CICapPixelType::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    TW_UINT16       currentPixelType;
    TWAIN_RESULT    twainResult =  GetCurrentValue(currentPixelType);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnOneValue(pCapability, currentPixelType);
}


/*******************************************************************************
DESCRIPTION:

CICapPixelType::HandleGetDefaultCap

PARAMETERS: 

pTW_CAPABILITY pCapability

RETURN VALUE:
TW_UINT16 

NOTES:

*******************************************************************************/
TW_UINT16 CICapPixelType::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    TW_UINT16 defaultPixelType;
    TWAIN_RESULT twainResult = GetDefaultValue(defaultPixelType);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnOneValue(pCapability, defaultPixelType);    
}


/*******************************************************************************
DESCRIPTION:

CICapPixelType::HandleResetCap

PARAMETERS: 

pTW_CAPABILITY pCapability

RETURN VALUE:
TW_UINT16 

NOTES:

*******************************************************************************/
TW_UINT16 CICapPixelType::HandleResetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();

    if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow resetting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    TW_UINT16       defaultValue;
    TWAIN_RESULT    twainResult = GetDefaultValue(defaultValue);

    if (twainResult == TWAIN_OK)
    {
        twainResult = SetCurrentPixelType(defaultValue);
    }
    
    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }
    
    return ReturnOneValue(pCapability, defaultValue);
}



/*******************************************************************************
DESCRIPTION:

CICapPixelType::HandleSetCap

PARAMETERS: 

pTW_CAPABILITY pCapability

RETURN VALUE:
TW_UINT16 

NOTES:

*******************************************************************************/
TW_UINT16 CICapPixelType::HandleSetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }


    // We will accept one value and enumeration containers however we will
    // only change the value for one value containers. For enumeration 
    // containers we will restrict the list of available choice if the 
    // enumeration is a subset of out supported cap values - This is the 
    // behavior required by the TWAIN spec
    switch (pCapability->ConType)
    {
        case TWON_ONEVALUE:
            {
                // Extract the value 
                TW_UINT16 newValue;
                TW_UINT16 conditionCode =  ExtractOneValue(pCapability, newValue);
                if (conditionCode == TWCC_SUCCESS)
                {
                    // Get the current settings and change the pixel type .
                    // We will accordingly adjust the bit depth
                    TWAIN_RESULT  twainResult = SetCurrentPixelType(newValue);

                    if (twainResult != TWAIN_OK)
                    {
                        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                                    TWCC_OPERATIONERROR);
                    }
                }
                else
                {
		            return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
                }
            }
            break;
        case TWON_ENUMERATION:
            return HandleSetWithEnumeration(pCapability);
        default:
            return pSession->HandleError(E_CAPABILITYMSG, TWCC_CAPBADOPERATION);
    }

    return TWRC_SUCCESS;
}


/*******************************************************************************
DESCRIPTION:

CICapPixelType::SetCurrentPixelType

PARAMETERS: 

TW_UINT16  pixelType

RETURN VALUE:
TWAIN_RESULT 

NOTES:

*******************************************************************************/
TWAIN_RESULT CICapPixelType::SetCurrentPixelType(TW_UINT16  pixelType)
{
    CScanInterface* pScanner = CScanInterface::GetInstance();

    // We also need to change the bit depth to reflect the supported 
    // values for this pixel type
    int     newPixelType = pixelType;
    int     pixelDepths[TWAIN_CAPID_MAXNUM];
    long    numOutputElements;

    // Get the supported bit depths for this pixel type
    TWAIN_RESULT result = pScanner->ScanGetCapability(  SCAN_CAP_PIXELDEPTH, 
                                                        &newPixelType , 
                                                        pixelDepths, 
                                                        &numOutputElements);    
    
    if ((result != TWAIN_OK) || (numOutputElements <= 0))
    {
        return TWAIN_ERROR;
    }


    TWAIN_PARAM_REC scanRec;
    result = pScanner->ScanGetScanParams(&scanRec);


    if (result == TWAIN_OK)
    {
        scanRec.pixelType       = pixelType;
        scanRec.bitsPerPixel    = pixelDepths[0];

        result = pScanner->ScanSetScanParams(&scanRec);
    }

    return result;
}

CICapUnits::CICapUnits()
{
    // We support only inches and pixels and the default has to be inches
    // as per the TWAIN spec
    m_capValues.push_back(TWUN_INCHES);
    m_capValues.push_back(TWUN_PIXELS);

    m_currentValueIndex = 0;
    m_defaultValueIndex = 0;

}

TW_UINT16 CICapUnits::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    return ReturnCapValuesAsEnumeration(pCapability);
}

TW_UINT16 CICapUnits::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CICapUnits::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_defaultValueIndex]);
}

TW_UINT16 CICapUnits::HandleResetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();

    if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow resetting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    m_currentValueIndex = m_defaultValueIndex;

    return ReturnOneValue(pCapability, m_capValues[m_defaultValueIndex]);
}

TW_UINT16 CICapUnits::HandleSetCap (pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    // We only accept one value container for this particular capability 
    if (pCapability->ConType != TWON_ONEVALUE)
    {
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_CAPBADOPERATION);
    }

    // Extract the value 
    TW_UINT16 newValue;
    TW_UINT16 conditionCode =  ExtractOneValue(pCapability, newValue);

    if (conditionCode == TWCC_SUCCESS)
    {
        // See if it is one of the accepted values otherwise 
        // send back a bad value
        switch (newValue)
        {
        case TWUN_PIXELS:
            m_currentValueIndex = 1;
            break;
        case TWUN_INCHES:
            m_currentValueIndex = 0;
            break;
        default:
            return pSession->HandleError(E_CAPABILITYMSG, TWCC_BADVALUE);
        }
    }
    else
    {
		return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
    }

    return TWRC_SUCCESS;
}


CICapXferMech::CICapXferMech()
{
    // We only support NATIVE and MEMORY (buffered) transfers at this point
    m_capValues.push_back(TWSX_NATIVE);
    m_capValues.push_back(TWSX_MEMORY);

    m_defaultValueIndex = 0;
    m_currentValueIndex = 0;
}

TW_UINT16 CICapXferMech::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    return ReturnCapValuesAsEnumeration(pCapability);
}

TW_UINT16 CICapXferMech::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_currentValueIndex]);
}

TW_UINT16 CICapXferMech::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    return ReturnOneValue(pCapability, m_capValues[m_defaultValueIndex]);
}

TW_UINT16 CICapXferMech::HandleResetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();

    if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow resetting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    m_currentValueIndex = m_defaultValueIndex;

    return ReturnOneValue(pCapability, m_capValues[m_defaultValueIndex]);
}

TW_UINT16 CICapXferMech::HandleSetCap (pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    // We will accept one value and enumeration containers however we will
    // only change the value for one value containers. For enumeration 
    // containers we will restrict the list of available choice if the 
    // enumeration is a subset of out supported cap values - This is the 
    // behavior required by the TWAIN spec
    switch (pCapability->ConType)
    {
        case TWON_ONEVALUE:
            {
                // Extract the value 
                TW_UINT16 newValue;
                TW_UINT16 conditionCode =  ExtractOneValue(pCapability, newValue);
                if (conditionCode == TWCC_SUCCESS)
                {
                    // See if it is one of the accepted values otherwise 
                    // send back a bad value
                    switch (newValue)
                    {
                    case TWSX_NATIVE:
                        m_currentValueIndex = 0;
                        break;
                    case TWSX_MEMORY:
                        m_currentValueIndex = 1;
                        break;
                    default:
                        return pSession->HandleError(E_CAPABILITYMSG, TWCC_BADVALUE);
                    }
                }
                else
                {
		            return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
                }
            }
            break;
        case TWON_ENUMERATION:
            return HandleSetWithEnumeration(pCapability);
        default:
            return pSession->HandleError(E_CAPABILITYMSG, TWCC_CAPBADOPERATION);
    }

    return TWRC_SUCCESS;
}


CICapXResolution::CICapXResolution()
{
    // The cap list and the indices will be updated the first time
    // the object is called upon for capability information
    m_currentValueIndex = -1;
    m_defaultValueIndex = -1;
}

TWAIN_RESULT  CICapXResolution::UpdateCapList()
{
    // Since the supported resolutions cannot change during the 
    // TWAIN session we will query the list of supported resolutions
    // only once - namely the firsr time that we asked for this list 
    // by the TWAIN application. However we will not maintain the 
    // current resolution value. this will be maintained by the scan 
    // module and we will query it on demand.    
    if (m_currentValueIndex < 0)
    {
        int     xResList[SCAN_CAPID_MAXNUM];
        long    numOutputElements;

        TWAIN_RESULT  resultCode =     
            CScanInterface::GetInstance()->ScanGetCapability(SCAN_CAP_HRES, 
                                                            NULL, 
                                                            xResList, 
                                                            &numOutputElements);    

        if ((resultCode != TWAIN_OK) || (numOutputElements <= 0))
        {
            return TWAIN_ERROR;
        }

        // now we need to store this list in our list of cap values  
        m_capValues.clear();
        for (int i=0; i < numOutputElements; i++)
            m_capValues.push_back(FloatToFIX32(xResList[i]));
    }

    return TWAIN_OK;
}

TWAIN_RESULT    CICapXResolution::UpdateIndices()
{
    if (UpdateCapList() != TWAIN_OK)
        return TWAIN_ERROR;

    TWAIN_PARAM_REC scanRec;
    
    TWAIN_RESULT result = 
        CScanInterface::GetInstance()->ScanGetScanParams(&scanRec);

    if (result == TWAIN_OK)
    {
        TW_FIX32 xResolution = FloatToFIX32(scanRec.hRes);
        m_defaultValueIndex = 0;
        m_currentValueIndex = GetIndexOfValue(xResolution);
        if (m_currentValueIndex < 0)
            result = TWAIN_ERROR;
    }

    return result;
}

TW_UINT16 CICapXResolution::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    TWAIN_RESULT twainResult = UpdateIndices();

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }
    
    return ReturnCapValuesAsEnumeration(pCapability);
}

TW_UINT16 CICapXResolution::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    TW_FIX32        xResolution;
    TWAIN_RESULT    twainResult =  GetCurrentValue(xResolution);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnOneValue(pCapability, xResolution);

}

TW_UINT16 CICapXResolution::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    TW_FIX32 defaultXResolution;
    TWAIN_RESULT twainResult = GetDefaultValue(defaultXResolution);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnOneValue(pCapability, defaultXResolution);    
}

TW_UINT16 CICapXResolution::HandleResetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();

    if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow resetting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    TW_FIX32 defaultXResolution;
    TWAIN_RESULT twainResult = GetDefaultValue(defaultXResolution);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }

    twainResult = SetCurrentXResolution(FIX32ToFloat(defaultXResolution));

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }
    
    // We don't have a default value so we assume the first value to be the 
    // default So if we are asked to reset we will return a success code
    // and return the current default value which is always the zeroth element
    return ReturnOneValue(pCapability, defaultXResolution);

}

TW_UINT16 CICapXResolution::HandleSetCap (pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    // We only accept one value container for this particular capability 
    if (pCapability->ConType != TWON_ONEVALUE)
    {
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_CAPBADOPERATION);
    }

    // Extract the value 
    TW_UINT16 newValue;
    TW_UINT16 conditionCode =  ExtractOneValue(pCapability, newValue);

    if (conditionCode == TWCC_SUCCESS)
    {
        TWAIN_RESULT  scanResult = SetCurrentXResolution(newValue);

        if (scanResult != TWAIN_OK)
        {
            return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                        TWCC_OPERATIONERROR);
        }
    }
    else
    {
		return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
    }

    return TWRC_SUCCESS;
}

TWAIN_RESULT CICapXResolution::SetCurrentXResolution(float xResolution)
{
    TWAIN_PARAM_REC scanRec;
    
    // Get the current settings and change just the xres Value
    TWAIN_RESULT result = 
        CScanInterface::GetInstance()->ScanGetScanParams(&scanRec);

    if (result == TWAIN_OK)
    {
        scanRec.hRes = xResolution;
        
        result = CScanInterface::GetInstance()->ScanSetScanParams(&scanRec);
    }

    return result;
}

CICapYResolution::CICapYResolution()
{
    m_currentValueIndex = -1;
    m_defaultValueIndex = -1;
}

TWAIN_RESULT    CICapYResolution::UpdateCapList()
{
    // Since the supported resolutions cannot change during the 
    // TWAIN session we will query the list of supported resolutions
    // only once - namely the firsr time that we asked for this list 
    // by the TWAIN application. However we will not maintain the 
    // current resolution value. this will be maintained by the scan 
    // module and we will query it on demand. 
    if (m_currentValueIndex < 0)
    {
        int     yResList[SCAN_CAPID_MAXNUM];
        long    numOutputElements;

        TWAIN_RESULT  resultCode =     
            CScanInterface::GetInstance()->ScanGetCapability(SCAN_CAP_VRES, 
                                                            NULL, 
                                                            yResList, 
                                                            &numOutputElements);    

        if ((resultCode != TWAIN_OK) || (numOutputElements <= 0))
        {
            return TWAIN_ERROR;
        }

        // now we need to store this list in our list of cap values  
        m_capValues.clear();
        for (int i=0; i < numOutputElements; i++)
            m_capValues.push_back(FloatToFIX32(yResList[i]));
    }

    return TWAIN_OK;
}

TWAIN_RESULT  CICapYResolution::UpdateIndices()
{
    if (UpdateCapList() != TWAIN_OK)
        return TWAIN_ERROR;

    TWAIN_PARAM_REC scanRec;
    
    TWAIN_RESULT result = 
        CScanInterface::GetInstance()->ScanGetScanParams(&scanRec);

    if (result == TWAIN_OK)
    {
        TW_FIX32 yResolution = FloatToFIX32(scanRec.vRes);
        m_defaultValueIndex = 0;
        m_currentValueIndex = GetIndexOfValue(yResolution);
        if (m_currentValueIndex < 0)
            result = TWAIN_ERROR;
    }

    return result;
}

   
TW_UINT16 CICapYResolution::HandleGetCapValues(pTW_CAPABILITY pCapability)
{
    TWAIN_RESULT twainResult = UpdateIndices();

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }
    
    return ReturnCapValuesAsEnumeration(pCapability);
}

TW_UINT16 CICapYResolution::HandleGetCurrentCap(pTW_CAPABILITY pCapability)
{
    TW_FIX32        yResolution;
    TWAIN_RESULT    twainResult =  GetCurrentValue(yResolution);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnOneValue(pCapability, yResolution);

}

TW_UINT16 CICapYResolution::HandleGetDefaultCap(pTW_CAPABILITY pCapability)
{
    TW_FIX32        yResolution;
    TWAIN_RESULT    twainResult =  GetDefaultValue(yResolution);

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }

    return ReturnOneValue(pCapability, yResolution);    
}

TW_UINT16 CICapYResolution::HandleResetCap(pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();

    if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow resetting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    TW_FIX32        yResolution;
    TWAIN_RESULT    twainResult =  GetDefaultValue(yResolution);
    
    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }

    twainResult = SetCurrentYResolution(FIX32ToFloat(yResolution));

    if (twainResult != TWAIN_OK)
    {
        return CSession::GetInstance()->HandleError(E_CAPABILITYMSG,
                                                    TWCC_OPERATIONERROR);
    }
    
    // We don't have a default value so we assume the first value to be the 
    // default So if we are asked to reset we will return a success code
    // and return the current default value which is always the zeroth element
    return ReturnOneValue(pCapability, yResolution);
}

TW_UINT16 CICapYResolution::HandleSetCap (pTW_CAPABILITY pCapability)
{
    CSession *pSession = CSession::GetInstance();
	if (pSession->GetState() != CSession::E_STATE4)
    {
        // We only allow setting the cap value in state 4. In all other 
        // states only queries are allowed. Therefore we will fail this
        // operation with a sequence error
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_SEQERROR);
    }

    // We only accept one value container for this particular capability 
    if (pCapability->ConType != TWON_ONEVALUE)
    {
        return pSession->HandleError(E_CAPABILITYMSG, TWCC_CAPBADOPERATION);
    }

    // Extract the value 
    TW_UINT16 newValue;
    TW_UINT16 conditionCode =  ExtractOneValue(pCapability, newValue);

    if (conditionCode == TWCC_SUCCESS)
    {
        TWAIN_RESULT  scanResult = SetCurrentYResolution(newValue);

        if (scanResult != SCAN_OK)
        {
            return CSession::GetInstance()->HandleError(E_CAPABILITYMSG, 
                                                        TWCC_OPERATIONERROR);
        }
    }
    else
    {
		return pSession->HandleError(E_CAPABILITYMSG, conditionCode);
    }

    return TWRC_SUCCESS;
}

TWAIN_RESULT CICapYResolution::SetCurrentYResolution(float yResolution)
{
    TWAIN_PARAM_REC scanRec;
    
    TWAIN_RESULT result = 
        CScanInterface::GetInstance()->ScanGetScanParams(&scanRec);

    if (result == TWAIN_OK)
    {
        scanRec.vRes = yResolution;
        
        result = CScanInterface::GetInstance()->ScanSetScanParams(&scanRec);
    }

    return result;
}
