/*******************************************************************************
FILE        : Twain_OperationTriplets.cpp
REVISION    : 1.0
DATE        : 9/12/99

DESCRIPTION :

Contains the implementaion of the operation triplet handler functions

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/
#include "stdafx.h"
#include "Twain_OperationTriplets.h"
#include "Twain_Session.h"
#include "Twain_CapHolder.h"
#include "Twain_UIInterface.h"
#include "Twain_ScanInterface.h"
#include "Twain_Utility.h"
#include "Twain_CapHelpers.h"

/*******************************************************************************
DESCRIPTION:

IdentityMsg - Handler for DG_CONTROL/DAT_IDENTITY messages

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call

RETURN VALUE:
TW_UINT16 - twRC TWAIN status response

NOTES:
Handle the parse of: (from DAT_IDENTITY msg's)
    - MSG_OPENDS    open the indicated Source
    - MSG_CLOSEDS   close the indicated Source
	- MSG_GET       get identity struct of Source
	- error         error message
*******************************************************************************/
TW_UINT16 IdentityMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In IdentityMsg() with msg - %s\n",  
            GetMsgName(pTWMsg).c_str());

    TW_UINT16 twRc = TWRC_SUCCESS;

	switch (pTWMsg->MSG)
	{
		case MSG_OPENDS:
			OpenDS(pTWMsg);
			break;

		case MSG_CLOSEDS:
			CloseDS();
			break;

		case MSG_GET:  // tell the SM who you are valid 3--7 states
			// copy the information into the message parameter
            *(pTW_IDENTITY)pTWMsg->pData = 
                CSession::GetInstance()->GetSourceIdentity();
			break;

		default:
		    twRc = CSession::GetInstance()->HandleError(E_IDENTITYMSG, 
                                                    TWCC_BADPROTOCOL);
			break;
	}

    return twRc;

} // IdentityMsg



/*******************************************************************************
DESCRIPTION:

InterfaceMsg - Handler for DG_CONTROL/DAT_USERINTERFACE messages

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call

RETURN VALUE:
TW_UINT16 - TWAIN status response

NOTES:
Handle the parse of: (from DAT_USERINTERFACE msg's)
    - MSG_ENABLEDS  enable the open Source
	- MSG_DISABLEDS disable the open Source
	- errors        error message
*******************************************************************************/
TW_UINT16 InterfaceMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In InterfaceMsg() with msg - %s\n",  
            GetMsgName(pTWMsg).c_str());

    TW_UINT16 twRc = TWRC_SUCCESS;

	switch (pTWMsg->MSG)
	{
		case MSG_ENABLEDS:
			twRc = EnableDS((pTW_USERINTERFACE)pTWMsg->pData);
			break;

		case MSG_DISABLEDS:
			twRc = DisableDS((pTW_USERINTERFACE)pTWMsg->pData);
			break;

		default:
            twRc = CSession::GetInstance()->HandleError(E_INTERFACEMSG, 
                                                        TWCC_BADPROTOCOL);
			break;
	}
	return twRc;
}  


/*******************************************************************************
DESCRIPTION:

EnableDS - Final Handler for the triplet 
                DG_CONTROL/DAT_USERINTERFACE/MSG_ENABLEDS

PARAMETERS: 

pTW_USERINTERFACE pUI - ptr to user interface struct (defined by TWAIN)

RETURN VALUE:
TW_UINT16 - TWAIN status response
	- TWRC_SUCCESS
	- TWRC_FAILURE
	- TWCC_SEQERROR
	- TWCC_LOWMEMORY

NOTES:
    Valid states: 4, transitions to state 5 if successful.
	ShowUI bit 1=Source user interface; 0=app supplied
	hParent field should contain a valid windows handle iff
	you are implementing a modal user interface were the hparent
	window will be disabled during access to the user IF dialog
	via EnableWindow(hParent, False).  Else use hParent == NULL
	for Modeless dialogs, see chapter 5 of TWAIN spec.
*******************************************************************************/
TW_UINT16 EnableDS(pTW_USERINTERFACE pUI)
{
    TRACE("Twain_LM983x : In EnableDS()\n");

    TW_UINT16 twRc = TWRC_SUCCESS;

    if (CSession::GetInstance()->GetState() == CSession::E_STATE4)
	{
		// Store this somewhere for later [VTODO] we need to change the
        // interface between the TWAIN and UI modules so that this
        // handle can be passed to the UI module for its use
		HWND hAppWnd = (HWND)pUI->hParent;  // need to know the app window's handle

		// Show the window, IFF user asks for it!
		if (pUI->ShowUI)
		{
            CUIInterface::GetInstance()->LoadUIModule();

            if (CUIInterface::GetInstance()->Activate(true) != TWAIN_OK)
            {
                twRc = CSession::GetInstance()->
                        HandleError(E_ENABLEDS,TWCC_LOWMEMORY);
            }
            else
            {
                CSession::GetInstance()->SetState(CSession::E_STATE5);
            }
		}
		else
		{
			NotifyXferReady();
			/*
			*	Move immediately to state 6 and wait for transfer message
			*/
            CSession::GetInstance()->SetState(CSession::E_STATE6);
		}
	}
	else
	{
        twRc = CSession::GetInstance()->HandleError(E_ENABLEDS,TWCC_SEQERROR);
	}

	return twRc;
} 



/*******************************************************************************
DESCRIPTION:

DisableDS - Final handler for operation triplet 
                DG_CONTROL / DAT_USERINTERFACE / MSG_DISABLEDS

PARAMETERS: 

pTW_USERINTERFACE pUI - ptr to user interface structure

RETURN VALUE:
TW_UINT16 - Twain status response
    - TWRC_SUCCESS
	- TWRC_FAILURE
	- TWCC_SEQERROR

NOTES:
    Valid states: 5 through 7, transitions to 4 if successful.  The 
	App sends down this call when the Source returns the TWMessage "MSG_CLOSEDSREQ".
	ShowUI bit 1=Source user interface; 0=app supplied
	hParent field should contain a valid windows handle

*******************************************************************************/
TW_UINT16 DisableDS(pTW_USERINTERFACE pUI)
{
    TRACE("Twain_LM983x : In DisableDS()\n");

    TW_UINT16 twRc = TWRC_SUCCESS;
    
    CSession *pSession = CSession::GetInstance();
	switch (pSession->GetState())
	{
		case CSession::E_STATE5:
            // hide the UI 
            CUIInterface::GetInstance()->Activate(false);
			// transition to new state
			pSession->SetState(CSession::E_STATE4); 
            break;

		default:
            twRc = pSession->HandleError(E_DISABLEDS, TWCC_SEQERROR);
	}
	return twRc;
} 

/*******************************************************************************
DESCRIPTION:

OpenDS - Final handler for the operation triplet 
    DG_CONTROL / DAT_IDENTITY / MSG_OPENDS

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call

RETURN VALUE:
TW_UINT16 - Twain status response code

NOTES:
This function will 'open' the Source, register it with the SM
	Valid state: 3 transitions to state 4 on success.

*******************************************************************************/
TW_UINT16 OpenDS(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In OpenDS()\n");

    TW_UINT16 twRc = TWRC_SUCCESS;
	OFSTRUCT of;

	// The log is only good for one session of twaining to this source before
	// it is reset
	OpenFile("c:\\twsrc.log", &of, OF_DELETE);

    CSession *pSession = CSession::GetInstance();

    if (pSession->GetState() == CSession::E_STATE3)
	{
		// Only allow one connection at a time.  If AppIdentity.Id is set,
		// then there is already a connection.  If not, then remember who
		// we're connected to so we can send messages to it.
        TW_IDENTITY appIdentity = pSession->GetAppIdentity();
        if (appIdentity.Id)
		{
			twRc = pSession->HandleError(E_OPENDS, TWCC_MAXCONNECTIONS);
		}
		else 
		{
			// This is to ensure that how we identify ourselves to the
			// SM is consistent with what it is keeping around.
			pSession->SetAppIdentity(*pTWMsg->pSrc);
			pSession->SetSourceIdentity(*(pTW_IDENTITY)pTWMsg->pData);
		}

		pSession->SetState(CSession::E_STATE4);

        CScanInterface::GetInstance()->LoadScanModule();

        CCapHolder::GetInstance()->LoadCapabilities();
	}
	else 
	{
		twRc = pSession->HandleError(E_OPENDS, TWCC_SEQERROR);
	}

	return twRc;
} // end OpenDS()



/*******************************************************************************
DESCRIPTION:

CloseDS -Final Handler for the triplet 
    DG_CONTROL / DAT_IDENTITY / MSG_CLOSEDS

PARAMETERS: 

VOID

RETURN VALUE:
TW_UINT16 - Twain status response code

NOTES:
This function will close the Source.

*******************************************************************************/
TW_UINT16 CloseDS(VOID)
{
    TRACE("Twain_LM983x : In CloseDS()\n");

    TW_UINT16 twRc = TWRC_SUCCESS;

    CSession *pSession = CSession::GetInstance();
	switch (pSession->GetState())
	{
		case CSession::E_STATE4:
		case CSession::E_STATE5:
		case CSession::E_STATE6:
		case CSession::E_STATE7:
            {
			    // clear out current connection
			    TW_IDENTITY appIdentity;
                appIdentity.Id = 0;
			    pSession->SetAppIdentity(appIdentity);
			    // Set the DS state
			    pSession->SetState(CSession::E_STATE3);
                CCapHolder::GetInstance()->UnloadCapabilities();
            }
			break;

		default:
			twRc = pSession->HandleError(E_CLOSEDS, TWCC_SEQERROR);
			break;
	}   

	return twRc;
} // end CloseDS()


/*******************************************************************************
DESCRIPTION:

StatusMsg - Handler for the DG_CONTROL / DAT_STATUS messages

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call

RETURN VALUE:
TW_UINT16 - Twain status response code

NOTES:
    Valid states 3 through 7.  Source fills in all fields of pDSMStatus 
	and clears it's condition code.

*******************************************************************************/
TW_UINT16 StatusMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In StatusMsg() with msg - %s\n",  
        GetMsgName(pTWMsg).c_str());

    TW_UINT16 twRc = TWRC_SUCCESS;

    CSession *pSession = CSession::GetInstance();
    switch (pTWMsg->MSG)
	{
		case MSG_GET:
			*(pTW_STATUS)pTWMsg->pData = pSession->GetSourceStatus();
			pSession->SetStatusCode(TWCC_SUCCESS); // clear condition code
			break;

			// Invalid triplet
		default:
			twRc = pSession->HandleError(E_STATUSMSG, TWCC_BADPROTOCOL);
			break;
	}

	return twRc;
} // StatusMsg()


/*******************************************************************************
DESCRIPTION:

PendingXferMsg - Handler for DG_CONTROL / DAT_PENDINGXFERS messages

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call

RETURN VALUE:
TW_UINT16 - Twain status response code 
    - TWRC_SUCCESS
	- TWRC_FAILURE
	- TWCC_SEQERROR
	- TWCC_BADPROTOCOL

NOTES:
    Valid in states 4 to 7

*******************************************************************************/
TW_UINT16 PendingXferMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In PendingXferMsg() with msg - %s\n",  
            GetMsgName(pTWMsg).c_str());

    TW_UINT16 twRc = TWRC_SUCCESS;
	TW_INT16 PendingImages = 0;

	switch (pTWMsg->MSG)
	{
		case MSG_GET:                    // valid states 4 -- 7
			switch (CSession::GetInstance()->GetState())
			{
				case CSession::E_STATE4:
				case CSession::E_STATE5:
				case CSession::E_STATE6:
					{
						((pTW_PENDINGXFERS)pTWMsg->pData)->Count=1;
					}
					break;
				case CSession::E_STATE7:
					{
						((pTW_PENDINGXFERS)pTWMsg->pData)->Count=0;
					}
					break;

				default:
					// fail get due to  seqerror
                    twRc = CSession::GetInstance()->HandleError(E_PENDINGXFERMSG, TWCC_SEQERROR);
					break; //break for any of the above states.
			}
			break;	//break for MSG_GET

		case MSG_ENDXFER: // clears ONLY current transfer, valid 6 or 7
			if((CSession::GetInstance()->GetState()==CSession::E_STATE6) ||
                (CSession::GetInstance()->GetState()==CSession::E_STATE7))
			{
                CScanInterface::GetInstance()->ScanStop();
			    /*
			    * Assume there are no more images to scan
			    */
			    ((pTW_PENDINGXFERS)pTWMsg->pData)->Count = 0;
                // Since we have no more images to transfer we go 
                // to state 5
                CSession::GetInstance()->SetState(CSession::E_STATE5);
			}
			else
			{
				twRc = CSession::GetInstance()->HandleError(E_PENDINGXFERMSG, TWCC_SEQERROR);
			}
			break;

		case MSG_RESET: // clears ALL transfers, valid only in state 6
			if ((CSession::GetInstance()->GetState()==CSession::E_STATE6))
			{
				/*
				* Stop the current scan, clear the rest of the pages, return no images left
				* Return to state 5
				*/
                CScanInterface::GetInstance()->ScanStop();

				CSession::GetInstance()->SetState(CSession::E_STATE5);

				((pTW_PENDINGXFERS)pTWMsg->pData)->Count = 0;
			}
			else 
			{
				twRc = CSession::GetInstance()->HandleError(E_PENDINGXFERMSG, TWCC_SEQERROR);
			}
			break;

		default:
			twRc = CSession::GetInstance()->HandleError(E_PENDINGXFERMSG, TWCC_BADPROTOCOL);
			break;
	}
	return twRc;
} // PendingXferMsg


/*******************************************************************************
DESCRIPTION:

SetupMemXferMsg - Handler for DG_CONTROL / DAT_SETUPMEMXFER messages

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call

RETURN VALUE:
TW_UINT16 - Twain status response code 
    TWRC_SUCCESS
	TWRC_FAILURE
	TWCC_SEQERROR
	TWCC_BADPROTOCOL        

NOTES:
Valid states: 4 -- 6.

*******************************************************************************/
TW_UINT16 SetupMemXferMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In SetupMemXferMsg() with msg - %s\n",  
            GetMsgName(pTWMsg).c_str());

    TW_UINT16 twRc = TWRC_SUCCESS;

	switch (CSession::GetInstance()->GetState())
	{
		case CSession::E_STATE4:
		case CSession::E_STATE5:
		case CSession::E_STATE6:
			switch (pTWMsg->MSG)
			{
				case MSG_GET:
					((pTW_SETUPMEMXFER)pTWMsg->pData)->MinBufSize = 16 * 1024;
					
					/*
					*	We use 64k - 1 since a 16bit application may call us and
					* have some problems allocating such a buffer with malloc
					*/
					((pTW_SETUPMEMXFER)pTWMsg->pData)->MaxBufSize = (64 * 1024) - 1;
					((pTW_SETUPMEMXFER)pTWMsg->pData)->Preferred = (64 * 1024) - 1;
					break;
				default:
					twRc = CSession::GetInstance()->HandleError(E_SETUPMEMXFER, 
                                                                TWCC_BADPROTOCOL);
					break;
			}
			break; 

		default: //case 4,5,6
			twRc = CSession::GetInstance()->HandleError(E_SETUPMEMXFER, 
                                                        TWCC_SEQERROR);
			break;
	}
	return twRc;
} 

/*******************************************************************************
DESCRIPTION:

XferGroupMsg - Handler for DG_CONTROL / DAT_XFERGROUP messages

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call 

RETURN VALUE:
TW_UINT16 - Twain status response code

NOTES:
    Valid states: 4 -- 6.  Since only the Data Group is supported 
    under Version 1.0 of the specification the application should only use 
	MSG_GET.
*******************************************************************************/
TW_UINT16 XferGroupMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In XferGroupMsg() with msg - %s\n",  
            GetMsgName(pTWMsg).c_str());

    TW_UINT16 twRc = TWRC_SUCCESS;

	switch (CSession::GetInstance()->GetState())
	{
		case CSession::E_STATE4:
		case CSession::E_STATE5:
		case CSession::E_STATE6:   
			switch (pTWMsg->MSG)
			{
				case MSG_GET:
				case MSG_GETDEFAULT:
				case MSG_GETCURRENT:
				case MSG_RESET:
					*(pTW_UINT16)pTWMsg->pData = DG_IMAGE;
					break;

		    case MSG_SET:
			    // ONLY images are define in current spec so you can not change DG_ type
			    break;

		    default:
					twRc = CSession::GetInstance()->HandleError(E_XFERGROUP, 
                                                                TWCC_BADPROTOCOL);
					break;
				}	//Inside switch
				break;

		default:  
			twRc = CSession::GetInstance()->HandleError(E_XFERGROUP, TWCC_SEQERROR);
			break;
	}//outside switch 
	return twRc;
} // XferGroupMsg()


/*******************************************************************************
DESCRIPTION:

ImageInfoMsg - Handler for the DG_IMAGE / DAT_IMAGEINFO messages

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call

RETURN VALUE:
TW_UINT16  - Twain status response code 
    TWRC_SUCCESS      
	TWRC_FAILURE
	TWCC_SEQERROR

NOTES:
Valid states 6.  Fills in all fields of imageinfo struct.  
Get info on pending image transfer.
*******************************************************************************/
TW_UINT16 ImageInfoMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In ImageInfoMsg() with msg - %s\n",  
            GetMsgName(pTWMsg).c_str());

    // This message can only be sent in state 6
    if (CSession::GetInstance()->GetState() != CSession::E_STATE6)
	{
		return CSession::GetInstance()->HandleError(E_IMAGEINFOMSG, TWCC_SEQERROR);
	}

    // and there is only one supported - the MSG_GET
    if (pTWMsg->MSG != MSG_GET)
    {
		return CSession::GetInstance()->HandleError(E_IMAGEINFOMSG, TWCC_BADPROTOCOL);
    }


    // First get the current scan parameters	
    float               xRes, yRes;
    TW_UINT16           pixelType, bitDepth;
    int                 imageWidth, imageHeight;

    TWAIN_RESULT result = TWAIN_OK;

    if (result == TWAIN_OK)
        result = GetCurrentXResolution(xRes);
    if (result == TWAIN_OK)
        result = GetCurrentYResolution(yRes);
    if (result == TWAIN_OK)
        result = GetCurrentPixelType(pixelType);
    if (result == TWAIN_OK)
        result = GetCurrentBitDepth(bitDepth);
    if (result == TWAIN_OK)
        result = GetScanWidthInPixels(imageWidth);
    if (result == TWAIN_OK)
        result = GetScanHeightInPixels(imageHeight);
    
    if (result != TWAIN_OK)
        return TWRC_FAILURE;


    // use the scan data to fill the ImageInfo struct
    // set up local pointer to IMAGEINFO struct
	pTW_IMAGEINFO pII   = (pTW_IMAGEINFO)pTWMsg->pData;
	
    pII->XResolution    = FloatToFIX32(xRes);
	pII->YResolution    = FloatToFIX32(yRes);
	pII->ImageWidth     = imageWidth;
    pII->ImageLength    = imageHeight;
    pII->PixelType      = pixelType; 
	pII->BitsPerPixel   = GetImageBitDepth(bitDepth);
	pII->Planar         = FALSE;
	pII->Compression    = TWCP_NONE;

	// clear first
	for (short i = 0; i <= 7; i++)
	{
		pII->BitsPerSample[i] = 0;
	}

	// assumes color .BMP are 1, 8, 24, 48 as spec. by MS
	if (pII->BitsPerPixel%3) 
	{
		//bw
		pII->SamplesPerPixel = 1;
		pII->BitsPerSample[0] = pII->BitsPerPixel;
	}
    else 
	{
		//color
		pII->SamplesPerPixel = 3;
		for (i = 0; i <= 3; i++)
		{
			pII->BitsPerSample[i] = pII->BitsPerPixel/3;
		}
	}

	return TWRC_SUCCESS;
} 



/*******************************************************************************
DESCRIPTION:

ImageLayoutMsg - Handler for the DG_IMAGE / DAT_IMAGELAYOUT messages

PARAMETERS: 

PTWMSG pTWMsg 

RETURN VALUE:
TW_UINT16 - Twain status response code

NOTES:

*******************************************************************************/
TW_UINT16 ImageLayoutMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In ImageLayoutMsg() with msg - %s\n",  
            GetMsgName(pTWMsg).c_str());

    TW_UINT16 twRc = TWRC_SUCCESS;
    // We deliberately initialize the frame struct with a negative value
    // so that we can set up this structure the first time this function is 
    // called
    static TW_IMAGELAYOUT s_currentLayout = {{{-1,0},{0,0},{0,0},{0,0}},0, 0,0};

	switch (pTWMsg->MSG)
	{
		case MSG_GET:
		case MSG_GETDEFAULT:
            twRc = GetImageLayout(&s_currentLayout, pTWMsg);
            break;
        case MSG_SET:
            twRc = SetImageLayout(&s_currentLayout, pTWMsg);
            break;
		case MSG_RESET:
            if (CSession::GetInstance()->GetState() != CSession::E_STATE4)  
            {
                twRc = CSession::GetInstance()->HandleError(E_IMAGELAYOUTMSG, 
                                                            TWCC_SEQERROR);
                break;
            }
            if (SetDefaultLayout(&s_currentLayout) != TRUE)
            {
                twRc = CSession::GetInstance()->HandleError(
                                        E_IMAGELAYOUTMSG, 
                                        TWCC_OPERATIONERROR);
                break;
        
            }
            else
            {
                *((pTW_IMAGELAYOUT)pTWMsg->pData) = s_currentLayout;
            }
            break;
		default:
			twRc = CSession::GetInstance()->HandleError(E_IMAGELAYOUTMSG, 
                                                        TWCC_BADPROTOCOL);
			break;
	}

	return twRc;
} 


/*******************************************************************************
DESCRIPTION:

SetDefaultLayout - Helper method used by the ImageLayout message handlers
sets the members TW_IMAGELAYOUT structure based on the current value
of the scan image height and width

PARAMETERS: 

pTW_IMAGELAYOUT currentLayout - the pointer to a TW_IMAGELAYOUT struct

RETURN VALUE:
BOOL - returns FALSE if an error occurs talking to the scan module, otherwise
returns TRUE.

NOTES:
The ImageLayout structure has to be filled in with length and width in the 
current units which in the default case would be inches. We also don't 
do much with the document number, frame number and page number fields which are
all set to 1.
*******************************************************************************/
BOOL SetDefaultLayout(pTW_IMAGELAYOUT currentLayout)
{
    TWAIN_RESULT result = TWAIN_OK;
    float imageWidth, imageHeight;

    if (result == TWAIN_OK)
        result = GetCurrentScanWidth(imageWidth);
    if (result == TWAIN_OK)
        result = GetCurrentScanHeight(imageHeight);
    
    if (result != TWAIN_OK)
        return FALSE;

    TW_FIX32 fix32 = {0,0};

    currentLayout->Frame.Top      = fix32;
    currentLayout->Frame.Left     = FloatToFIX32(imageWidth);
	currentLayout->Frame.Right    = fix32;
	currentLayout->Frame.Bottom   = FloatToFIX32(imageHeight);

	currentLayout->DocumentNumber = 1;  
	currentLayout->PageNumber     = 1;
	currentLayout->FrameNumber    = 1;

    return TRUE;
}


/*******************************************************************************
DESCRIPTION:

GetImageLayout - Final handler for the DG_IMAGE / DAT_IMAGELAYOUT / MSG_GET
and GET_DEFAULT messages 

PARAMETERS: 

pTW_IMAGELAYOUT currentLayout - pointer to the current layout struct
PTWMSG pTWMsg - pointer to the original message struct passed to DS_Entry

RETURN VALUE:
TW_UINT16 - Twain status response code 
    TWRC_SUCCESS or
    TWRC_FAILURE

NOTES:
    When this function is called with an uninitialized ImageLayout struct
    it will call SetDefaultLayout first to initialize the struct, otherwise
    it simply returns the current layout back in the message struct's
    data field
*******************************************************************************/
TW_UINT16 GetImageLayout(pTW_IMAGELAYOUT currentLayout, PTWMSG pTWMsg)
{
    TW_UINT16 twRc = TWRC_SUCCESS;

    switch (CSession::GetInstance()->GetState())
	{
		case CSession::E_STATE4:
		case CSession::E_STATE5:
		case CSession::E_STATE6:
        {
            if (currentLayout->Frame.Top.Whole < 0)
            {
                // We have an uninitialized layout struct so we 
                // initialize it here
                if (SetDefaultLayout(currentLayout) != TRUE)
                {
                    twRc = CSession::GetInstance()->HandleError(
                                            E_IMAGELAYOUTMSG, 
                                            TWCC_OPERATIONERROR);
                    break;
            
                }
            }

            *((pTW_IMAGELAYOUT)pTWMsg->pData) = *currentLayout;
            break;
        }
        default:
			// fail get due to  seqerror
            twRc = CSession::GetInstance()->HandleError(E_IMAGELAYOUTMSG, 
                                                        TWCC_SEQERROR);
			break; //break for any of the above states.
	}

    return twRc;
}


/*******************************************************************************
DESCRIPTION:

SetImageLayout - Final Handler for the DG_IMAGE / DAT_IMAGELAYOUT / MSG_SET
message

PARAMETERS: 

pTW_IMAGELAYOUT currentLayout - pointer to the current layout struct 
            maintained by this module
PTWMSG pTWMsg - pointer to the original message struct passed into DS_Entry

RETURN VALUE:
TW_UINT16 - Twain status response code
    TWRC_SUCCESS or
    TWRC_FAILURE

NOTES:
    This code checks the passed in struct for validity and then 
    sets up the current layout struct with the passed in struct
*******************************************************************************/
TW_UINT16 SetImageLayout(pTW_IMAGELAYOUT currentLayout, PTWMSG pTWMsg)
{
    CSession* pSession = CSession::GetInstance();
    if (pSession->GetState() != CSession::E_STATE4)  
    {
        return pSession->HandleError(E_IMAGELAYOUTMSG, TWCC_SEQERROR);
    }

    TWAIN_RESULT result = TWAIN_OK;
    float imageWidth, imageHeight;

    if (result == TWAIN_OK)
        result = GetMaxScanWidth(imageWidth);
    if (result == TWAIN_OK)
        result = GetMaxScanHeight(imageHeight);
    
    if (result != TWAIN_OK)
    {
        return pSession->HandleError(E_IMAGELAYOUTMSG, TWCC_OPERATIONERROR);
    }
	
    pTW_IMAGELAYOUT pNewLayout = (pTW_IMAGELAYOUT)pTWMsg->pData;
    
    if ((FIX32ToFloat(pNewLayout->Frame.Left) < 0.0) || 
        (FIX32ToFloat(pNewLayout->Frame.Top) < 0.0) )
    {
        TRACE("Twain_LM983x : SetImageLayout - Negative origin values\n");
        return pSession->HandleError(E_IMAGELAYOUTMSG, TWCC_BADVALUE);
    }

    float newWidth  = FIX32ToFloat(pNewLayout->Frame.Right) - 
                            FIX32ToFloat(pNewLayout->Frame.Left);
    float newHeight = FIX32ToFloat(pNewLayout->Frame.Bottom) - 
                            FIX32ToFloat(pNewLayout->Frame.Top);

    if ( (newWidth > imageWidth) || (newHeight > imageHeight))
    {
        TRACE("Twain_LM983x : SetImageLayout - Out of bound image dimensions\n");
        return pSession->HandleError(E_IMAGELAYOUTMSG, TWCC_BADVALUE);
    }

    *currentLayout = *pNewLayout;

    return TWRC_SUCCESS;
}

/*******************************************************************************
DESCRIPTION:

ImageMemXferMsg

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call

RETURN VALUE:
TW_UINT16 - Twain status response code 
    TWRC_SUCCESS
	TWRC_FAILURE
	TWCC_BADVALUE
	TWCC_SEQERROR
	TWRC_XFERDONE
	TWRC_CANCEL

NOTES:

*******************************************************************************/
TW_UINT16 ImageMemXferMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In ImageMemXferMsg() with msg - %s\n",  
            GetMsgName(pTWMsg).c_str());

    // Counters that maintian values between invocations
    static DWORD dwLineSize = 0l;
    static DWORD dwRowCount = 0l;

	TW_UINT16 twRc = TWRC_SUCCESS;
	pTW_IMAGEMEMXFER pMem = (pTW_IMAGEMEMXFER)pTWMsg->pData;


	switch (pTWMsg->MSG)
	{
		case MSG_GET:
            if ((CSession::GetInstance()->GetState() == CSession::E_STATE4) || 
                (CSession::GetInstance()->GetState() == CSession::E_STATE5))  
            {
                twRc = CSession::GetInstance()->HandleError(E_IMAGELAYOUTMSG, TWCC_SEQERROR);
                break;
            }

            if(CSession::GetInstance()->GetState()==CSession::E_STATE6)
			{
                /*
				*	Initiate the scan on the first read
				*/
                TWAIN_RESULT resCode = CScanInterface::GetInstance()->SetScanFormat(CScanInterface::E_RGBDATA); 
        		if(resCode != TWAIN_OK)
				{
					twRc = CSession::GetInstance()->HandleError(E_IMAGEMEMXFERMSG, 
																TWCC_OPERATIONERROR);
					CUIInterface::GetInstance()->EndScan(resCode);
					break;
				}
                resCode = CScanInterface::GetInstance()->ScanStart();
				/*
				*	There was a problem starting the scan
				*/
				if(resCode == TWAIN_ABORT)
				{
					TRACE("Twain_LM983x : ImageMemXferMsg() Aborted \n");
					twRc = TWRC_CANCEL;
					CUIInterface::GetInstance()->EndScan(TWAIN_ABORT);
					break;
				}
				else if(resCode != TWAIN_OK)
				{
					twRc = CSession::GetInstance()->HandleError(E_IMAGEMEMXFERMSG, 
																TWCC_OPERATIONERROR);
					CUIInterface::GetInstance()->EndScan(resCode);
					twRc = TWRC_CANCEL;
					break;
				}

                TWAIN_INFO_REC infoRec;
                CScanInterface::GetInstance()->ScanGetScanInfo(&infoRec);
            
                dwLineSize = infoRec.bytesPerLine;
				dwRowCount = 0;

			}
           
			pMem->BytesPerRow = dwLineSize;
			pMem->YOffset   = dwRowCount;

            int imageWidth;
            GetScanWidthInPixels(imageWidth);
			pMem->Columns = (TW_UINT16)(imageWidth);

			pMem->XOffset = 0;

			/*
			*	Really should check the memory flags here in case this is a
			* handle, not a pointer
			*/
			{
                TWAIN_RESULT resCode = CScanInterface::GetInstance()->
                                               ScanRead(pMem->Memory.TheMem, 
                                                        pMem->Memory.Length, 
                                                        (long*)&pMem->BytesWritten);


				if(resCode == TWAIN_DONE)
				{
					twRc = TWRC_XFERDONE;
                    CUIInterface::GetInstance()->EndScan(TWAIN_OK);
				}
				else if(resCode == TWAIN_ABORT)
				{
                    TRACE("Twain_LM983x : ImageMemXferMsg() Aborted \n");
					twRc = TWRC_CANCEL;
                    CUIInterface::GetInstance()->EndScan(TWAIN_ABORT);
                    break;
				}
                else if(resCode != TWAIN_OK)
				{
					twRc = CSession::GetInstance()->HandleError(E_IMAGEMEMXFERMSG, 
                                                                TWCC_OPERATIONERROR);
                    CUIInterface::GetInstance()->EndScan(resCode);
					twRc = TWRC_CANCEL;
					break;
				}
			}

			pMem->Rows = pMem->BytesWritten/pMem->BytesPerRow;
			dwRowCount += pMem->Rows;
			break;

		default:
			twRc = CSession::GetInstance()->HandleError(E_IMAGEMEMXFERMSG, 
                                                        TWCC_BADPROTOCOL);
			break;
	}	//turn off the switch


	/*
	*	Success means we move to state 7
	*/
	if(twRc != TWRC_FAILURE)
	{
		CSession::GetInstance()->SetState(CSession::E_STATE7);
	}
	return twRc;
} 

/*******************************************************************************
DESCRIPTION:

ImageNativeXferMsg - The handler for the DG_IMAGE / DAT_IMAGENATIVEXFER messages

PARAMETERS: 

PTWMSG pTWMsg - message struct, same as list of parms in DS_ENTRY call 

RETURN VALUE:
TW_UINT16 - Twain status response codes
    TWRC_SUCCESS
    TWRC_FAILURE
    TWCC_BADVALUE
    TWCC_SEQERROR
    TWRC_XFERDONE
    TWCC_CANCEL
    TWCC_BADPROTOCOL

NOTES:

*******************************************************************************/
TW_UINT16 ImageNativeXferMsg(PTWMSG pTWMsg)
{
    TRACE("Twain_LM983x : In ImageNativeXferMsg() with msg - %s\n",  
            GetMsgName(pTWMsg).c_str());

    TW_UINT16        twRc = TWRC_SUCCESS;

	if(CSession::GetInstance()->GetState()==CSession::E_STATE6)
	{
		switch (pTWMsg->MSG)
		{
			case MSG_GET:
                {
                    // First try to acquire the DIB  from the scanner module
                    HGLOBAL hDib;
                
                    TWAIN_RESULT result = CScanInterface::GetInstance()->ScanToDIB(hDib);
                
                    if (result == TWAIN_ABORT)
                    {
                        TRACE("Twain_LM983x : ImageNativeXfer() Aborted \n");
                        /// User aborted ,  just cancel this transfer
                        twRc = TWRC_CANCEL;
                    }
                    else if (result != TWAIN_OK)
                    {
                        twRc = CSession::GetInstance()->HandleError(E_IMAGENATIVEXFERMSG, 
                                                                        TWCC_OPERATIONERROR);
						twRc = TWRC_CANCEL;
                    }
                    else // result == TWAIN_OK
                    {
                        // Assuming evertything went ok , return hDib in low word
                        *(pTW_UINT32)pTWMsg->pData = (TW_UINT32)hDib;
				        twRc = TWRC_XFERDONE;
                    }
                    // Whatever happens inform the UI that the scan is done
                    CUIInterface::GetInstance()->EndScan(result);
                }
                break;
			default:
				twRc = CSession::GetInstance()->HandleError(E_IMAGENATIVEXFERMSG, 
                                                            TWCC_BADPROTOCOL);
				break;
		}
	}
	else
	{
		twRc = CSession::GetInstance()->HandleError(E_IMAGENATIVEXFERMSG, TWCC_SEQERROR);
	}

	/*
	*	Success means we move to state 7
	*/
	if(twRc != TWRC_FAILURE)
	{
		CSession::GetInstance()->SetState(CSession::E_STATE7);
	}
	return twRc;
} 

/*******************************************************************************
DESCRIPTION:

NotifyCloseDSReq - called by the TWAIN - UI interface code to inform the 
TWAIN DSM that we are waiting for a close request from the DSM

PARAMETERS: 

VOID

RETURN VALUE:
VOID 

NOTES:
Indirectly returns a MSG_CLOSEDSREQ to the Application, thru DSM_Entry call.

*******************************************************************************/
VOID NotifyCloseDSReq(VOID)
{
    TRACE("Twain_LM983x : In NotifyCloseDSReq()\n");

    HINSTANCE hDSMDLL = NULL;
	DSMENTRYPROC lpDSM_Entry = NULL;         // entry point to the SM
    
    CSession *pSession = CSession::GetInstance();

	if((hDSMDLL = LoadLibrary("TWAIN_32.DLL")) != NULL)
	{
		if((lpDSM_Entry = (DSMENTRYPROC)GetProcAddress(hDSMDLL, "DSM_Entry")) != NULL)
		{
            TW_IDENTITY dsIdentity  = pSession->GetSourceIdentity();
            TW_IDENTITY appIdentity = pSession->GetAppIdentity();

            (*lpDSM_Entry)( &dsIdentity, 
                            &appIdentity, 
                            (TW_UINT32)DG_CONTROL, 
                            DAT_NULL, 
                            MSG_CLOSEDSREQ, 
                            (TW_MEMREF) NULL);
		}
		FreeLibrary (hDSMDLL);
	}
	return;

} // NotifyCloseReq()

/*******************************************************************************
DESCRIPTION:

NotifyXferReady - called by the TWAIN - UI interface code to inform the 
TWAIN DSM that we are ready to transfer image data to the application

PARAMETERS: 

VOID

RETURN VALUE:
VOID 

NOTES:
Indirectly returns a MSG_XFERREADY to the Application, thru DSM_Entry call.
*******************************************************************************/
VOID NotifyXferReady(VOID)
{
    TRACE("Twain_LM983x : In NotifyXferReady()\n");

    HINSTANCE hDSMDLL = NULL;
    DSMENTRYPROC lpDSM_Entry = NULL;         // entry point to the SM

    CSession *pSession = CSession::GetInstance();

    // go to xferready state
    pSession->SetState(CSession::E_STATE6);

    if((hDSMDLL = LoadLibrary("TWAIN_32.DLL")) != NULL)
    {
	    if((lpDSM_Entry = (DSMENTRYPROC)GetProcAddress(hDSMDLL, "DSM_Entry")) != NULL)
	    {
            TW_IDENTITY dsIdentity  = pSession->GetSourceIdentity();
            TW_IDENTITY appIdentity = pSession->GetAppIdentity();

            (*lpDSM_Entry)( &dsIdentity, 
                            &appIdentity, 
                            (TW_UINT32)DG_CONTROL, 
                            DAT_NULL, 
                            MSG_XFERREADY, 
                            (TW_MEMREF) NULL);
	    }
	    FreeLibrary (hDSMDLL);
    }

    return;

} // NotifyXferReady()


/*******************************************************************************
DESCRIPTION:

CapabilityMsg - The handler for all capability messages i.e. messages with the 
data group and Data argument type DG_CONTROL / DAT_CAPABILITY 
All these messages are simply forwarded on to the CCapHolder singleton which is
part of the capability engine for this module. The CCapHolder object will 
then forward this on to the appropriate capability object which can handle
this message.

PARAMETERS: 

PTWMSG pDcMsg - message struct, same as list of parms in DS_ENTRY call

RETURN VALUE:
TW_UINT16 - Twain status response code

NOTES:
The application will create the container and destroy the container.
The Source will extract the value from the container 1st and use this value
to modify it's local value for the CAP.
The Source needs to be able to extract values from all of the different
container types and load as many as possible into it's local containers
for the particular capabilty.
NOTES: for the MSG_GET side of Source
The Source will create the container and the App will destroy container.
The App will extract the value from the container 1st and use this value
to modify it's local state or value for this CAP.
The App extracts values from all different container types and also
matches up these cap's when appropriate to find the value or values which
best suit the current mode of the application.
NOTES: for both App and Source in general:
You don't know what type of container you will be getting for any
particular cap until you get it.  

*******************************************************************************/
TW_UINT16 CapabilityMsg(PTWMSG pDcMsg)
{
    return CCapHolder::GetInstance()->HandleCapabilityMsg(pDcMsg);
}
