// Twain_ScanInterface.cpp: implementation of the CScanInterface class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Twain_ScanInterface.h"
#include "Twain_Session.h"
#include "Twain_Utility.h"
#include "Twain_UIInterface.h"


CScanInterface* CSingleton<CScanInterface>::m_theInstance = 0;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CScanInterface::CScanInterface()
{
    m_handleScanModule      = NULL;

    m_fnScanStart           = NULL;
    m_fnScanRead            = NULL;
    m_fnScanStop            = NULL;
    m_fnScanSetScanParams   = NULL;
    m_fnScanGetScanParams   = NULL;
    m_fnScanGetScanInfo     = NULL;
    m_fnScanGetScanHardware = NULL;
    m_fnScanGetCapability   = NULL;

    LoadCodeConversionTables();
}

void CScanInterface::LoadCodeConversionTables()
{
    // First map the TWAIN codes to the correspoding SCAN error codes
    m_twain2ScanTable[TWAIN_OK]     = SCAN_OK;
    m_twain2ScanTable[TWAIN_ERROR]  = SCAN_ERROR;
    m_twain2ScanTable[TWAIN_DONE]   = SCAN_DONE;
    m_twain2ScanTable[TWAIN_ABORT]  = SCAN_ABORT;
    m_twain2ScanTable[TWAIN_INVALID_CAPID] = SCAN_INVALID_CAPID;
//    m_twain2ScanTable[TWAIN_UNKNOWN_MSG] = SCAN_OK;- has no equivalent

    // Now map the SCAN codes to the corresponding TWAIN error codes
    m_scan2TwainTable[SCAN_OK]      = TWAIN_OK;
    m_scan2TwainTable[SCAN_ERROR]   = TWAIN_ERROR;
    m_scan2TwainTable[SCAN_DONE]    = TWAIN_DONE;
    m_scan2TwainTable[SCAN_ABORT]   = TWAIN_ABORT;
    m_scan2TwainTable[SCAN_INVALID_CAPID] = TWAIN_INVALID_CAPID;

    // Map of cap Ids from Scan Cap ID to Twain Cap ID
    m_scan2TwainCapMap[SCAN_CAP_HRES]           = TWAIN_CAP_HRES;
    m_scan2TwainCapMap[SCAN_CAP_VRES]           = TWAIN_CAP_VRES;
    m_scan2TwainCapMap[SCAN_CAP_PIXELTYPE]      = TWAIN_CAP_PIXELTYPE;
    m_scan2TwainCapMap[SCAN_CAP_PIXELDEPTH]     = TWAIN_CAP_PIXELDEPTH;
    m_scan2TwainCapMap[SCAN_MAX_SCAN_WIDTH]     = TWAIN_MAX_SCAN_WIDTH;
    m_scan2TwainCapMap[SCAN_MAX_SCAN_HEIGHT]    = TWAIN_MAX_SCAN_HEIGHT;

    // Map of cap Ids from Twain Cap ID to Scan Cap ID
    m_twain2ScanCapMap[TWAIN_CAP_HRES]          = SCAN_CAP_HRES;
    m_twain2ScanCapMap[TWAIN_CAP_VRES]          = SCAN_CAP_VRES;
    m_twain2ScanCapMap[TWAIN_CAP_PIXELTYPE]     = SCAN_CAP_PIXELTYPE;
    m_twain2ScanCapMap[TWAIN_CAP_PIXELDEPTH]    = SCAN_CAP_PIXELDEPTH;
    m_twain2ScanCapMap[TWAIN_MAX_SCAN_WIDTH]    = SCAN_MAX_SCAN_WIDTH;
    m_twain2ScanCapMap[TWAIN_MAX_SCAN_HEIGHT]   = SCAN_MAX_SCAN_HEIGHT;

}

CScanInterface::~CScanInterface()
{
    if (m_handleScanModule)
    {
        FreeLibrary(m_handleScanModule);
    }

    m_theInstance = 0;
}

TWAIN_RESULT CScanInterface::ScanStart()
{
    if (!m_fnScanStart)
        return TWAIN_ERROR;

    m_lastScanResult = (*m_fnScanStart)();

    return Scan2TwainResultCode(m_lastScanResult);
}

TWAIN_RESULT CScanInterface::ScanRead(void*	scanBuffer,
									long	reqCount,
									long*	actCount)
{
    if (!m_fnScanRead)
        return TWAIN_ERROR;
    
    m_lastScanResult = (*m_fnScanRead)(scanBuffer, reqCount, actCount);

    return Scan2TwainResultCode(m_lastScanResult);
}

TWAIN_RESULT CScanInterface::ScanStop()
{
    if (!m_fnScanStop)
        return TWAIN_ERROR;

    m_lastScanResult = (*m_fnScanStop)();

    return Scan2TwainResultCode(m_lastScanResult);
}

TWAIN_RESULT CScanInterface::ScanSetScanParams(TWAIN_PARAM_REC* paramRec)
{
    if (!m_fnScanSetScanParams)
        return TWAIN_ERROR;

    // pack the input parameter into a scan_param_rec
    SCAN_PARAM_REC scanRec;
    
    scanRec.boundingRect = paramRec->boundingRect;
    scanRec.scanRect     = paramRec->scanRect;
    scanRec.pixelType    = paramRec->pixelType;
    scanRec.bitsPerPixel = paramRec->bitsPerPixel;
    scanRec.hRes         = paramRec->hRes;
    scanRec.vRes         = paramRec->vRes;
    scanRec.hInterp      = paramRec->hInterp;
    scanRec.vInterp      = paramRec->vInterp;
	scanRec.scanToTwain  = FALSE;

    m_lastScanResult = (*m_fnScanSetScanParams)(&scanRec,TwainCallback);

    return Scan2TwainResultCode(m_lastScanResult);
}

TWAIN_RESULT CScanInterface::ScanGetScanParams(TWAIN_PARAM_REC* paramRec)
{
    if (!m_fnScanGetScanParams)
        return TWAIN_ERROR;

    SCAN_PARAM_REC  scanRec;


    m_lastScanResult = (*m_fnScanGetScanParams)(&scanRec);

    if (m_lastScanResult == SCAN_OK)
    {
        paramRec->boundingRect  = scanRec.boundingRect ;
        paramRec->scanRect      = scanRec.scanRect     ;
        paramRec->pixelType     = scanRec.pixelType    ;
        paramRec->bitsPerPixel  = scanRec.bitsPerPixel ;
        paramRec->hRes          = scanRec.hRes         ;
        paramRec->vRes          = scanRec.vRes         ;
        paramRec->hInterp       = scanRec.hInterp      ;
        paramRec->vInterp       = scanRec.vInterp      ;
        
    }

    return Scan2TwainResultCode(m_lastScanResult);
}

TWAIN_RESULT CScanInterface::ScanGetScanInfo(TWAIN_INFO_REC *infoRec)
{
    if (!m_fnScanGetScanInfo)
        return TWAIN_ERROR;

    SCAN_INFO_REC    scanInfoRec;
    m_lastScanResult = (*m_fnScanGetScanInfo)(&scanInfoRec);
    
    if (m_lastScanResult == SCAN_OK)
    {
        infoRec->totalBytes      = scanInfoRec.totalBytes;		
        infoRec->actBytes        = scanInfoRec.actBytes;		
        infoRec->bytesPerLine    = scanInfoRec.bytesPerLine;	
        infoRec->startTime       = scanInfoRec.startTime;		
        infoRec->scanStartTime   = scanInfoRec.scanStartTime;	
        infoRec->scanEndTime     = scanInfoRec.scanEndTime;	
        infoRec->endTime         = scanInfoRec.endTime;
		strcpy(infoRec->scanStatus,scanInfoRec.scanStatus);
        infoRec->rawTotalBytes      = scanInfoRec.rawTotalBytes;		
        infoRec->rawActBytes        = scanInfoRec.rawActBytes;	
    }
    
    return Scan2TwainResultCode(m_lastScanResult);
}

TWAIN_RESULT CScanInterface::ScanGetScanHardware(TWAIN_HARDWARE_REC *hardwareRec)
{
    if (!m_fnScanGetScanHardware)
        return TWAIN_ERROR;

    SCAN_HARDWARE_REC    scanHardwareRec;
    m_lastScanResult = (*m_fnScanGetScanHardware)(&scanHardwareRec);
    
    if (m_lastScanResult == SCAN_OK)
    {
		hardwareRec->widthOpticalPixels = scanHardwareRec.widthOpticalPixels;
		hardwareRec->heightOpticalPixels = scanHardwareRec.heightOpticalPixels;
		hardwareRec->opticalRes		= scanHardwareRec.opticalRes;
		hardwareRec->chip = scanHardwareRec.chip;
		strcpy(hardwareRec->device,scanHardwareRec.device);
		memcpy(hardwareRec->baseResolution,scanHardwareRec.baseResolution,
			sizeof(hardwareRec->baseResolution));
    }
    
    return Scan2TwainResultCode(m_lastScanResult);
}

TWAIN_RESULT CScanInterface::ScanGetCapability(TWAIN_CAPID capID, 
											 void*  input, 
											 void * output,
                                             long*  numOutputElements)
{
    if (!m_fnScanGetCapability)
        return TWAIN_ERROR;
    
    SCAN_CAPID  scanCapID;
        
    m_lastScanResult = Twain2ScanCapID(capID, &scanCapID);
    
    if (m_lastScanResult == SCAN_OK)
    {
        m_lastScanResult = (*m_fnScanGetCapability)(scanCapID, 
                                                    input, 
                                                    output, 
                                                    numOutputElements);
    }

    return Scan2TwainResultCode(m_lastScanResult);
}

TWAIN_RESULT    CScanInterface::Scan2TwainResultCode(SCAN_RESULT scanResultCode)
{
    SCAN2TWAIN_ITERATOR it = m_scan2TwainTable.find(scanResultCode);
    if (it == m_scan2TwainTable.end())
        return TWAIN_ERROR;

    return (*it).second;
}

SCAN_RESULT     CScanInterface::Twain2ScanResultCode(TWAIN_RESULT twainResultCode)
{
    TWAIN2SCAN_ITERATOR it = m_twain2ScanTable.find(twainResultCode);
    if (it == m_twain2ScanTable.end())
        return TWAIN_ERROR;

    return (*it).second;
}

TWAIN_RESULT    CScanInterface::Scan2TwainCapID(SCAN_CAPID scanCapID, 
                                TWAIN_CAPID* twainCapID)
{
    SCAN2TWAIN_CAPITERATOR it = m_scan2TwainCapMap.find(scanCapID);
    if (it == m_scan2TwainCapMap.end())
        return TWAIN_ERROR;

    *twainCapID = (*it).second;

    return TWAIN_OK;
}

TWAIN_RESULT    CScanInterface::Twain2ScanCapID(TWAIN_CAPID twainCapID, 
                                SCAN_CAPID* scanCapID)
{
    TWAIN2SCAN_CAPITERATOR it = m_twain2ScanCapMap.find(twainCapID);
    if (it == m_twain2ScanCapMap.end())
        return TWAIN_ERROR;

    *scanCapID = (*it).second;

    return TWAIN_OK;
}

void CScanInterface::LoadScanModule()
{
    if (m_handleScanModule)
        return;

	// Load the full filename with the absolute path for the Scan  module dll 
	char szModulePath[_MAX_PATH];
	char szCompletePath[_MAX_PATH];
	char szFileLocation[_MAX_PATH];
	char szDriveName[_MAX_DRIVE];
    
    HINSTANCE hDSInst = CSession::GetInstance()->GetInstanceHandle();
    
	GetModuleFileName(hDSInst, szModulePath, _MAX_PATH);
	_splitpath(szModulePath, szDriveName, szFileLocation, NULL, NULL);
	wsprintf(szCompletePath, "%s%s", szDriveName, szFileLocation);
    strcat(szCompletePath, "Scan_Lm983x.dll");

    m_handleScanModule= LoadLibrary(szCompletePath);

    if (m_handleScanModule!= NULL) 
    {
        m_fnScanStart = (pfnScanStart)
            GetProcAddress(m_handleScanModule, "ScanStart");

        m_fnScanRead = (pfnScanRead)
            GetProcAddress(m_handleScanModule, "ScanRead");

        m_fnScanStop = (pfnScanStop)
            GetProcAddress(m_handleScanModule, "ScanStop");

        m_fnScanSetScanParams = (pfnScanSetScanParams)
            GetProcAddress(m_handleScanModule, "ScanSetScanParams");

        m_fnScanGetScanParams = (pfnScanGetScanParams)
            GetProcAddress(m_handleScanModule, "ScanGetScanParams");

        m_fnScanGetScanInfo = (pfnScanGetScanInfo)
            GetProcAddress(m_handleScanModule, "ScanGetScanInfo");

        m_fnScanGetScanHardware = (pfnScanGetScanHardware)
            GetProcAddress(m_handleScanModule, "ScanGetScanHardware");

        m_fnScanGetCapability = (pfnScanGetCapability)
            GetProcAddress(m_handleScanModule, "ScanGetCapability");
    }

    return;
}

TWAIN_RESULT CScanInterface::SetScanFormat(SCAN_FORMAT eFormat)
{
    // We need scan data in DIB format so we set Scan to Twain to true
    // Get the current settings and change just the scan bit depth
    SCAN_PARAM_REC scanRec;
    m_lastScanResult = (*m_fnScanGetScanParams)(&scanRec);

    if (m_lastScanResult  != SCAN_OK)
        return Scan2TwainResultCode(m_lastScanResult);
    
    scanRec.scanToTwain = (eFormat == E_DIBDATA) ? TRUE : FALSE;

    m_lastScanResult = (*m_fnScanSetScanParams)(&scanRec,TwainCallback);

    return Scan2TwainResultCode(m_lastScanResult);
}

TWAIN_RESULT CScanInterface::ScanToDIB(HGLOBAL& hBmp)
{
    TWAIN_RESULT result  = SetScanFormat(E_DIBDATA);   
    if (result != TWAIN_OK)
        return result;

    // Simply get a DIB for preview from the scanner module
    // There will be no interaction with the TWAIN app itself
    // Figure out how big a buffer we need based on the current
    // scan parameters and then allocate memory for the DIB header
    // too
    HGLOBAL hDib;
    int     bytesToScan;
    BYTE*   pDataBytes;

    // Allocate memory for the dib header and the actual data bytes
    // but dont put anything in the data portion, just format the 
    // header to reflect the status of the dib that we will prepare 
    result = AllocateScanDib(  &hDib, &bytesToScan, &pDataBytes);
    if (result != TWAIN_OK)
		return result;

	result = ScanStart();
	if (result != TWAIN_OK)
	{
        GlobalUnlock(hDib);
        GlobalFree(hDib);
		return result;
	}
    
    int bytesScanned = 0;
    int maxStripSize = 64*1024; // 64K
    long actStripSize = 0;

	while (bytesScanned < bytesToScan)
    {
        // In a loop get the image data
        result = ScanRead(  pDataBytes+bytesScanned, 
                            maxStripSize, 
                            &actStripSize);

        if (result == TWAIN_DONE)
        {
            GlobalUnlock(hDib);
            hBmp    = hDib; 
            break;
        }
        else if ( (result == TWAIN_ABORT) || 
                    (result == TWAIN_ERROR) )
        {
	        GlobalUnlock(hDib);
		    GlobalFree(hDib);
			return result;
         }

        bytesScanned += actStripSize;
    }

    result = InvertDIB(hBmp);
    
    return result;
}

SCAN_RESULT CScanInterface::TwainCallback(BOOL check_messages)
{
    return CScanInterface::GetInstance()->HandleCallBack(check_messages);
}

SCAN_RESULT CScanInterface::HandleCallBack(BOOL check_messages)
{
    return Twain2ScanResultCode(CUIInterface::GetInstance()->Callback(check_messages));
}
