/*******************************************************************************
FILE        : Twain_Utility.cpp
REVISION    : 1.0
DATE        : 9/9/99

DESCRIPTION :

Contains the implementation miscellaneous helper functions used by the 
TWAIN module

Copyright (c) 1999 - National Semiconductor Ltd - All rights reserved
*******************************************************************************/
#include "stdafx.h"
#include "Twain_Utility.h"
#include "Twain_ScanInterface.h"

#include <stdio.h>
#include <stdarg.h>

//////////////////////////////////////////////////////////////////////////////
// FloatToFIX32 -- Convert a floating point value into a FIX32.
//
TW_FIX32 FloatToFIX32 (float floater)
{
  TW_FIX32 Fix32_value;

  TW_INT32 value = (TW_INT32) (floater * 65536.0 + 0.5);
  Fix32_value.Whole = LOWORD(value >> 16);
  Fix32_value.Frac = LOWORD(value & 0x0000ffffL);

  return Fix32_value;
}

//////////////////////////////////////////////////////////////////////////////
// FIX32ToFloat -- Convert a FIX32 value into a floating point value
//
float FIX32ToFloat (TW_FIX32 fix32)
{
	float   floater;
	floater = (float) fix32.Whole + (float) (fix32.Frac / 65536.0);
	return(floater);
}




//////////////////////////////////////////////////////////////////////////////
// DIB Support Routines 

//////////////////////////////////////////////////////////////////////////////
// AllocatePreviewDIB -- Allocate memory for a DIB based on the current
// scan settings and then format the DIB header to reflect this properly
// a pointer to the actual count of bytes to scan and the data area where 
// the image data has to be stored will be returned in out paramters
TWAIN_RESULT AllocateScanDib(   HANDLE* hDib,
                                int*    bytesToScan,
                                BYTE**  pDataBytes)
{
    TWAIN_PARAM_REC     scanParams;
    TWAIN_INFO_REC      scanInfo;
    TWAIN_RESULT        result;
        
    CScanInterface *pScanner = CScanInterface::GetInstance();

    // Get the current image scan setting and scan parameters 
    result = pScanner->ScanGetScanInfo(&scanInfo);

    if (result != TWAIN_OK)
        return result;

    result = pScanner->ScanGetScanParams(&scanParams);

    if (result != TWAIN_OK)
        return result;


    // Calculate the size of the dib memory needed 
    int imageSize           = scanInfo.totalBytes;
	int bitsPerPixel		= GetImageBitDepth(scanParams.bitsPerPixel);
    int colorTableEntries   = ComputePaletteSize(bitsPerPixel);
    int colorTableSize      = sizeof(RGBQUAD) * colorTableEntries;
                            
    // GlobalAlloc dibheader + image bytes
    int dibSize = sizeof(BITMAPINFOHEADER) + colorTableSize + imageSize;

    // Lock the memory and fill in the bytes
    HANDLE hDibHandle = ::GlobalAlloc(GPTR, dibSize);

    if (!hDibHandle)
        return TWAIN_ERROR;

    // We create the DIB in memory and this header describes it .
    LPBITMAPINFOHEADER lpBMIH = (LPBITMAPINFOHEADER) GlobalLock(hDibHandle);
	
	lpBMIH->biSize      = sizeof(BITMAPINFOHEADER);
	lpBMIH->biWidth     = scanParams.scanRect.right -  scanParams.scanRect.left; 
                            // scanInfo.bytesPerLine * 8 / scanParams.bitsPerPixel;
	lpBMIH->biHeight    = scanParams.scanRect.bottom - scanParams.scanRect.top;
                            // scanInfo.totalBytes / scanInfo.bytesPerLine;
	lpBMIH->biPlanes    = 1;
	lpBMIH->biBitCount  = bitsPerPixel;

	lpBMIH->biCompression   = BI_RGB;
	lpBMIH->biSizeImage     = imageSize;
	lpBMIH->biXPelsPerMeter = 0;
	lpBMIH->biYPelsPerMeter = 0;
	lpBMIH->biClrUsed       = colorTableEntries;
	lpBMIH->biClrImportant  = colorTableEntries;

	LPBYTE lpColorTable = (LPBYTE) lpBMIH + sizeof(BITMAPINFOHEADER);
    InitializeColorTable(colorTableEntries, lpColorTable);

	*pDataBytes     = (LPBYTE) lpBMIH + sizeof(BITMAPINFOHEADER) + 
					    (sizeof(RGBQUAD) * colorTableEntries); 
    *hDib           = hDibHandle;
    *bytesToScan    = imageSize;

    return TWAIN_OK;
}

void InitializeColorTable(int colorTableEntries, LPBYTE lpColorTable)
{
    if (colorTableEntries <= 0)
        return;

	// Initially set the Color Table Entries to 0
    memset(lpColorTable, 0, sizeof(RGBQUAD) * colorTableEntries);

    // This code initializes the color table which determines how the 
    // DIB bits will be interpreted by the UI and the TWAIN app.
    // We make quite a few assumptions which would be eliminated if 
    // we did not need to fill in the DIB header in the TWAIN module and
    // could instead get it from the SCAN module
    switch(colorTableEntries)
    {
        case 2:
        {
            // Assumption : This is only used for black and white 
            // so the palette has only black(0,0,0) and white (255,255,255)
            for (int j = 0; j < sizeof(RGBQUAD); j++)
			    *lpColorTable++ = 0;
		    for (j = 0; j < sizeof(RGBQUAD); j++)
			    *lpColorTable++ = 255;
        }
        case 16:
        {
            // Assumption : This is a 4 bit gray scale dib. Set the Color Table to 
            // reflect a gray scale scheme for 4 bit Dibs
	        for (int i = 0; i < colorTableEntries; i++)
		        for (int j = 0; j < sizeof(RGBQUAD); j++)
			        *lpColorTable++ = i*16; // Every 16th shade of gray will be 
                                            // chosen
        }
        case 256:
        {
            // Assumption : This is a gray scale dib.  Set the Color Table to 
            // reflect a gray scale scheme for 8 bit Dibs
	        for (int i = 0; i < colorTableEntries; i++)
		        for (int j = 0; j < sizeof(RGBQUAD); j++)
			        *lpColorTable++ = i;
        }
    }
}

UINT ComputePaletteSize(int nBitCount)
{
	// ASSUME lpBMIH->biClrUsed == 0
	UINT nColors = 0;
	if (nBitCount <=8)
		nColors=1<<nBitCount;
	ASSERT(nColors <= 256); 
	return nColors;
}

TWAIN_RESULT InvertDIB(HGLOBAL hDibHandle)
{
    TWAIN_RESULT    result = TWAIN_OK;
    BYTE            *pTop, *pBottom, *pTemp;
    int             nWidth = 0;

    // We create the DIB in memory and this header describes it .
    LPBITMAPINFOHEADER lpBMIH = (LPBITMAPINFOHEADER) GlobalLock(hDibHandle);

    nWidth  = lpBMIH->biSizeImage / lpBMIH->biHeight ;
    
    pTop    = (LPBYTE) lpBMIH + sizeof(BITMAPINFOHEADER) + 
					    (sizeof(RGBQUAD) * lpBMIH->biClrUsed); 
    pBottom = pTop + lpBMIH->biSizeImage - nWidth;
    
    pTemp   = new BYTE[nWidth];

    // Start with the last row and copy it to the first one using an 
    // internediate buffer which is one row wide 
    while (pTop < pBottom)
    {
        memcpy(pTemp, pBottom, nWidth);
        memcpy(pBottom, pTop, nWidth);
        memcpy(pTop, pTemp, nWidth);

        pTop += nWidth;
        pBottom -= nWidth;
    }
    
    delete [] pTemp;

    GlobalUnlock(hDibHandle);

    return result;
}


void SaveBitmap(HGLOBAL _hDIB, LPCSTR pszFileName)
{
	/*
	*	Declared locals
	*/
	BITMAPFILEHEADER bfh;
	BITMAPINFOHEADER *pBIH = NULL;

	HANDLE hf = NULL;
	UINT i = 0, cbString = 0;

	/*
	*	Initialize all structures
	*/
	memset(&bfh, 0, sizeof(BITMAPFILEHEADER));

	/*
	*	We are expecting a valid HWND
	*/
	ASSERT(pszFileName);
	
	if(_hDIB)
	{

		hf = CreateFile(pszFileName, 
										GENERIC_WRITE, 0, 
										(LPSECURITY_ATTRIBUTES)NULL,
										CREATE_ALWAYS,
										FILE_ATTRIBUTE_NORMAL,
										(HANDLE)NULL);

		if(hf)
		{
			/*
			*	Get a pointer to the beginning of the DIB
			*/
			pBIH = (BITMAPINFOHEADER*)GlobalLock(_hDIB);
			if(pBIH)
			{
				DWORD dwBytesWritten = 0l;

				/*
				* Fill in the Bitmap File Header
				*/
				bfh.bfType = ( (WORD) ('M' << 8) | 'B');
				
				/*
				*	Calculate the size of the bitmap including the palette
				*/
				bfh.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) +
					((((pBIH->biWidth * pBIH->biBitCount + 31)/32) * 4) * pBIH->biHeight)
						+ pBIH->biClrUsed * sizeof(RGBQUAD);
				bfh.bfReserved1 = 0;
				bfh.bfReserved2 = 0;
				
				/*
				*	Offset to actual bits is after palette
				*/
				bfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER)
					+ pBIH->biClrUsed * sizeof(RGBQUAD);

				ASSERT(bfh.bfSize > sizeof(BITMAPFILEHEADER));

				/*
				*	Write the results
				*/
				WriteFile(hf, &bfh, sizeof(BITMAPFILEHEADER), &dwBytesWritten, NULL);
				WriteFile(hf, pBIH, bfh.bfSize-sizeof(BITMAPFILEHEADER), &dwBytesWritten, NULL);
				
				/*
				* Done now
				*/
				GlobalUnlock(_hDIB);
			}
			CloseHandle(hf);
		}
	}
	return;
}

string GetMsgName(PTWMSG pTWMsg)
{
    switch (pTWMsg->MSG)
	{
        case MSG_NULL:
            return string("MSG_NULL");
        case MSG_CUSTOMBASE   :
            return string("MSG_CUSTOMBASE   ");
        case MSG_GET          :
            return string("MSG_GET          ");
        case MSG_GETCURRENT   :
            return string("MSG_GETCURRENT   ");
        case MSG_GETDEFAULT   :
            return string("MSG_GETDEFAULT   ");
        case MSG_GETFIRST     :
            return string("MSG_GETFIRST     ");
        case MSG_GETNEXT      :
            return string("MSG_GETNEXT      ");
        case MSG_SET          :
            return string("MSG_SET          ");
        case MSG_RESET        :
            return string("MSG_RESET        ");
        case MSG_QUERYSUPPORT :
            return string("MSG_QUERYSUPPORT ");
        case MSG_XFERREADY    :
            return string("MSG_XFERREADY    ");
        case MSG_CLOSEDSREQ   :
            return string("MSG_CLOSEDSREQ   ");
        case MSG_CLOSEDSOK    :
            return string("MSG_CLOSEDSOK    ");
        case MSG_DEVICEEVENT  :
            return string("MSG_DEVICEEVENT  ");
        case MSG_CHECKSTATUS  :
            return string("MSG_CHECKSTATUS  ");
        case MSG_OPENDSM      :
            return string("MSG_OPENDSM      ");
        case MSG_CLOSEDSM     :
            return string("MSG_CLOSEDSM     ");
        case MSG_OPENDS       :
            return string("MSG_OPENDS       ");
        case MSG_CLOSEDS      :
            return string("MSG_CLOSEDS      ");
        case MSG_USERSELECT   :
            return string("MSG_USERSELECT   ");
        case MSG_DISABLEDS    :
            return string("MSG_DISABLEDS    ");
        case MSG_ENABLEDS     :
            return string("MSG_ENABLEDS     ");
        case MSG_ENABLEDSUIONLY  :
            return string("MSG_ENABLEDSUIONLY  ");
        case MSG_PROCESSEVENT :
            return string("MSG_PROCESSEVENT ");
        case MSG_ENDXFER      :
            return string("MSG_ENDXFER      ");
        case MSG_CHANGEDIRECTORY   :
            return string("MSG_CHANGEDIRECTORY   ");
        case MSG_CREATEDIRECTORY   :
            return string("MSG_CREATEDIRECTORY   ");
        case MSG_DELETE            :
            return string("MSG_DELETE            ");
        case MSG_FORMATMEDIA       :
            return string("MSG_FORMATMEDIA       ");
        case MSG_GETCLOSE          :
            return string("MSG_GETCLOSE          ");
        case MSG_GETFIRSTFILE      :
            return string("MSG_GETFIRSTFILE      ");
        case MSG_GETINFO           :
            return string("MSG_GETINFO           ");
        case MSG_GETNEXTFILE       :
            return string("MSG_GETNEXTFILE       ");
        case MSG_RENAME            :
            return string("MSG_RENAME            ");
        case MSG_PASSTHRU          :
            return string("MSG_PASSTHRU          ");

    }
    return string("Unknown Msg");
}


/*
*	Remove this entire portion during release build
*/
#ifdef _DEBUG

/*
*	Maximum size of any string used for debugging
*/
#define MAXDEBUGSTRING 512

/*
* Function: TriggerAssertionFailed
* Inputs:
*		lpszFileName - contains name of file where assertion was triggered
*		nLine - containes the line index where assertion was triggered
* Outputs:
*		FALSE if the user chose to ignore the warning
* Comments:
*/
BOOL TriggerAssertionFailed(LPCSTR lpszFileName, int nLine)
{
	/*
	* Initialized locals, allocate string memory
	* (string is not allocated on the stack to reduce possible impact)
	*/
	int nCode = IDIGNORE;
	char *pszDebugString = (char*)malloc(MAXDEBUGSTRING);

	/*
	*	Ensure string is valid (don't assert...)
	*/
	if(pszDebugString)
	{
		/*
		*	Prepare and send debug window output message
		*/
		wsprintf(pszDebugString, "Assertion Failed. Line %d, File %s.\n", nLine, lpszFileName);
		OutputDebugString(pszDebugString);

		/*
		*	Prepare alternate message for alert box
		*/
		wsprintf(pszDebugString, "File %s\n Line %d", lpszFileName, nLine);

		/*
		*	Display alert message box and accept user response
		*/
		#ifdef _WIN32
		nCode = MessageBox(NULL, pszDebugString, "Assertion Failed!",
		MB_TASKMODAL|MB_ICONHAND|MB_RETRYCANCEL|MB_SETFOREGROUND);
		#else
		nCode = MessageBox(NULL, pszDebugString, "Assertion Failed!",
		MB_TASKMODAL|MB_ICONHAND|MB_RETRYCANCEL);
		#endif //_WIN32	

		/*
		*	Cleanup local string
		*/
		free(pszDebugString);
		pszDebugString = NULL;
	}

	if (nCode == IDIGNORE)
	{
		/*
		*	User chose to ignore, calling macro should proceed
		*/
		return FALSE;
	}

	/*
	*	User chose to calling macro should attempt a break for the debugger
	*/
	return TRUE; 
}

/*
* Function: DebugTraceMessage
* Inputs:
*		lpszFormat - printf style format string
*		... - variable # of stacked parameters
* Outputs: none
* Comments:
*/
void DebugTraceMessage(LPCSTR lpszFormat, ...)
{
	/*
	* duplicate printf functionality for OutputDebugString
	*/
	char *pszDebugString = (char*)malloc(MAXDEBUGSTRING);
	va_list args;
	va_start(args, lpszFormat);
	
	if(pszDebugString)
	{
		/*
		* format and output the debug string
		*/
		vsprintf(pszDebugString, lpszFormat, args);	
		OutputDebugString(pszDebugString);

		/*
		* Cleanup local string
		*/
		free(pszDebugString);
		pszDebugString = NULL;
	}
	va_end(args);
	return;
}

#endif //_DEBUG

//convert scan bit depth to image bit depth
int GetImageBitDepth(int scanBitDepth)
{
	if (scanBitDepth%3)
	{
		//bw
		if (scanBitDepth <= 1) return 1;
		else if (scanBitDepth <= 8) return 8;
		return 16;
	}
	//color
	if (scanBitDepth <= 24) return 24;
	return 48;
}

