// FILE		: UI_GAMMADLG.CPP

// DESCRIPTION	: Implementation of the CGammaDlg class
//		  and the CGammaWnd class

// This file contains the code for the gamma curve window in the ui
// and the gamma curve window displayed in the tab control

// REVISIONS	: WS 7/27/98, first version
//		: ws 8/19/98, gamma file name is loaded from ini file
//		: ws 8/19/98, redo ui per specs
//		: ws 8/20/98, move gamma file to mb/ui.ini
//		: ws 8/20/98, use global gamma file name
//		: ws 8/20/98, when ok is pressed, ask to save or download
//		: ws 8/21/98, only cancel button exits without save/write to merlin
//		: ws 8/21/98, remove all tab if gamma curve not modifiable
//		: ws 8/24/98, save gamma file in ini file
//		: ws 8/26/98, gamma file goes in setup not file settings section
//		: ws 8/26/98, remove revert, load, save, save as, read merlin buttons
//		: ws 8/26/98, always save file onok if curve modified
//		: ws 8/27/98, change gamma dlg to modal
//		: ws 9/9/98, set mclk to min while writing gamma values
//		: ws 9/14/98, set mclk to 1 before writing gamma table
//		: sr 8/12/99, cleanup file (remove write regs, external refs, etc.)
//-------------------------------------------------------------------



#include "stdafx.h"
#include "math.h"
#include "resource.h"
#include "UI_GammaDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define INIFILE		GetFullFilePath("LM983x.DEF")


/////////////////////////////////////////////////////////////////////////////
// CGammaDlg dialog
//-------------------------------------------------------
// FUNCTION	:	CGammaDlg class constructor

// DESCRIPTION	:	Create the gamma window and display
//			the gamma window displayed by the ui is a modless
//			dialog whose parent is the ui window.
//			This window allows the user to adjust the gamma
//			curves, send the values to merlin, and adjust
//			the image based on the gamma curve.

// INPUTS	:	pCurve	- GAMMACURVE structure
//		:	pImg	- the image window
//		:	pParent	- window to use as the gamma window parent

// RETURNS	:	nothing
//--------------------------------------------------------
CGammaDlg::CGammaDlg( CString csGammaFileName, GAMMACURVE *pCurve, CDibImage *pImg, CWnd* pParent /*=NULL*/)
	: CDialog(CGammaDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CGammaDlg)
	//}}AFX_DATA_INIT

	iSameGamma = 0;		// same gamma check box, indicates use all or individual curves
	curve = *pCurve;	// current gamma curve
	pImage = pImg;		// image window
	bIsModifiable = TRUE;	// use curve values or use values loaded from file
	bIsChanged = FALSE;	// no changes to gamma curve
	m_filename = csGammaFileName;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	windows control and data object mapper

// DESCRIPTION	:	map the windows controls to their respective objects
//			created and managed by the framework

// INPUTS	:	pDX	- MFC data object

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CGammaDlg)
	DDX_Control(pDX, IDC_CURVESELECT, m_TabControl);
	DDX_Control(pDX, IDC_GRAPHWND, m_GraphWnd);
	//}}AFX_DATA_MAP
}


/////////////////////////////////////////////////////////////////////////////
BEGIN_MESSAGE_MAP(CGammaDlg, CDialog)
	//{{AFX_MSG_MAP(CGammaDlg)
	ON_WM_HSCROLL()
	ON_NOTIFY(NM_CLICK, IDC_CURVESELECT, OnClickCurveselect)
	ON_NOTIFY(TCN_SELCHANGE, IDC_CURVESELECT, OnCurveSelchange)
	ON_EN_KILLFOCUS(IDC_BRIGHTLABEL, OnKillfocusBright)
	ON_EN_KILLFOCUS(IDC_CONTRASTLABEL, OnKillfocusContrast)
	ON_EN_KILLFOCUS(IDC_GAMMALABEL, OnKillfocusGamma)
	ON_EN_KILLFOCUS(IDC_HIGHLIGHTLABEL, OnKillfocusHighlight)
	ON_EN_KILLFOCUS(IDC_SHADOWLABEL, OnKillfocusShadow)
	ON_BN_CLICKED(IDC_RESET, OnReset)
	ON_BN_CLICKED(IDC_REALLYCANCEL, OnReallycancel)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnInitDialog

// DESCRIPTION	:	Initialize ui window controls, position gamma window where last used

// RETURNS	:	FALSE if focus is set to a control, TRUE otherwise
//--------------------------------------------------------
BOOL CGammaDlg::OnInitDialog()
{
char sBuf[20];		// ini file buffer
CScrollBar *pScroll;	// scroll bar
FILE *fptr;		// file handle of initial gamma file

	CDialog::OnInitDialog();

	// get the last position of this window
	GetPrivateProfileString("GAMMA","POS","20,10",sBuf,sizeof(sBuf),INIFILE);
	// and restore it

	// init scroll bars
	pScroll = (CScrollBar *) GetDlgItem(IDC_BRIGHTSCROLL);
	pScroll->SetScrollRange(BRIGHT_MIN,BRIGHT_MAX,TRUE);
	pScroll = (CScrollBar *) GetDlgItem(IDC_CONTRASTSCROLL);
	pScroll->SetScrollRange(CONTRAST_MIN,CONTRAST_MAX,TRUE);
	pScroll = (CScrollBar *) GetDlgItem(IDC_HIGHLIGHTSCROLL);
	pScroll->SetScrollRange(HIGHLIGHT_MIN,HIGHLIGHT_MAX,TRUE);
	pScroll = (CScrollBar *) GetDlgItem(IDC_SHADOWSCROLL);
	pScroll->SetScrollRange(SHADOW_MIN,SHADOW_MAX,TRUE);
	pScroll = (CScrollBar *) GetDlgItem(IDC_GAMMASCROLL);
	pScroll->SetScrollRange(1,40,TRUE);

	// tell the tab window how to display curves
	m_GraphWnd.SetGammaValues(bIsModifiable,&curve);

	// get the last selected color
	GetPrivateProfileString("SETUP","DRAWCOLOR","3",sBuf,sizeof(sBuf),INIFILE);
	drawColor = atoi(sBuf);
	if (drawColor < DRAW_RED || drawColor > DRAW_ALL)
		drawColor = DRAW_ALL;

	SetDlgItemText(IDC_FILENAME,(LPCSTR) m_filename);

	// setup the tab window
	TC_ITEM TabCtrlItem;
	TabCtrlItem.mask = TCIF_TEXT;
	TabCtrlItem.pszText = "Red";
	m_TabControl.InsertItem( 0, &TabCtrlItem );
	TabCtrlItem.pszText = "Green";
	m_TabControl.InsertItem( 1, &TabCtrlItem );
	TabCtrlItem.pszText = "Blue";
	m_TabControl.InsertItem( 2, &TabCtrlItem );
	TabCtrlItem.pszText = "All";
	m_TabControl.InsertItem( 3, &TabCtrlItem );

	// draw gamma curves in tab window
	m_GraphWnd.InitGraph(drawColor);

	// open or create gamma table
	if(!(fptr = fopen((LPCSTR) m_filename,"r")))
	{
		// if INI gamma file not available, try default.gam
		m_filename = "default.gam";
		if(!(fptr = fopen((LPCSTR) m_filename,"r")))
		{
			// if default.gam not available, create it
			if(!(fptr = fopen((LPCSTR) m_filename,"w"))) 
			{
				m_filename = "";
diskerr:
				if (fptr) fclose(fptr);
				ResetAll();
				SetDlgItemText(IDC_FILENAME,(LPCSTR) m_filename);
				return TRUE;
			}
			if (!SaveDefaultGamma(fptr)) goto diskerr;
			fclose(fptr);
			if(!(fptr = fopen((LPCSTR) m_filename,"r"))) goto diskerr;
		}
	}
	if (!ReadGammaFile(fptr)) ResetAll();
	fclose(fptr);
	SetDlgItemText(IDC_FILENAME,(LPCSTR) m_filename);

	// initialize scroll bars
	SetupScrollControls(drawColor);

	return TRUE;
}
//-------------------------------------------------------
// FUNCTION	:	ResetAll()

// DESCRIPTION	:	reset all gamma curve parameters

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::ResetAll()
{
int i;	// work variable

	for (i = 0; i < 4; i++) { // init parameters
		curve.bright[i] = 0;
		curve.contrast[i] = 0;
		curve.highlight[i] = 0;
		curve.shadow[i] = 0;
		curve.gamma[i] = 10;
	}
	bIsModifiable = TRUE;
	bIsChanged = TRUE;
}
//-------------------------------------------------------
// FUNCTION	:	SaveDefaultGamma(FILE *fp)

// DESCRIPTION	:	save a default gamma file

// INPUTS	:	fp	- file handle of file to write

// RETURNS	:	TRUE if file stored ok
//			FALSE otherwise
//--------------------------------------------------------
BOOL CGammaDlg::SaveDefaultGamma(FILE* fp)
{
int i,j,y;	// work variables

	ResetAll();

	// write the brightness values
	if (fprintf(fp,"; B=%d,%d,%d,%d\n",curve.bright[0],curve.bright[1],curve.bright[2],curve.bright[3]) < 0) {
outputerror:
		return FALSE;
	}

	// write contrast parameters
	if (fprintf(fp,"; C=%d,%d,%d,%d\n",curve.contrast[0],curve.contrast[1],curve.contrast[2],curve.contrast[3]) < 0) goto outputerror;
	// write highlight parameters
	if (fprintf(fp,"; H=%d,%d,%d,%d\n",curve.highlight[0],curve.highlight[1],curve.highlight[2],curve.highlight[3]) < 0) goto outputerror;
	// write shadow parameters
	if (fprintf(fp,"; S=%d,%d,%d,%d\n",curve.shadow[0],curve.shadow[1],curve.shadow[2],curve.shadow[3]) < 0) goto outputerror;
	// write gamma parameters
	if (fprintf(fp,"; G=%d,%d,%d,%d\n",curve.gamma[0],curve.gamma[1],curve.gamma[2],curve.gamma[3]) < 0) goto outputerror;

	for (j = 0; j < 4; j++) { // save r,g,b, and all
		switch (j) {
			case 0: fprintf(fp, ";Gamma Data (red)\n"); break;
			case 1: fprintf(fp, ";Gamma Data (green)\n"); break;
			case 2: fprintf(fp, ";Gamma Data (blue)\n"); break;
			case 3: fprintf(fp, ";Gamma Data (all)\n"); break;
		}
		for (i = 0; i < GAMMA_SIZE; i++) {
			y = i/16;
			if (fprintf(fp,"%x\t%x\n",i,y) < 0) goto outputerror;
		}
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CGammaDlg message handlers

//-------------------------------------------------------
// FUNCTION	:	OnCancel()

// DESCRIPTION	:	message processor for the close button (upper right)
//			save window position and close dialog
//			same as onok

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnCancel()
{
	OnOK();
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnOK()

// DESCRIPTION	:	message processor for the ok button
//			save window position and current screen settings
//			then close dialog

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnOK()
{
CRect rect;	// current window position and size
char sBuf[MAX_PATH];	// ini file buffer

	// get current window size and position and save in ini file
	GetWindowRect(&rect);
	wsprintf(sBuf,"%d,%d,%d,%d",rect.left,rect.top,rect.right,rect.bottom);
	WritePrivateProfileString("GAMMA","POS",sBuf,INIFILE);

	// save current window settings
	wsprintf(sBuf,"%d",drawColor);
	WritePrivateProfileString("SETUP","DRAWCOLOR",sBuf,INIFILE);

	if (bIsChanged) {
		OnFilesave(); // save current gamma settings
	}

	// save selected gamma file name in ini file
	m_filename = m_filename;
	if (m_filename.GetLength() > 0)
		WritePrivateProfileString("SETUP","GAMMAFILE",(LPCSTR)m_filename,INIFILE);

	CDialog::OnOK();
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SetGammaValues(int iSame,GAMMACURVE *pCurve)

// DESCRIPTION	:	set the gamma curves

// INPUTS	:	iSame	- 0 if use individual color curves, 1 if use all curve
//			pCurve	- gamma curve parameters to use

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::SetGammaValues(int iSame,GAMMACURVE *pCurve)
{
	// save the gamma curve parameters
	iSameGamma = iSame;
	curve = *pCurve;
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	GetGammaValues(int *iSame,GAMMACURVE *pCurve)

// DESCRIPTION	:	retrieve the current gamma curve settings

// INPUTS	:	iSame	- 0 if use individual color curves, 1 if use all curve
//			pCurve	- gamma curve parameters to use

// RETURNS	:	iSame and pCurve contain the current gamma window settings
//--------------------------------------------------------
void CGammaDlg::GetGammaValues(int *iSame,GAMMACURVE *pCurve)
{
	// set user supplied variables with current values
	*iSame = iSameGamma;
	*pCurve = curve;
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)

// DESCRIPTION	:	message processor for the horizontal scroll bars used
//			to set the vertical/horizontal resolutions

// INPUTS	:	sBCode	- type of scroll
//			nPos	- current scroll bar position
//			pScrollBar - scroll bar to receive the message

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
int i;			// work variable
char sBuf[20];		// buffer for converting double to text
int min,max;		// min and max values for the scroll bar
int pagewidth;		// distance to travel on page commands
int ctrlid;		// id of the scroll bar control

	// get the scroll bar being modified
	ctrlid = pScrollBar->GetDlgCtrlID();

	min = 1;
	max = 9600;
	pagewidth = 50;
	// set parameters for this scroll bar
	switch (ctrlid) {
		case IDC_BRIGHTSCROLL:
			min = BRIGHT_MIN;
			max = BRIGHT_MAX;
			pagewidth = 10;
			break;
		case IDC_CONTRASTSCROLL:
			min = CONTRAST_MIN;
			max = CONTRAST_MAX;
			pagewidth = 10;
			break;
		case IDC_HIGHLIGHTSCROLL:
			min = HIGHLIGHT_MIN;
			max = HIGHLIGHT_MAX;
			pagewidth = 10;
			break;
		case IDC_SHADOWSCROLL:
			min = SHADOW_MIN;
			max = SHADOW_MAX;
			pagewidth = 10;
			break;
		case IDC_GAMMASCROLL:
			min = 1;
			max = 40;
			pagewidth = 10;
			break;
	}

	// get current position of this scroll bar
	int pos = pScrollBar->GetScrollPos();

	// update current position
	switch ( nSBCode ) {

		case SB_LINELEFT:
			if (pos > min) --pos;
			else pos = min;
			break;

		case SB_LINERIGHT:
			if (pos < max) ++pos;
			else pos = max;
			break;

		case SB_PAGELEFT:
			if (pos >= pagewidth+min) pos -= pagewidth;
			else pos = min;
			pos = (pos/10) * 10;
			if (pos < min) pos = min;
			break;

		case SB_PAGERIGHT:
			if (pos <= max-pagewidth) pos += pagewidth;
			else pos = max;
			pos = (pos/10) * 10;
			break;

		case SB_THUMBPOSITION:
		case SB_THUMBTRACK:
			pos = nPos;
			break;

		case SB_ENDSCROLL:
		default:
			return;
	}

	// set current value and update the labels/edit controls of the scroll bar
	switch (ctrlid) {
		case IDC_BRIGHTSCROLL:
			SetDlgItemInt(IDC_BRIGHTLABEL,pos);
			curve.bright[drawColor] = pos;
			break;
		case IDC_CONTRASTSCROLL:
			SetDlgItemInt(IDC_CONTRASTLABEL,pos);
			curve.contrast[drawColor] = pos;
			break;
		case IDC_HIGHLIGHTSCROLL:
			// highlight + shadow must be < max
			i = ((CScrollBar *)GetDlgItem(IDC_SHADOWSCROLL))->GetScrollPos();
			if (i >= HIGHLIGHT_PLUS_SHADOW_MAX - pos) {
				curve.shadow[drawColor] = HIGHLIGHT_PLUS_SHADOW_MAX - pos - 1;
				((CScrollBar *)GetDlgItem(IDC_SHADOWSCROLL))->SetScrollPos(curve.shadow[drawColor]);
				SetDlgItemInt(IDC_SHADOWLABEL,curve.shadow[drawColor]);
			}
			SetDlgItemInt(IDC_HIGHLIGHTLABEL,pos);
			curve.highlight[drawColor] = pos;
			break;
		case IDC_SHADOWSCROLL:
			i = ((CScrollBar *)GetDlgItem(IDC_HIGHLIGHTSCROLL))->GetScrollPos();
			if (i >= HIGHLIGHT_PLUS_SHADOW_MAX - pos)
				pos = HIGHLIGHT_PLUS_SHADOW_MAX - i - 1;
			SetDlgItemInt(IDC_SHADOWLABEL,pos);
			curve.shadow[drawColor] = pos;
			break;
		case IDC_GAMMASCROLL:
			sprintf(sBuf,"%3.1f",(float)pos/10.);
			SetDlgItemText(IDC_GAMMALABEL,sBuf);
			curve.gamma[drawColor] = pos;
			break;
	}

	pScrollBar->SetScrollPos( pos );
	// if the image is adjusted per these curves, update image
	if (bIsModifiable) {
		bIsChanged = TRUE;
		m_GraphWnd.SetGammaValues(bIsModifiable,&curve);
		m_GraphWnd.Invalidate();
	}

	if (drawColor == 3) OnSetall(); // if all curve selected, set all curves
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SetupScrollControls(int color)

// DESCRIPTION	:	init gamma curve parameters in scroll bars

// INPUTS	:	color	- current color to display in tab window

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::SetupScrollControls(int color)
{
char sBuf[20];	// double to text buffer

	// tell the tab control which color to display
	drawColor = color;
	m_TabControl.SetCurSel(drawColor);

	// set the gamma curve parameters on the scroll bars
	CScrollBar *pScroll = (CScrollBar *) GetDlgItem(IDC_BRIGHTSCROLL);
	pScroll->SetScrollPos(curve.bright[drawColor]);
	SetDlgItemInt(IDC_BRIGHTLABEL,curve.bright[drawColor]);

	pScroll = (CScrollBar *) GetDlgItem(IDC_CONTRASTSCROLL);
	pScroll->SetScrollPos(curve.contrast[drawColor]);
	SetDlgItemInt(IDC_CONTRASTLABEL,curve.contrast[drawColor]);

	pScroll = (CScrollBar *) GetDlgItem(IDC_HIGHLIGHTSCROLL);
	pScroll->SetScrollPos(curve.highlight[drawColor]);
	SetDlgItemInt(IDC_HIGHLIGHTLABEL,curve.highlight[drawColor]);

	pScroll = (CScrollBar *) GetDlgItem(IDC_SHADOWSCROLL);
	pScroll->SetScrollPos(curve.shadow[drawColor]);
	SetDlgItemInt(IDC_SHADOWLABEL,curve.shadow[drawColor]);

	pScroll = (CScrollBar *) GetDlgItem(IDC_GAMMASCROLL);
	pScroll->SetScrollPos(curve.gamma[drawColor]);
	sprintf(sBuf,"%3.1f",(float)curve.gamma[drawColor]/10.);
	SetDlgItemText(IDC_GAMMALABEL,sBuf);

	if (bIsModifiable) {
		// enable controls
		pScroll = (CScrollBar *) GetDlgItem(IDC_BRIGHTSCROLL);
		pScroll->EnableWindow(TRUE);
		GetDlgItem(IDC_BRIGHTLABEL)->EnableWindow(TRUE);

		pScroll = (CScrollBar *) GetDlgItem(IDC_CONTRASTSCROLL);
		pScroll->EnableWindow(TRUE);
		GetDlgItem(IDC_CONTRASTLABEL)->EnableWindow(TRUE);

		pScroll = (CScrollBar *) GetDlgItem(IDC_HIGHLIGHTSCROLL);
		pScroll->EnableWindow(TRUE);
		GetDlgItem(IDC_HIGHLIGHTLABEL)->EnableWindow(TRUE);

		pScroll = (CScrollBar *) GetDlgItem(IDC_SHADOWSCROLL);
		pScroll->EnableWindow(TRUE);
		GetDlgItem(IDC_SHADOWLABEL)->EnableWindow(TRUE);

		pScroll = (CScrollBar *) GetDlgItem(IDC_GAMMASCROLL);
		pScroll->EnableWindow(TRUE);
		GetDlgItem(IDC_GAMMALABEL)->EnableWindow(TRUE);
		SetDlgItemText(IDC_STATUS,"Status: Data is modifiable.");
		if (m_TabControl.GetItemCount() < 4) {
			TC_ITEM TabCtrlItem;
			TabCtrlItem.mask = TCIF_TEXT;
			TabCtrlItem.pszText = "All";
			m_TabControl.InsertItem( 3, &TabCtrlItem );
		}
	}
	else {
		// disable controls
		pScroll = (CScrollBar *) GetDlgItem(IDC_BRIGHTSCROLL);
		pScroll->EnableWindow(FALSE);
		GetDlgItem(IDC_BRIGHTLABEL)->EnableWindow(FALSE);

		pScroll = (CScrollBar *) GetDlgItem(IDC_CONTRASTSCROLL);
		pScroll->EnableWindow(FALSE);
		GetDlgItem(IDC_CONTRASTLABEL)->EnableWindow(FALSE);

		pScroll = (CScrollBar *) GetDlgItem(IDC_HIGHLIGHTSCROLL);
		pScroll->EnableWindow(FALSE);
		GetDlgItem(IDC_HIGHLIGHTLABEL)->EnableWindow(FALSE);

		pScroll = (CScrollBar *) GetDlgItem(IDC_SHADOWSCROLL);
		pScroll->EnableWindow(FALSE);
		GetDlgItem(IDC_SHADOWLABEL)->EnableWindow(FALSE);

		pScroll = (CScrollBar *) GetDlgItem(IDC_GAMMASCROLL);
		pScroll->EnableWindow(FALSE);
		GetDlgItem(IDC_GAMMALABEL)->EnableWindow(FALSE);
		SetDlgItemText(IDC_STATUS,"Status: Data is unmodifiable (Press Reset to modify default data).");
		if (m_TabControl.GetItemCount() > 3)
			m_TabControl.DeleteItem( 3);
		if (drawColor == DRAW_ALL) {
			drawColor = DRAW_RED;
			m_GraphWnd.InitGraph(drawColor);
			m_TabControl.SetCurSel(drawColor);
		}
	}
} 
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	ReadGammaFile(FILE *fp)

// DESCRIPTION	:	read gamma tables and parameters
//			and set the values on screen

//		:	the file contains the gamma curve parameters for the
//			4 curves (red, green, blue, all) at the beginning of
//			the file in the following order:
//			B=brightness parameters
//			C=contrast parameters
//			H=highlight parameters
//			S=shadow parameters
//			G=gamma parameters

// INPUTS	:	fp	- handle of file to read

// RETURNS	:	true if read ok, false otherwise
//--------------------------------------------------------
BOOL CGammaDlg::ReadGammaFile(FILE *fp)
{
int bright[4];		// brightness coefficients
int contrast[4];	// contrast coefficients
int highlight[4];	// highlight coefficients
int shadow[4];		// shadow coefficients
int gamma[4];		// gamma coefficients
int i,j;		// work variables
LPSTR ptr;		// file input buffer pointer
char sBuf[128];		// file input buffer

	// read the coefficients from the file
	if (!fgets(sBuf,sizeof(sBuf),fp)) {
inputerror:
		MessageBox("Can't read file!","File Error",MB_OK | MB_ICONHAND);
		return FALSE;
	}
	if (strstr(sBuf,"B=") != NULL) {
		sscanf(sBuf,"; B=%d,%d,%d,%d",&bright[0],&bright[1],&bright[2],&bright[3]);
		bIsModifiable = TRUE;
	}
	else {
		// file doesnt contain coefficients, so set to default
		for (i = 0; i < 4; i++) {
			bright[i] = 0;
			contrast[i] = 0;
			highlight[i] = 0;
			shadow[i] = 0;
			gamma[i] = 10;
		}
		bIsModifiable = FALSE;
		goto skipread;
	}

	// read contrast parameters
	if (!fgets(sBuf,128,fp)) goto inputerror;
	if (strstr(sBuf,"C=") != NULL) {
		sscanf(sBuf,"; C=%d,%d,%d,%d",&contrast[0],&contrast[1],&contrast[2],&contrast[3]);
		bIsModifiable = TRUE;
	}
	else {
		// file doesnt contain coefficients, so set to default
		for (i = 0; i < 4; i++) {
			contrast[i] = 0;
			highlight[i] = 0;
			shadow[i] = 0;
			gamma[i] = 10;
		}
		bIsModifiable = FALSE;
		goto skipread;
	}

	if (!fgets(sBuf,128,fp)) goto inputerror;
	// read highlight parameters
	if (strstr(sBuf,"H=") != NULL) {
		sscanf(sBuf,"; H=%d,%d,%d,%d",&highlight[0],&highlight[1],&highlight[2],&highlight[3]);
		bIsModifiable = TRUE;
	}
	else {
		// file doesnt contain coefficients, so set to default
		for (i = 0; i < 4; i++) {
			highlight[i] = 0;
			shadow[i] = 0;
			gamma[i] = 10;
		}
		bIsModifiable = FALSE;
		goto skipread;
	}

	if (!fgets(sBuf,128,fp)) goto inputerror;
	// read shadow parameters
	if (strstr(sBuf,"S=") != NULL) {
		sscanf(sBuf,"; S=%d,%d,%d,%d",&shadow[0],&shadow[1],&shadow[2],&shadow[3]);
		bIsModifiable = TRUE;
	}
	else {
		// file doesnt contain coefficients, so set to default
		for (i = 0; i < 4; i++) {
			shadow[i] = 0;
			gamma[i] = 10;
		}
		bIsModifiable = FALSE;
		goto skipread;
	}

	if (!fgets(sBuf,128,fp)) goto inputerror;
	// read gamma parameters
	if (strstr(sBuf,"G=") != NULL) {
		sscanf(sBuf,"; G=%d,%d,%d,%d",&gamma[0],&gamma[1],&gamma[2],&gamma[3]);
		bIsModifiable = TRUE;
	}
	else {
		// file doesnt contain coefficients, so set to default
		for (i = 0; i < 4; i++) {
			gamma[i] = 10;
		}
		bIsModifiable = FALSE;
		goto skipread;
	}
	// read line from gamma file
	if (!fgets(sBuf,128,fp)) goto inputerror;

// at this point we have set the coefficients and have the next line of input
skipread:
	// set the gamma curve parameters from the file or default
	// and check values are valid
	for (i = 0; i < 4; i++) {
		curve.bright[i] = bright[i];
		if (curve.bright[i] < BRIGHT_MIN || 
		    curve.bright[i] > BRIGHT_MAX) curve.bright[i] = 0;
	
		curve.contrast[i] = contrast[i];
		if (curve.contrast[i] < CONTRAST_MIN || 
		    curve.contrast[i] > CONTRAST_MAX) 
			curve.contrast[i] = 0;
	
		curve.highlight[i] = highlight[i];
		if (curve.highlight[i] < HIGHLIGHT_MIN || 
		    curve.highlight[i] > HIGHLIGHT_MAX) 
			curve.highlight[i] = 0;
	
		curve.shadow[i] = shadow[i];
		if (curve.shadow[i] < SHADOW_MIN || 
		    curve.shadow[i] > SHADOW_MAX) 
			curve.shadow[i] = 0;
		if (curve.shadow[i] >= HIGHLIGHT_PLUS_SHADOW_MAX-curve.highlight[i]) 
			curve.shadow[i] = HIGHLIGHT_PLUS_SHADOW_MAX - curve.highlight[i] - 1;
	
		curve.gamma[i] = gamma[i];
		if (curve.gamma[i] < 1 || 
		    curve.gamma[i] > 40) curve.gamma[i] = 10;
	}

	// initialize the gamma curves
	for (i = 0; i < GAMMA_SIZE; i++) {
		redy[i] = 0xFF;
		greeny[i] = 0xFF;
		bluey[i] = 0xFF;
		ally[i] = 0xFF;
	}

	// read red, first line already read
	ptr = sBuf; // skip comments and blank lines
	while (isspace(*ptr) && *ptr != '\n') ++ptr;
	if (*ptr != '\n' && *ptr != ';') {
		sscanf(ptr,"%x %x",&i,&j); // line is: offset gamma value
		if (j > 255) j = 255; // range check gamma value
		if (j < 0) j = 0;
		if (i >= 0 && i < GAMMA_SIZE) redy[i] = j; // range check offset
	}
	while (fgets(sBuf,128,fp)) {
		ptr = sBuf; // skip comments and blank lines
		while (isspace(*ptr) && *ptr != '\n') ++ptr;
		if (*ptr == '\n' || *ptr == ';') continue;
		sscanf(ptr,"%x %x",&i,&j);
		if (j > 255) j = 255;
		if (j < 0) j = 0;
		if (i >= 0 && i < GAMMA_SIZE) redy[i] = j;
		if (i == GAMMA_SIZE-1) break; // all done with red when offset = GAMMA_SIZE-1
	}

	// read green
	while (fgets(sBuf,128,fp)) {
		ptr = sBuf; // skip comments and blank lines
		while (isspace(*ptr) && *ptr != '\n') ++ptr;
		if (*ptr == '\n' || *ptr == ';') continue;
		sscanf(ptr,"%x %x",&i,&j);
		if (j > 255) j = 255;
		if (j < 0) j = 0;
		if (i >= 0 && i < GAMMA_SIZE) greeny[i] = j;
		if (i == GAMMA_SIZE-1) break; // all done with green when offset = GAMMA_SIZE-1
	}
	// read blue
	while (fgets(sBuf,128,fp)) {
		ptr = sBuf; // skip comments and blank lines
		while (isspace(*ptr) && *ptr != '\n') ++ptr;
		if (*ptr == '\n' || *ptr == ';') continue;
		sscanf(ptr,"%x %x",&i,&j);
		if (j > 255) j = 255;
		if (j < 0) j = 0;
		if (i >= 0 && i < GAMMA_SIZE) bluey[i] = j;
		if (i == GAMMA_SIZE-1) break; // all done with blue
	}

	// read all
	while (fgets(sBuf,128,fp)) {
		ptr = sBuf; // skip comments and blank lines
		while (isspace(*ptr) && *ptr != '\n') ++ptr;
		if (*ptr == '\n' || *ptr == ';') continue;
		sscanf(ptr,"%x %x",&i,&j);
		if (j > 255) j = 255;
		if (j < 0) j = 0;
		if (i >= 0 && i < GAMMA_SIZE) ally[i] = j;
		if (i == GAMMA_SIZE-1) break; // all done with all curve
	}

	// init scroll bars with current values
	SetupScrollControls(drawColor);
	// set the gamma values in the tab window
	m_GraphWnd.SetGammaValues(bIsModifiable,&curve,GAMMA_SIZE,redy,greeny,bluey,ally);
	m_GraphWnd.InitGraph(drawColor);
	// update the image with the new gamma values
	bIsChanged = FALSE;
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SaveGammaFile() 

// DESCRIPTION	:	save the computed gamma curves or the previously read
//			gamma values in the currently selected file

//		:	the file contains the gamma curve parameters for the
//			4 curves (red, green, blue, all) at the beginning of
//			the file in the following order:
//			B=brightness parameters
//			C=contrast parameters
//			H=highlight parameters
//			S=shadow parameters
//			G=gamma parameters

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::SaveGammaFile() 
{
FILE *fp;	// file handle for gamma file
int i,j,y;	// work variables

	// rewrite the selected file
	if ((fp = fopen((LPCSTR)m_filename,"wt")) == NULL) {
		MessageBox("Can't create file!","File Error",MB_OK | MB_ICONHAND);
		return;
	}

	if (bIsModifiable) { // if the data is modifiable, we have the gamma parameters
		// write the brightness values
		if (fprintf(fp,"; B=%d,%d,%d,%d\n",curve.bright[0],curve.bright[1],curve.bright[2],curve.bright[3]) < 0) {
outputerror:
			fclose(fp);
			MessageBox("Can't write file!","File Error",MB_OK | MB_ICONHAND);
			return;
		}
	
		// write contrast parameters
		if (fprintf(fp,"; C=%d,%d,%d,%d\n",curve.contrast[0],curve.contrast[1],curve.contrast[2],curve.contrast[3]) < 0) goto outputerror;
		// write highlight parameters
		if (fprintf(fp,"; H=%d,%d,%d,%d\n",curve.highlight[0],curve.highlight[1],curve.highlight[2],curve.highlight[3]) < 0) goto outputerror;
		// write shadow parameters
		if (fprintf(fp,"; S=%d,%d,%d,%d\n",curve.shadow[0],curve.shadow[1],curve.shadow[2],curve.shadow[3]) < 0) goto outputerror;
		// write gamma parameters
		if (fprintf(fp,"; G=%d,%d,%d,%d\n",curve.gamma[0],curve.gamma[1],curve.gamma[2],curve.gamma[3]) < 0) goto outputerror;
	}

	// if save from curves, compute curve parameters
	if (bIsModifiable) {
		for (i = 0; i < 4; i++) {
			double xx = (double)GAMMA_SIZE - (4*curve.highlight[i] + 4*curve.shadow[i]);
			double ig = 10./(double) curve.gamma[i];
			if (xx <= 0) m_GraphWnd.constantPart[i] = 0;
			else m_GraphWnd.constantPart[i] = 255./pow(xx,ig);
		}
	}

	int *py;	// pointer to current array to save
	for (j = 0; j < 4; j++) { // save r,g,b, and all
		switch (j) {
			case 0: py = redy; fprintf(fp, ";Gamma Data (red)\n"); break;
			case 1: py = greeny; fprintf(fp, ";Gamma Data (green)\n"); break;
			case 2: py = bluey; fprintf(fp, ";Gamma Data (blue)\n"); break;
			case 3: py = ally; fprintf(fp, ";Gamma Data (all)\n"); break;
		}
		for (i = 0; i < GAMMA_SIZE; i++) {
			if (bIsModifiable) // if save curve, evaluate gamma value
				y = m_GraphWnd.EvalCurve(i,j);
			else
				y = py[i]; // otherwise, use value read from file
			if (fprintf(fp,"%x\t%x\n",i,y) < 0) goto outputerror;
		}
	}
	fclose(fp);
	SetDlgItemText(IDC_STATUS,"File saved.");
	SetupScrollControls(drawColor);
	bIsChanged = FALSE;
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnFilesave() 

// DESCRIPTION	:	message processor for the save file button
//			save the gamma values in the selected gamma file
//			and write the values to merlin

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnFilesave() 
{
	// if a file is not selected, get file name then save
	if (m_filename.GetLength() == 0)
		return;
	else { // otherwise, save the file and write it to merlin
		SaveGammaFile();
	}
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnClickCurveselect(NMHDR* pNMHDR, LRESULT* pResult) 

// DESCRIPTION	:	called by the framework when a tab is changed
//			not used

// INPUTS	:	NMHDR* pNMHDR - see MFC message description for tab control
//		:	LRESULT* pResult

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnClickCurveselect(NMHDR* pNMHDR, LRESULT* pResult) 
{
	*pResult = 0;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnCurveSelchange(NMHDR* pNMHDR, LRESULT* pResult) 

// DESCRIPTION	:	message processor for tab control
//			this message is sent when the user clicks on a tab
//			a color tab on the graph window has been pressed

// INPUTS	:	NMHDR* pNMHDR - see MFC message description for tab control
//		:	LRESULT* pResult

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnCurveSelchange(NMHDR* pNMHDR, LRESULT* pResult) 
{
	// get the newly selected tab
	int nColor = m_TabControl.GetCurSel();

	// set the color of the selected tab
	switch(nColor)
	{
	case 0:
		drawColor = DRAW_RED;
		break;
	case 1:
		drawColor = DRAW_GREEN;
		break;
	case 2:
		drawColor = DRAW_BLUE;
		break;
	case 3:
		drawColor = DRAW_ALL;
		break;
	}
	// display the scroll bars for this color		
	SetupScrollControls(drawColor);
	// redraw the graph for this color
	m_GraphWnd.InitGraph(drawColor);
	*pResult = 0;
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnGraphcurve() 

// DESCRIPTION	:	message processor for the graph computed curve check box
//			set the graphs in the tab window to show the computed
//			curve or the gamma values in the file.

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnGraphcurve() 
{
	// set the graphs depending on the check box
	m_GraphWnd.SetGammaValues(bIsModifiable,&curve);
	m_GraphWnd.Invalidate(); // redraw graphs in tab window
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnKillfocusBright() 

// DESCRIPTION	:	message processor for the kill focus message on the brightness edit box
//			update the scroll bar with the entered value

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnKillfocusBright() 
{
	int i = m_TabControl.GetCurSel();	// get the current color
	int x = GetDlgItemInt(IDC_BRIGHTLABEL);	// get value in edit control
	if (x == curve.bright[i]) return;	// no change, so nothing to do
	curve.bright[i] = x;			// update the current value
	if (curve.bright[i] < BRIGHT_MIN || 
	    curve.bright[i] > BRIGHT_MAX) 
		curve.bright[i] = 0;		// range check
	SetupScrollControls(i);			// adjust the scroll bar
	if (bIsModifiable) {			// if drawing computed curve, recalc and show
		m_GraphWnd.SetGammaValues(bIsModifiable,&curve);
		m_GraphWnd.Invalidate();
	}
	if (drawColor == 3) OnSetall(); // if all curve selected, set all curves
	bIsChanged = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnKillfocusContrast() 

// DESCRIPTION	:	message processor for the kill focus message on the contrast edit box
//			update the scroll bar with the entered value

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnKillfocusContrast() 
{
	int i = m_TabControl.GetCurSel();	// get current color
	int x = GetDlgItemInt(IDC_CONTRASTLABEL);// get entered value
	if (x == curve.contrast[i]) return;	// no change, so nothing to do
	curve.contrast[i] = x;			// save new value
	if (curve.contrast[i] < CONTRAST_MIN || 
	    curve.contrast[i] > CONTRAST_MAX) 
		curve.contrast[i] = 0;		// range check
	SetupScrollControls(i);			// update scroll bar
	if (bIsModifiable) {			// update graph in tab window
		m_GraphWnd.SetGammaValues(bIsModifiable,&curve);
		m_GraphWnd.Invalidate();
	}
	if (drawColor == 3) OnSetall(); // if all curve selected, set all curves
	bIsChanged = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnKillfocusGamma() 

// DESCRIPTION	:	message processor for the kill focus message on the gamma edit box
//			update the scroll bar with the entered value

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnKillfocusGamma() 
{
char sBuf[20];	// text to float buffer
int x;		// 10 * gamma value
float fx;	// gamma value

	int i = m_TabControl.GetCurSel();	// get current color
	GetDlgItemText(IDC_GAMMALABEL,sBuf,sizeof(sBuf)); // get entered value
	sscanf(sBuf,"%f",&fx);			// convert to float
	x = (int) (10.*fx);			// gamma value is 10*entered value
	if (x == curve.gamma[i]) return;	// no change, so nothing to do
	curve.gamma[i] = x;
	if (curve.gamma[i] < 1 || 
	    curve.gamma[i] > 40) curve.gamma[i] = 10; // range check
	SetupScrollControls(i);			// set scroll bar
	if (bIsModifiable) {			// update curve in tab window
		m_GraphWnd.SetGammaValues(bIsModifiable,&curve);
		m_GraphWnd.Invalidate();
	}
	if (drawColor == 3) OnSetall(); // if all curve selected, set all curves
	bIsChanged = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnKillfocusHighlight() 

// DESCRIPTION	:	message processor for the kill focus message on the highlight edit box
//			update the scroll bar with the entered value

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnKillfocusHighlight() 
{
	int i = m_TabControl.GetCurSel();	// get color
	int x = GetDlgItemInt(IDC_HIGHLIGHTLABEL); // get entered value
	if (x == curve.highlight[i]) return;	// no change, so nothing to do
	curve.highlight[i] = x;
	if (curve.highlight[i] < HIGHLIGHT_MIN || 
	    curve.highlight[i] > HIGHLIGHT_MAX) 
		curve.highlight[i] = 0;		// range check
	int s = ((CScrollBar *)GetDlgItem(IDC_SHADOWSCROLL))->GetScrollPos();
	if (s >= HIGHLIGHT_PLUS_SHADOW_MAX - x) { // adjust shadow if necessary
		curve.shadow[i] = HIGHLIGHT_PLUS_SHADOW_MAX - x - 1;
		((CScrollBar *)GetDlgItem(IDC_SHADOWSCROLL))->SetScrollPos(curve.shadow[i]);
		SetDlgItemInt(IDC_SHADOWLABEL,curve.shadow[i]);
	}
	SetupScrollControls(i);	// update scroll bar
	if (bIsModifiable) {	// update graph in tab window
		m_GraphWnd.SetGammaValues(bIsModifiable,&curve);
		m_GraphWnd.Invalidate();
	}
	if (drawColor == 3) OnSetall(); // if all curve selected, set all curves
	bIsChanged = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnKillfocusShadow() 

// DESCRIPTION	:	message processor for the kill focus message on the shadow edit box
//			update the scroll bar with the entered value

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnKillfocusShadow() 
{
	int i = m_TabControl.GetCurSel();	// current color
	int x = GetDlgItemInt(IDC_SHADOWLABEL);	// current value
	if (x == curve.shadow[i]) return;	// no change, so nothing to do
	curve.shadow[i] = x;
	if (curve.shadow[i] < SHADOW_MIN || 
	    curve.shadow[i] > SHADOW_MAX) 
		curve.shadow[i] = 0;		// range check
	// shadow+highlight must be < max
	if (curve.shadow[i] >= HIGHLIGHT_PLUS_SHADOW_MAX-curve.highlight[i]) 
		curve.shadow[i] = HIGHLIGHT_PLUS_SHADOW_MAX - curve.highlight[i] - 1;

	SetupScrollControls(i);	// update scroll bars
	if (bIsModifiable) {	// update graph in tab window
		m_GraphWnd.SetGammaValues(bIsModifiable,&curve);
		m_GraphWnd.Invalidate();
	}
	if (drawColor == 3) OnSetall(); // if all curve selected, set all curves
	bIsChanged = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnReset() 

// DESCRIPTION	:	message processor for the reset button
//			reset the gamma coefficients to the default

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnReset() 
{
	ResetAll();
	bIsModifiable = TRUE;

	SetupScrollControls(drawColor);		// update scroll bars
	if (bIsModifiable) {		// update graph in tab window
		m_GraphWnd.SetGammaValues(bIsModifiable,&curve);
		m_GraphWnd.Invalidate();
	}
	bIsChanged = TRUE;
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnSetall() 

// DESCRIPTION	:	message processor for the set all menu entry
//			this is the right button menu to set the individual
//			color coefficients from the all curve.
//			Set the coefficients for r,g,b from the coefficients
//			now set for the all curve.

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaDlg::OnSetall() 
{
	// copy the coefficients to r,g,b curves
	for (int i = 0; i < 3; i++) {
		curve.bright[i] = curve.bright[3];
		curve.contrast[i] = curve.contrast[3];
		curve.highlight[i] = curve.highlight[3];
		curve.shadow[i] = curve.shadow[3];
		curve.gamma[i] = curve.gamma[3];
	}
	if (bIsModifiable) { // update the graph in the tab window
		m_GraphWnd.SetGammaValues(bIsModifiable,&curve);
		m_GraphWnd.Invalidate();
	}
	bIsChanged = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CGammaWnd
//-------------------------------------------------------
// FUNCTION	:	CGammaWnd class constructor

// DESCRIPTION	:	Create the gamma curve window
//			the gamma curve window sits inside of the tab control
//			and displays the currently selected gamma curve

// RETURNS	:	nothing
//--------------------------------------------------------
CGammaWnd::CGammaWnd()
{
	bDrawCircles = TRUE;	// points for dragging around gamma curve
	bEvalCurve = TRUE;	// display computed curve
	drawColor = DRAW_ALL;	// default graph to display
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	CGammaWnd class destructor

// DESCRIPTION	:	clean up gamma wnd memory

// RETURNS	:	nothing
//--------------------------------------------------------
CGammaWnd::~CGammaWnd()
{
}


/////////////////////////////////////////////////////////////////////////////
BEGIN_MESSAGE_MAP(CGammaWnd, CWnd)
	//{{AFX_MSG_MAP(CGammaWnd)
	ON_WM_ERASEBKGND()
	ON_WM_LBUTTONDOWN()
	ON_WM_PAINT()
	ON_WM_SETCURSOR()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	InitGraph(int color)

// DESCRIPTION	:	set which graph to display

// INPUTS	:	color	- color of graph to draw, r,g,b or all

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaWnd::InitGraph(int color)
{
	drawColor = color;	// save graph to draw 0..3
	Invalidate();		// redraw graph
}
/////////////////////////////////////////////////////////////////////////////
// CGammaWnd message handlers
//-------------------------------------------------------
// FUNCTION	:	OnEraseBkgnd(CDC* pDC)

// DESCRIPTION	:	erase the gamma curve background

// INPUTS	:	pDC	- dc of gamma curve window

// RETURNS	:	TRUE if message processed, FALSE otherwise
//--------------------------------------------------------
BOOL CGammaWnd::OnEraseBkgnd(CDC* pDC)
{
CRect urect;	// rectangle to update

	GetUpdateRect(&urect);		// get area to update
	if (urect.IsRectEmpty())	// no area specified means entire window
		GetClientRect(&urect);	// so get entire client area
	CBrush cbrush(RGB(0xff,0xff,0xff));	// fill update region with white
	pDC->FillRect(&urect,&cbrush);

	return TRUE;	// no additional processing is needed
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnLButtonDown(UINT nFlags, CPoint pt)

// DESCRIPTION	:	message processor for the lbutton down msg
//			if the mouse is pointing at a dot on the curve,
//			adjust the curve until the mouse button is released.

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaWnd::OnLButtonDown(UINT nFlags, CPoint pt)
{
	CWnd::OnLButtonDown(nFlags, pt);
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnPaint()

// DESCRIPTION	:	message processor for the paint message
//			draw the gamma curve for the currently selected color

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaWnd::OnPaint()
{
	CPaintDC dc(this); // device context for painting

	DrawAxes(&dc);		// draw the x and y axis
	DrawCircles(&dc);	// draw dots on the graph (not used)
	DrawCurve(&dc);		// draw the gamma curve
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	MapXToScreen(int x,int w)

// DESCRIPTION	:	convert an x value in the range of 0..GAMMA_SIZE
//			to the corresponding screen value.

// INPUTS	:	x	- value to map to the screen
//			w	- width of graph window

// RETURNS	:	screen coordinate
//--------------------------------------------------------
int CGammaWnd::MapXToScreen(int x,int w)
{
	return yaxis + x*(w-yaxis)/GAMMA_SIZE; // map 0..GAMMA_SIZE->yaxis..window width
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	MapYToScreen(int y)

// DESCRIPTION	:	convert a y value in the range of 0..256
//			to the corresponding screen value.

// INPUTS	:	y	- value to map to the screen

// RETURNS	:	screen coordinate
//--------------------------------------------------------
int CGammaWnd::MapYToScreen(int y)
{
	return xaxis - y*xaxis/256; // map 0..256->xaxis..0

}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	MapScreenToX(int x,int w)

// DESCRIPTION	:	convert a screen value in the range yaxis..window width
//			to the corresponding x value.

// INPUTS	:	x	- screen value to map to 0..GAMMA_SIZE
//			w	- width of graph window

// RETURNS	:	x coordinate
//--------------------------------------------------------
int CGammaWnd::MapScreenToX(int x,int w)
{
	return (x-yaxis)*GAMMA_SIZE/(w-yaxis); // map yaxis..window width->0..GAMMA_SIZE
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	MapScreenToY(int y)

// DESCRIPTION	:	convert a screen y value to 0..256

// INPUTS	:	y	- screen value to map to 0..256

// RETURNS	:	y coordinate
//--------------------------------------------------------
int CGammaWnd::MapScreenToY(int y)
{
	return 256 - y*256/xaxis; // map xaxis..0->0..256
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	EvalCurve(int x,int color)

// DESCRIPTION	:	evaluate the gamma curve at the point x.
//			before calling this, compute 
//			constantPart = 255/((GAMMA_SIZE - (s+h))^(1/g)
//			where	s = shadow
//				h = highlight
//				g = gamma coefficient
//			y = constant part * (x-4*shadow)^(1/gamma)
//				+ brightness - contrast * cos(x/GAMMA_SIZE * pi)

// INPUTS	:	x	- 0..GAMMA_SIZE
//			color	- curve to compute

// RETURNS	:	gamma curve value
//--------------------------------------------------------
int CGammaWnd::EvalCurve(int x,int color)
{
	// 0..4*shadow -> 0
	if (x < 4*curve.shadow[color]) return 0;
	// if shadow/highlight out of range, gamma value = 0
	if (curve.shadow[color]+curve.highlight[color] >= HIGHLIGHT_PLUS_SHADOW_MAX) return 0;
	double ig = 10./(double) curve.gamma[color];
	double y = (int) (constantPart[color] * pow((double)(x-4*curve.shadow[color]),ig));
	y += (double)curve.bright[color] - (double)curve.contrast[color] * cos(3.1415926536*(double) x/GAMMA_SIZE);
	if (y < 0) return 0;
	if (y > 255) return 255;
	return (int) (y+.5);
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	DrawAxes(CDC *pdc)

// DESCRIPTION	:	draw the axes in the graph window
//			origin is in the lower left
//			graph domain = 0..GAMMA_SIZE
//			graph range = 0..256

// INPUTS	:	pdc	- cdc to draw on

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaWnd::DrawAxes(CDC *pdc)
{
CRect 	textRect;	// rectangle for drawing axis labels
CFont 	*poldfont;	// current font in the cdc
CFont 	*pfont;		// font to draw axis labels
CPen	*poldpen;	// current pen in the cdc
TEXTMETRIC tm;		// text metrics for label font
CRect	pRect;		// window rectangle
char	sBuf[20];	// buffer for creating labels

	if (!pdc) return;	// invalid cdc

	GetClientRect(&pRect);	// get drawing rect

	/* select the font for the axis labels*/
	pfont = new CFont();
	pfont->CreatePointFont(60,"Arial Narrow"); // 6 pt arial narrow
	poldfont = pdc->SelectObject( pfont);
   	poldpen = (CPen *) pdc->SelectStockObject(BLACK_PEN);
	pdc->GetTextMetrics(&tm);

	// x axis is above the bottom of the window 1.5 lines
	xaxis = pRect.Height() - 3*tm.tmHeight/2;
	// y axis is in from the left 5 character widths
	yaxis = 5*tm.tmAveCharWidth;
	// move to the origin
	pdc->MoveTo(yaxis, xaxis);
	pdc->LineTo(pRect.right,xaxis); // draw the x axis

	pdc->MoveTo(yaxis, xaxis+tm.tmHeight/2);
	pdc->LineTo(yaxis,0); // draw the y axis

	pdc->SetBkMode(TRANSPARENT);	// dont draw any background with text
	pdc->SetTextColor(RGB(0,0,0));	// black labels
	textRect.left = tm.tmAveCharWidth/2; // y axis label area
	textRect.right = yaxis - tm.tmAveCharWidth/2;
	int ypos = xaxis; // draw labels up from the x axis
	for (int i=0; i < 256; i += 50) { // draw labels on y axis
		ypos = MapYToScreen(i);
		pdc->MoveTo(yaxis,ypos);
		pdc->LineTo(yaxis-tm.tmAveCharWidth,ypos); // hash mark
		textRect.top = ypos - tm.tmHeight/2; // draw text in middle of hash mark
		textRect.bottom = textRect.top + tm.tmHeight;
		wsprintf(sBuf,"%d",i);
		pdc->DrawText((LPCSTR) sBuf, -1, &textRect,
			DT_RIGHT | DT_NOPREFIX);
	}

	// draw labels on x axis
	textRect.top = xaxis+tm.tmAveCharWidth/2; // just below x axis
	textRect.bottom = textRect.top + tm.tmHeight;
	int xpos = yaxis;
	for (i=0; i < GAMMA_SIZE; i += 500) { // draw labels on x axis
		xpos = MapXToScreen(i,pRect.right);
		pdc->MoveTo(xpos,xaxis);
		pdc->LineTo(xpos,xaxis+tm.tmHeight/2); // draw has mark
		textRect.left = xpos - 2*tm.tmMaxCharWidth;
		textRect.right = textRect.left + 4*tm.tmMaxCharWidth;
		wsprintf(sBuf,"%d",i);
		pdc->DrawText((LPCSTR) sBuf, -1, &textRect,
			DT_CENTER | DT_NOPREFIX | DT_NOCLIP | DT_SINGLELINE);
	}

	/* Clean up. */
	pdc->SelectObject(poldfont);
	pdc->SelectObject(poldpen);
	delete pfont;
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	DrawCircles(CDC *pdc)

// DESCRIPTION	:	draw circles on gamma curve
//			not currently used
//			the dots are used to grab the curve with the mouse
//			and adjust the curve

// INPUTS	:	pdc	- cdc to draw on

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaWnd::DrawCircles(CDC *pdc)
{
//	CPen	*poldpen;
//	CRect	pRect;
//	int	xpos,ypos;

//	if (!pdc) return;
//
//	GetClientRect(&pRect);
//
//	/* create the pen and select */
//	CBrush black_brush(RGB(0,0,0));
//
//   	poldpen = (CPen *) pdc->SelectStockObject(BLACK_PEN);
//
//	xpos = MapXToScreen(0,pRect.right);
//	ypos = MapYToScreen(0);
//	pdc->MoveTo(xpos,ypos);
//	for (int i=0; i < 5; i++) {
//		xpos = MapXToScreen(curve.x[i],pRect.right);
//		ypos = MapYToScreen(curve.y[i]);
//		pdc->LineTo(xpos,ypos); // draw curve between points
//	}
//	xpos = MapXToScreen(GAMMA_SIZE,pRect.right);
//	ypos = MapYToScreen(256);
//	pdc->LineTo(xpos,ypos); // draw curve between points
//
//	if (bDrawCircles) {
//		CRgn *pRegion = new CRgn();
//		for (i=0; i < 5; i++) {
//			xpos = MapXToScreen(curve.x[i],pRect.right);
//			ypos = MapYToScreen(curve.y[i]);
//			pRegion->CreateEllipticRgn(xpos-6,ypos-6,xpos+6,ypos+6);
//			pdc->FillRgn(pRegion,&black_brush);
//			pRegion->DeleteObject();
//		}
//		delete pRegion;
//	}
//
//	pdc->SelectObject(poldpen);
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	DrawCurve(CDC *pdc)

// DESCRIPTION	:	draw the curve corresponding to the currently
//			selected tab (r,g,b,all)

// INPUTS	:	pdc	- dc to draw on

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaWnd::DrawCurve(CDC *pdc)
{
CPen	*poldpen;	// pen currently selected in the dc
CRect	pRect;		// client area of window
int	i,xpos,ypos;	// work variables
double ig;		// intermediate variable for computing gamma curve
COLORREF color;		// color to use for the curve
int *py;		// pointer to the array of gamma values

	if (!pdc) return;	// nothing to draw on

	GetClientRect(&pRect);	// get client area of window

	color = RGB(0,0,0);	// default is black
	switch (drawColor) {	// get color of currently selected curve
		case 0: color = RGB(0xFF,0,0); break;
		case 1: color = RGB(0,0xFF,0); break;
		case 2: color = RGB(0,0,0xFF); break;
		case 3: color = RGB(0,0,0); break;
	}
	// create a pen with the color of the curve
	CPen pen( PS_SOLID, 1, color);
   	poldpen = (CPen *) pdc->SelectObject(&pen);

	// if draw computed curve, compute the curve and draw it
	if (bEvalCurve) {
		for (i = 0; i < 4; i++) { // intermediate value for computing gamma curve
			double xx = (double)GAMMA_SIZE - (4*curve.highlight[i] + 4*curve.shadow[i]);
			ig = 10./(double) curve.gamma[i];
			if (xx <= 0) constantPart[i] = 0;
			else constantPart[i] = 255./pow(xx,ig);
		}
		xpos = 0;	// start at the origin
		ypos = EvalCurve(xpos,drawColor);
		xpos = MapXToScreen(0,pRect.right);
		ypos = MapYToScreen(ypos);
		pdc->MoveTo(xpos,ypos); // move to the first point
		for (i=1; i < GAMMA_SIZE; i++) {
			ypos = EvalCurve(i,drawColor); // compute gamma value
			xpos = MapXToScreen(i,pRect.right);
			ypos = MapYToScreen(ypos);
			pdc->LineTo(xpos,ypos); // draw curve
		}
	}
	else { // otherwise, draw the array of gamma values
		py = NULL;
		switch (drawColor) { // get the array for the current color
			case 0: py = pRedY; break;
			case 1: py = pGreenY; break;
			case 2: py = pBlueY; break;
			case 3: py = pAllY; break;
		}
		if (py) {	// if the array is defined, draw it
			xpos = MapXToScreen(0,pRect.right);
			ypos = MapYToScreen(*py++);
			pdc->MoveTo(xpos,ypos);
			for (i=1; i < GAMMA_SIZE; i++) {
				xpos = MapXToScreen(i,pRect.right);
				ypos = MapYToScreen(*py++);
				pdc->LineTo(xpos,ypos); // draw curve
			}
		}
	}

	// clean up
	pdc->SelectObject(poldpen);
}

/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message)

// DESCRIPTION	:	message processor for the set cursor msg
//			not currently used, use default

// RETURNS	:	mfc default
//--------------------------------------------------------
BOOL CGammaWnd::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message)
{
	return CWnd::OnSetCursor(pWnd, nHitTest, message);
}
/////////////////////////////////////////////////////////////////////////////
//-------------------------------------------------------
// FUNCTION	:	SetGammaValues(BOOL bEval,GAMMACURVE *pCurve,int nPts,
//				int *predy, int *pgreeny,int *pbluey,int *pally)

// DESCRIPTION	:	set the gamma curve coefficients for computing the curves
//			and set the gamma arrays if curve is not computed

// INPUTS	:	bEval	- TRUE to evaluate computed gamma curve,
//				  false to draw array of gamma values
//			pCurve	- gamma curve coefficients
//			pRedy	- red gamma table
//			pGreeny	- green gamma table
//			pBluey	- blue gamma table
//			pAlly	- all gamma table

// RETURNS	:	nothing
//--------------------------------------------------------
void CGammaWnd::SetGammaValues(BOOL bEval,GAMMACURVE *pCurve,int nPts,
				int *predy,
				int *pgreeny,
				int *pbluey,
				int *pally)
{
	// save current settings
	bEvalCurve = bEval;
	curve = *pCurve;
	// save gamma curves
	if (predy) {
		npts = nPts;
		pRedY = predy;
	}
	if (pgreeny) {
		npts = nPts;
		pGreenY = pgreeny;
	}
	if (pbluey) {
		npts = nPts;
		pBlueY = pbluey;
	}
	if (pally) {
		npts = nPts;
		pAllY = pally;
	}
}

/////////////////////////////////////////////////////////////////////////////
void CGammaDlg::OnReallycancel() 
{
CRect rect;	// current window position and size
char sBuf[MAX_PATH];	// ini file buffer

	// get current window size and position and save in ini file
	GetWindowRect(&rect);
	wsprintf(sBuf,"%d,%d,%d,%d",rect.left,rect.top,rect.right,rect.bottom);
	WritePrivateProfileString("GAMMA","POS",sBuf,INIFILE);

	CDialog::OnCancel();
}

/*******************************************************************************
DESCRIPTION:	Returns full path name
PARAMETERS:		none

RETURN VALUE:	CString containing the path.  If error, then a null
				string is returned.
*******************************************************************************/
CString CGammaDlg::GetFullFilePath(CString file)
{
	CString		path;
	char		directory[MAX_PATH];

	if (!GetWindowsDirectory(directory,MAX_PATH))
	{
		AfxMessageBox("Can't find Windows directory!");
		return "";
	}
	path = directory;
	if (path.Right(1) != "\\") path += "\\";
	path += "twain_32\\lm983x\\";
	path += file;
	return( path );
}

