/*******************************************************************************
 *	@doc
 *
 *  @module NSCDevice.cpp | 
 *
 *	This module contains <o UsdNSCDevice> object methods and helpers that 
 *	implement the IStiUSDDevice interface which is one of the standard 
 *	STI interfaces
 *
 *  Copyright (c) National Semiconductor Corporation 1999-2000
 *  All rights reserved
 *
 ******************************************************************************/

#include "NSCStiU.h"
#include "NscDevice.h"
#include "NSCUtils.h"
#include "NSCUsbIO.h"

// Uncomment this if the device has to be polled for events otherwise it will
// be assumed that the device will generate events
#define NS_POLL_FOR_EVENTS   

const DWORD NS_READ_DATA_TIMEOUT	= 	20000; // in milliseconds
const DWORD NS_WRITE_DATA_TIMEOUT	=	20000; // in milliseconds
const DWORD NS_READ_CMD_TIMEOUT		=	20000; // in milliseconds
const DWORD NS_WRITE_CMD_TIMEOUT	=	20000; // in milliseconds
const DWORD NS_THREADWAIT_TIMEOUT	=	5000;  // in milliseconds
const DWORD NS_MAX_ERRORSTRING		=	254;   // characters
const UCHAR NS_MAX_INTERRUPT_PACKET	=	1;	   // bytes	
const DWORD NS_MAX_LOCKTIMEOUT		=	1000;  // in milliseconds
const DWORD NS_POLLING_INTERVAL     =   500;   // in milliseconds
const DWORD NS_DEFAULT_KICKSTART	=	250;   // in milliseconds

const DWORD NS_DEFAULT_DEBOUNCE     = 5000;     // in milliseconds
const DWORD NS_MIN_DEBOUNCE         = 500;      // in milliseconds
const DWORD NS_MAX_DEBOUNCE         = 10000;    // in milliseconds

// The "Significant Events" recognized by the interrupt thread
const DWORD NS_DEVICE_INTERRUPT    = 0x0001; // Process a device event
const DWORD NS_SHUTDOWN_REQUEST    = 0x0002; // Shutdown the interrupt thread
const DWORD NS_WAIT_ERROR          = 0x0003; // Error while polling/waiting 
                                            // for events
const DWORD NS_IOCONTROL_ERROR     = 0x004;  // Error while using the IOcontrol
                                            // call to talk to the device

// Temporary define for the device descriptor the next version 
// of the DDK should have this value defined
const UCHAR NS_USB_DEVICE	= 1;

// Constants for reading and writing registers from the 
// LM9831 using the bulk endpoints
const UCHAR COMMAND_BYTE_COUNT	= 4;
const BYTE	LM_COMMAND_REGISTER	= 7;
const BYTE	LM_IOSTATUS_REGISTER = 2;

const BYTE	NS_MISC_IO_BIT[]	= { 0x04, 0x08, 0x10};
const BYTE	NS_EVENT_BIT[]		= { 0x01, 0x02, 0x04};
const BYTE	NS_RESET_BIT		=  0x20 ;

// The first byte for the bulk register reads is the mode 
// byte and this is bit mapped 
// Bit 0 , 1 - Read, 0 - Write
// Bit 1 , 1 - Inc, 0 - No Inc
const BYTE MODE_INC_READ	= 0x03;
const BYTE MODE_NOINC_READ	= 0x01;
const BYTE MODE_INC_WRITE	= 0x02;
const BYTE MODE_NOINC_WRITE = 0x00;

const TCHAR	NSCUSD_MUTEX[] = TEXT("NSCSTIU_MUTEX_");
const TCHAR	NSCUSD_POLLING_MUTEX[] = TEXT("NSCSTIU_POLLING_MUTEX_"); 
const TCHAR NSCUSD_KICKSTART_MUTEX[] = TEXT("NSCSTIU_KICKSTART_MUTEX_");

const BYTE NS_ENABLE_REMOTEWAKEUP   = 3;
const BYTE NS_DISABLE_REMOTEWAKEUP  = 1;

// The prototype for the interrupt thread fucntion
VOID	InterruptThreadFunc(LPVOID  lpParameter);

/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::UsdNSCDevice is the constructor for this class 

PARAMETERS: 

LPUNKNOWN punkOuter - This is the pointer to the controlling unknown, if we are
being aggregated otherwise this will be NULL - For more info on COM 
aggregation, see Inside COM by Dale Rogerson of other articles on the MSDN

RETURN VALUE:
none

NOTES:
    This constructor handles COM delegation using the non delgating unknonw 
    base class. The only other thing that is done here is the creation of the
    event objects that we will need to use later on to signal the event thread 
    to shutdown
*******************************************************************************/
UsdNSCDevice::UsdNSCDevice( LPUNKNOWN punkOuter ):
    m_cRef(1),
    m_punkOuter(NULL),
    m_fValid(FALSE),
    m_pDcb(NULL),
    m_DeviceDataHandle(INVALID_HANDLE_VALUE),
	m_pszDeviceNameA(NULL),
    m_hSignalEvent(INVALID_HANDLE_VALUE),
    m_hThread(NULL),
    m_guidLastEvent(GUID_NULL),
	m_bEventMask(0),
	m_bEventsInitialized(false),
	m_hMutex(NULL),
    m_dwDebounceTime(NS_DEFAULT_DEBOUNCE),
	m_bPolling(false),
	m_bRemoteWakeup(false),
	m_bWHQLTest(false),
	m_dwKickStartTime(NS_DEFAULT_KICKSTART),
	m_hKickStartMutex(NULL)
{
	NSC_TRACE("In UsdNSCDevice::UsdNSCDevice");

	for (int index = 0; index < NS_MAX_OPEN_APPS ; index++)
		m_hPollingMutex[index] = NULL;
    //
    // See if we are aggregated. If we are ( which will be almost always the case )
    // save pointer to controlling Unknown , so subsequent calls will be delegated
    // If not, set the same pointer to "this" .
    // N.b. cast below is important in order to point to right virtual table
    //
    if (punkOuter) 
	{
        m_punkOuter = punkOuter;
    }
    else 
	{
        m_punkOuter = reinterpret_cast<IUnknown*>
                      (static_cast<INonDelegatingUnknown*>
                      (this));
    }

	m_hShutdownEvent =  CreateEvent( NULL,   // Attributes
                                   TRUE,     // Manual reset
                                   FALSE,    // Initial state - not set
                                   NULL );   // Anonymous

    if (m_hShutdownEvent && (INVALID_HANDLE_VALUE !=m_hShutdownEvent)) 
	{
		// We are now properly initialized and ready for action
        m_fValid = TRUE;
    }
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::~UsdNSCDevice is the destructor for this object class

PARAMETERS: 

VOID

RETURN VALUE:
none

NOTES:

This method cleans up after this object. The main actions performed here are 
1) shutting down the event thread in case it is active 
2) Closing the handle to the actual device 
3) Destroying the mutex used for synchronization

*******************************************************************************/
UsdNSCDevice::~UsdNSCDevice( VOID )
{
	NSC_TRACE("In UsdNSCDevice::~UsdNSCDevice");

    // Kill the notification thread if it exists by 
	// making sure that any notification handles are 
	// NULLed out.
    if (m_hSignalEvent && (m_hSignalEvent != INVALID_HANDLE_VALUE))
	{
		SetNotificationHandle(NULL);
	}

    if (m_hShutdownEvent && (m_hShutdownEvent!=INVALID_HANDLE_VALUE)) 
	{
        CloseHandle(m_hShutdownEvent);
    }

	CloseDevice();

    if (m_pszDeviceNameA) 
	{
        delete [] m_pszDeviceNameA;
        m_pszDeviceNameA = NULL;
    }

	// Close the mutex object 
	if (m_hMutex != NULL)
	{
		ReleaseMutex(m_hMutex);
	    CloseHandle(m_hMutex);
	}

	for (int index = 0; index < NS_MAX_OPEN_APPS; index++)
	{
		if (m_hPollingMutex[index] != NULL)
		{
			ReleaseMutex(m_hPollingMutex[index]);
			CloseHandle(m_hPollingMutex[index]);
		}
	}
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::GetCapabilities is called by STI to determine the capabilities
supported by this device 

PARAMETERS: 

PSTI_USD_CAPS pUsdCaps - This structure has to be filled in with data 
indicting the capabilitoes of the device

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT 

NOTES:
    The only capability which we want the STI to know about at this 
    point is the fact that we don't need polling and can generate our 
    own events

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::GetCapabilities( PSTI_USD_CAPS pUsdCaps )
{
	NSC_TRACE("In UsdNSCDevice::GetCapabilities");

	HRESULT hres = STI_OK;

    ZeroMemory(pUsdCaps,sizeof(*pUsdCaps));

    pUsdCaps->dwVersion = STI_VERSION;

    // We support device notifications without polling
    pUsdCaps->dwGenericCaps = STI_USD_GENCAP_NATIVE_PUSHSUPPORT;

    return hres;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::GetStatus is called by the STI to determine if the device that we 
    represent is still online

PARAMETERS: 

PSTI_DEVICE_STATUS pDevStatus

RETURN VALUE:
STDMETHODIMP - This is a macro that indicates that we return HRESULT 

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::GetStatus( PSTI_DEVICE_STATUS pDevStatus )
{
	WriteToLog(	STI_TRACE_INFORMATION,
				L"In UsdNSCDevice::GetStatus - Online Status : %d, Event Status : %d", 
				pDevStatus->StatusMask & STI_DEVSTATUS_ONLINE_STATE,
				pDevStatus->StatusMask & STI_DEVSTATUS_EVENTS_STATE);

	HRESULT hres = STI_OK;

    //
    // If we are asked, verify whether device is online
    //
    pDevStatus->dwOnlineState = 0L;
    if( pDevStatus->StatusMask & STI_DEVSTATUS_ONLINE_STATE )  
	{
		hres = GetDeviceStatus(pDevStatus);
	}

    //
    // If we are asked, verify state of event
    //
    pDevStatus->dwEventHandlingState = 0L;
    if( pDevStatus->StatusMask & STI_DEVSTATUS_EVENTS_STATE ) 
	{
		if(m_guidLastEvent != GUID_NULL)
		{
			pDevStatus->dwEventHandlingState |= STI_EVENTHANDLING_PENDING;
		}
		
		if (m_hSignalEvent && (m_hSignalEvent!= INVALID_HANDLE_VALUE))
		{
			pDevStatus->dwEventHandlingState |= STI_EVENTHANDLING_ENABLED;
		}

    }

    return hres;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::DeviceReset

PARAMETERS: 

VOID

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT 

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::DeviceReset( VOID )
{
	NSC_TRACE("In UsdNSCDevice::DeviceReset");

	HRESULT hres = STIERR_NOT_INITIALIZED;

    // Reset the current active device - this can be 
	// accomplished by a DeviceIOControl call with an 
	// IOCTL of RESET_PIPE and a value of all pipes
	// Use the escape command to handle all the niceties
	// of the overlapped IO on this call
    if (INVALID_HANDLE_VALUE != m_DeviceDataHandle) 
	{
		
		DWORD		cbBytesWritten;
		PIPE_TYPE	ePipes = ALL_PIPE;
		hres = Escape(	IOCTL_RESET_PIPE,
						&ePipes,
						sizeof(ePipes),
						NULL,
						0,
						&cbBytesWritten);
    }
	return hres;
}

/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::Diagnostic

PARAMETERS: 

LPDIAG pBuffer

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT 

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::Diagnostic( LPDIAG pBuffer )
{
	NSC_TRACE("In UsdNSCDevice::Diagnostic");

    WriteToLog(	STI_TRACE_INFORMATION,
						L"%s : %s : 0x%X", // Message, Error, Code
						L"UsdNSCDevice::Diagnostic",
						L"testing",
						0) ;

	HRESULT hres = STI_OK;

    // We will try to get the maximum packet size for the 
	// USB pipes. This must obviously be greater than zero
	// If there is an error during this operation or the 
	// size is not greater than zero then the device is offline
    if (INVALID_HANDLE_VALUE != m_DeviceDataHandle) 
	{
		// Get the maximum packet size - this has to be 
		// greater than zero for all the pipes 
		CHANNEL_INFO	channelInfo = {0,0,0};
		DWORD			cbBytesWritten(0);

		hres = Escape(	IOCTL_GET_CHANNEL_ALIGN_RQST,
						NULL,
						0,
						&channelInfo,
						sizeof(channelInfo),
						&cbBytesWritten);

		if (FAILED(hres) ||  !channelInfo.EventChannelSize)
		{
			hres = STIERR_GENERIC;
		    m_dwLastOperationError = ::GetLastError();
			// Fill in the error info using the GetLastErrorInfo call
			GetLastErrorInfo(&pBuffer->sErrorInfo);
		}
		else
		{
		    m_dwLastOperationError = 0;
			GetLastErrorInfo(&pBuffer->sErrorInfo);
		}
    }
	else
	{
		hres = STIERR_NOT_INITIALIZED;		
	}

    return hres;
}


/*******************************************************************************
DESCRIPTION:

SetNotificationHandle

PARAMETERS: 

HANDLE hEvent

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT 

NOTES:
// SYNCHRONIZED

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::SetNotificationHandle( HANDLE hEvent )
{
	WriteToLog(	STI_TRACE_INFORMATION,
				L"In UsdNSCDevice::SetNotificationHandle with handle Ox%X", 
				hEvent);

	HRESULT hres = STI_OK;

    TAKE_CRIT_SECT t(m_cs);

    if (hEvent && (hEvent !=INVALID_HANDLE_VALUE)) 
	{
        m_hSignalEvent = hEvent;

        if (m_DeviceDataHandle != INVALID_HANDLE_VALUE) 
		{
            m_guidLastEvent = GUID_NULL;

            if (!m_hThread) 
			{
			    DWORD   dwThread;
				m_hThread = ::CreateThread(NULL,
                                       2*1024,
                                       (LPTHREAD_START_ROUTINE)InterruptThreadFunc,
                                       (LPVOID)this,
                                       0,
                                       &dwThread);

				WriteToLog(	STI_TRACE_INFORMATION,
							L"%s : %s : 0x%X", // Message, Error, Code
							L"UsdNSCDevice::Enabling notification monitoring",
							L"Created Thread (handle) ",
							m_hThread) ;
            }
        }
        else 
		{
            // If we have not been initialized (like when we are 
			// opened in status mode) we will start listening
			// when we are actually opened in data mode so there is
			// no need to return an error code at this point
			WriteToLog(	STI_TRACE_INFORMATION,
						L"%s : %s : 0x%X", // Message, Error, Code
						L"UsdNSCDevice::Device not initialized",
						L"Deferring notification monitoring",
						0) ;
        }
    }
    else
	{
	    //
        // Disable hardware notifications
        //
        m_hSignalEvent = INVALID_HANDLE_VALUE;

		WriteToLog(	STI_TRACE_INFORMATION,
					L"%s : %s : 0x%X", // Message, Error, Code
					L"UsdNSCDevice::Disabling notification monitoring",
					L"Closing Thread (handle)",
					m_hThread) ;

        if ( m_hThread ) 
		{
			::SetEvent(m_hShutdownEvent);
            WaitForSingleObject(m_hThread, NS_THREADWAIT_TIMEOUT);
            CloseHandle(m_hThread);
            m_hThread = NULL;
            m_guidLastEvent = GUID_NULL;
        }
	}

    return hres;
}



/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::GetNotificationData

PARAMETERS: 

LPSTINOTIFY pBuffer

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT  

NOTES:
// SYNCHRONIZED

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::GetNotificationData( LPSTINOTIFY pBuffer )
{
	NSC_TRACE("In UsdNSCDevice::GetNotificationData");

    HRESULT hres = STI_OK;
	
    TAKE_CRIT_SECT t(m_cs);

    //
    // If we have notification ready - return it's guid
    //
    if (!IsEqualIID(m_guidLastEvent,GUID_NULL)) 
	{
        pBuffer->guidNotificationCode  = m_guidLastEvent;
        m_guidLastEvent = GUID_NULL;
        pBuffer->dwSize = sizeof(STINOTIFY);
    }
    else 
	{
        hres = STIERR_NOEVENTS;
    }

	return hres ;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::Escape

PARAMETERS: 

 STI_RAW_CONTROL_CODE    EscapeFunction
LPVOID                  pInData
DWORD                   cbInDataSize
LPVOID                  pOutData
DWORD                   cbOutDataSize
LPDWORD                 pcbActualData

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT  

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::Escape( STI_RAW_CONTROL_CODE    EscapeFunction,
                                    LPVOID                  pInData,
                                    DWORD                   cbInDataSize,
                                    LPVOID                  pOutData,
                                    DWORD                   cbOutDataSize,
                                    LPDWORD                 pcbActualData )
{
//	NSC_TRACE("In UsdNSCDevice::Escape");

	if (INVALID_HANDLE_VALUE == m_DeviceDataHandle)
    {
        return STIERR_NOT_INITIALIZED;
    }

    HRESULT hres = STI_OK;
    m_dwLastOperationError = NOERROR;

	m_bWHQLTest=false; //for tracking WHQL test
    switch (EscapeFunction)
    {
		case 0:
				m_bWHQLTest=true; //for tracking WHQL test
				break; // for WHQL test
        case IOCTL_ENABLE_REMOTEWAKEUP:
                hres = SetRemoteWakeup(true);
                break;
        case IOCTL_DISABLE_REMOTEWAKEUP:
                hres = SetRemoteWakeup(false);
                break;
        default :
        {
			// The escape command will be translated directly into an 
            // overlapped DeviceIOControl call
            OVERLAPPED	stOverlapped;
            ZeroMemory(&stOverlapped, sizeof(OVERLAPPED));
            stOverlapped.hEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);
			BOOL fRet = ::DeviceIoControl(	m_DeviceDataHandle,
							            EscapeFunction,
							            pInData,
							            cbInDataSize,                    
							            pOutData,
							            cbOutDataSize,
							            pcbActualData,
							            &stOverlapped );
			
            if (!fRet) 
            {
				hres = GetNscOverlappedResult(  &stOverlapped,  
								                NS_WRITE_CMD_TIMEOUT, 
								                pcbActualData);
            }
			CloseHandle(stOverlapped.hEvent);

        }
    } 
	return hres;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::GetLastError

PARAMETERS: 

LPDWORD pdwLastDeviceError

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT  

NOTES:
// SYNCHRONIZED

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::GetLastError( LPDWORD pdwLastDeviceError )
{
	NSC_TRACE("In UsdNSCDevice::GetLastError");

    HRESULT hres = STI_OK;

    TAKE_CRIT_SECT t(m_cs);

    if ( IsBadWritePtr( pdwLastDeviceError,4 ))
    {
        hres = STIERR_INVALID_PARAM;
    }
    else
    {
        *pdwLastDeviceError = m_dwLastOperationError;
    }

    return hres;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::GetLastErrorInfo

PARAMETERS: 

STI_ERROR_INFO *pLastErrorInfo

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT  

NOTES:
// SYNCHRONIZED

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::GetLastErrorInfo(STI_ERROR_INFO *pLastErrorInfo)
{
	NSC_TRACE("In UsdNSCDevice::GetLastErrorInfo");

    HRESULT hres = STI_OK;

    TAKE_CRIT_SECT t(m_cs);

    if ( IsBadWritePtr( pLastErrorInfo,4 ))
    {
        hres = STIERR_INVALID_PARAM;
    }
    else
    {
        pLastErrorInfo->dwGenericError	= m_dwLastOperationError;
        pLastErrorInfo->dwVendorError	= 0;

        // Get the message string from the operating system 
		// for the error code
		::FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM,
						NULL,
						m_dwLastOperationError,
						MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
						pLastErrorInfo->szExtendedErrorText,
						NS_MAX_ERRORSTRING,
						NULL );

		// pLastErrorInfo->szExtendedErrorText[0] = L'\0';
    }

    return hres;
}



/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::LockDevice

PARAMETERS: 

VOID

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT  

NOTES:
// SYNCHRONIZED

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::LockDevice( VOID )
{
	//NSC_TRACE("In UsdNSCDevice::LockDevice");

	// The LockDevice function may be called from multiple processes
	// at the same tiem to gain access to the scanner so it is necessary
	// to use a synchronization mechanism that works across process
	// so we use a named Mutex here to make sure that we can 
	// synchronize cross process access

    HRESULT hres = STI_OK;
	
	if (m_hMutex != NULL)
	{

		// Try to acquire the mutex within the timeout period
		DWORD dwRetCode = ::WaitForSingleObject(m_hMutex, NS_MAX_LOCKTIMEOUT);
		switch (dwRetCode)
		{
			case WAIT_ABANDONED:
				//deliberate fallthrough
			case WAIT_OBJECT_0:
				break;
			case WAIT_TIMEOUT:
				m_dwLastOperationError = hres = STIERR_DEVICE_LOCKED;
				break;
			case WAIT_FAILED:
				m_dwLastOperationError = ::GetLastError();
				hres = HRESULT_FROM_WIN32(m_dwLastOperationError);
				break;
			default:
				hres = STIERR_GENERIC;
		}
	}
	else
	{	
		hres = STIERR_NOT_INITIALIZED;
	}

    return hres;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::UnLockDevice

PARAMETERS: 

VOID

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT  

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::UnLockDevice( VOID )
{
	//NSC_TRACE("In UsdNSCDevice::UnLockDevice");

    HRESULT hres = STI_OK;
	
	if (m_hMutex != NULL)
	{
		hres = ::ReleaseMutex(m_hMutex) ? STI_OK : STIERR_GENERIC;
	}
	else
	{
		hres = STIERR_NOT_INITIALIZED;
	}

	return hres;
}

	BYTE bWrite[5];//TEST!!
	DWORD dwWriteLen;//TEST!!

/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::RawReadData

PARAMETERS: 

 LPVOID lpBuffer
LPDWORD lpdwNumberOfBytes
LPOVERLAPPED lpOverlapped

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT 

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::RawReadData( LPVOID lpBuffer, 
                                        LPDWORD lpdwNumberOfBytes,
                                        LPOVERLAPPED lpOverlapped )
{
	//NSC_TRACE("In UsdNSCDevice::RawReadData");

	HRESULT hres = STI_OK;

	if (m_bWHQLTest) return WHQLRawWriteReadTest();	

	DWORD Logct = *lpdwNumberOfBytes; //TEST!!!

    if (INVALID_HANDLE_VALUE != m_DeviceDataHandle)
    {
		//mutex should always be available due to locks
		if (::WaitForSingleObject(m_hKickStartMutex,m_dwKickStartTime) == WAIT_TIMEOUT)
		{
			hres = STIERR_GENERIC;
		}
		else
		{
			BOOL    fRet = FALSE;

			// Used only when the overlapped structure passed in is not valid
			OVERLAPPED	stOverlapped;
			bool		bPrivateEvent(false);

			if (lpOverlapped == NULL)
			{
				lpOverlapped = &stOverlapped;
				ZeroMemory(lpOverlapped, sizeof(OVERLAPPED));
				lpOverlapped->hEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);
				bPrivateEvent = true;
			}

			m_dwLastOperationError = NOERROR;

			fRet = ::ReadFile(m_DeviceDataHandle,
                    lpBuffer,
                    *lpdwNumberOfBytes,
                    lpdwNumberOfBytes,
                    lpOverlapped);

			if (!fRet) 
			{
				hres = GetNscOverlappedResult(  lpOverlapped,  
										  	NS_READ_DATA_TIMEOUT, 
											lpdwNumberOfBytes);
			}

			if (bPrivateEvent)
			{
				::CloseHandle(lpOverlapped->hEvent);
			}
			ReleaseMutex(m_hKickStartMutex);
		}
     } 
    else
    {
        hres = STIERR_NOT_INITIALIZED;
    }

	if (!SUCCEEDED(hres))
	{
		DeviceReset();
		LogError("RawWriteData",dwWriteLen,bWrite[0],bWrite[1],bWrite[2],
			bWrite[3],bWrite[4],0); //TEST!!!
		BYTE b[5] = {0,0,0,0,0};
		for (DWORD i = 0; i < 5; i++)
			b[i] = ((BYTE*)lpBuffer)[i];
		LogError("RawReadData",Logct,b[0],b[1],b[2],b[3],b[4],hres); //TEST!!!
	}

	return hres;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::RawWriteData

PARAMETERS: 

 LPVOID lpBuffer
DWORD dwNumberOfBytes
LPOVERLAPPED lpOverlapped

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT 

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::RawWriteData( LPVOID lpBuffer, 
                                         DWORD dwNumberOfBytes,
                                         LPOVERLAPPED lpOverlapped )
{
	//NSC_TRACE("In UsdNSCDevice::RawWriteData");

    HRESULT hres = STI_OK;

	if (m_bWHQLTest) return WHQLRawWriteReadTest();

	dwWriteLen = dwNumberOfBytes; //TEST!!!

	for (DWORD i = 0; i < 5; i++)
	{
		if (i < dwWriteLen) bWrite[i] = ((BYTE*)lpBuffer)[i]; //TEST!!!
		else bWrite[i] = 0;
	}

    if (INVALID_HANDLE_VALUE != m_DeviceDataHandle)
    {
		BOOL    fRet = FALSE;;
		DWORD   dwBytesReturned = 0;

		// USed only when the overlapped structure passed in is not valid
		OVERLAPPED	stOverlapped;
		bool		bPrivateEvent(false);

		if (lpOverlapped == NULL)
		{
			lpOverlapped = &stOverlapped;
			ZeroMemory(lpOverlapped, sizeof(OVERLAPPED));
			lpOverlapped->hEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);
			bPrivateEvent = true;
		}

        fRet = ::WriteFile(m_DeviceDataHandle,
                            lpBuffer,
                            dwNumberOfBytes,
                            &dwBytesReturned,
                            lpOverlapped);

        if (!fRet) 
		{
			hres = GetNscOverlappedResult(  lpOverlapped,  
										  	NS_WRITE_DATA_TIMEOUT, 
											&dwBytesReturned);
		}

		if (bPrivateEvent)
		{
			::CloseHandle(lpOverlapped->hEvent);
		}

    }
    else
    {
        hres = STIERR_NOT_INITIALIZED;
    }

	if (!SUCCEEDED(hres))
	{
		DeviceReset();
		LogError("RawWriteData",dwWriteLen,bWrite[0],bWrite[1],bWrite[2],
			bWrite[3],bWrite[4],hres); //TEST!!!
	}
	
	return hres;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::RawReadCommand

PARAMETERS: 

LPVOID lpBuffer
LPDWORD lpdwNumberOfBytes
LPOVERLAPPED lpOverlapped

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT 

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::RawReadCommand(LPVOID lpBuffer, 
                                            LPDWORD lpdwNumberOfBytes,
                                            LPOVERLAPPED lpOverlapped )
{   
	HRESULT hres = STI_OK;

	if (m_bWHQLTest) return WHQLRawWriteReadTest();

	//NSC_TRACE("In UsdNSCDevice::RawReadCommand");

    if (INVALID_HANDLE_VALUE != m_DeviceDataHandle)
    {
		BOOL    fRet = FALSE;

		IO_BLOCK* pIoBlock = (IO_BLOCK*)lpBuffer;

		// USed only when the overlapped structure passed in is not valid
		OVERLAPPED	stOverlapped;
		bool		bPrivateEvent(false);

		if (lpOverlapped == NULL)
		{
			lpOverlapped = &stOverlapped;
			ZeroMemory(lpOverlapped, sizeof(OVERLAPPED));
			lpOverlapped->hEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);
			bPrivateEvent = true;
		}

		m_dwLastOperationError = NOERROR;

        fRet = ::DeviceIoControl(m_DeviceDataHandle,
								(DWORD) IOCTL_READ_REGISTERS,
								(PVOID) pIoBlock,
								(DWORD) sizeof(IO_BLOCK),                    
								(PVOID) pIoBlock->pbyData,
								(DWORD) pIoBlock->uLength*sizeof(BYTE),
								lpdwNumberOfBytes,
								lpOverlapped);
        if (!fRet) 
		{
			hres = GetNscOverlappedResult(  lpOverlapped,  
										  	NS_READ_CMD_TIMEOUT, 
											lpdwNumberOfBytes);
		}

		if (bPrivateEvent)
		{
			::CloseHandle(lpOverlapped->hEvent);
		}
    } 
    else
    {
        hres = STIERR_NOT_INITIALIZED;
    }

	if (!SUCCEEDED(hres))
	{
		DeviceReset();
		LogError("RawReadCommand",*lpdwNumberOfBytes,0,0,0,0,0,hres);
	}

    return hres;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::RawWriteCommand

PARAMETERS: 

 LPVOID lpBuffer
DWORD dwNumberOfBytes
LPOVERLAPPED lpOverlapped

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT 

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::RawWriteCommand( LPVOID lpBuffer, 
                                            DWORD dwNumberOfBytes,
                                            LPOVERLAPPED lpOverlapped )
{
	//NSC_TRACE("In UsdNSCDevice::RawWriteCommand");

	HRESULT hres = STI_OK;

	if (m_bWHQLTest) return WHQLRawWriteReadTest();

    if (INVALID_HANDLE_VALUE != m_DeviceDataHandle)
    {
		BOOL    fRet = FALSE;;
		DWORD   dwBytesReturned = 0;

		// so we will verify this with the actual hardware
		IO_BLOCK* pIoBlock = (IO_BLOCK*)lpBuffer;

		// USed only when the overlapped structure passed in is not valid
		OVERLAPPED	stOverlapped;
		bool		bPrivateEvent(false);

		if (lpOverlapped == NULL)
		{
			lpOverlapped = &stOverlapped;
			ZeroMemory(lpOverlapped, sizeof(OVERLAPPED));
			lpOverlapped->hEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);
			bPrivateEvent = true;
		}

		fRet = ::DeviceIoControl(m_DeviceDataHandle,
								(DWORD) IOCTL_WRITE_REGISTERS,
								(PVOID) pIoBlock,
								(DWORD) sizeof(IO_BLOCK),                    
								NULL,
								0, 
								&dwBytesReturned,
								lpOverlapped);

        if (!fRet) 
		{
			hres = GetNscOverlappedResult(  lpOverlapped,  
										  	NS_WRITE_CMD_TIMEOUT, 
											&dwBytesReturned);
		}

		if (bPrivateEvent)
		{
			::CloseHandle(lpOverlapped->hEvent);
		}
    }
    else
    {
        hres = STIERR_NOT_INITIALIZED;
    }

	if (!SUCCEEDED(hres))
	{
		DeviceReset();
		LogError("RawWriteCommand",dwNumberOfBytes,0,0,0,0,0,hres);
	}

    return hres;
}



/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::Initialize

PARAMETERS: 

PSTIDEVICECONTROL pDcb
DWORD dwStiVersion
HKEY hParametersKey

RETURN VALUE:
STDMETHODIMP  - This is a macro that indicates that we return HRESULT 

NOTES:

*******************************************************************************/
STDMETHODIMP UsdNSCDevice::Initialize(PSTIDEVICECONTROL pDcb, 
                                        DWORD dwStiVersion,
                                        HKEY hParametersKey )
{
	NSC_TRACE("In UsdNSCDevice::Initialize");

	HRESULT hres = STI_OK;

    if (!pDcb) 
	{
        return STIERR_INVALID_PARAM;
    }
	// Before creating the mutxt it is necessary that we create 
	// a NULL DACL for the security descriptor if this code is running
	// on W2k - otherwise we will use a NULL pointer for the 
	// security attributes

    SECURITY_ATTRIBUTES     sa;
	{
		char bSecurity[SECURITY_DESCRIPTOR_MIN_LENGTH];
		memset(bSecurity,0,SECURITY_DESCRIPTOR_MIN_LENGTH);

		PSECURITY_DESCRIPTOR pSD = (PSECURITY_DESCRIPTOR) bSecurity;

		if (!InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION))
		{
			WriteToLog(	STI_TRACE_ERROR,
						L"%s : %s : 0x%X", // Message, Error, Code
						L"UsdNSCDevice::Initializing the SD failed",
						L"Device not Initialized in Thread (id)",
						::GetCurrentThreadId()) ;
			return STIERR_GENERIC;
		}

		// Add a NULL DACL to the security descriptor..

		if (!SetSecurityDescriptorDacl(pSD, TRUE, (PACL) NULL, FALSE))
		{
			WriteToLog(	STI_TRACE_ERROR,
						L"%s : %s : 0x%X", // Message, Error, Code
						L"UsdNSCDevice::Setting the NULL DACL failed",
						L"Device not Initialized in Thread (id)",
						::GetCurrentThreadId()) ;
		  return STIERR_GENERIC;
		}

		sa.nLength = sizeof(sa);
		sa.lpSecurityDescriptor = pSD;
		sa.bInheritHandle = TRUE;
	}

	// Create the mutex object that will be used for synchronizing access
	// to the scanner
	m_hMutex = CreateMutex(&sa, FALSE, NSCUSD_MUTEX);

	if (m_hMutex == NULL)
	{
	    m_dwLastOperationError = ::GetLastError();
		hres = HRESULT_FROM_WIN32(m_dwLastOperationError);
		return hres;
	}

    // Increment the refcount for the device control object that has 
	// been passed in to us.
    m_pDcb = pDcb;
    m_pDcb->AddRef();

	hres = OpenDevice(GENERIC_READ | GENERIC_WRITE, 
                        FILE_SHARE_READ | FILE_SHARE_WRITE, sa);

    if (SUCCEEDED(hres))
	{
        // Get the currenlty selected event settings (which bits 
        // to recognize ) and other parameters from the registry
        LoadDeviceData(hParametersKey);

        // If the scanner is in reset mode bring it out of reset mode
        hres = ResetModeCheckAndClear();    
        
	}

	return hres;
}

///////////////////////////////////////////////////////////////
// Helper Functions 
///////////////////////////////////////////////////////////////

/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::OpenDevice

PARAMETERS: 

DWORD dwAccess
DWORD dwShare
SECURITY_ATTRIBUTES& sa

RETURN VALUE:
HRESULT 

NOTES:

*******************************************************************************/
HRESULT UsdNSCDevice::OpenDevice(DWORD dwAccess, DWORD dwShare,
								    SECURITY_ATTRIBUTES& sa)
{
	NSC_TRACE("In UsdNSCDevice::OpenDevice");
	m_bWHQLTest=false; //for tracking WHQL test

    if (!m_pDcb) 
	{
        return STIERR_NOT_INITIALIZED;
    }

	HRESULT hres = STI_OK;
    UINT    uiNameLen = 0;
    WCHAR   szDeviceNameW[MAX_PATH];

    *szDeviceNameW = L'\0';

    // Get the name of the device port we need to open
    hres = m_pDcb->GetMyDevicePortName(szDeviceNameW,sizeof(szDeviceNameW)/sizeof(WCHAR));
    if (!SUCCEEDED(hres) || !*szDeviceNameW) {
        return hres;
    }

    // Convert name to SBCS for use in WIn95
    uiNameLen = WideCharToMultiByte(CP_ACP, 0, szDeviceNameW, -1, NULL, NULL, 0, 0);
    if (!uiNameLen) {
        return STIERR_INVALID_PARAM;
    }

    // If name already exists delete it 
	delete m_pszDeviceNameA;
	m_pszDeviceNameA = new CHAR[uiNameLen+1];
    if (!m_pszDeviceNameA) {
        return STIERR_INVALID_PARAM;
    }

    WideCharToMultiByte(CP_ACP, 0, szDeviceNameW, -1, m_pszDeviceNameA, uiNameLen, 0, 0);

    // Open device using the CreateFile API Call 
	m_DeviceDataHandle = CreateFileA( m_pszDeviceNameA,
                                 dwAccess,							// Access mask
                                 dwShare,							// Share mode
                                 NULL,                              // SA
                                 OPEN_EXISTING,                     // Create disposition
                                 FILE_FLAG_OVERLAPPED,              // Attributes
                                 NULL );

    m_dwLastOperationError = ::GetLastError();

    hres = (m_DeviceDataHandle != INVALID_HANDLE_VALUE) ?
                S_OK : MAKE_HRESULT(SEVERITY_ERROR,FACILITY_WIN32,m_dwLastOperationError);

	if (SUCCEEDED(hres))
	{
		//Create the bulk in mutex
		m_hKickStartMutex = CreateMutex(&sa, FALSE, NSCUSD_KICKSTART_MUTEX);

		if (m_hKickStartMutex == NULL)
		{
			m_dwLastOperationError = ::GetLastError();
			hres = HRESULT_FROM_WIN32(m_dwLastOperationError);
			return hres;
		}

		//Create unique polling mutex name based on index and port
		CHAR device_name[MAX_PATH];
		strcpy(device_name,m_pszDeviceNameA);
		for (UINT i = 0; device_name[i]; i++)
			if (device_name[i] == '\\') device_name[i] = '_';

		for (int index = 0; index < NS_MAX_OPEN_APPS; index++)
		{
			CHAR mutex_name[MAX_PATH];
			sprintf(mutex_name,"%s%s%d",
				NSCUSD_POLLING_MUTEX,device_name,index);

			::SetLastError(0);
			m_hPollingMutex[index] = CreateMutex(&sa, FALSE, mutex_name);

			if (m_hPollingMutex[index] == NULL)
			{
				m_dwLastOperationError = ::GetLastError();
				hres = HRESULT_FROM_WIN32(m_dwLastOperationError);
				return hres;
			}

			//if mutex already exists, block polling
			if (::GetLastError() == ERROR_ALREADY_EXISTS)
			{
				if (::WaitForSingleObject(m_hPollingMutex[index], 0)
					== WAIT_OBJECT_0) break; //have ownership, done
			}
		}
	}

    return hres;

}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::CloseDevice

PARAMETERS: 


RETURN VALUE:
HRESULT 

NOTES:

*******************************************************************************/
HRESULT UsdNSCDevice::CloseDevice()
{
	NSC_TRACE("In UsdNSCDevice::CloseDevice");
	m_bWHQLTest=false; //for tracking WHQL test

    if( INVALID_HANDLE_VALUE != m_DeviceDataHandle ) 
	{
        CloseHandle( m_DeviceDataHandle );
		m_DeviceDataHandle  = INVALID_HANDLE_VALUE;
    }

	return STI_OK;
}



/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::GetNscOverlappedResult

PARAMETERS: 

LPOVERLAPPED lpOverlapped
DWORD       dwTimeout
LPDWORD     lpBytesRead

RETURN VALUE:


NOTES:

*******************************************************************************/
HRESULT	UsdNSCDevice::GetNscOverlappedResult(LPOVERLAPPED lpOverlapped, 
                                            DWORD       dwTimeout, 
                                            LPDWORD     lpBytesRead)
{
	HRESULT hres = STI_OK;

    m_dwLastOperationError = ::GetLastError();

	if (m_dwLastOperationError == ERROR_IO_PENDING)
	{
		// Wait for the overlapped operation to complete
		if (::WaitForSingleObject(lpOverlapped->hEvent, dwTimeout)
				== WAIT_TIMEOUT)
		{
			// Timeout error 
			hres = STIERR_GENERIC;
		}
		else
		{
			// Get the actual number of bytes read using the 
			// GetOverlappedResult API
			DWORD dwBytesRead = 0;
			if (::GetOverlappedResult(m_DeviceDataHandle,
									lpOverlapped,
									&dwBytesRead,
									FALSE))
			{
				ResetEvent(lpOverlapped->hEvent);
				*lpBytesRead = dwBytesRead;
			}
			else
			{
			    m_dwLastOperationError = ::GetLastError();
				hres = HRESULT_FROM_WIN32(m_dwLastOperationError);
			}
		}
	}
	else
	{
		hres = HRESULT_FROM_WIN32(m_dwLastOperationError);
	}

	return hres;
}


/*******************************************************************************
DESCRIPTION:

InterruptThreadFunc

PARAMETERS: 

LPVOID  lpParameter

RETURN VALUE:
VOID 

NOTES:

*******************************************************************************/
VOID InterruptThreadFunc(LPVOID  lpParameter)
{
//	NSC_TRACE("In InterruptThreadFunc - Interrupt thread started");

	::CoInitializeEx(NULL, COINIT_MULTITHREADED);

	UsdNSCDevice   *pThisDevice = (UsdNSCDevice*)lpParameter;

    pThisDevice->ListenForInterrupts();

	::CoUninitialize();
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::ListenForInterrupts

PARAMETERS: 


RETURN VALUE:
void 

NOTES:

*******************************************************************************/
void UsdNSCDevice::ListenForInterrupts()
{
	NSC_TRACE("In UsdNSCDevice::ListenForInterrupts");

	WriteToLog(	STI_TRACE_INFORMATION,
				L"%s : %s : 0x%X", // Message, Error, Code
				L"UsdNSCDevice::Listening for Interrupts Started",
				L"In Thread (id)",
                ::GetCurrentThreadId()) ;

	if (!m_DeviceDataHandle || (m_DeviceDataHandle == INVALID_HANDLE_VALUE))
	{
		WriteToLog(	STI_TRACE_ERROR,
					L"%s : %s : 0x%X", // Message, Error, Code
					L"UsdNSCDevice::Listening for Interrupts Aborted",
					L"Device not Initialized in Thread (id)",
                    ::GetCurrentThreadId()) ;
		return;
	}

    // We dont have to create a overlapped event if we are only polling
    // the device but we do it anyway just to make the code simpler to read
    // and modify . The alternatice would be to bracket all this code and
    // the code that closes the overlapped handle with #ifdef commands
    // and that makes the code really messy
    OVERLAPPED	 stOverlapped;
	LPOVERLAPPED lpOverlapped = &stOverlapped;
	ZeroMemory(lpOverlapped, sizeof(OVERLAPPED));
	lpOverlapped->hEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);

	BOOL	fLooping = TRUE;
    DWORD   dwEventCode;

	while(fLooping)
	{

#ifdef NS_POLL_FOR_EVENTS
        dwEventCode = PollForDriverEvent();
#else
        dwEventCode = WaitForDriverEvent(lpOverlapped);
#endif
		switch(dwEventCode) 
		{
			case NS_DEVICE_INTERRUPT: 

				// Find out what kind of event has occured and then 
                // generate a event for the STI if necessary
				// an event has occured 
				GenerateScannerEvent();

				// We will be back through this loop to wait on device
				// events so it is critical that we reset the event object
				// as this is a manual reset event
				::ResetEvent(lpOverlapped->hEvent);
				break;
			
			case NS_SHUTDOWN_REQUEST:
				// We have to shutdown the interrupt thread
				WriteToLog(	STI_TRACE_INFORMATION,
							L"%s : %s : 0x%X", // Message, Error, Code
							L"UsdNSCDevice::Listening for Interrupts Aborted",
							L"Shutdown Request",
							0) ;

				::ResetEvent(m_hShutdownEvent);
				fLooping  = false;
				break;

			case NS_WAIT_ERROR:
				WriteToLog(	STI_TRACE_ERROR,
							L"%s : %s : 0x%X", // Message, Error, Code
							L"UsdNSCDevice::Listening for Interrupts Aborted",
							L"Wait Error ",
							::GetLastError()) ;

                fLooping  = false;
                break;
            case NS_IOCONTROL_ERROR:
                WriteToLog(	STI_TRACE_ERROR, 
						    L"%s : %s : 0x%X", // Message, Error, Code
						    L"UsdNSCDevice::Listening for Interrupts Aborted",
						    L"DeviceIoControl() Error",
						    ::GetLastError()) ;

			    fLooping = FALSE;
                break;
		}
	}

	::CloseHandle(lpOverlapped->hEvent);
}


/*******************************************************************************
DESCRIPTION:

 UsdNSCDevice::WaitForDriverEvent

PARAMETERS: 

LPOVERLAPPED lpOverlapped

RETURN VALUE:
DWORD  

NOTES:

*******************************************************************************/
DWORD   UsdNSCDevice::WaitForDriverEvent(LPOVERLAPPED lpOverlapped)
{
	DWORD   dwBytesReturned;
	BYTE	pStatus[NS_MAX_INTERRUPT_PACKET];
    DWORD   dwReturnCode;
	BOOL    fRet = TRUE;

    fRet = ::DeviceIoControl(m_DeviceDataHandle,
							(DWORD) IOCTL_WAIT_ON_DEVICE_EVENT,
							NULL, 
							0, 
							pStatus,
							NS_MAX_INTERRUPT_PACKET, 
							&dwBytesReturned,
							lpOverlapped);

    // If the deviceIOControl call returned false and it was not
    // because of a pending IO then we have a real error. In this case 
    // the wait will be aborted with an error return otherwise we 
    // can process the results of the call
    if ( fRet || (::GetLastError() == ERROR_IO_PENDING))
    {
	    HANDLE	EventHandleArray[2] = {m_hShutdownEvent, lpOverlapped->hEvent};
	    int numEvents = sizeof(EventHandleArray) / sizeof(EventHandleArray[0]);

	    // Wait for an event to occur - one of the events could be a 
	    // request for this thread to shutdown because our minidriver
	    // is being unloaded - this is not exactly a device event but it 
	    // is handled here anyway as it is easy to do so
	    DWORD dwIndex = WaitForMultipleObjects(	numEvents, 
											    EventHandleArray, 
											    FALSE, 
											    INFINITE);


        switch (dwIndex)
        {
            case WAIT_OBJECT_0 + 1:
                // We have an interrupt, Trace the event 
				WriteToLog(	STI_TRACE_INFORMATION,
							L"%s : %s : 0x%X", // Message, Error, Code
							L"UsdNSCDevice::Status Register Changed Event",
							L"In Thread (id)",
							::GetCurrentThreadId()) ;

                dwReturnCode = NS_DEVICE_INTERRUPT;
                break;
            case WAIT_OBJECT_0:
                dwReturnCode = NS_SHUTDOWN_REQUEST;
                break;
            default:
                dwReturnCode = NS_WAIT_ERROR;
        }
    }
    else // actual error in the device io control call
    {
        dwReturnCode = NS_IOCONTROL_ERROR;
    }

    return dwReturnCode;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::PollForDriverEvent

PARAMETERS: 


RETURN VALUE:
DWORD 

NOTES:

*******************************************************************************/
DWORD UsdNSCDevice::PollForDriverEvent()
{
    // if some app has opened the driver then we will 
    // be stuck here till we either get a shutdown notice or
    // the app closes the driver
	bool mutex_wait = false;
	for (int index = 0; index < NS_MAX_OPEN_APPS; index++)
	{
		while (::WaitForSingleObject(m_hPollingMutex[index], 
				min(m_dwKickStartTime,NS_POLLING_INTERVAL))
					== WAIT_TIMEOUT)
		{
			// Check to make sure that we don't have a shutdown request
			if (WaitForSingleObject(m_hShutdownEvent, 0) == WAIT_OBJECT_0)
				return NS_SHUTDOWN_REQUEST;
			// Check command reg = reset
			mutex_wait = true;
			// If waiting on bulkin, do some command to kick start
			// any pause due to step counter termination
			if (::WaitForSingleObject(m_hKickStartMutex,m_dwKickStartTime) == WAIT_TIMEOUT)
			{
				DoKickStart(); 
			}
			else
			{
				ReleaseMutex(m_hKickStartMutex);
			}
		}
		ReleaseMutex(m_hPollingMutex[index]);
	}

	//flush status when exit mutex wait so don't get any events
	if (mutex_wait)
	{
		ResetModeCheckAndClear();
		BYTE bValue;
        ReadStatusRegister(bValue); //flush first after wait
	}

	Sleep (NS_POLLING_INTERVAL);
    
    // Check to make sure that we don't have a shutdown request
    if (WaitForSingleObject(m_hShutdownEvent, 0) == WAIT_OBJECT_0)
        return NS_SHUTDOWN_REQUEST;
    
    // Pretend we just had a device event
    return NS_DEVICE_INTERRUPT;

}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::WriteToLog

PARAMETERS: 

DWORD dwType
LPWSTR  lpszFormat
...

RETURN VALUE:
void 

NOTES:

*******************************************************************************/
void UsdNSCDevice::WriteToLog(DWORD dwType, LPWSTR  lpszFormat, ...)
{
	va_list args;
	va_start(args, lpszFormat);

	int nBuf;
	WCHAR szBuffer[512];
	nBuf = _vsnwprintf(szBuffer, 511, lpszFormat, args);

	va_end(args);

	if (nBuf < 0) 
		return;
	
	if (m_pDcb)
	{
		m_pDcb->WriteToErrorLog(dwType, szBuffer, NOERROR) ;	
	}

#ifdef _DEBUG
	// Also trace the message to the debugger if in debug mode
	// Convert name to SBCS as using OutputDebugStringW does not seem
	// to work
    UINT uiNameLen = WideCharToMultiByte(CP_ACP, 0, szBuffer, -1, NULL, NULL, 0, 0);
    if (uiNameLen) 
	{
		LPSTR lpszBuffer = new CHAR[uiNameLen+1];
	    if (lpszBuffer)
		{
		    WideCharToMultiByte(CP_ACP, 0, szBuffer, -1, lpszBuffer, uiNameLen, 0, 0);
			::OutputDebugString(lpszBuffer);
			::OutputDebugString("\n");
			delete lpszBuffer;
		}
	}

#endif    

	return;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::GetDeviceStatus

PARAMETERS: 

PSTI_DEVICE_STATUS pDevStatus

RETURN VALUE:
HRESULT 

NOTES:

*******************************************************************************/
HRESULT UsdNSCDevice::GetDeviceStatus(PSTI_DEVICE_STATUS pDevStatus)
{
	// The device status can be one of the following 
	// and these conditions are checked for in the 
	// given order 
	// STI_ONLINESTATE_POWER_SAVE (reg 7, bit 4)
	// STI_ONLINESTATE_BUSY (reg 7, bit 0-2 are NOT zero)
	// STI_ONLINESTATE_OPERATIONAL  (if none of the above)

	HRESULT hres = STI_OK;
	// Always lock the device before trying to access the scanner
	hres = LockDevice();

	// So the first order of business is to make sure that we 
	// can read register 7 and then based on the value we will
	// determine the actual status of the device
    if ( SUCCEEDED(hres) && (INVALID_HANDLE_VALUE != m_DeviceDataHandle)) 
	{

		BYTE pCommandBytes[COMMAND_BYTE_COUNT];
		BYTE bValue;

		// Write the command bytes for a register read 
		// without increment 
		pCommandBytes[0] = MODE_NOINC_READ;
		pCommandBytes[1] = LM_COMMAND_REGISTER;
		pCommandBytes[2] = 0; 
		pCommandBytes[3] = 1;

		hres = 	RawWriteData(&pCommandBytes, COMMAND_BYTE_COUNT, NULL);
		
		if (SUCCEEDED(hres))
		{
			// Then read the register in question
			DWORD cbBytes(1);
			hres = RawReadData(&bValue, &cbBytes, NULL);
		}

		if (SUCCEEDED(hres))
		{
			pDevStatus->dwOnlineState |= STI_ONLINESTATE_OPERATIONAL;
			if (bValue & 0x10)
				pDevStatus->dwOnlineState |= STI_ONLINESTATE_POWER_SAVE;
			else if (bValue & 0x07) 
				pDevStatus->dwOnlineState |= STI_ONLINESTATE_BUSY;
		}

		UnLockDevice();
    }

	return hres;
}



/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::GenerateScannerEvent

PARAMETERS: 


RETURN VALUE:
void 

NOTES:

*******************************************************************************/
void UsdNSCDevice::GenerateScannerEvent()
{
	// Make sure that somebody is actually interested in hearing 
	// about this event
	if ((m_hSignalEvent == INVALID_HANDLE_VALUE) || (!m_bEventsInitialized))
		return;

	// Find out if the scanner is online and is the scanner is in 
	// reset mode - if the scanner is in reset mode we will simply
	// bring it out of this mode and ignore the event
	HRESULT hres = S_OK;

	//do every 20 times for safety
	static int ct = 0;
	if (!ct--) 
	{
		ct = 20;	
		hres = ResetModeCheckAndClear();
	}

	BYTE bStatusRegister;
	if SUCCEEDED(hres)
	{
		hres = ReadStatusRegister(bStatusRegister);
	}

	// If there was a scanner IO error then in all probability scanner
	// is offline and this event should be ignored
	if SUCCEEDED(hres)
	{
		// Get the event GUID corresponding to the current state
		// of the register
		m_guidLastEvent = GetEventGUID(bStatusRegister);
		if (!IsEqualGUID(m_guidLastEvent,GUID_NULL))
		{
			// The event is forwarded onto the STI only if the 
			// status register bit is set and the event mask for that 
			// IO pin is turned on
			::SetEvent(m_hSignalEvent);

            // Once we successfully activate an event, we want to make 
            // sure that we debounce the input that is we dont want
            // any more false positives till the debounce interval 
            // is finished so we just let this thread sleep for the
            // debounce interval and then read and clear the status 
            // register
            ::Sleep(m_dwDebounceTime);
            
            // Now clear any bits that were set when we were asleep
            ReadStatusRegister(bStatusRegister);

		}
	}

}



/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::ReadStatusRegister

PARAMETERS: 

BYTE& bValue

RETURN VALUE:
HRESULT 

NOTES:

*******************************************************************************/
HRESULT UsdNSCDevice::ReadStatusRegister(BYTE& bValue)
{
	HRESULT hres = STI_OK;

	if (!m_bPolling)
	{
		bValue = 0;
		return hres;
	}

	hres = LockDevice();

	// Follow the standard procedure of writing command bytes
	// to the bulk out endpoint to read the io status register and then 
	// actually reading in the value from the bulk in endpoint
    if ( SUCCEEDED(hres) && (INVALID_HANDLE_VALUE != m_DeviceDataHandle)) 
	{
		BYTE	pCommandBytes[COMMAND_BYTE_COUNT];

		// Write the command bytes for a register read 
		// without increment 
		pCommandBytes[0] = MODE_NOINC_READ;
		pCommandBytes[1] = LM_IOSTATUS_REGISTER;
		pCommandBytes[2] = 0; 
		pCommandBytes[3] = 1;

		hres = 	RawWriteData(&pCommandBytes, COMMAND_BYTE_COUNT, NULL);
		
		if (SUCCEEDED(hres))
		{
			// Then read the register in question
			DWORD cbBytes(1);
			hres = RawReadData(&bValue, &cbBytes, NULL);
		// Write To Log
		WriteToLog(	STI_TRACE_INFORMATION,
					L"%s=0x%X", // Message, Error, Code
					L"Reg2",
					bValue) ;
		}
		// If the read failed bValue should be unchanged 

		UnLockDevice();
    }
	
	return hres;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::ResetModeCheckAndClear checks to see if the scanner is in 
the reset mode and if it is, attempts to bring it out of the reset mode

PARAMETERS: 
-

RETURN VALUE:
HRESULT - return STI_OK if the operation succeeded otherwise it returns
an error code

NOTES:
Please note  If the scanner is in reset mode this function will attempt will 
bring it out of the reset mode by clearing the reset bit *** BUT *** 
the return value does not indicate if the scanner was successfully brought
out of the reset mode , the return value simply indicates if the scanner
was in reset mode or not when this method was called.

*******************************************************************************/
HRESULT UsdNSCDevice::ResetModeCheckAndClear()
{
	HRESULT hres = STI_OK;

	if (!m_bPolling)
	{
		return hres;
	}

	hres = LockDevice();

	const int MAX_RETRY =  20;
	for (int retry = 0; retry < MAX_RETRY; retry++)
	{
	// Read the command register and check that the reset bit is not set 
	// If it is set clear it and return false to indicate that 
	// the bit has only now been cleared
    if ( SUCCEEDED(hres) && (INVALID_HANDLE_VALUE != m_DeviceDataHandle)) 
	{
		BYTE	pCommandBytes[COMMAND_BYTE_COUNT];
		BYTE	bValue;

		// Write the command bytes for a register read 
		// without increment 
		pCommandBytes[0] = MODE_NOINC_READ;
		pCommandBytes[1] = LM_COMMAND_REGISTER;
		pCommandBytes[2] = 0; 
		pCommandBytes[3] = 1;

		hres = 	RawWriteData(pCommandBytes, COMMAND_BYTE_COUNT, NULL);
		
		if (SUCCEEDED(hres))
		{
			// Then read the register in question
			DWORD cbBytes(1);
			hres = RawReadData(&bValue, &cbBytes, NULL);
		// Write To Log
		WriteToLog(	STI_TRACE_INFORMATION,
					L"%s=0x%X,0x%X", // Message, Error, Code
					L"Reg7,Retry",
					bValue,retry) ;
		}

		if (SUCCEEDED(hres))
		{
			if ( (bValue & NS_RESET_BIT) == NS_RESET_BIT)
			{
				hres = STIERR_GENERIC;

				BYTE	pWriteData[COMMAND_BYTE_COUNT+1];
				// Write the command bytes for a register read 
				// without increment 
				pWriteData[0] = MODE_NOINC_WRITE;
				pWriteData[1] = LM_COMMAND_REGISTER;
				pWriteData[2] = 0; 
				pWriteData[3] = 1;
				pWriteData[4] = 0; // <--- The data for the register

				// We will attempt to reset it but we really don't do 
				// anything if this fails
				hres = RawWriteData(pWriteData, COMMAND_BYTE_COUNT+1, NULL);
			}
			else break; //done
		}
    }
	} 

	UnLockDevice();

	if (SUCCEEDED(hres))
	{
		hres = SetRemoteWakeup(true);	//enable
	}
	
	return (hres);
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::GetEventGUID

PARAMETERS: 

BYTE bStatusRegister

RETURN VALUE:
GUID 

NOTES:

*******************************************************************************/
GUID UsdNSCDevice::GetEventGUID(BYTE bStatusRegister)
{
	// For each MISC IO bit in the status register we will 
	// check if there is a corresponding event registered
	// If an event is registered for a particular IO bit and
	// the bit is turned on then fire that event
	// if the status register bit corresponding to a event is
	// turned on and that event is currently registered with the 
	// STI then return the event code for that event otherwise 
	// return a NULL guid 	
	// NEW: only accept one bit set at once to avoid problem with
	// remote wakeup
	int savei,ct = 0;
	for (int i=0 ; i < NS_MAX_EVENTS; i++)
	{
		if ((bStatusRegister & NS_MISC_IO_BIT[i]) == NS_MISC_IO_BIT[i])
		{
			ct++;
			savei = i;
		}
	}
	if ((ct == 1) && 
		((m_bEventMask & NS_EVENT_BIT[savei]) == NS_EVENT_BIT[savei]))
			return m_bEventGuidArray[savei];
	return GUID_NULL;
}


/*******************************************************************************
DESCRIPTION:

  UsdNSCDevice::LoadDeviceData

PARAMETERS: 

HKEY     hParameters

RETURN VALUE:
void  

NOTES:

*******************************************************************************/
void    UsdNSCDevice::LoadDeviceData(HKEY     hParameters)
{
	HKEY hDeviceData;

	LONG nRes = RegOpenKey(	hParameters,
							"DeviceData",
							&hDeviceData );

	if (nRes != ERROR_SUCCESS)
	{
		// Write To Log
		WriteToLog(	STI_TRACE_ERROR,
					L"%s : %s : 0x%X", // Message, Error, Code
					L"UsdNSCDevice::Loading Device Data Failed - No Events will be fired",
					L"Registry Read Error",
					::GetLastError()) ;
		return ;
	}

    LoadEventMask(hDeviceData);
    LoadDebounceTime(hDeviceData);
	LoadPollingFlag(hDeviceData);
	LoadRemoteWakeupFlag(hDeviceData);
	LoadKickStartTime(hDeviceData);

	RegCloseKey(hDeviceData);

    return;
}



/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::LoadEventMask

PARAMETERS: 

HKEY hDeviceData

RETURN VALUE:
void 

NOTES:

*******************************************************************************/
void UsdNSCDevice::LoadEventMask(HKEY hDeviceData)
{
    // Initially we assume that all events are turned off
    m_bEventMask = 0;

    DWORD dwType;
	char  szKey[16];
	char  szEventMask[8];
	DWORD cbData = sizeof(szEventMask)-1;

	for (int i=1; i <= NS_MAX_EVENTS; i++)
	{
		sprintf(szKey, "%s%d", "EventMask", i);
		LONG nRes = RegQueryValueEx( hDeviceData,
								szKey,
								NULL,
								&dwType,
								(LPBYTE)szEventMask, 
								&cbData );

		if ((nRes == ERROR_SUCCESS) && !strcmp(szEventMask, "1"))
		{
			m_bEventMask |= NS_EVENT_BIT[i-1];	
		}
	}
	
	// Also initialize the event guids for all the events
	m_bEventGuidArray[0] = guidEvent1;
	m_bEventGuidArray[1] = guidEvent2;
	m_bEventGuidArray[2] = guidEvent3;

	m_bEventsInitialized = true;

	return;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::LoadDebounceTime

PARAMETERS: 

HKEY   hDeviceData

RETURN VALUE:
void 

NOTES:

*******************************************************************************/
void UsdNSCDevice::LoadDebounceTime(HKEY   hDeviceData)
{
    DWORD   dwType;
	char*   szKey = "DebounceTime";
	char    szDebounceTime[8];
	DWORD   cbData = sizeof(szDebounceTime)-1;

    LONG nRes = RegQueryValueEx( hDeviceData,
								szKey,
								NULL,
								&dwType,
								(LPBYTE)szDebounceTime, 
								&cbData );

    // If we successsfully load the debounce value and it turns out 
    // to be within our limits  then we will change the debounce
    // time otherwise we just let it be the default value.
    if ((nRes == ERROR_SUCCESS))
	{
		DWORD dwDebounceTime = atoi(szDebounceTime);
        
        if ((dwDebounceTime >= NS_MIN_DEBOUNCE) &&
            (dwDebounceTime <= NS_MAX_DEBOUNCE))
        {
            m_dwDebounceTime = dwDebounceTime;
        }
	}
	
	return;
}


/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::SetRemoteWakeup - enables or disables the remote wakeup feature
on the scanner based on the input parameter. Uses the Win32 power managment fns
to accomplish this.

PARAMETERS: 

bool bEnable

RETURN VALUE:
HRESULT 

NOTES:
It is assumed that when this function is called, the device handle is 
valid. Since this is a protected function and cannot be invoked outside
the class this is reasoanable as only class memebers will have to make
sure that the devide handle is valid before invoking this function
The result code will be returned based on the return from the Win32 power
management fucntions that are invoked and the last error code will also 
be set appropriately

*******************************************************************************/
HRESULT UsdNSCDevice::SetRemoteWakeup(bool bEnable)
{
	if (!m_bRemoteWakeup) return STI_OK;

	UCHAR data = 0;
	IO_BLOCK	stIoBlock;
    stIoBlock.uOffset = 1;
    stIoBlock.uLength = 1;
    stIoBlock.pbyData = &data;
	stIoBlock.uIndex  = bEnable?NS_ENABLE_REMOTEWAKEUP:NS_DISABLE_REMOTEWAKEUP;

	DWORD cbBytes = sizeof(UCHAR);

	HRESULT hres = LockDevice();

	if (SUCCEEDED(hres)) 
	{
		hres = 	RawWriteCommand(&stIoBlock, cbBytes, NULL);
	
		UnLockDevice();
	
	}

	return (hres);
}


void UsdNSCDevice::LogError(char *string,DWORD ct,BYTE b0,BYTE b1,BYTE b2,
							BYTE b3,BYTE b4,HRESULT hres)
{
	//open file, save error
	char dir[MAX_PATH];		// file name 
	dir[0] = '\0';
	GetWindowsDirectory(dir,sizeof(dir));
	if (dir[sizeof(dir)-1] != '\\') strcat(dir,"\\");
	strcat(dir,"nscstiu_error.txt");

	FILE *fp;
	fp = fopen(dir,"at"); //open text file 
	if (fp)
	{
		fprintf(fp,"%s,len=%x,data=%x,%x,%x,%x,%x,err=%x\n",string,(int)ct,
			(int)b0,(int)b1,(int)b2,(int)b3,(int)b4,(int)hres);	//write to file
		fclose(fp); //close so keep if crash
	}
}

/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::LoadPollingFlag

PARAMETERS: 

HKEY   hDeviceData

RETURN VALUE:
void 

NOTES:

*******************************************************************************/
void UsdNSCDevice::LoadPollingFlag(HKEY   hDeviceData)
{
    DWORD   dwType;
	char*   szKey = "Polling";
	char    szPolling[8];
	DWORD   cbData = sizeof(szPolling)-1;

    LONG nRes = RegQueryValueEx( hDeviceData,
								szKey,
								NULL,
								&dwType,
								(LPBYTE)szPolling, 
								&cbData );

    // If we successsfully load the debounce value and it turns out 
    // to be within our limits  then we will change the debounce
    // time otherwise we just let it be the default value.
    if ((nRes == ERROR_SUCCESS))
	{
		DWORD dwPolling = atoi(szPolling);
        
        if ((dwPolling == 0) ||
            (dwPolling == 1))
        {
            m_bPolling = (dwPolling == 1);
        }
	}
	
	return;
}

/*******************************************************************************
DESCRIPTION:

UsdNSCDevice::LoadRemoteWakeupFlag

PARAMETERS: 

HKEY   hDeviceData

RETURN VALUE:
void 

NOTES:

*******************************************************************************/
void UsdNSCDevice::LoadRemoteWakeupFlag(HKEY   hDeviceData)
{
    DWORD   dwType;
	char*   szKey = "RemoteWakeup";
	char    szRemoteWakeup[8];
	DWORD   cbData = sizeof(szRemoteWakeup)-1;

    LONG nRes = RegQueryValueEx( hDeviceData,
								szKey,
								NULL,
								&dwType,
								(LPBYTE)szRemoteWakeup, 
								&cbData );

    // If we successsfully load the debounce value and it turns out 
    // to be within our limits  then we will change the debounce
    // time otherwise we just let it be the default value.
    if ((nRes == ERROR_SUCCESS))
	{
		DWORD dwRemoteWakeup = atoi(szRemoteWakeup);
        
        if ((dwRemoteWakeup == 0) ||
            (dwRemoteWakeup == 1))
        {
            m_bRemoteWakeup = (dwRemoteWakeup == 1);
        }
	}
	
	return;
}

//For WHQL test, do RawWriteData,RawReadData
HRESULT UsdNSCDevice::WHQLRawWriteReadTest(void)
{
		BYTE	pCommandBytes[COMMAND_BYTE_COUNT];
		BYTE	bValue;

		// Write the command bytes for a register read 
		// without increment 
		pCommandBytes[0] = MODE_NOINC_READ;
		pCommandBytes[1] = LM_COMMAND_REGISTER;
		pCommandBytes[2] = 0; 
		pCommandBytes[3] = 1;
		m_bWHQLTest = false;
		HRESULT hres = 	RawWriteData(pCommandBytes, COMMAND_BYTE_COUNT, NULL);
		
		if (SUCCEEDED(hres))
		{
			// Then read the register in question
			DWORD cbBytes(1);
			hres = RawReadData(&bValue, &cbBytes, NULL);
		}
		m_bWHQLTest = true;
		return hres;
}

//Do kick start
HRESULT UsdNSCDevice::DoKickStart()
{	
	NSC_TRACE("In UsdNSCDevice::DoKickStart");

	HRESULT hres = STIERR_NOT_INITIALIZED;

	// Do some none intrusive endpoint zero command to 
	// kick start any pausing bulk in
    if (INVALID_HANDLE_VALUE != m_DeviceDataHandle) 
	{
		UCHAR data = 0;
		IO_BLOCK	stIoBlock;
		stIoBlock.uOffset = 1;
		stIoBlock.uLength = 1;
		stIoBlock.pbyData = &data;
		stIoBlock.uIndex  = m_bRemoteWakeup?NS_ENABLE_REMOTEWAKEUP:NS_DISABLE_REMOTEWAKEUP;
		DWORD cbBytes = sizeof(UCHAR);

		hres = 	RawWriteCommand(&stIoBlock, cbBytes, NULL);
	
	}
    return hres;
}

/*******************************************************************************/
void UsdNSCDevice::LoadKickStartTime(HKEY   hDeviceData)
{
    DWORD   dwType;
	char*   szKey = "KickStart";
	char    szKickStart[8];
	DWORD   cbData = sizeof(szKickStart)-1;

    LONG nRes = RegQueryValueEx( hDeviceData,
								szKey,
								NULL,
								&dwType,
								(LPBYTE)szKickStart, 
								&cbData );

    if ((nRes == ERROR_SUCCESS))
	{
		m_dwKickStartTime = atoi(szKickStart);
 	}
	
	return;
}